/* Generated by Cython 3.0.12 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "include_dirs": [
            "/tmp/build-env-iwk05xn9/lib/python3.10/site-packages/numpy/_core/include"
        ],
        "name": "pyretechnics.fuel_models",
        "sources": [
            "src/pyretechnics/fuel_models.py"
        ]
    },
    "module_name": "pyretechnics.fuel_models"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#if defined(CYTHON_LIMITED_API) && 0
  #ifndef Py_LIMITED_API
    #if CYTHON_LIMITED_API+0 > 0x03030000
      #define Py_LIMITED_API CYTHON_LIMITED_API
    #else
      #define Py_LIMITED_API 0x03030000
    #endif
  #endif
#endif

#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02070000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.7+ or Python 3.3+.
#else
#if defined(CYTHON_LIMITED_API) && CYTHON_LIMITED_API
#define __PYX_EXTRA_ABI_MODULE_NAME "limited"
#else
#define __PYX_EXTRA_ABI_MODULE_NAME ""
#endif
#define CYTHON_ABI "3_0_12" __PYX_EXTRA_ABI_MODULE_NAME
#define __PYX_ABI_MODULE_NAME "_cython_" CYTHON_ABI
#define __PYX_TYPE_MODULE_PREFIX __PYX_ABI_MODULE_NAME "."
#define CYTHON_HEX_VERSION 0x03000CF0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(_WIN32) && !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #define HAVE_LONG_LONG
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#define __PYX_LIMITED_VERSION_HEX PY_VERSION_HEX
#if defined(GRAALVM_PYTHON)
  /* For very preliminary testing purposes. Most variables are set the same as PyPy.
     The existence of this section does not imply that anything works or is even tested */
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(PYPY_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #if PY_VERSION_HEX < 0x03090000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1 && PYPY_VERSION_NUM >= 0x07030C00)
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(CYTHON_LIMITED_API)
  #ifdef Py_LIMITED_API
    #undef __PYX_LIMITED_VERSION_HEX
    #define __PYX_LIMITED_VERSION_HEX Py_LIMITED_API
  #endif
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_CLINE_IN_TRACEBACK
  #define CYTHON_CLINE_IN_TRACEBACK 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 1
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 1
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(Py_GIL_DISABLED) || defined(Py_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL 1
  #endif
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS 0
  #endif
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL (PY_MAJOR_VERSION < 3 || PY_VERSION_HEX >= 0x03060000 && PY_VERSION_HEX < 0x030C00A6)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL (PY_VERSION_HEX >= 0x030700A1)
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #if PY_VERSION_HEX < 0x030400a1
    #undef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #elif !defined(CYTHON_USE_TP_FINALIZE)
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #if PY_VERSION_HEX < 0x030600B1
    #undef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS 0
  #elif !defined(CYTHON_USE_DICT_VERSIONS)
    #define CYTHON_USE_DICT_VERSIONS  (PY_VERSION_HEX < 0x030C00A5)
  #endif
  #if PY_VERSION_HEX < 0x030700A3
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK 1
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if !defined(CYTHON_VECTORCALL)
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL && PY_VERSION_HEX >= 0x030800B1)
#endif
#define CYTHON_BACKPORT_VECTORCALL (CYTHON_METH_FASTCALL && PY_VERSION_HEX < 0x030800B1)
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(maybe_unused) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(maybe_unused)
        #define CYTHON_UNUSED [[maybe_unused]]
      #endif
    #endif
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
  #define CYTHON_MAYBE_UNUSED_VAR(x) CYTHON_UNUSED_VAR(x)
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_USE_CPP_STD_MOVE
  #if defined(__cplusplus) && (\
    __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600))
    #define CYTHON_USE_CPP_STD_MOVE 1
  #else
    #define CYTHON_USE_CPP_STD_MOVE 0
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
            typedef unsigned char     uint8_t;
            typedef unsigned short    uint16_t;
            typedef unsigned int      uint32_t;
        #else
            typedef unsigned __int8   uint8_t;
            typedef unsigned __int16  uint16_t;
            typedef unsigned __int32  uint32_t;
        #endif
    #endif
    #if _MSC_VER < 1300
        #ifdef _WIN64
            typedef unsigned long long  __pyx_uintptr_t;
        #else
            typedef unsigned int        __pyx_uintptr_t;
        #endif
    #else
        #ifdef _WIN64
            typedef unsigned __int64    __pyx_uintptr_t;
        #else
            typedef unsigned __int32    __pyx_uintptr_t;
        #endif
    #endif
#else
    #include <stdint.h>
    typedef uintptr_t  __pyx_uintptr_t;
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(fallthrough) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(fallthrough)
        #define CYTHON_FALLTHROUGH [[fallthrough]]
      #endif
    #endif
    #ifndef CYTHON_FALLTHROUGH
      #if __has_cpp_attribute(clang::fallthrough)
        #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
      #elif __has_cpp_attribute(gnu::fallthrough)
        #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
      #endif
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif
#ifdef __cplusplus
  template <typename T>
  struct __PYX_IS_UNSIGNED_IMPL {static const bool value = T(0) < T(-1);};
  #define __PYX_IS_UNSIGNED(type) (__PYX_IS_UNSIGNED_IMPL<type>::value)
#else
  #define __PYX_IS_UNSIGNED(type) (((type)-1) > 0)
#endif
#if CYTHON_COMPILING_IN_PYPY == 1
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x030A0000)
#else
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000)
#endif
#define __PYX_REINTERPRET_FUNCION(func_pointer, other_pointer) ((func_pointer)(void(*)(void))(other_pointer))

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_DefaultClassType PyClass_Type
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if CYTHON_COMPILING_IN_LIMITED_API
    static CYTHON_INLINE PyObject* __Pyx_PyCode_New(int a, int p, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *exception_table = NULL;
        PyObject *types_module=NULL, *code_type=NULL, *result=NULL;
        #if __PYX_LIMITED_VERSION_HEX < 0x030B0000
        PyObject *version_info;
        PyObject *py_minor_version = NULL;
        #endif
        long minor_version = 0;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        #if __PYX_LIMITED_VERSION_HEX >= 0x030B0000
        minor_version = 11;
        #else
        if (!(version_info = PySys_GetObject("version_info"))) goto end;
        if (!(py_minor_version = PySequence_GetItem(version_info, 1))) goto end;
        minor_version = PyLong_AsLong(py_minor_version);
        Py_DECREF(py_minor_version);
        if (minor_version == -1 && PyErr_Occurred()) goto end;
        #endif
        if (!(types_module = PyImport_ImportModule("types"))) goto end;
        if (!(code_type = PyObject_GetAttrString(types_module, "CodeType"))) goto end;
        if (minor_version <= 7) {
            (void)p;
            result = PyObject_CallFunction(code_type, "iiiiiOOOOOOiOO", a, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else if (minor_version <= 10) {
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOiOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else {
            if (!(exception_table = PyBytes_FromStringAndSize(NULL, 0))) goto end;
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOOiOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, name, fline, lnos, exception_table, fv, cell);
        }
    end:
        Py_XDECREF(code_type);
        Py_XDECREF(exception_table);
        Py_XDECREF(types_module);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return result;
    }
    #ifndef CO_OPTIMIZED
    #define CO_OPTIMIZED 0x0001
    #endif
    #ifndef CO_NEWLOCALS
    #define CO_NEWLOCALS 0x0002
    #endif
    #ifndef CO_VARARGS
    #define CO_VARARGS 0x0004
    #endif
    #ifndef CO_VARKEYWORDS
    #define CO_VARKEYWORDS 0x0008
    #endif
    #ifndef CO_ASYNC_GENERATOR
    #define CO_ASYNC_GENERATOR 0x0200
    #endif
    #ifndef CO_GENERATOR
    #define CO_GENERATOR 0x0020
    #endif
    #ifndef CO_COROUTINE
    #define CO_COROUTINE 0x0080
    #endif
#elif PY_VERSION_HEX >= 0x030B0000
  static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int p, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
    PyCodeObject *result;
    PyObject *empty_bytes = PyBytes_FromStringAndSize("", 0);
    if (!empty_bytes) return NULL;
    result =
      #if PY_VERSION_HEX >= 0x030C0000
        PyUnstable_Code_NewWithPosOnlyArgs
      #else
        PyCode_NewWithPosOnlyArgs
      #endif
        (a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, name, fline, lnos, empty_bytes);
    Py_DECREF(empty_bytes);
    return result;
  }
#elif PY_VERSION_HEX >= 0x030800B2 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_NewWithPosOnlyArgs(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
#endif
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_Is)
  #define __Pyx_Py_Is(x, y)  Py_Is(x, y)
#else
  #define __Pyx_Py_Is(x, y) ((x) == (y))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsNone)
  #define __Pyx_Py_IsNone(ob) Py_IsNone(ob)
#else
  #define __Pyx_Py_IsNone(ob) __Pyx_Py_Is((ob), Py_None)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsTrue)
  #define __Pyx_Py_IsTrue(ob) Py_IsTrue(ob)
#else
  #define __Pyx_Py_IsTrue(ob) __Pyx_Py_Is((ob), Py_True)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsFalse)
  #define __Pyx_Py_IsFalse(ob) Py_IsFalse(ob)
#else
  #define __Pyx_Py_IsFalse(ob) __Pyx_Py_Is((ob), Py_False)
#endif
#define __Pyx_NoneAsNull(obj)  (__Pyx_Py_IsNone(obj) ? NULL : (obj))
#if PY_VERSION_HEX >= 0x030900F0 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyObject_GC_IsFinalized(o) PyObject_GC_IsFinalized(o)
#else
  #define __Pyx_PyObject_GC_IsFinalized(o) _PyGC_FINALIZED(o)
#endif
#ifndef CO_COROUTINE
  #define CO_COROUTINE 0x80
#endif
#ifndef CO_ASYNC_GENERATOR
  #define CO_ASYNC_GENERATOR 0x200
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef Py_TPFLAGS_SEQUENCE
  #define Py_TPFLAGS_SEQUENCE 0
#endif
#ifndef Py_TPFLAGS_MAPPING
  #define Py_TPFLAGS_MAPPING 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #if PY_VERSION_HEX >= 0x030d00A4
  #  define __Pyx_PyCFunctionFast PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords PyCFunctionFastWithKeywords
  #else
  #  define __Pyx_PyCFunctionFast _PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
  #endif
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS((size_t)(n))
#elif CYTHON_BACKPORT_VECTORCALL
  typedef PyObject *(*__pyx_vectorcallfunc)(PyObject *callable, PyObject *const *args,
                                            size_t nargsf, PyObject *kwnames);
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  ((size_t)1 << (8 * sizeof(size_t) - 1))
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(((size_t)(n)) & ~__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET))
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(n))
#endif
#if PY_MAJOR_VERSION >= 0x030900B1
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_CheckExact(func)
#else
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_Check(func)
#endif
#define __Pyx_CyOrPyCFunction_Check(func)  PyCFunction_Check(func)
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  (((PyCFunctionObject*)(func))->m_ml->ml_meth)
#elif !CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  PyCFunction_GET_FUNCTION(func)
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FLAGS(func)  (((PyCFunctionObject*)(func))->m_ml->ml_flags)
static CYTHON_INLINE PyObject* __Pyx_CyOrPyCFunction_GET_SELF(PyObject *func) {
    return (__Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_STATIC) ? NULL : ((PyCFunctionObject*)func)->m_self;
}
#endif
static CYTHON_INLINE int __Pyx__IsSameCFunction(PyObject *func, void *cfunc) {
#if CYTHON_COMPILING_IN_LIMITED_API
    return PyCFunction_Check(func) && PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
#else
    return PyCFunction_Check(func) && PyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
#endif
}
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCFunction(func, cfunc)
#if __PYX_LIMITED_VERSION_HEX < 0x030900B1
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  ((void)m, PyType_FromSpecWithBases(s, b))
  typedef PyObject *(*__Pyx_PyCMethod)(PyObject *, PyTypeObject *, PyObject *const *, size_t, PyObject *);
#else
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  PyType_FromModuleAndSpec(m, s, b)
  #define __Pyx_PyCMethod  PyCMethod
#endif
#ifndef METH_METHOD
  #define METH_METHOD 0x200
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyThreadState_Current PyThreadState_GetUnchecked()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE void *__Pyx_PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#endif
#define __Pyx_PyObject_GetSlot(obj, name, func_ctype)  __Pyx_PyType_GetSlot(Py_TYPE(obj), name, func_ctype)
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((func_ctype) PyType_GetSlot((type), Py_##name))
#else
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((type)->name)
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if PY_MAJOR_VERSION < 3
    #if CYTHON_COMPILING_IN_PYPY
        #if PYPY_VERSION_NUM < 0x07030600
            #if defined(__cplusplus) && __cplusplus >= 201402L
                [[deprecated("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6")]]
            #elif defined(__GNUC__) || defined(__clang__)
                __attribute__ ((__deprecated__("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6")))
            #elif defined(_MSC_VER)
                __declspec(deprecated("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6"))
            #endif
            static CYTHON_INLINE int PyGILState_Check(void) {
                return 0;
            }
        #else  // PYPY_VERSION_NUM < 0x07030600
        #endif  // PYPY_VERSION_NUM < 0x07030600
    #else
        static CYTHON_INLINE int PyGILState_Check(void) {
            PyThreadState * tstate = _PyThreadState_Current;
            return tstate && (tstate == PyGILState_GetThisThreadState());
        }
    #endif
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000 || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX > 0x030600B4 && PY_VERSION_HEX < 0x030d0000 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif PY_MAJOR_VERSION >= 3 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000)
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
  #define __Pyx_PyType_HasFeature(type, feature)  ((__Pyx_PyType_GetFlags(type) & (feature)) != 0)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  (Py_TYPE(obj)->tp_iternext)
#else
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
  #define __Pyx_PyType_HasFeature(type, feature)  PyType_HasFeature(type, feature)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  PyIter_Next
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_SetItemOnTypeDict(tp, k, v) PyObject_GenericSetAttr((PyObject*)tp, k, v)
#else
  #define __Pyx_SetItemOnTypeDict(tp, k, v) PyDict_SetItem(tp->tp_dict, k, v)
#endif
#if CYTHON_USE_TYPE_SPECS && PY_VERSION_HEX >= 0x03080000
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  {\
    PyTypeObject *type = Py_TYPE((PyObject*)obj);\
    assert(__Pyx_PyType_HasFeature(type, Py_TPFLAGS_HEAPTYPE));\
    PyObject_GC_Del(obj);\
    Py_DECREF(type);\
}
#else
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  PyObject_GC_Del(obj)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GetLength(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#elif PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         ((int)PyUnicode_KIND(u))
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, (Py_UCS4) ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535U : 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((int)sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = (Py_UNICODE) ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #if !defined(PyUnicode_DecodeUnicodeEscape)
    #define PyUnicode_DecodeUnicodeEscape(s, size, errors)  PyUnicode_Decode(s, size, "unicode_escape", errors)
  #endif
  #if !defined(PyUnicode_Contains) || (PY_MAJOR_VERSION == 2 && PYPY_VERSION_NUM < 0x07030500)
    #undef PyUnicode_Contains
    #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
  #endif
  #if !defined(PyByteArray_Check)
    #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
  #endif
  #if !defined(PyObject_Format)
    #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
  #endif
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#if CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_ITEM(o, i) PySequence_ITEM(o, i)
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) (PyTuple_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyList_SET_ITEM(o, i, v) (PyList_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_GET_SIZE(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_GET_SIZE(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_GET_SIZE(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_GET_SIZE(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_GET_SIZE(o)
#else
  #define __Pyx_PySequence_ITEM(o, i) PySequence_GetItem(o, i)
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) PyTuple_SetItem(o, i, v)
  #define __Pyx_PyList_SET_ITEM(o, i, v) PyList_SetItem(o, i, v)
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_Size(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_Size(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_Size(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_Size(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_Size(o)
#endif
#if __PYX_LIMITED_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#else
  static CYTHON_INLINE PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *module = PyImport_AddModule(name);
      Py_XINCREF(module);
      return module;
  }
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define __Pyx_Py3Int_Check(op)       PyLong_Check(op)
  #define __Pyx_Py3Int_CheckExact(op)  PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#else
  #define __Pyx_Py3Int_Check(op)       (PyLong_Check(op) || PyInt_Check(op))
  #define __Pyx_Py3Int_CheckExact(op)  (PyLong_CheckExact(op) || PyInt_CheckExact(op))
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__;  (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifdef CYTHON_EXTERN_C
    #undef __PYX_EXTERN_C
    #define __PYX_EXTERN_C CYTHON_EXTERN_C
#elif defined(__PYX_EXTERN_C)
    #ifdef _MSC_VER
    #pragma message ("Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.")
    #else
    #warning Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.
    #endif
#else
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__pyretechnics__fuel_models
#define __PYX_HAVE_API__pyretechnics__fuel_models
/* Early includes */
#include <math.h>
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s);
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char*);
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#define __Pyx_PyUnicode_FromOrdinal(o)       PyUnicode_FromOrdinal((int)o)
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_VERSION_HEX >= 0x030C00A7
  #ifndef _PyLong_SIGN_MASK
    #define _PyLong_SIGN_MASK 3
  #endif
  #ifndef _PyLong_NON_SIZE_BITS
    #define _PyLong_NON_SIZE_BITS 3
  #endif
  #define __Pyx_PyLong_Sign(x)  (((PyLongObject*)x)->long_value.lv_tag & _PyLong_SIGN_MASK)
  #define __Pyx_PyLong_IsNeg(x)  ((__Pyx_PyLong_Sign(x) & 2) != 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (!__Pyx_PyLong_IsNeg(x))
  #define __Pyx_PyLong_IsZero(x)  (__Pyx_PyLong_Sign(x) & 1)
  #define __Pyx_PyLong_IsPos(x)  (__Pyx_PyLong_Sign(x) == 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  (__Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  ((Py_ssize_t) (((PyLongObject*)x)->long_value.lv_tag >> _PyLong_NON_SIZE_BITS))
  #define __Pyx_PyLong_SignedDigitCount(x)\
        ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * __Pyx_PyLong_DigitCount(x))
  #if defined(PyUnstable_Long_IsCompact) && defined(PyUnstable_Long_CompactValue)
    #define __Pyx_PyLong_IsCompact(x)     PyUnstable_Long_IsCompact((PyLongObject*) x)
    #define __Pyx_PyLong_CompactValue(x)  PyUnstable_Long_CompactValue((PyLongObject*) x)
  #else
    #define __Pyx_PyLong_IsCompact(x)     (((PyLongObject*)x)->long_value.lv_tag < (2 << _PyLong_NON_SIZE_BITS))
    #define __Pyx_PyLong_CompactValue(x)  ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * (Py_ssize_t) __Pyx_PyLong_Digits(x)[0])
  #endif
  typedef Py_ssize_t  __Pyx_compact_pylong;
  typedef size_t  __Pyx_compact_upylong;
  #else
  #define __Pyx_PyLong_IsNeg(x)  (Py_SIZE(x) < 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (Py_SIZE(x) >= 0)
  #define __Pyx_PyLong_IsZero(x)  (Py_SIZE(x) == 0)
  #define __Pyx_PyLong_IsPos(x)  (Py_SIZE(x) > 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  ((Py_SIZE(x) == 0) ? 0 : __Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  __Pyx_sst_abs(Py_SIZE(x))
  #define __Pyx_PyLong_SignedDigitCount(x)  Py_SIZE(x)
  #define __Pyx_PyLong_IsCompact(x)  (Py_SIZE(x) == 0 || Py_SIZE(x) == 1 || Py_SIZE(x) == -1)
  #define __Pyx_PyLong_CompactValue(x)\
        ((Py_SIZE(x) == 0) ? (sdigit) 0 : ((Py_SIZE(x) < 0) ? -(sdigit)__Pyx_PyLong_Digits(x)[0] : (sdigit)__Pyx_PyLong_Digits(x)[0]))
  typedef sdigit  __Pyx_compact_pylong;
  typedef digit  __Pyx_compact_upylong;
  #endif
  #if PY_VERSION_HEX >= 0x030C00A5
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->long_value.ob_digit)
  #else
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->ob_digit)
  #endif
#endif
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
#include <string.h>
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = (char) c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#include <string.h>
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_m = NULL;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* #### Code section: filename_table ### */

static const char *__pyx_f[] = {
  "<stringsource>",
  "src/pyretechnics/fuel_models.py",
};
/* #### Code section: utility_code_proto_before_types ### */
/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* #### Code section: numeric_typedefs ### */

/* "pyretechnics/cy_types.pxd":6
 * #==================================================
 * 
 * ctypedef Py_ssize_t pyidx             # <<<<<<<<<<<<<<
 * ctypedef (float, float) vec_xy
 * ctypedef (float, float, float) vec_xyz
 */
typedef Py_ssize_t __pyx_t_12pyretechnics_8cy_types_pyidx;
/* #### Code section: complex_type_declarations ### */
/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
struct __pyx_ctuple_float__and_float;
typedef struct __pyx_ctuple_float__and_float __pyx_ctuple_float__and_float;
struct __pyx_ctuple_float__and_float__and_float;
typedef struct __pyx_ctuple_float__and_float__and_float __pyx_ctuple_float__and_float__and_float;
struct __pyx_ctuple_f4ae04__65e98e__12pyretechnics_8cy_types__dunder_pyx_t__etc;
typedef struct __pyx_ctuple_f4ae04__65e98e__12pyretechnics_8cy_types__dunder_pyx_t__etc __pyx_ctuple_f4ae04__65e98e__12pyretechnics_8cy_types__dunder_pyx_t__etc;
struct __pyx_ctuple_ea8a79__65e98e__12pyretechnics_8cy_types__dunder_pyx_t__etc;
typedef struct __pyx_ctuple_ea8a79__65e98e__12pyretechnics_8cy_types__dunder_pyx_t__etc __pyx_ctuple_ea8a79__65e98e__12pyretechnics_8cy_types__dunder_pyx_t__etc;
struct __pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float;
typedef struct __pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float __pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float;
struct __pyx_ctuple_a8db9c__float__and_float__and_float__and_float__and_fl__etc;
typedef struct __pyx_ctuple_a8db9c__float__and_float__and_float__and_float__and_fl__etc __pyx_ctuple_a8db9c__float__and_float__and_float__and_float__and_fl__etc;
struct __pyx_t_12pyretechnics_8cy_types_FuelModel;
struct __pyx_t_12pyretechnics_8cy_types_ProjectedVectors;
struct __pyx_t_12pyretechnics_8cy_types_FireBehaviorMin;
struct __pyx_t_12pyretechnics_8cy_types_FireBehaviorMax;
struct __pyx_t_12pyretechnics_8cy_types_SpreadBehavior;
struct __pyx_t_12pyretechnics_8cy_types_CrownSpreadInfo;
struct __pyx_t_12pyretechnics_8cy_types_SpotConfig;
struct __pyx_t_12pyretechnics_8cy_types_JumpDistribution;
struct __pyx_t_12pyretechnics_8cy_types_PartialedEllWavelet;
struct __pyx_t_12pyretechnics_8cy_types_CellInputs;
struct __pyx_t_12pyretechnics_8cy_types_EllipticalInfo;
struct __pyx_t_12pyretechnics_8cy_types_Pass1CellOutput;

/* "pyretechnics/cy_types.pxd":7
 * 
 * ctypedef Py_ssize_t pyidx
 * ctypedef (float, float) vec_xy             # <<<<<<<<<<<<<<
 * ctypedef (float, float, float) vec_xyz
 * ctypedef (pyidx, pyidx) coord_yx
 */
struct __pyx_ctuple_float__and_float {
  float f0;
  float f1;
};
typedef __pyx_ctuple_float__and_float __pyx_t_12pyretechnics_8cy_types_vec_xy;

/* "pyretechnics/cy_types.pxd":8
 * ctypedef Py_ssize_t pyidx
 * ctypedef (float, float) vec_xy
 * ctypedef (float, float, float) vec_xyz             # <<<<<<<<<<<<<<
 * ctypedef (pyidx, pyidx) coord_yx
 * ctypedef (pyidx, pyidx, pyidx) coord_tyx
 */
struct __pyx_ctuple_float__and_float__and_float {
  float f0;
  float f1;
  float f2;
};
typedef __pyx_ctuple_float__and_float__and_float __pyx_t_12pyretechnics_8cy_types_vec_xyz;

/* "pyretechnics/cy_types.pxd":9
 * ctypedef (float, float) vec_xy
 * ctypedef (float, float, float) vec_xyz
 * ctypedef (pyidx, pyidx) coord_yx             # <<<<<<<<<<<<<<
 * ctypedef (pyidx, pyidx, pyidx) coord_tyx
 * ctypedef (float, float) fcatarr
 */
struct __pyx_ctuple_f4ae04__65e98e__12pyretechnics_8cy_types__dunder_pyx_t__etc {
  __pyx_t_12pyretechnics_8cy_types_pyidx f0;
  __pyx_t_12pyretechnics_8cy_types_pyidx f1;
};
typedef __pyx_ctuple_f4ae04__65e98e__12pyretechnics_8cy_types__dunder_pyx_t__etc __pyx_t_12pyretechnics_8cy_types_coord_yx;

/* "pyretechnics/cy_types.pxd":10
 * ctypedef (float, float, float) vec_xyz
 * ctypedef (pyidx, pyidx) coord_yx
 * ctypedef (pyidx, pyidx, pyidx) coord_tyx             # <<<<<<<<<<<<<<
 * ctypedef (float, float) fcatarr
 * ctypedef (float, float, float, float, float, float) fclaarr
 */
struct __pyx_ctuple_ea8a79__65e98e__12pyretechnics_8cy_types__dunder_pyx_t__etc {
  __pyx_t_12pyretechnics_8cy_types_pyidx f0;
  __pyx_t_12pyretechnics_8cy_types_pyidx f1;
  __pyx_t_12pyretechnics_8cy_types_pyidx f2;
};
typedef __pyx_ctuple_ea8a79__65e98e__12pyretechnics_8cy_types__dunder_pyx_t__etc __pyx_t_12pyretechnics_8cy_types_coord_tyx;

/* "pyretechnics/cy_types.pxd":11
 * ctypedef (pyidx, pyidx) coord_yx
 * ctypedef (pyidx, pyidx, pyidx) coord_tyx
 * ctypedef (float, float) fcatarr             # <<<<<<<<<<<<<<
 * ctypedef (float, float, float, float, float, float) fclaarr
 * ctypedef (float, float, float, float, float, float, float, float, float, float, float, float, float) CompactFuelModel
 */
typedef __pyx_ctuple_float__and_float __pyx_t_12pyretechnics_8cy_types_fcatarr;

/* "pyretechnics/cy_types.pxd":12
 * ctypedef (pyidx, pyidx, pyidx) coord_tyx
 * ctypedef (float, float) fcatarr
 * ctypedef (float, float, float, float, float, float) fclaarr             # <<<<<<<<<<<<<<
 * ctypedef (float, float, float, float, float, float, float, float, float, float, float, float, float) CompactFuelModel
 * 
 */
struct __pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float {
  float f0;
  float f1;
  float f2;
  float f3;
  float f4;
  float f5;
};
typedef __pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float __pyx_t_12pyretechnics_8cy_types_fclaarr;

/* "pyretechnics/cy_types.pxd":13
 * ctypedef (float, float) fcatarr
 * ctypedef (float, float, float, float, float, float) fclaarr
 * ctypedef (float, float, float, float, float, float, float, float, float, float, float, float, float) CompactFuelModel             # <<<<<<<<<<<<<<
 * 
 * cdef struct FuelModel:
 */
struct __pyx_ctuple_a8db9c__float__and_float__and_float__and_float__and_fl__etc {
  float f0;
  float f1;
  float f2;
  float f3;
  float f4;
  float f5;
  float f6;
  float f7;
  float f8;
  float f9;
  float f10;
  float f11;
  float f12;
};
typedef __pyx_ctuple_a8db9c__float__and_float__and_float__and_float__and_fl__etc __pyx_t_12pyretechnics_8cy_types_CompactFuelModel;

/* "pyretechnics/cy_types.pxd":15
 * ctypedef (float, float, float, float, float, float, float, float, float, float, float, float, float) CompactFuelModel
 * 
 * cdef struct FuelModel:             # <<<<<<<<<<<<<<
 *     int number
 *     float delta
 */
struct __pyx_t_12pyretechnics_8cy_types_FuelModel {
  int number;
  float delta;
  __pyx_t_12pyretechnics_8cy_types_fclaarr M_x;
  __pyx_t_12pyretechnics_8cy_types_fclaarr M_f;
  __pyx_t_12pyretechnics_8cy_types_fclaarr w_o;
  __pyx_t_12pyretechnics_8cy_types_fclaarr sigma;
  __pyx_t_12pyretechnics_8cy_types_fclaarr h;
  __pyx_t_12pyretechnics_8cy_types_fclaarr rho_p;
  __pyx_t_12pyretechnics_8cy_types_fclaarr S_T;
  __pyx_t_12pyretechnics_8cy_types_fclaarr S_e;
  int dynamic;
  int burnable;
  __pyx_t_12pyretechnics_8cy_types_fclaarr exp_A_sigma;
  __pyx_t_12pyretechnics_8cy_types_fclaarr firemod_size_classes;
  __pyx_t_12pyretechnics_8cy_types_fclaarr f_ij;
  __pyx_t_12pyretechnics_8cy_types_fcatarr f_i;
  __pyx_t_12pyretechnics_8cy_types_fclaarr g_ij;
};

/* "pyretechnics/cy_types.pxd":34
 *     fclaarr g_ij
 * 
 * cdef struct ProjectedVectors:             # <<<<<<<<<<<<<<
 *     vec_xyz wind_vector_3d
 *     vec_xyz slope_vector_3d
 */
struct __pyx_t_12pyretechnics_8cy_types_ProjectedVectors {
  __pyx_t_12pyretechnics_8cy_types_vec_xyz wind_vector_3d;
  __pyx_t_12pyretechnics_8cy_types_vec_xyz slope_vector_3d;
};

/* "pyretechnics/cy_types.pxd":38
 *     vec_xyz slope_vector_3d
 * 
 * cdef struct FireBehaviorMin:             # <<<<<<<<<<<<<<
 *     float base_spread_rate
 *     float base_fireline_intensity
 */
struct __pyx_t_12pyretechnics_8cy_types_FireBehaviorMin {
  float base_spread_rate;
  float base_fireline_intensity;
  float max_effective_wind_speed;
  float _phiS_G;
  float _phiW_scalr;
  float _phiW_expnt;
  float _ws_scalr;
  float _ws_expnt;
};

/* "pyretechnics/cy_types.pxd":48
 *     float _ws_expnt
 * 
 * cdef struct FireBehaviorMax:             # <<<<<<<<<<<<<<
 *     int max_fire_type
 *     float max_spread_rate
 */
struct __pyx_t_12pyretechnics_8cy_types_FireBehaviorMax {
  int max_fire_type;
  float max_spread_rate;
  __pyx_t_12pyretechnics_8cy_types_vec_xyz max_spread_direction;
  float max_fireline_intensity;
  float max_flame_length;
  float length_to_width_ratio;
  float eccentricity;
  float critical_spread_rate;
};

/* "pyretechnics/cy_types.pxd":58
 *     float critical_spread_rate
 * 
 * cdef struct SpreadBehavior:             # <<<<<<<<<<<<<<
 *     float dphi_dt
 *     int fire_type
 */
struct __pyx_t_12pyretechnics_8cy_types_SpreadBehavior {
  float dphi_dt;
  int fire_type;
  float spread_rate;
  __pyx_t_12pyretechnics_8cy_types_vec_xyz spread_direction;
  float fireline_intensity;
  float flame_length;
};

/* "pyretechnics/cy_types.pxd":66
 *     float flame_length
 * 
 * cdef struct CrownSpreadInfo:             # <<<<<<<<<<<<<<
 *     int fire_type
 *     float spread_rate
 */
struct __pyx_t_12pyretechnics_8cy_types_CrownSpreadInfo {
  int fire_type;
  float spread_rate;
  float critical_spread_rate;
};

/* "pyretechnics/cy_types.pxd":71
 *     float critical_spread_rate
 * 
 * cdef struct SpotConfig:             # <<<<<<<<<<<<<<
 *     long random_seed
 *     float firebrands_per_unit_heat
 */
struct __pyx_t_12pyretechnics_8cy_types_SpotConfig {
  long random_seed;
  float firebrands_per_unit_heat;
  float downwind_distance_mean;
  float fireline_intensity_exponent;
  float wind_speed_exponent;
  float downwind_variance_mean_ratio;
  float crosswind_distance_stdev;
  float decay_distance;
};

/* "pyretechnics/cy_types.pxd":81
 *     float decay_distance
 * 
 * cdef struct JumpDistribution:             # <<<<<<<<<<<<<<
 *     float mu_x
 *     float sigma_x
 */
struct __pyx_t_12pyretechnics_8cy_types_JumpDistribution {
  float mu_x;
  float sigma_x;
  float sigma_y;
};

/* "pyretechnics/cy_types.pxd":86
 *     float sigma_y
 * 
 * cdef struct PartialedEllWavelet:             # <<<<<<<<<<<<<<
 *     vec_xyz Vh_3d
 *     float ewc_A
 */
struct __pyx_t_12pyretechnics_8cy_types_PartialedEllWavelet {
  __pyx_t_12pyretechnics_8cy_types_vec_xyz Vh_3d;
  float ewc_A;
  float ewc_B;
  float ewc_C;
};

/* "pyretechnics/cy_types.pxd":92
 *     float ewc_C
 * 
 * cdef struct CellInputs:             # <<<<<<<<<<<<<<
 *     float slope
 *     float aspect
 */
struct __pyx_t_12pyretechnics_8cy_types_CellInputs {
  float slope;
  float aspect;
  float fuel_model_number;
  float canopy_cover;
  float canopy_height;
  float canopy_base_height;
  float canopy_bulk_density;
  float wind_speed_10m;
  float upwind_direction;
  float fuel_moisture_dead_1hr;
  float fuel_moisture_dead_10hr;
  float fuel_moisture_dead_100hr;
  float fuel_moisture_live_herbaceous;
  float fuel_moisture_live_woody;
  float foliar_moisture;
  float fuel_spread_adjustment;
  float weather_spread_adjustment;
};

/* "pyretechnics/cy_types.pxd":111
 *     float weather_spread_adjustment
 * 
 * cdef struct EllipticalInfo:             # <<<<<<<<<<<<<<
 *     coord_yx cell_index
 *     vec_xy elevation_gradient
 */
struct __pyx_t_12pyretechnics_8cy_types_EllipticalInfo {
  __pyx_t_12pyretechnics_8cy_types_coord_yx cell_index;
  __pyx_t_12pyretechnics_8cy_types_vec_xy elevation_gradient;
  struct __pyx_t_12pyretechnics_8cy_types_PartialedEllWavelet surface_wavelet;
  struct __pyx_t_12pyretechnics_8cy_types_PartialedEllWavelet crown_wavelet;
  float crowning_spread_rate;
};

/* "pyretechnics/cy_types.pxd":118
 *     float crowning_spread_rate
 * 
 * cdef struct Pass1CellOutput:             # <<<<<<<<<<<<<<
 *     coord_yx cell_index
 *     vec_xy phi_gradient_xy
 */
struct __pyx_t_12pyretechnics_8cy_types_Pass1CellOutput {
  __pyx_t_12pyretechnics_8cy_types_coord_yx cell_index;
  __pyx_t_12pyretechnics_8cy_types_vec_xy phi_gradient_xy;
  float dphi_dt_flim;
  float phi_old;
};
/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__))
  #define __Pyx_RefNannyFinishContextNogil() __Pyx_RefNannyFinishContext()
#endif
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_XINCREF(r)  do { if((r) == NULL); else {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) == NULL); else {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) == NULL); else {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) == NULL); else {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#if PY_VERSION_HEX >= 0x030C00A6
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->current_exception != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->current_exception ? (PyObject*) Py_TYPE(__pyx_tstate->current_exception) : (PyObject*) NULL)
#else
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->curexc_type != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->curexc_type)
#endif
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  (PyErr_Occurred() != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A6
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* py_dict_keys.proto */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Keys(PyObject* d);

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod0.proto */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CallUnboundCMethod0(cfunc, self)\
    (likely((cfunc)->func) ?\
        (likely((cfunc)->flag == METH_NOARGS) ?  (*((cfunc)->func))(self, NULL) :\
         (PY_VERSION_HEX >= 0x030600B1 && likely((cfunc)->flag == METH_FASTCALL) ?\
            (PY_VERSION_HEX >= 0x030700A0 ?\
                (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0) :\
                (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL)) :\
          (PY_VERSION_HEX >= 0x030700A0 && (cfunc)->flag == (METH_FASTCALL | METH_KEYWORDS) ?\
            (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL) :\
            (likely((cfunc)->flag == (METH_VARARGS | METH_KEYWORDS)) ?  ((*(PyCFunctionWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, __pyx_empty_tuple, NULL)) :\
               ((cfunc)->flag == METH_VARARGS ?  (*((cfunc)->func))(self, __pyx_empty_tuple) :\
               __Pyx__CallUnboundCMethod0(cfunc, self)))))) :\
        __Pyx__CallUnboundCMethod0(cfunc, self))
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* TupleAndListFromArray.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n);
static CYTHON_INLINE PyObject* __Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n);
#endif

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* fastcall.proto */
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_Arg_VARARGS(args, i) PySequence_GetItem(args, i)
#elif CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_Arg_VARARGS(args, i) PyTuple_GET_ITEM(args, i)
#else
    #define __Pyx_Arg_VARARGS(args, i) PyTuple_GetItem(args, i)
#endif
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_Arg_NewRef_VARARGS(arg) __Pyx_NewRef(arg)
    #define __Pyx_Arg_XDECREF_VARARGS(arg) Py_XDECREF(arg)
#else
    #define __Pyx_Arg_NewRef_VARARGS(arg) arg
    #define __Pyx_Arg_XDECREF_VARARGS(arg)
#endif
#define __Pyx_NumKwargs_VARARGS(kwds) PyDict_Size(kwds)
#define __Pyx_KwValues_VARARGS(args, nargs) NULL
#define __Pyx_GetKwValue_VARARGS(kw, kwvalues, s) __Pyx_PyDict_GetItemStrWithError(kw, s)
#define __Pyx_KwargsAsDict_VARARGS(kw, kwvalues) PyDict_Copy(kw)
#if CYTHON_METH_FASTCALL
    #define __Pyx_Arg_FASTCALL(args, i) args[i]
    #define __Pyx_NumKwargs_FASTCALL(kwds) PyTuple_GET_SIZE(kwds)
    #define __Pyx_KwValues_FASTCALL(args, nargs) ((args) + (nargs))
    static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
    CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues);
  #else
    #define __Pyx_KwargsAsDict_FASTCALL(kw, kwvalues) _PyStack_AsDict(kwvalues, kw)
  #endif
    #define __Pyx_Arg_NewRef_FASTCALL(arg) arg  /* no-op, __Pyx_Arg_FASTCALL is direct and this needs
                                                   to have the same reference counting */
    #define __Pyx_Arg_XDECREF_FASTCALL(arg)
#else
    #define __Pyx_Arg_FASTCALL __Pyx_Arg_VARARGS
    #define __Pyx_NumKwargs_FASTCALL __Pyx_NumKwargs_VARARGS
    #define __Pyx_KwValues_FASTCALL __Pyx_KwValues_VARARGS
    #define __Pyx_GetKwValue_FASTCALL __Pyx_GetKwValue_VARARGS
    #define __Pyx_KwargsAsDict_FASTCALL __Pyx_KwargsAsDict_VARARGS
    #define __Pyx_Arg_NewRef_FASTCALL(arg) __Pyx_Arg_NewRef_VARARGS(arg)
    #define __Pyx_Arg_XDECREF_FASTCALL(arg) __Pyx_Arg_XDECREF_VARARGS(arg)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_VARARGS(args, start), stop - start)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_FASTCALL(args, start), stop - start)
#else
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) PyTuple_GetSlice(args, start, stop)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) PyTuple_GetSlice(args, start, stop)
#endif

/* py_dict_values.proto */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Values(PyObject* d);

/* PyDictContains.proto */
static CYTHON_INLINE int __Pyx_PyDict_ContainsTF(PyObject* item, PyObject* dict, int eq) {
    int result = PyDict_Contains(dict, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,
    const char* function_name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#if !CYTHON_VECTORCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if !CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03080000
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API && !defined(PYPY_VERSION)
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()
  #define __Pyx_PyFrame_GetLocalsplus(frame)  ((frame)->f_localsplus)
#else
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif
#endif
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, (size_t)(nargs), NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, size_t nargs, PyObject *kwargs);

/* PyObjectCallNoArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* UnpackTupleError.proto */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto */
#define __Pyx_unpack_tuple2(tuple, value1, value2, is_tuple, has_known_size, decref_tuple)\
    (likely(is_tuple || PyTuple_Check(tuple)) ?\
        (likely(has_known_size || PyTuple_GET_SIZE(tuple) == 2) ?\
            __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple) :\
            (__Pyx_UnpackTupleError(tuple, 2), -1)) :\
        __Pyx_unpack_tuple2_generic(tuple, value1, value2, has_known_size, decref_tuple))
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
    PyObject* tuple, PyObject** value1, PyObject** value2, int decref_tuple);
static int __Pyx_unpack_tuple2_generic(
    PyObject* tuple, PyObject** value1, PyObject** value2, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* IncludeStructmemberH.proto */
#include <structmember.h>

/* FixUpExtensionType.proto */
#if CYTHON_USE_TYPE_SPECS
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type);
#endif

/* FetchSharedCythonModule.proto */
static PyObject *__Pyx_FetchSharedCythonABIModule(void);

/* FetchCommonType.proto */
#if !CYTHON_USE_TYPE_SPECS
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);
#else
static PyTypeObject* __Pyx_FetchCommonTypeFromSpec(PyObject *module, PyType_Spec *spec, PyObject *bases);
#endif

/* PyMethodNew.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    PyObject *typesModule=NULL, *methodType=NULL, *result=NULL;
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    typesModule = PyImport_ImportModule("types");
    if (!typesModule) return NULL;
    methodType = PyObject_GetAttrString(typesModule, "MethodType");
    Py_DECREF(typesModule);
    if (!methodType) return NULL;
    result = PyObject_CallFunctionObjArgs(methodType, func, self, NULL);
    Py_DECREF(methodType);
    return result;
}
#elif PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    return PyMethod_New(func, self);
}
#else
    #define __Pyx_PyMethod_New PyMethod_New
#endif

/* PyVectorcallFastCallDict.proto */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw);
#endif

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CYFUNCTION_COROUTINE     0x08
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_CyFunction_GetClassObj(f)\
      (((__pyx_CyFunctionObject *) (f))->func_classobj)
#else
  #define __Pyx_CyFunction_GetClassObj(f)\
      ((PyObject*) ((PyCMethodObject *) (f))->mm_class)
#endif
#define __Pyx_CyFunction_SetClassObj(f, classobj)\
    __Pyx__CyFunction_SetClassObj((__pyx_CyFunctionObject *) (f), (classobj))
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject_HEAD
    PyObject *func;
#elif PY_VERSION_HEX < 0x030900B1
    PyCFunctionObject func;
#else
    PyCMethodObject func;
#endif
#if CYTHON_BACKPORT_VECTORCALL
    __pyx_vectorcallfunc func_vectorcall;
#endif
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_classobj;
#endif
    void *defaults;
    int defaults_pyobjects;
    size_t defaults_size;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
    PyObject *func_is_coroutine;
} __pyx_CyFunctionObject;
#undef __Pyx_CyOrPyCFunction_Check
#define __Pyx_CyFunction_Check(obj)  __Pyx_TypeCheck(obj, __pyx_CyFunctionType)
#define __Pyx_CyOrPyCFunction_Check(obj)  __Pyx_TypeCheck2(obj, __pyx_CyFunctionType, &PyCFunction_Type)
#define __Pyx_CyFunction_CheckExact(obj)  __Pyx_IS_TYPE(obj, __pyx_CyFunctionType)
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc);
#undef __Pyx_IsSameCFunction
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCyOrCFunction(func, cfunc)
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(PyObject *module);
#if CYTHON_METH_FASTCALL
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#if CYTHON_BACKPORT_VECTORCALL
#define __Pyx_CyFunction_func_vectorcall(f) (((__pyx_CyFunctionObject*)f)->func_vectorcall)
#else
#define __Pyx_CyFunction_func_vectorcall(f) (((PyCFunctionObject*)f)->vectorcall)
#endif
#endif

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);
#endif

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if !defined(__INTEL_COMPILER) && defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* FromPyCTupleUtility.proto */
static CYTHON_INLINE __pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float __pyx_convert__from_py___pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float(PyObject *);

/* FromPyCTupleUtility.proto */
static CYTHON_INLINE __pyx_ctuple_float__and_float __pyx_convert__from_py___pyx_ctuple_float__and_float(PyObject *);

/* RaiseUnexpectedTypeError.proto */
static int __Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj);

/* FromPyCTupleUtility.proto */
static CYTHON_INLINE __pyx_ctuple_a8db9c__float__and_float__and_float__and_float__and_fl__etc __pyx_convert__from_py___pyx_ctuple_a8db9c__float__and_float__and_float__and_float__and_fl__etc(PyObject *);

/* ToPyCTupleUtility.proto */
static PyObject* __pyx_convert__to_py___pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float(__pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float);

/* ToPyCTupleUtility.proto */
static PyObject* __pyx_convert__to_py___pyx_ctuple_float__and_float(__pyx_ctuple_float__and_float);

struct __pyx_t_12pyretechnics_8cy_types_FuelModel;
static PyObject* __pyx_convert__to_py_struct____pyx_t_12pyretechnics_8cy_types_FuelModel(struct __pyx_t_12pyretechnics_8cy_types_FuelModel s);
/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* FormatTypeName.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%U"
static __Pyx_TypeName __Pyx_PyType_GetName(PyTypeObject* tp);
#define __Pyx_DECREF_TypeName(obj) Py_XDECREF(obj)
#else
typedef const char *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%.200s"
#define __Pyx_PyType_GetName(tp) ((tp)->tp_name)
#define __Pyx_DECREF_TypeName(obj)
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) __Pyx_IsAnySubtype2(Py_TYPE(obj), (PyTypeObject *)type1, (PyTypeObject *)type2)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) (PyObject_TypeCheck(obj, (PyTypeObject *)type1) || PyObject_TypeCheck(obj, (PyTypeObject *)type2))
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_CurrentExceptionType(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static unsigned long __Pyx_get_runtime_version(void);
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer);

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* VoidPtrExport.proto */
static int __Pyx_ExportVoidPtr(PyObject *name, void *p, const char *sig);

/* FunctionExport.proto */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

/* #### Code section: module_declarations ### */

/* Module declarations from "pyretechnics.cy_types" */

/* Module declarations from "cython" */

/* Module declarations from "libc.math" */

/* Module declarations from "pyretechnics.fuel_models" */
static PyObject *__pyx_v_12pyretechnics_11fuel_models_compact_fuel_model_table = 0;
static float __pyx_v_12pyretechnics_11fuel_models_almost_zero;
static PyObject *__pyx_v_12pyretechnics_11fuel_models_fuel_model_table = 0;
static PyObject *__pyx_7genexpr__pyx_v_12pyretechnics_11fuel_models_k;
static CYTHON_INLINE int __pyx_f_12pyretechnics_11fuel_models_is_burnable_fuel_model_number(int); /*proto*/
static float __pyx_f_12pyretechnics_11fuel_models_compute_exp_A_sigma(float, float); /*proto*/
static CYTHON_INLINE float __pyx_f_12pyretechnics_11fuel_models_compute_firemod_size_class(float); /*proto*/
static struct __pyx_t_12pyretechnics_8cy_types_FuelModel __pyx_f_12pyretechnics_11fuel_models_expand_compact_fuel_model(int); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_12pyretechnics_11fuel_models_list_fuel_model_numbers(int __pyx_skip_dispatch); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_12pyretechnics_11fuel_models_list_fuel_models(int __pyx_skip_dispatch); /*proto*/
static CYTHON_INLINE int __pyx_f_12pyretechnics_11fuel_models_fuel_model_exists(int, int __pyx_skip_dispatch); /*proto*/
static CYTHON_INLINE struct __pyx_t_12pyretechnics_8cy_types_FuelModel __pyx_f_12pyretechnics_11fuel_models_get_fuel_model(int, int __pyx_skip_dispatch); /*proto*/
static struct __pyx_t_12pyretechnics_8cy_types_FuelModel __pyx_f_12pyretechnics_11fuel_models_add_dynamic_fuel_loading(struct __pyx_t_12pyretechnics_8cy_types_FuelModel, __pyx_t_12pyretechnics_8cy_types_fclaarr); /*proto*/
static float __pyx_f_12pyretechnics_11fuel_models_compute_gij(__pyx_t_12pyretechnics_8cy_types_fclaarr, __pyx_t_12pyretechnics_8cy_types_fclaarr, float, int); /*proto*/
static struct __pyx_t_12pyretechnics_8cy_types_FuelModel __pyx_f_12pyretechnics_11fuel_models_add_weighting_factors(struct __pyx_t_12pyretechnics_8cy_types_FuelModel); /*proto*/
static struct __pyx_t_12pyretechnics_8cy_types_FuelModel __pyx_f_12pyretechnics_11fuel_models_add_live_moisture_of_extinction(struct __pyx_t_12pyretechnics_8cy_types_FuelModel); /*proto*/
static struct __pyx_t_12pyretechnics_8cy_types_FuelModel __pyx_f_12pyretechnics_11fuel_models_moisturize(struct __pyx_t_12pyretechnics_8cy_types_FuelModel, __pyx_t_12pyretechnics_8cy_types_fclaarr, int __pyx_skip_dispatch); /*proto*/
static struct __pyx_t_12pyretechnics_8cy_types_FuelModel __pyx_convert__from_py_struct____pyx_t_12pyretechnics_8cy_types_FuelModel(PyObject *); /*proto*/
/* #### Code section: typeinfo ### */
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "pyretechnics.fuel_models"
extern int __pyx_module_is_main_pyretechnics__fuel_models;
int __pyx_module_is_main_pyretechnics__fuel_models = 0;

/* Implementation of "pyretechnics.fuel_models" */
/* #### Code section: global_var ### */
static PyObject *__pyx_builtin_KeyError;
static PyObject *__pyx_builtin_ValueError;
/* #### Code section: string_decls ### */
static const char __pyx_k_h[] = "h";
static const char __pyx_k_M_f[] = "M_f";
static const char __pyx_k_M_x[] = "M_x";
static const char __pyx_k_S_T[] = "S_T";
static const char __pyx_k_S_e[] = "S_e";
static const char __pyx_k__79[] = "?";
static const char __pyx_k_f_i[] = "f_i";
static const char __pyx_k_w_o[] = "w_o";
static const char __pyx_k_f_ij[] = "f_ij";
static const char __pyx_k_g_ij[] = "g_ij";
static const char __pyx_k_keys[] = "keys";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_delta[] = "delta";
static const char __pyx_k_rho_p[] = "rho_p";
static const char __pyx_k_sigma[] = "sigma";
static const char __pyx_k_cy_int[] = "cy.int";
static const char __pyx_k_number[] = "number";
static const char __pyx_k_values[] = "values";
static const char __pyx_k_dynamic[] = "dynamic";
static const char __pyx_k_fclaarr[] = "fclaarr";
static const char __pyx_k_KeyError[] = "KeyError";
static const char __pyx_k_burnable[] = "burnable";
static const char __pyx_k_pyx_capi[] = "__pyx_capi__";
static const char __pyx_k_FuelModel[] = "FuelModel";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_fuel_model[] = "fuel_model";
static const char __pyx_k_moisturize[] = "moisturize";
static const char __pyx_k_almost_zero[] = "almost_zero";
static const char __pyx_k_exp_A_sigma[] = "exp_A_sigma";
static const char __pyx_k_is_coroutine[] = "_is_coroutine";
static const char __pyx_k_fuel_moisture[] = "fuel_moisture";
static const char __pyx_k_get_fuel_model[] = "get_fuel_model";
static const char __pyx_k_fuel_model_table[] = "fuel_model_table";
static const char __pyx_k_list_fuel_models[] = "list_fuel_models";
static const char __pyx_k_fuel_model_exists[] = "fuel_model_exists";
static const char __pyx_k_fuel_model_number[] = "fuel_model_number";
static const char __pyx_k_asyncio_coroutines[] = "asyncio.coroutines";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_firemod_size_classes[] = "firemod_size_classes";
static const char __pyx_k_list_fuel_model_numbers[] = "list_fuel_model_numbers";
static const char __pyx_k_compact_fuel_model_table[] = "compact_fuel_model_table";
static const char __pyx_k_pyretechnics_fuel_models[] = "pyretechnics.fuel_models";
static const char __pyx_k_src_pyretechnics_fuel_models_py[] = "src/pyretechnics/fuel_models.py";
static const char __pyx_k_No_value_specified_for_struct_at[] = "No value specified for struct attribute 'number'";
static const char __pyx_k_No_value_specified_for_struct_at_2[] = "No value specified for struct attribute 'delta'";
static const char __pyx_k_No_value_specified_for_struct_at_3[] = "No value specified for struct attribute 'M_x'";
static const char __pyx_k_No_value_specified_for_struct_at_4[] = "No value specified for struct attribute 'M_f'";
static const char __pyx_k_No_value_specified_for_struct_at_5[] = "No value specified for struct attribute 'w_o'";
static const char __pyx_k_No_value_specified_for_struct_at_6[] = "No value specified for struct attribute 'sigma'";
static const char __pyx_k_No_value_specified_for_struct_at_7[] = "No value specified for struct attribute 'h'";
static const char __pyx_k_No_value_specified_for_struct_at_8[] = "No value specified for struct attribute 'rho_p'";
static const char __pyx_k_No_value_specified_for_struct_at_9[] = "No value specified for struct attribute 'S_T'";
static const char __pyx_k_No_value_specified_for_struct_at_10[] = "No value specified for struct attribute 'S_e'";
static const char __pyx_k_No_value_specified_for_struct_at_11[] = "No value specified for struct attribute 'dynamic'";
static const char __pyx_k_No_value_specified_for_struct_at_12[] = "No value specified for struct attribute 'burnable'";
static const char __pyx_k_No_value_specified_for_struct_at_13[] = "No value specified for struct attribute 'exp_A_sigma'";
static const char __pyx_k_No_value_specified_for_struct_at_14[] = "No value specified for struct attribute 'firemod_size_classes'";
static const char __pyx_k_No_value_specified_for_struct_at_15[] = "No value specified for struct attribute 'f_ij'";
static const char __pyx_k_No_value_specified_for_struct_at_16[] = "No value specified for struct attribute 'f_i'";
static const char __pyx_k_No_value_specified_for_struct_at_17[] = "No value specified for struct attribute 'g_ij'";
/* #### Code section: decls ### */
static PyObject *__pyx_pf_12pyretechnics_11fuel_models_list_fuel_model_numbers(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_12pyretechnics_11fuel_models_2list_fuel_models(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_12pyretechnics_11fuel_models_4fuel_model_exists(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_fuel_model_number); /* proto */
static PyObject *__pyx_pf_12pyretechnics_11fuel_models_6get_fuel_model(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_fuel_model_number); /* proto */
static PyObject *__pyx_pf_12pyretechnics_11fuel_models_8moisturize(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_t_12pyretechnics_8cy_types_FuelModel __pyx_v_fuel_model, __pyx_t_12pyretechnics_8cy_types_fclaarr __pyx_v_fuel_moisture); /* proto */
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_keys = {0, 0, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_values = {0, 0, 0, 0, 0};
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
typedef struct {
  PyObject *__pyx_d;
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  #ifdef __Pyx_CyFunction_USED
  PyTypeObject *__pyx_CyFunctionType;
  #endif
  #ifdef __Pyx_FusedFunction_USED
  PyTypeObject *__pyx_FusedFunctionType;
  #endif
  #ifdef __Pyx_Generator_USED
  PyTypeObject *__pyx_GeneratorType;
  #endif
  #ifdef __Pyx_IterableCoroutine_USED
  PyTypeObject *__pyx_IterableCoroutineType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineAwaitType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineType;
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyObject *__pyx_n_s_FuelModel;
  PyObject *__pyx_n_s_KeyError;
  PyObject *__pyx_n_s_M_f;
  PyObject *__pyx_n_s_M_x;
  PyObject *__pyx_kp_s_No_value_specified_for_struct_at;
  PyObject *__pyx_kp_s_No_value_specified_for_struct_at_10;
  PyObject *__pyx_kp_s_No_value_specified_for_struct_at_11;
  PyObject *__pyx_kp_s_No_value_specified_for_struct_at_12;
  PyObject *__pyx_kp_s_No_value_specified_for_struct_at_13;
  PyObject *__pyx_kp_s_No_value_specified_for_struct_at_14;
  PyObject *__pyx_kp_s_No_value_specified_for_struct_at_15;
  PyObject *__pyx_kp_s_No_value_specified_for_struct_at_16;
  PyObject *__pyx_kp_s_No_value_specified_for_struct_at_17;
  PyObject *__pyx_kp_s_No_value_specified_for_struct_at_2;
  PyObject *__pyx_kp_s_No_value_specified_for_struct_at_3;
  PyObject *__pyx_kp_s_No_value_specified_for_struct_at_4;
  PyObject *__pyx_kp_s_No_value_specified_for_struct_at_5;
  PyObject *__pyx_kp_s_No_value_specified_for_struct_at_6;
  PyObject *__pyx_kp_s_No_value_specified_for_struct_at_7;
  PyObject *__pyx_kp_s_No_value_specified_for_struct_at_8;
  PyObject *__pyx_kp_s_No_value_specified_for_struct_at_9;
  PyObject *__pyx_n_s_S_T;
  PyObject *__pyx_n_s_S_e;
  PyObject *__pyx_n_s_ValueError;
  PyObject *__pyx_n_s__79;
  PyObject *__pyx_n_s_almost_zero;
  PyObject *__pyx_n_s_asyncio_coroutines;
  PyObject *__pyx_n_s_burnable;
  PyObject *__pyx_n_s_cline_in_traceback;
  PyObject *__pyx_n_s_compact_fuel_model_table;
  PyObject *__pyx_kp_s_cy_int;
  PyObject *__pyx_n_s_delta;
  PyObject *__pyx_n_s_dynamic;
  PyObject *__pyx_n_s_exp_A_sigma;
  PyObject *__pyx_n_s_f_i;
  PyObject *__pyx_n_s_f_ij;
  PyObject *__pyx_n_s_fclaarr;
  PyObject *__pyx_n_s_firemod_size_classes;
  PyObject *__pyx_n_s_fuel_model;
  PyObject *__pyx_n_s_fuel_model_exists;
  PyObject *__pyx_n_s_fuel_model_number;
  PyObject *__pyx_n_s_fuel_model_table;
  PyObject *__pyx_n_s_fuel_moisture;
  PyObject *__pyx_n_s_g_ij;
  PyObject *__pyx_n_s_get_fuel_model;
  PyObject *__pyx_n_s_h;
  PyObject *__pyx_n_s_is_coroutine;
  PyObject *__pyx_n_s_keys;
  PyObject *__pyx_n_s_list_fuel_model_numbers;
  PyObject *__pyx_n_s_list_fuel_models;
  PyObject *__pyx_n_s_main;
  PyObject *__pyx_n_s_moisturize;
  PyObject *__pyx_n_s_name;
  PyObject *__pyx_n_s_number;
  PyObject *__pyx_n_s_pyretechnics_fuel_models;
  PyObject *__pyx_n_s_pyx_capi;
  PyObject *__pyx_n_s_rho_p;
  PyObject *__pyx_n_s_sigma;
  PyObject *__pyx_kp_s_src_pyretechnics_fuel_models_py;
  PyObject *__pyx_n_s_test;
  PyObject *__pyx_n_s_values;
  PyObject *__pyx_n_s_w_o;
  PyObject *__pyx_float_0_0;
  PyObject *__pyx_float_0_2;
  PyObject *__pyx_float_0_3;
  PyObject *__pyx_float_0_4;
  PyObject *__pyx_float_0_5;
  PyObject *__pyx_float_0_6;
  PyObject *__pyx_float_0_9;
  PyObject *__pyx_float_1_0;
  PyObject *__pyx_float_1_2;
  PyObject *__pyx_float_1_3;
  PyObject *__pyx_float_1_5;
  PyObject *__pyx_float_1_8;
  PyObject *__pyx_float_2_0;
  PyObject *__pyx_float_2_1;
  PyObject *__pyx_float_2_3;
  PyObject *__pyx_float_2_4;
  PyObject *__pyx_float_2_5;
  PyObject *__pyx_float_2_7;
  PyObject *__pyx_float_3_0;
  PyObject *__pyx_float_4_0;
  PyObject *__pyx_float_4_4;
  PyObject *__pyx_float_5_0;
  PyObject *__pyx_float_6_0;
  PyObject *__pyx_float_8_0;
  PyObject *__pyx_float_9_0;
  PyObject *__pyx_float_12_0;
  PyObject *__pyx_float_15_0;
  PyObject *__pyx_float_20_0;
  PyObject *__pyx_float_25_0;
  PyObject *__pyx_float_30_0;
  PyObject *__pyx_float_35_0;
  PyObject *__pyx_float_40_0;
  PyObject *__pyx_float_109_0;
  PyObject *__pyx_float_750_0;
  PyObject *__pyx_float_0_0000;
  PyObject *__pyx_float_0_0046;
  PyObject *__pyx_float_0_0069;
  PyObject *__pyx_float_0_0070;
  PyObject *__pyx_float_0_0092;
  PyObject *__pyx_float_0_0115;
  PyObject *__pyx_float_0_0138;
  PyObject *__pyx_float_0_0170;
  PyObject *__pyx_float_0_0184;
  PyObject *__pyx_float_0_0190;
  PyObject *__pyx_float_0_0207;
  PyObject *__pyx_float_0_0230;
  PyObject *__pyx_float_0_0275;
  PyObject *__pyx_float_0_0298;
  PyObject *__pyx_float_0_0340;
  PyObject *__pyx_float_0_0344;
  PyObject *__pyx_float_0_0390;
  PyObject *__pyx_float_0_0413;
  PyObject *__pyx_float_0_0436;
  PyObject *__pyx_float_0_0459;
  PyObject *__pyx_float_0_0460;
  PyObject *__pyx_float_0_0505;
  PyObject *__pyx_float_0_0520;
  PyObject *__pyx_float_0_0528;
  PyObject *__pyx_float_0_0551;
  PyObject *__pyx_float_0_0574;
  PyObject *__pyx_float_0_0597;
  PyObject *__pyx_float_0_0620;
  PyObject *__pyx_float_0_0643;
  PyObject *__pyx_float_0_0666;
  PyObject *__pyx_float_0_0689;
  PyObject *__pyx_float_0_0690;
  PyObject *__pyx_float_0_0712;
  PyObject *__pyx_float_0_0826;
  PyObject *__pyx_float_0_0860;
  PyObject *__pyx_float_0_0872;
  PyObject *__pyx_float_0_0918;
  PyObject *__pyx_float_0_0920;
  PyObject *__pyx_float_0_0941;
  PyObject *__pyx_float_0_0964;
  PyObject *__pyx_float_0_1010;
  PyObject *__pyx_float_0_1056;
  PyObject *__pyx_float_0_1102;
  PyObject *__pyx_float_0_1125;
  PyObject *__pyx_float_0_1148;
  PyObject *__pyx_float_0_1150;
  PyObject *__pyx_float_0_1171;
  PyObject *__pyx_float_0_1263;
  PyObject *__pyx_float_0_1286;
  PyObject *__pyx_float_0_1331;
  PyObject *__pyx_float_0_1340;
  PyObject *__pyx_float_0_1377;
  PyObject *__pyx_float_0_1380;
  PyObject *__pyx_float_0_1515;
  PyObject *__pyx_float_0_1561;
  PyObject *__pyx_float_0_1607;
  PyObject *__pyx_float_0_1653;
  PyObject *__pyx_float_0_1768;
  PyObject *__pyx_float_0_1837;
  PyObject *__pyx_float_0_1840;
  PyObject *__pyx_float_0_1905;
  PyObject *__pyx_float_0_1928;
  PyObject *__pyx_float_0_1951;
  PyObject *__pyx_float_0_1997;
  PyObject *__pyx_float_0_2020;
  PyObject *__pyx_float_0_2066;
  PyObject *__pyx_float_0_2070;
  PyObject *__pyx_float_0_2300;
  PyObject *__pyx_float_0_2410;
  PyObject *__pyx_float_0_2433;
  PyObject *__pyx_float_0_2479;
  PyObject *__pyx_float_0_2525;
  PyObject *__pyx_float_0_2530;
  PyObject *__pyx_float_0_2663;
  PyObject *__pyx_float_0_2847;
  PyObject *__pyx_float_0_3053;
  PyObject *__pyx_float_0_3214;
  PyObject *__pyx_float_0_3220;
  PyObject *__pyx_float_0_3260;
  PyObject *__pyx_float_0_3352;
  PyObject *__pyx_float_0_3719;
  PyObject *__pyx_float_0_4132;
  PyObject *__pyx_float_0_5051;
  PyObject *__pyx_float_0_6440;
  PyObject *__pyx_float_0_7590;
  PyObject *__pyx_float_1_0580;
  PyObject *__pyx_float_1_2880;
  PyObject *__pyx_float_1300_0;
  PyObject *__pyx_float_1400_0;
  PyObject *__pyx_float_1500_0;
  PyObject *__pyx_float_1550_0;
  PyObject *__pyx_float_1600_0;
  PyObject *__pyx_float_1750_0;
  PyObject *__pyx_float_1800_0;
  PyObject *__pyx_float_2000_0;
  PyObject *__pyx_float_2200_0;
  PyObject *__pyx_float_2300_0;
  PyObject *__pyx_float_2500_0;
  PyObject *__pyx_float_3000_0;
  PyObject *__pyx_float_3500_0;
  PyObject *__pyx_int_1;
  PyObject *__pyx_int_2;
  PyObject *__pyx_int_3;
  PyObject *__pyx_int_4;
  PyObject *__pyx_int_5;
  PyObject *__pyx_int_6;
  PyObject *__pyx_int_7;
  PyObject *__pyx_int_8;
  PyObject *__pyx_int_9;
  PyObject *__pyx_int_10;
  PyObject *__pyx_int_11;
  PyObject *__pyx_int_12;
  PyObject *__pyx_int_13;
  PyObject *__pyx_int_91;
  PyObject *__pyx_int_92;
  PyObject *__pyx_int_93;
  PyObject *__pyx_int_98;
  PyObject *__pyx_int_99;
  PyObject *__pyx_int_101;
  PyObject *__pyx_int_102;
  PyObject *__pyx_int_103;
  PyObject *__pyx_int_104;
  PyObject *__pyx_int_105;
  PyObject *__pyx_int_106;
  PyObject *__pyx_int_107;
  PyObject *__pyx_int_108;
  PyObject *__pyx_int_109;
  PyObject *__pyx_int_121;
  PyObject *__pyx_int_122;
  PyObject *__pyx_int_123;
  PyObject *__pyx_int_124;
  PyObject *__pyx_int_141;
  PyObject *__pyx_int_142;
  PyObject *__pyx_int_143;
  PyObject *__pyx_int_144;
  PyObject *__pyx_int_145;
  PyObject *__pyx_int_146;
  PyObject *__pyx_int_147;
  PyObject *__pyx_int_148;
  PyObject *__pyx_int_149;
  PyObject *__pyx_int_161;
  PyObject *__pyx_int_162;
  PyObject *__pyx_int_163;
  PyObject *__pyx_int_164;
  PyObject *__pyx_int_165;
  PyObject *__pyx_int_181;
  PyObject *__pyx_int_182;
  PyObject *__pyx_int_183;
  PyObject *__pyx_int_184;
  PyObject *__pyx_int_185;
  PyObject *__pyx_int_186;
  PyObject *__pyx_int_187;
  PyObject *__pyx_int_188;
  PyObject *__pyx_int_189;
  PyObject *__pyx_int_201;
  PyObject *__pyx_int_202;
  PyObject *__pyx_int_203;
  PyObject *__pyx_int_204;
  PyObject *__pyx_tuple_;
  PyObject *__pyx_tuple__2;
  PyObject *__pyx_tuple__3;
  PyObject *__pyx_tuple__4;
  PyObject *__pyx_tuple__5;
  PyObject *__pyx_tuple__6;
  PyObject *__pyx_tuple__7;
  PyObject *__pyx_tuple__8;
  PyObject *__pyx_tuple__9;
  PyObject *__pyx_tuple__10;
  PyObject *__pyx_tuple__11;
  PyObject *__pyx_tuple__12;
  PyObject *__pyx_tuple__13;
  PyObject *__pyx_tuple__14;
  PyObject *__pyx_tuple__15;
  PyObject *__pyx_tuple__16;
  PyObject *__pyx_tuple__17;
  PyObject *__pyx_tuple__18;
  PyObject *__pyx_tuple__19;
  PyObject *__pyx_tuple__20;
  PyObject *__pyx_tuple__21;
  PyObject *__pyx_tuple__22;
  PyObject *__pyx_tuple__23;
  PyObject *__pyx_tuple__24;
  PyObject *__pyx_tuple__25;
  PyObject *__pyx_tuple__26;
  PyObject *__pyx_tuple__27;
  PyObject *__pyx_tuple__28;
  PyObject *__pyx_tuple__29;
  PyObject *__pyx_tuple__30;
  PyObject *__pyx_tuple__31;
  PyObject *__pyx_tuple__32;
  PyObject *__pyx_tuple__33;
  PyObject *__pyx_tuple__34;
  PyObject *__pyx_tuple__35;
  PyObject *__pyx_tuple__36;
  PyObject *__pyx_tuple__37;
  PyObject *__pyx_tuple__38;
  PyObject *__pyx_tuple__39;
  PyObject *__pyx_tuple__40;
  PyObject *__pyx_tuple__41;
  PyObject *__pyx_tuple__42;
  PyObject *__pyx_tuple__43;
  PyObject *__pyx_tuple__44;
  PyObject *__pyx_tuple__45;
  PyObject *__pyx_tuple__46;
  PyObject *__pyx_tuple__47;
  PyObject *__pyx_tuple__48;
  PyObject *__pyx_tuple__49;
  PyObject *__pyx_tuple__50;
  PyObject *__pyx_tuple__51;
  PyObject *__pyx_tuple__52;
  PyObject *__pyx_tuple__53;
  PyObject *__pyx_tuple__54;
  PyObject *__pyx_tuple__55;
  PyObject *__pyx_tuple__56;
  PyObject *__pyx_tuple__57;
  PyObject *__pyx_tuple__58;
  PyObject *__pyx_tuple__59;
  PyObject *__pyx_tuple__60;
  PyObject *__pyx_tuple__61;
  PyObject *__pyx_tuple__62;
  PyObject *__pyx_tuple__63;
  PyObject *__pyx_tuple__64;
  PyObject *__pyx_tuple__65;
  PyObject *__pyx_tuple__66;
  PyObject *__pyx_tuple__67;
  PyObject *__pyx_tuple__68;
  PyObject *__pyx_tuple__69;
  PyObject *__pyx_tuple__70;
  PyObject *__pyx_tuple__71;
  PyObject *__pyx_tuple__74;
  PyObject *__pyx_tuple__77;
  PyObject *__pyx_codeobj__72;
  PyObject *__pyx_codeobj__73;
  PyObject *__pyx_codeobj__75;
  PyObject *__pyx_codeobj__76;
  PyObject *__pyx_codeobj__78;
} __pyx_mstate;

#if CYTHON_USE_MODULE_STATE
#ifdef __cplusplus
namespace {
  extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate(o) ((__pyx_mstate *)__Pyx_PyModule_GetState(o))

#define __pyx_mstate_global (__pyx_mstate(PyState_FindModule(&__pyx_moduledef)))

#define __pyx_m (PyState_FindModule(&__pyx_moduledef))
#else
static __pyx_mstate __pyx_mstate_global_static =
#ifdef __cplusplus
    {};
#else
    {0};
#endif
static __pyx_mstate *__pyx_mstate_global = &__pyx_mstate_global_static;
#endif
/* #### Code section: module_state_clear ### */
#if CYTHON_USE_MODULE_STATE
static int __pyx_m_clear(PyObject *m) {
  __pyx_mstate *clear_module_state = __pyx_mstate(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_d);
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_CLEAR(clear_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_CLEAR(clear_module_state->__pyx_FusedFunctionType);
  #endif
  Py_CLEAR(clear_module_state->__pyx_n_s_FuelModel);
  Py_CLEAR(clear_module_state->__pyx_n_s_KeyError);
  Py_CLEAR(clear_module_state->__pyx_n_s_M_f);
  Py_CLEAR(clear_module_state->__pyx_n_s_M_x);
  Py_CLEAR(clear_module_state->__pyx_kp_s_No_value_specified_for_struct_at);
  Py_CLEAR(clear_module_state->__pyx_kp_s_No_value_specified_for_struct_at_10);
  Py_CLEAR(clear_module_state->__pyx_kp_s_No_value_specified_for_struct_at_11);
  Py_CLEAR(clear_module_state->__pyx_kp_s_No_value_specified_for_struct_at_12);
  Py_CLEAR(clear_module_state->__pyx_kp_s_No_value_specified_for_struct_at_13);
  Py_CLEAR(clear_module_state->__pyx_kp_s_No_value_specified_for_struct_at_14);
  Py_CLEAR(clear_module_state->__pyx_kp_s_No_value_specified_for_struct_at_15);
  Py_CLEAR(clear_module_state->__pyx_kp_s_No_value_specified_for_struct_at_16);
  Py_CLEAR(clear_module_state->__pyx_kp_s_No_value_specified_for_struct_at_17);
  Py_CLEAR(clear_module_state->__pyx_kp_s_No_value_specified_for_struct_at_2);
  Py_CLEAR(clear_module_state->__pyx_kp_s_No_value_specified_for_struct_at_3);
  Py_CLEAR(clear_module_state->__pyx_kp_s_No_value_specified_for_struct_at_4);
  Py_CLEAR(clear_module_state->__pyx_kp_s_No_value_specified_for_struct_at_5);
  Py_CLEAR(clear_module_state->__pyx_kp_s_No_value_specified_for_struct_at_6);
  Py_CLEAR(clear_module_state->__pyx_kp_s_No_value_specified_for_struct_at_7);
  Py_CLEAR(clear_module_state->__pyx_kp_s_No_value_specified_for_struct_at_8);
  Py_CLEAR(clear_module_state->__pyx_kp_s_No_value_specified_for_struct_at_9);
  Py_CLEAR(clear_module_state->__pyx_n_s_S_T);
  Py_CLEAR(clear_module_state->__pyx_n_s_S_e);
  Py_CLEAR(clear_module_state->__pyx_n_s_ValueError);
  Py_CLEAR(clear_module_state->__pyx_n_s__79);
  Py_CLEAR(clear_module_state->__pyx_n_s_almost_zero);
  Py_CLEAR(clear_module_state->__pyx_n_s_asyncio_coroutines);
  Py_CLEAR(clear_module_state->__pyx_n_s_burnable);
  Py_CLEAR(clear_module_state->__pyx_n_s_cline_in_traceback);
  Py_CLEAR(clear_module_state->__pyx_n_s_compact_fuel_model_table);
  Py_CLEAR(clear_module_state->__pyx_kp_s_cy_int);
  Py_CLEAR(clear_module_state->__pyx_n_s_delta);
  Py_CLEAR(clear_module_state->__pyx_n_s_dynamic);
  Py_CLEAR(clear_module_state->__pyx_n_s_exp_A_sigma);
  Py_CLEAR(clear_module_state->__pyx_n_s_f_i);
  Py_CLEAR(clear_module_state->__pyx_n_s_f_ij);
  Py_CLEAR(clear_module_state->__pyx_n_s_fclaarr);
  Py_CLEAR(clear_module_state->__pyx_n_s_firemod_size_classes);
  Py_CLEAR(clear_module_state->__pyx_n_s_fuel_model);
  Py_CLEAR(clear_module_state->__pyx_n_s_fuel_model_exists);
  Py_CLEAR(clear_module_state->__pyx_n_s_fuel_model_number);
  Py_CLEAR(clear_module_state->__pyx_n_s_fuel_model_table);
  Py_CLEAR(clear_module_state->__pyx_n_s_fuel_moisture);
  Py_CLEAR(clear_module_state->__pyx_n_s_g_ij);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_fuel_model);
  Py_CLEAR(clear_module_state->__pyx_n_s_h);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_coroutine);
  Py_CLEAR(clear_module_state->__pyx_n_s_keys);
  Py_CLEAR(clear_module_state->__pyx_n_s_list_fuel_model_numbers);
  Py_CLEAR(clear_module_state->__pyx_n_s_list_fuel_models);
  Py_CLEAR(clear_module_state->__pyx_n_s_main);
  Py_CLEAR(clear_module_state->__pyx_n_s_moisturize);
  Py_CLEAR(clear_module_state->__pyx_n_s_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_number);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyretechnics_fuel_models);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_capi);
  Py_CLEAR(clear_module_state->__pyx_n_s_rho_p);
  Py_CLEAR(clear_module_state->__pyx_n_s_sigma);
  Py_CLEAR(clear_module_state->__pyx_kp_s_src_pyretechnics_fuel_models_py);
  Py_CLEAR(clear_module_state->__pyx_n_s_test);
  Py_CLEAR(clear_module_state->__pyx_n_s_values);
  Py_CLEAR(clear_module_state->__pyx_n_s_w_o);
  Py_CLEAR(clear_module_state->__pyx_float_0_0);
  Py_CLEAR(clear_module_state->__pyx_float_0_2);
  Py_CLEAR(clear_module_state->__pyx_float_0_3);
  Py_CLEAR(clear_module_state->__pyx_float_0_4);
  Py_CLEAR(clear_module_state->__pyx_float_0_5);
  Py_CLEAR(clear_module_state->__pyx_float_0_6);
  Py_CLEAR(clear_module_state->__pyx_float_0_9);
  Py_CLEAR(clear_module_state->__pyx_float_1_0);
  Py_CLEAR(clear_module_state->__pyx_float_1_2);
  Py_CLEAR(clear_module_state->__pyx_float_1_3);
  Py_CLEAR(clear_module_state->__pyx_float_1_5);
  Py_CLEAR(clear_module_state->__pyx_float_1_8);
  Py_CLEAR(clear_module_state->__pyx_float_2_0);
  Py_CLEAR(clear_module_state->__pyx_float_2_1);
  Py_CLEAR(clear_module_state->__pyx_float_2_3);
  Py_CLEAR(clear_module_state->__pyx_float_2_4);
  Py_CLEAR(clear_module_state->__pyx_float_2_5);
  Py_CLEAR(clear_module_state->__pyx_float_2_7);
  Py_CLEAR(clear_module_state->__pyx_float_3_0);
  Py_CLEAR(clear_module_state->__pyx_float_4_0);
  Py_CLEAR(clear_module_state->__pyx_float_4_4);
  Py_CLEAR(clear_module_state->__pyx_float_5_0);
  Py_CLEAR(clear_module_state->__pyx_float_6_0);
  Py_CLEAR(clear_module_state->__pyx_float_8_0);
  Py_CLEAR(clear_module_state->__pyx_float_9_0);
  Py_CLEAR(clear_module_state->__pyx_float_12_0);
  Py_CLEAR(clear_module_state->__pyx_float_15_0);
  Py_CLEAR(clear_module_state->__pyx_float_20_0);
  Py_CLEAR(clear_module_state->__pyx_float_25_0);
  Py_CLEAR(clear_module_state->__pyx_float_30_0);
  Py_CLEAR(clear_module_state->__pyx_float_35_0);
  Py_CLEAR(clear_module_state->__pyx_float_40_0);
  Py_CLEAR(clear_module_state->__pyx_float_109_0);
  Py_CLEAR(clear_module_state->__pyx_float_750_0);
  Py_CLEAR(clear_module_state->__pyx_float_0_0000);
  Py_CLEAR(clear_module_state->__pyx_float_0_0046);
  Py_CLEAR(clear_module_state->__pyx_float_0_0069);
  Py_CLEAR(clear_module_state->__pyx_float_0_0070);
  Py_CLEAR(clear_module_state->__pyx_float_0_0092);
  Py_CLEAR(clear_module_state->__pyx_float_0_0115);
  Py_CLEAR(clear_module_state->__pyx_float_0_0138);
  Py_CLEAR(clear_module_state->__pyx_float_0_0170);
  Py_CLEAR(clear_module_state->__pyx_float_0_0184);
  Py_CLEAR(clear_module_state->__pyx_float_0_0190);
  Py_CLEAR(clear_module_state->__pyx_float_0_0207);
  Py_CLEAR(clear_module_state->__pyx_float_0_0230);
  Py_CLEAR(clear_module_state->__pyx_float_0_0275);
  Py_CLEAR(clear_module_state->__pyx_float_0_0298);
  Py_CLEAR(clear_module_state->__pyx_float_0_0340);
  Py_CLEAR(clear_module_state->__pyx_float_0_0344);
  Py_CLEAR(clear_module_state->__pyx_float_0_0390);
  Py_CLEAR(clear_module_state->__pyx_float_0_0413);
  Py_CLEAR(clear_module_state->__pyx_float_0_0436);
  Py_CLEAR(clear_module_state->__pyx_float_0_0459);
  Py_CLEAR(clear_module_state->__pyx_float_0_0460);
  Py_CLEAR(clear_module_state->__pyx_float_0_0505);
  Py_CLEAR(clear_module_state->__pyx_float_0_0520);
  Py_CLEAR(clear_module_state->__pyx_float_0_0528);
  Py_CLEAR(clear_module_state->__pyx_float_0_0551);
  Py_CLEAR(clear_module_state->__pyx_float_0_0574);
  Py_CLEAR(clear_module_state->__pyx_float_0_0597);
  Py_CLEAR(clear_module_state->__pyx_float_0_0620);
  Py_CLEAR(clear_module_state->__pyx_float_0_0643);
  Py_CLEAR(clear_module_state->__pyx_float_0_0666);
  Py_CLEAR(clear_module_state->__pyx_float_0_0689);
  Py_CLEAR(clear_module_state->__pyx_float_0_0690);
  Py_CLEAR(clear_module_state->__pyx_float_0_0712);
  Py_CLEAR(clear_module_state->__pyx_float_0_0826);
  Py_CLEAR(clear_module_state->__pyx_float_0_0860);
  Py_CLEAR(clear_module_state->__pyx_float_0_0872);
  Py_CLEAR(clear_module_state->__pyx_float_0_0918);
  Py_CLEAR(clear_module_state->__pyx_float_0_0920);
  Py_CLEAR(clear_module_state->__pyx_float_0_0941);
  Py_CLEAR(clear_module_state->__pyx_float_0_0964);
  Py_CLEAR(clear_module_state->__pyx_float_0_1010);
  Py_CLEAR(clear_module_state->__pyx_float_0_1056);
  Py_CLEAR(clear_module_state->__pyx_float_0_1102);
  Py_CLEAR(clear_module_state->__pyx_float_0_1125);
  Py_CLEAR(clear_module_state->__pyx_float_0_1148);
  Py_CLEAR(clear_module_state->__pyx_float_0_1150);
  Py_CLEAR(clear_module_state->__pyx_float_0_1171);
  Py_CLEAR(clear_module_state->__pyx_float_0_1263);
  Py_CLEAR(clear_module_state->__pyx_float_0_1286);
  Py_CLEAR(clear_module_state->__pyx_float_0_1331);
  Py_CLEAR(clear_module_state->__pyx_float_0_1340);
  Py_CLEAR(clear_module_state->__pyx_float_0_1377);
  Py_CLEAR(clear_module_state->__pyx_float_0_1380);
  Py_CLEAR(clear_module_state->__pyx_float_0_1515);
  Py_CLEAR(clear_module_state->__pyx_float_0_1561);
  Py_CLEAR(clear_module_state->__pyx_float_0_1607);
  Py_CLEAR(clear_module_state->__pyx_float_0_1653);
  Py_CLEAR(clear_module_state->__pyx_float_0_1768);
  Py_CLEAR(clear_module_state->__pyx_float_0_1837);
  Py_CLEAR(clear_module_state->__pyx_float_0_1840);
  Py_CLEAR(clear_module_state->__pyx_float_0_1905);
  Py_CLEAR(clear_module_state->__pyx_float_0_1928);
  Py_CLEAR(clear_module_state->__pyx_float_0_1951);
  Py_CLEAR(clear_module_state->__pyx_float_0_1997);
  Py_CLEAR(clear_module_state->__pyx_float_0_2020);
  Py_CLEAR(clear_module_state->__pyx_float_0_2066);
  Py_CLEAR(clear_module_state->__pyx_float_0_2070);
  Py_CLEAR(clear_module_state->__pyx_float_0_2300);
  Py_CLEAR(clear_module_state->__pyx_float_0_2410);
  Py_CLEAR(clear_module_state->__pyx_float_0_2433);
  Py_CLEAR(clear_module_state->__pyx_float_0_2479);
  Py_CLEAR(clear_module_state->__pyx_float_0_2525);
  Py_CLEAR(clear_module_state->__pyx_float_0_2530);
  Py_CLEAR(clear_module_state->__pyx_float_0_2663);
  Py_CLEAR(clear_module_state->__pyx_float_0_2847);
  Py_CLEAR(clear_module_state->__pyx_float_0_3053);
  Py_CLEAR(clear_module_state->__pyx_float_0_3214);
  Py_CLEAR(clear_module_state->__pyx_float_0_3220);
  Py_CLEAR(clear_module_state->__pyx_float_0_3260);
  Py_CLEAR(clear_module_state->__pyx_float_0_3352);
  Py_CLEAR(clear_module_state->__pyx_float_0_3719);
  Py_CLEAR(clear_module_state->__pyx_float_0_4132);
  Py_CLEAR(clear_module_state->__pyx_float_0_5051);
  Py_CLEAR(clear_module_state->__pyx_float_0_6440);
  Py_CLEAR(clear_module_state->__pyx_float_0_7590);
  Py_CLEAR(clear_module_state->__pyx_float_1_0580);
  Py_CLEAR(clear_module_state->__pyx_float_1_2880);
  Py_CLEAR(clear_module_state->__pyx_float_1300_0);
  Py_CLEAR(clear_module_state->__pyx_float_1400_0);
  Py_CLEAR(clear_module_state->__pyx_float_1500_0);
  Py_CLEAR(clear_module_state->__pyx_float_1550_0);
  Py_CLEAR(clear_module_state->__pyx_float_1600_0);
  Py_CLEAR(clear_module_state->__pyx_float_1750_0);
  Py_CLEAR(clear_module_state->__pyx_float_1800_0);
  Py_CLEAR(clear_module_state->__pyx_float_2000_0);
  Py_CLEAR(clear_module_state->__pyx_float_2200_0);
  Py_CLEAR(clear_module_state->__pyx_float_2300_0);
  Py_CLEAR(clear_module_state->__pyx_float_2500_0);
  Py_CLEAR(clear_module_state->__pyx_float_3000_0);
  Py_CLEAR(clear_module_state->__pyx_float_3500_0);
  Py_CLEAR(clear_module_state->__pyx_int_1);
  Py_CLEAR(clear_module_state->__pyx_int_2);
  Py_CLEAR(clear_module_state->__pyx_int_3);
  Py_CLEAR(clear_module_state->__pyx_int_4);
  Py_CLEAR(clear_module_state->__pyx_int_5);
  Py_CLEAR(clear_module_state->__pyx_int_6);
  Py_CLEAR(clear_module_state->__pyx_int_7);
  Py_CLEAR(clear_module_state->__pyx_int_8);
  Py_CLEAR(clear_module_state->__pyx_int_9);
  Py_CLEAR(clear_module_state->__pyx_int_10);
  Py_CLEAR(clear_module_state->__pyx_int_11);
  Py_CLEAR(clear_module_state->__pyx_int_12);
  Py_CLEAR(clear_module_state->__pyx_int_13);
  Py_CLEAR(clear_module_state->__pyx_int_91);
  Py_CLEAR(clear_module_state->__pyx_int_92);
  Py_CLEAR(clear_module_state->__pyx_int_93);
  Py_CLEAR(clear_module_state->__pyx_int_98);
  Py_CLEAR(clear_module_state->__pyx_int_99);
  Py_CLEAR(clear_module_state->__pyx_int_101);
  Py_CLEAR(clear_module_state->__pyx_int_102);
  Py_CLEAR(clear_module_state->__pyx_int_103);
  Py_CLEAR(clear_module_state->__pyx_int_104);
  Py_CLEAR(clear_module_state->__pyx_int_105);
  Py_CLEAR(clear_module_state->__pyx_int_106);
  Py_CLEAR(clear_module_state->__pyx_int_107);
  Py_CLEAR(clear_module_state->__pyx_int_108);
  Py_CLEAR(clear_module_state->__pyx_int_109);
  Py_CLEAR(clear_module_state->__pyx_int_121);
  Py_CLEAR(clear_module_state->__pyx_int_122);
  Py_CLEAR(clear_module_state->__pyx_int_123);
  Py_CLEAR(clear_module_state->__pyx_int_124);
  Py_CLEAR(clear_module_state->__pyx_int_141);
  Py_CLEAR(clear_module_state->__pyx_int_142);
  Py_CLEAR(clear_module_state->__pyx_int_143);
  Py_CLEAR(clear_module_state->__pyx_int_144);
  Py_CLEAR(clear_module_state->__pyx_int_145);
  Py_CLEAR(clear_module_state->__pyx_int_146);
  Py_CLEAR(clear_module_state->__pyx_int_147);
  Py_CLEAR(clear_module_state->__pyx_int_148);
  Py_CLEAR(clear_module_state->__pyx_int_149);
  Py_CLEAR(clear_module_state->__pyx_int_161);
  Py_CLEAR(clear_module_state->__pyx_int_162);
  Py_CLEAR(clear_module_state->__pyx_int_163);
  Py_CLEAR(clear_module_state->__pyx_int_164);
  Py_CLEAR(clear_module_state->__pyx_int_165);
  Py_CLEAR(clear_module_state->__pyx_int_181);
  Py_CLEAR(clear_module_state->__pyx_int_182);
  Py_CLEAR(clear_module_state->__pyx_int_183);
  Py_CLEAR(clear_module_state->__pyx_int_184);
  Py_CLEAR(clear_module_state->__pyx_int_185);
  Py_CLEAR(clear_module_state->__pyx_int_186);
  Py_CLEAR(clear_module_state->__pyx_int_187);
  Py_CLEAR(clear_module_state->__pyx_int_188);
  Py_CLEAR(clear_module_state->__pyx_int_189);
  Py_CLEAR(clear_module_state->__pyx_int_201);
  Py_CLEAR(clear_module_state->__pyx_int_202);
  Py_CLEAR(clear_module_state->__pyx_int_203);
  Py_CLEAR(clear_module_state->__pyx_int_204);
  Py_CLEAR(clear_module_state->__pyx_tuple_);
  Py_CLEAR(clear_module_state->__pyx_tuple__2);
  Py_CLEAR(clear_module_state->__pyx_tuple__3);
  Py_CLEAR(clear_module_state->__pyx_tuple__4);
  Py_CLEAR(clear_module_state->__pyx_tuple__5);
  Py_CLEAR(clear_module_state->__pyx_tuple__6);
  Py_CLEAR(clear_module_state->__pyx_tuple__7);
  Py_CLEAR(clear_module_state->__pyx_tuple__8);
  Py_CLEAR(clear_module_state->__pyx_tuple__9);
  Py_CLEAR(clear_module_state->__pyx_tuple__10);
  Py_CLEAR(clear_module_state->__pyx_tuple__11);
  Py_CLEAR(clear_module_state->__pyx_tuple__12);
  Py_CLEAR(clear_module_state->__pyx_tuple__13);
  Py_CLEAR(clear_module_state->__pyx_tuple__14);
  Py_CLEAR(clear_module_state->__pyx_tuple__15);
  Py_CLEAR(clear_module_state->__pyx_tuple__16);
  Py_CLEAR(clear_module_state->__pyx_tuple__17);
  Py_CLEAR(clear_module_state->__pyx_tuple__18);
  Py_CLEAR(clear_module_state->__pyx_tuple__19);
  Py_CLEAR(clear_module_state->__pyx_tuple__20);
  Py_CLEAR(clear_module_state->__pyx_tuple__21);
  Py_CLEAR(clear_module_state->__pyx_tuple__22);
  Py_CLEAR(clear_module_state->__pyx_tuple__23);
  Py_CLEAR(clear_module_state->__pyx_tuple__24);
  Py_CLEAR(clear_module_state->__pyx_tuple__25);
  Py_CLEAR(clear_module_state->__pyx_tuple__26);
  Py_CLEAR(clear_module_state->__pyx_tuple__27);
  Py_CLEAR(clear_module_state->__pyx_tuple__28);
  Py_CLEAR(clear_module_state->__pyx_tuple__29);
  Py_CLEAR(clear_module_state->__pyx_tuple__30);
  Py_CLEAR(clear_module_state->__pyx_tuple__31);
  Py_CLEAR(clear_module_state->__pyx_tuple__32);
  Py_CLEAR(clear_module_state->__pyx_tuple__33);
  Py_CLEAR(clear_module_state->__pyx_tuple__34);
  Py_CLEAR(clear_module_state->__pyx_tuple__35);
  Py_CLEAR(clear_module_state->__pyx_tuple__36);
  Py_CLEAR(clear_module_state->__pyx_tuple__37);
  Py_CLEAR(clear_module_state->__pyx_tuple__38);
  Py_CLEAR(clear_module_state->__pyx_tuple__39);
  Py_CLEAR(clear_module_state->__pyx_tuple__40);
  Py_CLEAR(clear_module_state->__pyx_tuple__41);
  Py_CLEAR(clear_module_state->__pyx_tuple__42);
  Py_CLEAR(clear_module_state->__pyx_tuple__43);
  Py_CLEAR(clear_module_state->__pyx_tuple__44);
  Py_CLEAR(clear_module_state->__pyx_tuple__45);
  Py_CLEAR(clear_module_state->__pyx_tuple__46);
  Py_CLEAR(clear_module_state->__pyx_tuple__47);
  Py_CLEAR(clear_module_state->__pyx_tuple__48);
  Py_CLEAR(clear_module_state->__pyx_tuple__49);
  Py_CLEAR(clear_module_state->__pyx_tuple__50);
  Py_CLEAR(clear_module_state->__pyx_tuple__51);
  Py_CLEAR(clear_module_state->__pyx_tuple__52);
  Py_CLEAR(clear_module_state->__pyx_tuple__53);
  Py_CLEAR(clear_module_state->__pyx_tuple__54);
  Py_CLEAR(clear_module_state->__pyx_tuple__55);
  Py_CLEAR(clear_module_state->__pyx_tuple__56);
  Py_CLEAR(clear_module_state->__pyx_tuple__57);
  Py_CLEAR(clear_module_state->__pyx_tuple__58);
  Py_CLEAR(clear_module_state->__pyx_tuple__59);
  Py_CLEAR(clear_module_state->__pyx_tuple__60);
  Py_CLEAR(clear_module_state->__pyx_tuple__61);
  Py_CLEAR(clear_module_state->__pyx_tuple__62);
  Py_CLEAR(clear_module_state->__pyx_tuple__63);
  Py_CLEAR(clear_module_state->__pyx_tuple__64);
  Py_CLEAR(clear_module_state->__pyx_tuple__65);
  Py_CLEAR(clear_module_state->__pyx_tuple__66);
  Py_CLEAR(clear_module_state->__pyx_tuple__67);
  Py_CLEAR(clear_module_state->__pyx_tuple__68);
  Py_CLEAR(clear_module_state->__pyx_tuple__69);
  Py_CLEAR(clear_module_state->__pyx_tuple__70);
  Py_CLEAR(clear_module_state->__pyx_tuple__71);
  Py_CLEAR(clear_module_state->__pyx_tuple__74);
  Py_CLEAR(clear_module_state->__pyx_tuple__77);
  Py_CLEAR(clear_module_state->__pyx_codeobj__72);
  Py_CLEAR(clear_module_state->__pyx_codeobj__73);
  Py_CLEAR(clear_module_state->__pyx_codeobj__75);
  Py_CLEAR(clear_module_state->__pyx_codeobj__76);
  Py_CLEAR(clear_module_state->__pyx_codeobj__78);
  return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_USE_MODULE_STATE
static int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstate *traverse_module_state = __pyx_mstate(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_d);
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  Py_VISIT(traverse_module_state->__pyx_empty_tuple);
  Py_VISIT(traverse_module_state->__pyx_empty_bytes);
  Py_VISIT(traverse_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_VISIT(traverse_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_VISIT(traverse_module_state->__pyx_FusedFunctionType);
  #endif
  Py_VISIT(traverse_module_state->__pyx_n_s_FuelModel);
  Py_VISIT(traverse_module_state->__pyx_n_s_KeyError);
  Py_VISIT(traverse_module_state->__pyx_n_s_M_f);
  Py_VISIT(traverse_module_state->__pyx_n_s_M_x);
  Py_VISIT(traverse_module_state->__pyx_kp_s_No_value_specified_for_struct_at);
  Py_VISIT(traverse_module_state->__pyx_kp_s_No_value_specified_for_struct_at_10);
  Py_VISIT(traverse_module_state->__pyx_kp_s_No_value_specified_for_struct_at_11);
  Py_VISIT(traverse_module_state->__pyx_kp_s_No_value_specified_for_struct_at_12);
  Py_VISIT(traverse_module_state->__pyx_kp_s_No_value_specified_for_struct_at_13);
  Py_VISIT(traverse_module_state->__pyx_kp_s_No_value_specified_for_struct_at_14);
  Py_VISIT(traverse_module_state->__pyx_kp_s_No_value_specified_for_struct_at_15);
  Py_VISIT(traverse_module_state->__pyx_kp_s_No_value_specified_for_struct_at_16);
  Py_VISIT(traverse_module_state->__pyx_kp_s_No_value_specified_for_struct_at_17);
  Py_VISIT(traverse_module_state->__pyx_kp_s_No_value_specified_for_struct_at_2);
  Py_VISIT(traverse_module_state->__pyx_kp_s_No_value_specified_for_struct_at_3);
  Py_VISIT(traverse_module_state->__pyx_kp_s_No_value_specified_for_struct_at_4);
  Py_VISIT(traverse_module_state->__pyx_kp_s_No_value_specified_for_struct_at_5);
  Py_VISIT(traverse_module_state->__pyx_kp_s_No_value_specified_for_struct_at_6);
  Py_VISIT(traverse_module_state->__pyx_kp_s_No_value_specified_for_struct_at_7);
  Py_VISIT(traverse_module_state->__pyx_kp_s_No_value_specified_for_struct_at_8);
  Py_VISIT(traverse_module_state->__pyx_kp_s_No_value_specified_for_struct_at_9);
  Py_VISIT(traverse_module_state->__pyx_n_s_S_T);
  Py_VISIT(traverse_module_state->__pyx_n_s_S_e);
  Py_VISIT(traverse_module_state->__pyx_n_s_ValueError);
  Py_VISIT(traverse_module_state->__pyx_n_s__79);
  Py_VISIT(traverse_module_state->__pyx_n_s_almost_zero);
  Py_VISIT(traverse_module_state->__pyx_n_s_asyncio_coroutines);
  Py_VISIT(traverse_module_state->__pyx_n_s_burnable);
  Py_VISIT(traverse_module_state->__pyx_n_s_cline_in_traceback);
  Py_VISIT(traverse_module_state->__pyx_n_s_compact_fuel_model_table);
  Py_VISIT(traverse_module_state->__pyx_kp_s_cy_int);
  Py_VISIT(traverse_module_state->__pyx_n_s_delta);
  Py_VISIT(traverse_module_state->__pyx_n_s_dynamic);
  Py_VISIT(traverse_module_state->__pyx_n_s_exp_A_sigma);
  Py_VISIT(traverse_module_state->__pyx_n_s_f_i);
  Py_VISIT(traverse_module_state->__pyx_n_s_f_ij);
  Py_VISIT(traverse_module_state->__pyx_n_s_fclaarr);
  Py_VISIT(traverse_module_state->__pyx_n_s_firemod_size_classes);
  Py_VISIT(traverse_module_state->__pyx_n_s_fuel_model);
  Py_VISIT(traverse_module_state->__pyx_n_s_fuel_model_exists);
  Py_VISIT(traverse_module_state->__pyx_n_s_fuel_model_number);
  Py_VISIT(traverse_module_state->__pyx_n_s_fuel_model_table);
  Py_VISIT(traverse_module_state->__pyx_n_s_fuel_moisture);
  Py_VISIT(traverse_module_state->__pyx_n_s_g_ij);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_fuel_model);
  Py_VISIT(traverse_module_state->__pyx_n_s_h);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_coroutine);
  Py_VISIT(traverse_module_state->__pyx_n_s_keys);
  Py_VISIT(traverse_module_state->__pyx_n_s_list_fuel_model_numbers);
  Py_VISIT(traverse_module_state->__pyx_n_s_list_fuel_models);
  Py_VISIT(traverse_module_state->__pyx_n_s_main);
  Py_VISIT(traverse_module_state->__pyx_n_s_moisturize);
  Py_VISIT(traverse_module_state->__pyx_n_s_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_number);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyretechnics_fuel_models);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_capi);
  Py_VISIT(traverse_module_state->__pyx_n_s_rho_p);
  Py_VISIT(traverse_module_state->__pyx_n_s_sigma);
  Py_VISIT(traverse_module_state->__pyx_kp_s_src_pyretechnics_fuel_models_py);
  Py_VISIT(traverse_module_state->__pyx_n_s_test);
  Py_VISIT(traverse_module_state->__pyx_n_s_values);
  Py_VISIT(traverse_module_state->__pyx_n_s_w_o);
  Py_VISIT(traverse_module_state->__pyx_float_0_0);
  Py_VISIT(traverse_module_state->__pyx_float_0_2);
  Py_VISIT(traverse_module_state->__pyx_float_0_3);
  Py_VISIT(traverse_module_state->__pyx_float_0_4);
  Py_VISIT(traverse_module_state->__pyx_float_0_5);
  Py_VISIT(traverse_module_state->__pyx_float_0_6);
  Py_VISIT(traverse_module_state->__pyx_float_0_9);
  Py_VISIT(traverse_module_state->__pyx_float_1_0);
  Py_VISIT(traverse_module_state->__pyx_float_1_2);
  Py_VISIT(traverse_module_state->__pyx_float_1_3);
  Py_VISIT(traverse_module_state->__pyx_float_1_5);
  Py_VISIT(traverse_module_state->__pyx_float_1_8);
  Py_VISIT(traverse_module_state->__pyx_float_2_0);
  Py_VISIT(traverse_module_state->__pyx_float_2_1);
  Py_VISIT(traverse_module_state->__pyx_float_2_3);
  Py_VISIT(traverse_module_state->__pyx_float_2_4);
  Py_VISIT(traverse_module_state->__pyx_float_2_5);
  Py_VISIT(traverse_module_state->__pyx_float_2_7);
  Py_VISIT(traverse_module_state->__pyx_float_3_0);
  Py_VISIT(traverse_module_state->__pyx_float_4_0);
  Py_VISIT(traverse_module_state->__pyx_float_4_4);
  Py_VISIT(traverse_module_state->__pyx_float_5_0);
  Py_VISIT(traverse_module_state->__pyx_float_6_0);
  Py_VISIT(traverse_module_state->__pyx_float_8_0);
  Py_VISIT(traverse_module_state->__pyx_float_9_0);
  Py_VISIT(traverse_module_state->__pyx_float_12_0);
  Py_VISIT(traverse_module_state->__pyx_float_15_0);
  Py_VISIT(traverse_module_state->__pyx_float_20_0);
  Py_VISIT(traverse_module_state->__pyx_float_25_0);
  Py_VISIT(traverse_module_state->__pyx_float_30_0);
  Py_VISIT(traverse_module_state->__pyx_float_35_0);
  Py_VISIT(traverse_module_state->__pyx_float_40_0);
  Py_VISIT(traverse_module_state->__pyx_float_109_0);
  Py_VISIT(traverse_module_state->__pyx_float_750_0);
  Py_VISIT(traverse_module_state->__pyx_float_0_0000);
  Py_VISIT(traverse_module_state->__pyx_float_0_0046);
  Py_VISIT(traverse_module_state->__pyx_float_0_0069);
  Py_VISIT(traverse_module_state->__pyx_float_0_0070);
  Py_VISIT(traverse_module_state->__pyx_float_0_0092);
  Py_VISIT(traverse_module_state->__pyx_float_0_0115);
  Py_VISIT(traverse_module_state->__pyx_float_0_0138);
  Py_VISIT(traverse_module_state->__pyx_float_0_0170);
  Py_VISIT(traverse_module_state->__pyx_float_0_0184);
  Py_VISIT(traverse_module_state->__pyx_float_0_0190);
  Py_VISIT(traverse_module_state->__pyx_float_0_0207);
  Py_VISIT(traverse_module_state->__pyx_float_0_0230);
  Py_VISIT(traverse_module_state->__pyx_float_0_0275);
  Py_VISIT(traverse_module_state->__pyx_float_0_0298);
  Py_VISIT(traverse_module_state->__pyx_float_0_0340);
  Py_VISIT(traverse_module_state->__pyx_float_0_0344);
  Py_VISIT(traverse_module_state->__pyx_float_0_0390);
  Py_VISIT(traverse_module_state->__pyx_float_0_0413);
  Py_VISIT(traverse_module_state->__pyx_float_0_0436);
  Py_VISIT(traverse_module_state->__pyx_float_0_0459);
  Py_VISIT(traverse_module_state->__pyx_float_0_0460);
  Py_VISIT(traverse_module_state->__pyx_float_0_0505);
  Py_VISIT(traverse_module_state->__pyx_float_0_0520);
  Py_VISIT(traverse_module_state->__pyx_float_0_0528);
  Py_VISIT(traverse_module_state->__pyx_float_0_0551);
  Py_VISIT(traverse_module_state->__pyx_float_0_0574);
  Py_VISIT(traverse_module_state->__pyx_float_0_0597);
  Py_VISIT(traverse_module_state->__pyx_float_0_0620);
  Py_VISIT(traverse_module_state->__pyx_float_0_0643);
  Py_VISIT(traverse_module_state->__pyx_float_0_0666);
  Py_VISIT(traverse_module_state->__pyx_float_0_0689);
  Py_VISIT(traverse_module_state->__pyx_float_0_0690);
  Py_VISIT(traverse_module_state->__pyx_float_0_0712);
  Py_VISIT(traverse_module_state->__pyx_float_0_0826);
  Py_VISIT(traverse_module_state->__pyx_float_0_0860);
  Py_VISIT(traverse_module_state->__pyx_float_0_0872);
  Py_VISIT(traverse_module_state->__pyx_float_0_0918);
  Py_VISIT(traverse_module_state->__pyx_float_0_0920);
  Py_VISIT(traverse_module_state->__pyx_float_0_0941);
  Py_VISIT(traverse_module_state->__pyx_float_0_0964);
  Py_VISIT(traverse_module_state->__pyx_float_0_1010);
  Py_VISIT(traverse_module_state->__pyx_float_0_1056);
  Py_VISIT(traverse_module_state->__pyx_float_0_1102);
  Py_VISIT(traverse_module_state->__pyx_float_0_1125);
  Py_VISIT(traverse_module_state->__pyx_float_0_1148);
  Py_VISIT(traverse_module_state->__pyx_float_0_1150);
  Py_VISIT(traverse_module_state->__pyx_float_0_1171);
  Py_VISIT(traverse_module_state->__pyx_float_0_1263);
  Py_VISIT(traverse_module_state->__pyx_float_0_1286);
  Py_VISIT(traverse_module_state->__pyx_float_0_1331);
  Py_VISIT(traverse_module_state->__pyx_float_0_1340);
  Py_VISIT(traverse_module_state->__pyx_float_0_1377);
  Py_VISIT(traverse_module_state->__pyx_float_0_1380);
  Py_VISIT(traverse_module_state->__pyx_float_0_1515);
  Py_VISIT(traverse_module_state->__pyx_float_0_1561);
  Py_VISIT(traverse_module_state->__pyx_float_0_1607);
  Py_VISIT(traverse_module_state->__pyx_float_0_1653);
  Py_VISIT(traverse_module_state->__pyx_float_0_1768);
  Py_VISIT(traverse_module_state->__pyx_float_0_1837);
  Py_VISIT(traverse_module_state->__pyx_float_0_1840);
  Py_VISIT(traverse_module_state->__pyx_float_0_1905);
  Py_VISIT(traverse_module_state->__pyx_float_0_1928);
  Py_VISIT(traverse_module_state->__pyx_float_0_1951);
  Py_VISIT(traverse_module_state->__pyx_float_0_1997);
  Py_VISIT(traverse_module_state->__pyx_float_0_2020);
  Py_VISIT(traverse_module_state->__pyx_float_0_2066);
  Py_VISIT(traverse_module_state->__pyx_float_0_2070);
  Py_VISIT(traverse_module_state->__pyx_float_0_2300);
  Py_VISIT(traverse_module_state->__pyx_float_0_2410);
  Py_VISIT(traverse_module_state->__pyx_float_0_2433);
  Py_VISIT(traverse_module_state->__pyx_float_0_2479);
  Py_VISIT(traverse_module_state->__pyx_float_0_2525);
  Py_VISIT(traverse_module_state->__pyx_float_0_2530);
  Py_VISIT(traverse_module_state->__pyx_float_0_2663);
  Py_VISIT(traverse_module_state->__pyx_float_0_2847);
  Py_VISIT(traverse_module_state->__pyx_float_0_3053);
  Py_VISIT(traverse_module_state->__pyx_float_0_3214);
  Py_VISIT(traverse_module_state->__pyx_float_0_3220);
  Py_VISIT(traverse_module_state->__pyx_float_0_3260);
  Py_VISIT(traverse_module_state->__pyx_float_0_3352);
  Py_VISIT(traverse_module_state->__pyx_float_0_3719);
  Py_VISIT(traverse_module_state->__pyx_float_0_4132);
  Py_VISIT(traverse_module_state->__pyx_float_0_5051);
  Py_VISIT(traverse_module_state->__pyx_float_0_6440);
  Py_VISIT(traverse_module_state->__pyx_float_0_7590);
  Py_VISIT(traverse_module_state->__pyx_float_1_0580);
  Py_VISIT(traverse_module_state->__pyx_float_1_2880);
  Py_VISIT(traverse_module_state->__pyx_float_1300_0);
  Py_VISIT(traverse_module_state->__pyx_float_1400_0);
  Py_VISIT(traverse_module_state->__pyx_float_1500_0);
  Py_VISIT(traverse_module_state->__pyx_float_1550_0);
  Py_VISIT(traverse_module_state->__pyx_float_1600_0);
  Py_VISIT(traverse_module_state->__pyx_float_1750_0);
  Py_VISIT(traverse_module_state->__pyx_float_1800_0);
  Py_VISIT(traverse_module_state->__pyx_float_2000_0);
  Py_VISIT(traverse_module_state->__pyx_float_2200_0);
  Py_VISIT(traverse_module_state->__pyx_float_2300_0);
  Py_VISIT(traverse_module_state->__pyx_float_2500_0);
  Py_VISIT(traverse_module_state->__pyx_float_3000_0);
  Py_VISIT(traverse_module_state->__pyx_float_3500_0);
  Py_VISIT(traverse_module_state->__pyx_int_1);
  Py_VISIT(traverse_module_state->__pyx_int_2);
  Py_VISIT(traverse_module_state->__pyx_int_3);
  Py_VISIT(traverse_module_state->__pyx_int_4);
  Py_VISIT(traverse_module_state->__pyx_int_5);
  Py_VISIT(traverse_module_state->__pyx_int_6);
  Py_VISIT(traverse_module_state->__pyx_int_7);
  Py_VISIT(traverse_module_state->__pyx_int_8);
  Py_VISIT(traverse_module_state->__pyx_int_9);
  Py_VISIT(traverse_module_state->__pyx_int_10);
  Py_VISIT(traverse_module_state->__pyx_int_11);
  Py_VISIT(traverse_module_state->__pyx_int_12);
  Py_VISIT(traverse_module_state->__pyx_int_13);
  Py_VISIT(traverse_module_state->__pyx_int_91);
  Py_VISIT(traverse_module_state->__pyx_int_92);
  Py_VISIT(traverse_module_state->__pyx_int_93);
  Py_VISIT(traverse_module_state->__pyx_int_98);
  Py_VISIT(traverse_module_state->__pyx_int_99);
  Py_VISIT(traverse_module_state->__pyx_int_101);
  Py_VISIT(traverse_module_state->__pyx_int_102);
  Py_VISIT(traverse_module_state->__pyx_int_103);
  Py_VISIT(traverse_module_state->__pyx_int_104);
  Py_VISIT(traverse_module_state->__pyx_int_105);
  Py_VISIT(traverse_module_state->__pyx_int_106);
  Py_VISIT(traverse_module_state->__pyx_int_107);
  Py_VISIT(traverse_module_state->__pyx_int_108);
  Py_VISIT(traverse_module_state->__pyx_int_109);
  Py_VISIT(traverse_module_state->__pyx_int_121);
  Py_VISIT(traverse_module_state->__pyx_int_122);
  Py_VISIT(traverse_module_state->__pyx_int_123);
  Py_VISIT(traverse_module_state->__pyx_int_124);
  Py_VISIT(traverse_module_state->__pyx_int_141);
  Py_VISIT(traverse_module_state->__pyx_int_142);
  Py_VISIT(traverse_module_state->__pyx_int_143);
  Py_VISIT(traverse_module_state->__pyx_int_144);
  Py_VISIT(traverse_module_state->__pyx_int_145);
  Py_VISIT(traverse_module_state->__pyx_int_146);
  Py_VISIT(traverse_module_state->__pyx_int_147);
  Py_VISIT(traverse_module_state->__pyx_int_148);
  Py_VISIT(traverse_module_state->__pyx_int_149);
  Py_VISIT(traverse_module_state->__pyx_int_161);
  Py_VISIT(traverse_module_state->__pyx_int_162);
  Py_VISIT(traverse_module_state->__pyx_int_163);
  Py_VISIT(traverse_module_state->__pyx_int_164);
  Py_VISIT(traverse_module_state->__pyx_int_165);
  Py_VISIT(traverse_module_state->__pyx_int_181);
  Py_VISIT(traverse_module_state->__pyx_int_182);
  Py_VISIT(traverse_module_state->__pyx_int_183);
  Py_VISIT(traverse_module_state->__pyx_int_184);
  Py_VISIT(traverse_module_state->__pyx_int_185);
  Py_VISIT(traverse_module_state->__pyx_int_186);
  Py_VISIT(traverse_module_state->__pyx_int_187);
  Py_VISIT(traverse_module_state->__pyx_int_188);
  Py_VISIT(traverse_module_state->__pyx_int_189);
  Py_VISIT(traverse_module_state->__pyx_int_201);
  Py_VISIT(traverse_module_state->__pyx_int_202);
  Py_VISIT(traverse_module_state->__pyx_int_203);
  Py_VISIT(traverse_module_state->__pyx_int_204);
  Py_VISIT(traverse_module_state->__pyx_tuple_);
  Py_VISIT(traverse_module_state->__pyx_tuple__2);
  Py_VISIT(traverse_module_state->__pyx_tuple__3);
  Py_VISIT(traverse_module_state->__pyx_tuple__4);
  Py_VISIT(traverse_module_state->__pyx_tuple__5);
  Py_VISIT(traverse_module_state->__pyx_tuple__6);
  Py_VISIT(traverse_module_state->__pyx_tuple__7);
  Py_VISIT(traverse_module_state->__pyx_tuple__8);
  Py_VISIT(traverse_module_state->__pyx_tuple__9);
  Py_VISIT(traverse_module_state->__pyx_tuple__10);
  Py_VISIT(traverse_module_state->__pyx_tuple__11);
  Py_VISIT(traverse_module_state->__pyx_tuple__12);
  Py_VISIT(traverse_module_state->__pyx_tuple__13);
  Py_VISIT(traverse_module_state->__pyx_tuple__14);
  Py_VISIT(traverse_module_state->__pyx_tuple__15);
  Py_VISIT(traverse_module_state->__pyx_tuple__16);
  Py_VISIT(traverse_module_state->__pyx_tuple__17);
  Py_VISIT(traverse_module_state->__pyx_tuple__18);
  Py_VISIT(traverse_module_state->__pyx_tuple__19);
  Py_VISIT(traverse_module_state->__pyx_tuple__20);
  Py_VISIT(traverse_module_state->__pyx_tuple__21);
  Py_VISIT(traverse_module_state->__pyx_tuple__22);
  Py_VISIT(traverse_module_state->__pyx_tuple__23);
  Py_VISIT(traverse_module_state->__pyx_tuple__24);
  Py_VISIT(traverse_module_state->__pyx_tuple__25);
  Py_VISIT(traverse_module_state->__pyx_tuple__26);
  Py_VISIT(traverse_module_state->__pyx_tuple__27);
  Py_VISIT(traverse_module_state->__pyx_tuple__28);
  Py_VISIT(traverse_module_state->__pyx_tuple__29);
  Py_VISIT(traverse_module_state->__pyx_tuple__30);
  Py_VISIT(traverse_module_state->__pyx_tuple__31);
  Py_VISIT(traverse_module_state->__pyx_tuple__32);
  Py_VISIT(traverse_module_state->__pyx_tuple__33);
  Py_VISIT(traverse_module_state->__pyx_tuple__34);
  Py_VISIT(traverse_module_state->__pyx_tuple__35);
  Py_VISIT(traverse_module_state->__pyx_tuple__36);
  Py_VISIT(traverse_module_state->__pyx_tuple__37);
  Py_VISIT(traverse_module_state->__pyx_tuple__38);
  Py_VISIT(traverse_module_state->__pyx_tuple__39);
  Py_VISIT(traverse_module_state->__pyx_tuple__40);
  Py_VISIT(traverse_module_state->__pyx_tuple__41);
  Py_VISIT(traverse_module_state->__pyx_tuple__42);
  Py_VISIT(traverse_module_state->__pyx_tuple__43);
  Py_VISIT(traverse_module_state->__pyx_tuple__44);
  Py_VISIT(traverse_module_state->__pyx_tuple__45);
  Py_VISIT(traverse_module_state->__pyx_tuple__46);
  Py_VISIT(traverse_module_state->__pyx_tuple__47);
  Py_VISIT(traverse_module_state->__pyx_tuple__48);
  Py_VISIT(traverse_module_state->__pyx_tuple__49);
  Py_VISIT(traverse_module_state->__pyx_tuple__50);
  Py_VISIT(traverse_module_state->__pyx_tuple__51);
  Py_VISIT(traverse_module_state->__pyx_tuple__52);
  Py_VISIT(traverse_module_state->__pyx_tuple__53);
  Py_VISIT(traverse_module_state->__pyx_tuple__54);
  Py_VISIT(traverse_module_state->__pyx_tuple__55);
  Py_VISIT(traverse_module_state->__pyx_tuple__56);
  Py_VISIT(traverse_module_state->__pyx_tuple__57);
  Py_VISIT(traverse_module_state->__pyx_tuple__58);
  Py_VISIT(traverse_module_state->__pyx_tuple__59);
  Py_VISIT(traverse_module_state->__pyx_tuple__60);
  Py_VISIT(traverse_module_state->__pyx_tuple__61);
  Py_VISIT(traverse_module_state->__pyx_tuple__62);
  Py_VISIT(traverse_module_state->__pyx_tuple__63);
  Py_VISIT(traverse_module_state->__pyx_tuple__64);
  Py_VISIT(traverse_module_state->__pyx_tuple__65);
  Py_VISIT(traverse_module_state->__pyx_tuple__66);
  Py_VISIT(traverse_module_state->__pyx_tuple__67);
  Py_VISIT(traverse_module_state->__pyx_tuple__68);
  Py_VISIT(traverse_module_state->__pyx_tuple__69);
  Py_VISIT(traverse_module_state->__pyx_tuple__70);
  Py_VISIT(traverse_module_state->__pyx_tuple__71);
  Py_VISIT(traverse_module_state->__pyx_tuple__74);
  Py_VISIT(traverse_module_state->__pyx_tuple__77);
  Py_VISIT(traverse_module_state->__pyx_codeobj__72);
  Py_VISIT(traverse_module_state->__pyx_codeobj__73);
  Py_VISIT(traverse_module_state->__pyx_codeobj__75);
  Py_VISIT(traverse_module_state->__pyx_codeobj__76);
  Py_VISIT(traverse_module_state->__pyx_codeobj__78);
  return 0;
}
#endif
/* #### Code section: module_state_defines ### */
#define __pyx_d __pyx_mstate_global->__pyx_d
#define __pyx_b __pyx_mstate_global->__pyx_b
#define __pyx_cython_runtime __pyx_mstate_global->__pyx_cython_runtime
#define __pyx_empty_tuple __pyx_mstate_global->__pyx_empty_tuple
#define __pyx_empty_bytes __pyx_mstate_global->__pyx_empty_bytes
#define __pyx_empty_unicode __pyx_mstate_global->__pyx_empty_unicode
#ifdef __Pyx_CyFunction_USED
#define __pyx_CyFunctionType __pyx_mstate_global->__pyx_CyFunctionType
#endif
#ifdef __Pyx_FusedFunction_USED
#define __pyx_FusedFunctionType __pyx_mstate_global->__pyx_FusedFunctionType
#endif
#ifdef __Pyx_Generator_USED
#define __pyx_GeneratorType __pyx_mstate_global->__pyx_GeneratorType
#endif
#ifdef __Pyx_IterableCoroutine_USED
#define __pyx_IterableCoroutineType __pyx_mstate_global->__pyx_IterableCoroutineType
#endif
#ifdef __Pyx_Coroutine_USED
#define __pyx_CoroutineAwaitType __pyx_mstate_global->__pyx_CoroutineAwaitType
#endif
#ifdef __Pyx_Coroutine_USED
#define __pyx_CoroutineType __pyx_mstate_global->__pyx_CoroutineType
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_n_s_FuelModel __pyx_mstate_global->__pyx_n_s_FuelModel
#define __pyx_n_s_KeyError __pyx_mstate_global->__pyx_n_s_KeyError
#define __pyx_n_s_M_f __pyx_mstate_global->__pyx_n_s_M_f
#define __pyx_n_s_M_x __pyx_mstate_global->__pyx_n_s_M_x
#define __pyx_kp_s_No_value_specified_for_struct_at __pyx_mstate_global->__pyx_kp_s_No_value_specified_for_struct_at
#define __pyx_kp_s_No_value_specified_for_struct_at_10 __pyx_mstate_global->__pyx_kp_s_No_value_specified_for_struct_at_10
#define __pyx_kp_s_No_value_specified_for_struct_at_11 __pyx_mstate_global->__pyx_kp_s_No_value_specified_for_struct_at_11
#define __pyx_kp_s_No_value_specified_for_struct_at_12 __pyx_mstate_global->__pyx_kp_s_No_value_specified_for_struct_at_12
#define __pyx_kp_s_No_value_specified_for_struct_at_13 __pyx_mstate_global->__pyx_kp_s_No_value_specified_for_struct_at_13
#define __pyx_kp_s_No_value_specified_for_struct_at_14 __pyx_mstate_global->__pyx_kp_s_No_value_specified_for_struct_at_14
#define __pyx_kp_s_No_value_specified_for_struct_at_15 __pyx_mstate_global->__pyx_kp_s_No_value_specified_for_struct_at_15
#define __pyx_kp_s_No_value_specified_for_struct_at_16 __pyx_mstate_global->__pyx_kp_s_No_value_specified_for_struct_at_16
#define __pyx_kp_s_No_value_specified_for_struct_at_17 __pyx_mstate_global->__pyx_kp_s_No_value_specified_for_struct_at_17
#define __pyx_kp_s_No_value_specified_for_struct_at_2 __pyx_mstate_global->__pyx_kp_s_No_value_specified_for_struct_at_2
#define __pyx_kp_s_No_value_specified_for_struct_at_3 __pyx_mstate_global->__pyx_kp_s_No_value_specified_for_struct_at_3
#define __pyx_kp_s_No_value_specified_for_struct_at_4 __pyx_mstate_global->__pyx_kp_s_No_value_specified_for_struct_at_4
#define __pyx_kp_s_No_value_specified_for_struct_at_5 __pyx_mstate_global->__pyx_kp_s_No_value_specified_for_struct_at_5
#define __pyx_kp_s_No_value_specified_for_struct_at_6 __pyx_mstate_global->__pyx_kp_s_No_value_specified_for_struct_at_6
#define __pyx_kp_s_No_value_specified_for_struct_at_7 __pyx_mstate_global->__pyx_kp_s_No_value_specified_for_struct_at_7
#define __pyx_kp_s_No_value_specified_for_struct_at_8 __pyx_mstate_global->__pyx_kp_s_No_value_specified_for_struct_at_8
#define __pyx_kp_s_No_value_specified_for_struct_at_9 __pyx_mstate_global->__pyx_kp_s_No_value_specified_for_struct_at_9
#define __pyx_n_s_S_T __pyx_mstate_global->__pyx_n_s_S_T
#define __pyx_n_s_S_e __pyx_mstate_global->__pyx_n_s_S_e
#define __pyx_n_s_ValueError __pyx_mstate_global->__pyx_n_s_ValueError
#define __pyx_n_s__79 __pyx_mstate_global->__pyx_n_s__79
#define __pyx_n_s_almost_zero __pyx_mstate_global->__pyx_n_s_almost_zero
#define __pyx_n_s_asyncio_coroutines __pyx_mstate_global->__pyx_n_s_asyncio_coroutines
#define __pyx_n_s_burnable __pyx_mstate_global->__pyx_n_s_burnable
#define __pyx_n_s_cline_in_traceback __pyx_mstate_global->__pyx_n_s_cline_in_traceback
#define __pyx_n_s_compact_fuel_model_table __pyx_mstate_global->__pyx_n_s_compact_fuel_model_table
#define __pyx_kp_s_cy_int __pyx_mstate_global->__pyx_kp_s_cy_int
#define __pyx_n_s_delta __pyx_mstate_global->__pyx_n_s_delta
#define __pyx_n_s_dynamic __pyx_mstate_global->__pyx_n_s_dynamic
#define __pyx_n_s_exp_A_sigma __pyx_mstate_global->__pyx_n_s_exp_A_sigma
#define __pyx_n_s_f_i __pyx_mstate_global->__pyx_n_s_f_i
#define __pyx_n_s_f_ij __pyx_mstate_global->__pyx_n_s_f_ij
#define __pyx_n_s_fclaarr __pyx_mstate_global->__pyx_n_s_fclaarr
#define __pyx_n_s_firemod_size_classes __pyx_mstate_global->__pyx_n_s_firemod_size_classes
#define __pyx_n_s_fuel_model __pyx_mstate_global->__pyx_n_s_fuel_model
#define __pyx_n_s_fuel_model_exists __pyx_mstate_global->__pyx_n_s_fuel_model_exists
#define __pyx_n_s_fuel_model_number __pyx_mstate_global->__pyx_n_s_fuel_model_number
#define __pyx_n_s_fuel_model_table __pyx_mstate_global->__pyx_n_s_fuel_model_table
#define __pyx_n_s_fuel_moisture __pyx_mstate_global->__pyx_n_s_fuel_moisture
#define __pyx_n_s_g_ij __pyx_mstate_global->__pyx_n_s_g_ij
#define __pyx_n_s_get_fuel_model __pyx_mstate_global->__pyx_n_s_get_fuel_model
#define __pyx_n_s_h __pyx_mstate_global->__pyx_n_s_h
#define __pyx_n_s_is_coroutine __pyx_mstate_global->__pyx_n_s_is_coroutine
#define __pyx_n_s_keys __pyx_mstate_global->__pyx_n_s_keys
#define __pyx_n_s_list_fuel_model_numbers __pyx_mstate_global->__pyx_n_s_list_fuel_model_numbers
#define __pyx_n_s_list_fuel_models __pyx_mstate_global->__pyx_n_s_list_fuel_models
#define __pyx_n_s_main __pyx_mstate_global->__pyx_n_s_main
#define __pyx_n_s_moisturize __pyx_mstate_global->__pyx_n_s_moisturize
#define __pyx_n_s_name __pyx_mstate_global->__pyx_n_s_name
#define __pyx_n_s_number __pyx_mstate_global->__pyx_n_s_number
#define __pyx_n_s_pyretechnics_fuel_models __pyx_mstate_global->__pyx_n_s_pyretechnics_fuel_models
#define __pyx_n_s_pyx_capi __pyx_mstate_global->__pyx_n_s_pyx_capi
#define __pyx_n_s_rho_p __pyx_mstate_global->__pyx_n_s_rho_p
#define __pyx_n_s_sigma __pyx_mstate_global->__pyx_n_s_sigma
#define __pyx_kp_s_src_pyretechnics_fuel_models_py __pyx_mstate_global->__pyx_kp_s_src_pyretechnics_fuel_models_py
#define __pyx_n_s_test __pyx_mstate_global->__pyx_n_s_test
#define __pyx_n_s_values __pyx_mstate_global->__pyx_n_s_values
#define __pyx_n_s_w_o __pyx_mstate_global->__pyx_n_s_w_o
#define __pyx_float_0_0 __pyx_mstate_global->__pyx_float_0_0
#define __pyx_float_0_2 __pyx_mstate_global->__pyx_float_0_2
#define __pyx_float_0_3 __pyx_mstate_global->__pyx_float_0_3
#define __pyx_float_0_4 __pyx_mstate_global->__pyx_float_0_4
#define __pyx_float_0_5 __pyx_mstate_global->__pyx_float_0_5
#define __pyx_float_0_6 __pyx_mstate_global->__pyx_float_0_6
#define __pyx_float_0_9 __pyx_mstate_global->__pyx_float_0_9
#define __pyx_float_1_0 __pyx_mstate_global->__pyx_float_1_0
#define __pyx_float_1_2 __pyx_mstate_global->__pyx_float_1_2
#define __pyx_float_1_3 __pyx_mstate_global->__pyx_float_1_3
#define __pyx_float_1_5 __pyx_mstate_global->__pyx_float_1_5
#define __pyx_float_1_8 __pyx_mstate_global->__pyx_float_1_8
#define __pyx_float_2_0 __pyx_mstate_global->__pyx_float_2_0
#define __pyx_float_2_1 __pyx_mstate_global->__pyx_float_2_1
#define __pyx_float_2_3 __pyx_mstate_global->__pyx_float_2_3
#define __pyx_float_2_4 __pyx_mstate_global->__pyx_float_2_4
#define __pyx_float_2_5 __pyx_mstate_global->__pyx_float_2_5
#define __pyx_float_2_7 __pyx_mstate_global->__pyx_float_2_7
#define __pyx_float_3_0 __pyx_mstate_global->__pyx_float_3_0
#define __pyx_float_4_0 __pyx_mstate_global->__pyx_float_4_0
#define __pyx_float_4_4 __pyx_mstate_global->__pyx_float_4_4
#define __pyx_float_5_0 __pyx_mstate_global->__pyx_float_5_0
#define __pyx_float_6_0 __pyx_mstate_global->__pyx_float_6_0
#define __pyx_float_8_0 __pyx_mstate_global->__pyx_float_8_0
#define __pyx_float_9_0 __pyx_mstate_global->__pyx_float_9_0
#define __pyx_float_12_0 __pyx_mstate_global->__pyx_float_12_0
#define __pyx_float_15_0 __pyx_mstate_global->__pyx_float_15_0
#define __pyx_float_20_0 __pyx_mstate_global->__pyx_float_20_0
#define __pyx_float_25_0 __pyx_mstate_global->__pyx_float_25_0
#define __pyx_float_30_0 __pyx_mstate_global->__pyx_float_30_0
#define __pyx_float_35_0 __pyx_mstate_global->__pyx_float_35_0
#define __pyx_float_40_0 __pyx_mstate_global->__pyx_float_40_0
#define __pyx_float_109_0 __pyx_mstate_global->__pyx_float_109_0
#define __pyx_float_750_0 __pyx_mstate_global->__pyx_float_750_0
#define __pyx_float_0_0000 __pyx_mstate_global->__pyx_float_0_0000
#define __pyx_float_0_0046 __pyx_mstate_global->__pyx_float_0_0046
#define __pyx_float_0_0069 __pyx_mstate_global->__pyx_float_0_0069
#define __pyx_float_0_0070 __pyx_mstate_global->__pyx_float_0_0070
#define __pyx_float_0_0092 __pyx_mstate_global->__pyx_float_0_0092
#define __pyx_float_0_0115 __pyx_mstate_global->__pyx_float_0_0115
#define __pyx_float_0_0138 __pyx_mstate_global->__pyx_float_0_0138
#define __pyx_float_0_0170 __pyx_mstate_global->__pyx_float_0_0170
#define __pyx_float_0_0184 __pyx_mstate_global->__pyx_float_0_0184
#define __pyx_float_0_0190 __pyx_mstate_global->__pyx_float_0_0190
#define __pyx_float_0_0207 __pyx_mstate_global->__pyx_float_0_0207
#define __pyx_float_0_0230 __pyx_mstate_global->__pyx_float_0_0230
#define __pyx_float_0_0275 __pyx_mstate_global->__pyx_float_0_0275
#define __pyx_float_0_0298 __pyx_mstate_global->__pyx_float_0_0298
#define __pyx_float_0_0340 __pyx_mstate_global->__pyx_float_0_0340
#define __pyx_float_0_0344 __pyx_mstate_global->__pyx_float_0_0344
#define __pyx_float_0_0390 __pyx_mstate_global->__pyx_float_0_0390
#define __pyx_float_0_0413 __pyx_mstate_global->__pyx_float_0_0413
#define __pyx_float_0_0436 __pyx_mstate_global->__pyx_float_0_0436
#define __pyx_float_0_0459 __pyx_mstate_global->__pyx_float_0_0459
#define __pyx_float_0_0460 __pyx_mstate_global->__pyx_float_0_0460
#define __pyx_float_0_0505 __pyx_mstate_global->__pyx_float_0_0505
#define __pyx_float_0_0520 __pyx_mstate_global->__pyx_float_0_0520
#define __pyx_float_0_0528 __pyx_mstate_global->__pyx_float_0_0528
#define __pyx_float_0_0551 __pyx_mstate_global->__pyx_float_0_0551
#define __pyx_float_0_0574 __pyx_mstate_global->__pyx_float_0_0574
#define __pyx_float_0_0597 __pyx_mstate_global->__pyx_float_0_0597
#define __pyx_float_0_0620 __pyx_mstate_global->__pyx_float_0_0620
#define __pyx_float_0_0643 __pyx_mstate_global->__pyx_float_0_0643
#define __pyx_float_0_0666 __pyx_mstate_global->__pyx_float_0_0666
#define __pyx_float_0_0689 __pyx_mstate_global->__pyx_float_0_0689
#define __pyx_float_0_0690 __pyx_mstate_global->__pyx_float_0_0690
#define __pyx_float_0_0712 __pyx_mstate_global->__pyx_float_0_0712
#define __pyx_float_0_0826 __pyx_mstate_global->__pyx_float_0_0826
#define __pyx_float_0_0860 __pyx_mstate_global->__pyx_float_0_0860
#define __pyx_float_0_0872 __pyx_mstate_global->__pyx_float_0_0872
#define __pyx_float_0_0918 __pyx_mstate_global->__pyx_float_0_0918
#define __pyx_float_0_0920 __pyx_mstate_global->__pyx_float_0_0920
#define __pyx_float_0_0941 __pyx_mstate_global->__pyx_float_0_0941
#define __pyx_float_0_0964 __pyx_mstate_global->__pyx_float_0_0964
#define __pyx_float_0_1010 __pyx_mstate_global->__pyx_float_0_1010
#define __pyx_float_0_1056 __pyx_mstate_global->__pyx_float_0_1056
#define __pyx_float_0_1102 __pyx_mstate_global->__pyx_float_0_1102
#define __pyx_float_0_1125 __pyx_mstate_global->__pyx_float_0_1125
#define __pyx_float_0_1148 __pyx_mstate_global->__pyx_float_0_1148
#define __pyx_float_0_1150 __pyx_mstate_global->__pyx_float_0_1150
#define __pyx_float_0_1171 __pyx_mstate_global->__pyx_float_0_1171
#define __pyx_float_0_1263 __pyx_mstate_global->__pyx_float_0_1263
#define __pyx_float_0_1286 __pyx_mstate_global->__pyx_float_0_1286
#define __pyx_float_0_1331 __pyx_mstate_global->__pyx_float_0_1331
#define __pyx_float_0_1340 __pyx_mstate_global->__pyx_float_0_1340
#define __pyx_float_0_1377 __pyx_mstate_global->__pyx_float_0_1377
#define __pyx_float_0_1380 __pyx_mstate_global->__pyx_float_0_1380
#define __pyx_float_0_1515 __pyx_mstate_global->__pyx_float_0_1515
#define __pyx_float_0_1561 __pyx_mstate_global->__pyx_float_0_1561
#define __pyx_float_0_1607 __pyx_mstate_global->__pyx_float_0_1607
#define __pyx_float_0_1653 __pyx_mstate_global->__pyx_float_0_1653
#define __pyx_float_0_1768 __pyx_mstate_global->__pyx_float_0_1768
#define __pyx_float_0_1837 __pyx_mstate_global->__pyx_float_0_1837
#define __pyx_float_0_1840 __pyx_mstate_global->__pyx_float_0_1840
#define __pyx_float_0_1905 __pyx_mstate_global->__pyx_float_0_1905
#define __pyx_float_0_1928 __pyx_mstate_global->__pyx_float_0_1928
#define __pyx_float_0_1951 __pyx_mstate_global->__pyx_float_0_1951
#define __pyx_float_0_1997 __pyx_mstate_global->__pyx_float_0_1997
#define __pyx_float_0_2020 __pyx_mstate_global->__pyx_float_0_2020
#define __pyx_float_0_2066 __pyx_mstate_global->__pyx_float_0_2066
#define __pyx_float_0_2070 __pyx_mstate_global->__pyx_float_0_2070
#define __pyx_float_0_2300 __pyx_mstate_global->__pyx_float_0_2300
#define __pyx_float_0_2410 __pyx_mstate_global->__pyx_float_0_2410
#define __pyx_float_0_2433 __pyx_mstate_global->__pyx_float_0_2433
#define __pyx_float_0_2479 __pyx_mstate_global->__pyx_float_0_2479
#define __pyx_float_0_2525 __pyx_mstate_global->__pyx_float_0_2525
#define __pyx_float_0_2530 __pyx_mstate_global->__pyx_float_0_2530
#define __pyx_float_0_2663 __pyx_mstate_global->__pyx_float_0_2663
#define __pyx_float_0_2847 __pyx_mstate_global->__pyx_float_0_2847
#define __pyx_float_0_3053 __pyx_mstate_global->__pyx_float_0_3053
#define __pyx_float_0_3214 __pyx_mstate_global->__pyx_float_0_3214
#define __pyx_float_0_3220 __pyx_mstate_global->__pyx_float_0_3220
#define __pyx_float_0_3260 __pyx_mstate_global->__pyx_float_0_3260
#define __pyx_float_0_3352 __pyx_mstate_global->__pyx_float_0_3352
#define __pyx_float_0_3719 __pyx_mstate_global->__pyx_float_0_3719
#define __pyx_float_0_4132 __pyx_mstate_global->__pyx_float_0_4132
#define __pyx_float_0_5051 __pyx_mstate_global->__pyx_float_0_5051
#define __pyx_float_0_6440 __pyx_mstate_global->__pyx_float_0_6440
#define __pyx_float_0_7590 __pyx_mstate_global->__pyx_float_0_7590
#define __pyx_float_1_0580 __pyx_mstate_global->__pyx_float_1_0580
#define __pyx_float_1_2880 __pyx_mstate_global->__pyx_float_1_2880
#define __pyx_float_1300_0 __pyx_mstate_global->__pyx_float_1300_0
#define __pyx_float_1400_0 __pyx_mstate_global->__pyx_float_1400_0
#define __pyx_float_1500_0 __pyx_mstate_global->__pyx_float_1500_0
#define __pyx_float_1550_0 __pyx_mstate_global->__pyx_float_1550_0
#define __pyx_float_1600_0 __pyx_mstate_global->__pyx_float_1600_0
#define __pyx_float_1750_0 __pyx_mstate_global->__pyx_float_1750_0
#define __pyx_float_1800_0 __pyx_mstate_global->__pyx_float_1800_0
#define __pyx_float_2000_0 __pyx_mstate_global->__pyx_float_2000_0
#define __pyx_float_2200_0 __pyx_mstate_global->__pyx_float_2200_0
#define __pyx_float_2300_0 __pyx_mstate_global->__pyx_float_2300_0
#define __pyx_float_2500_0 __pyx_mstate_global->__pyx_float_2500_0
#define __pyx_float_3000_0 __pyx_mstate_global->__pyx_float_3000_0
#define __pyx_float_3500_0 __pyx_mstate_global->__pyx_float_3500_0
#define __pyx_int_1 __pyx_mstate_global->__pyx_int_1
#define __pyx_int_2 __pyx_mstate_global->__pyx_int_2
#define __pyx_int_3 __pyx_mstate_global->__pyx_int_3
#define __pyx_int_4 __pyx_mstate_global->__pyx_int_4
#define __pyx_int_5 __pyx_mstate_global->__pyx_int_5
#define __pyx_int_6 __pyx_mstate_global->__pyx_int_6
#define __pyx_int_7 __pyx_mstate_global->__pyx_int_7
#define __pyx_int_8 __pyx_mstate_global->__pyx_int_8
#define __pyx_int_9 __pyx_mstate_global->__pyx_int_9
#define __pyx_int_10 __pyx_mstate_global->__pyx_int_10
#define __pyx_int_11 __pyx_mstate_global->__pyx_int_11
#define __pyx_int_12 __pyx_mstate_global->__pyx_int_12
#define __pyx_int_13 __pyx_mstate_global->__pyx_int_13
#define __pyx_int_91 __pyx_mstate_global->__pyx_int_91
#define __pyx_int_92 __pyx_mstate_global->__pyx_int_92
#define __pyx_int_93 __pyx_mstate_global->__pyx_int_93
#define __pyx_int_98 __pyx_mstate_global->__pyx_int_98
#define __pyx_int_99 __pyx_mstate_global->__pyx_int_99
#define __pyx_int_101 __pyx_mstate_global->__pyx_int_101
#define __pyx_int_102 __pyx_mstate_global->__pyx_int_102
#define __pyx_int_103 __pyx_mstate_global->__pyx_int_103
#define __pyx_int_104 __pyx_mstate_global->__pyx_int_104
#define __pyx_int_105 __pyx_mstate_global->__pyx_int_105
#define __pyx_int_106 __pyx_mstate_global->__pyx_int_106
#define __pyx_int_107 __pyx_mstate_global->__pyx_int_107
#define __pyx_int_108 __pyx_mstate_global->__pyx_int_108
#define __pyx_int_109 __pyx_mstate_global->__pyx_int_109
#define __pyx_int_121 __pyx_mstate_global->__pyx_int_121
#define __pyx_int_122 __pyx_mstate_global->__pyx_int_122
#define __pyx_int_123 __pyx_mstate_global->__pyx_int_123
#define __pyx_int_124 __pyx_mstate_global->__pyx_int_124
#define __pyx_int_141 __pyx_mstate_global->__pyx_int_141
#define __pyx_int_142 __pyx_mstate_global->__pyx_int_142
#define __pyx_int_143 __pyx_mstate_global->__pyx_int_143
#define __pyx_int_144 __pyx_mstate_global->__pyx_int_144
#define __pyx_int_145 __pyx_mstate_global->__pyx_int_145
#define __pyx_int_146 __pyx_mstate_global->__pyx_int_146
#define __pyx_int_147 __pyx_mstate_global->__pyx_int_147
#define __pyx_int_148 __pyx_mstate_global->__pyx_int_148
#define __pyx_int_149 __pyx_mstate_global->__pyx_int_149
#define __pyx_int_161 __pyx_mstate_global->__pyx_int_161
#define __pyx_int_162 __pyx_mstate_global->__pyx_int_162
#define __pyx_int_163 __pyx_mstate_global->__pyx_int_163
#define __pyx_int_164 __pyx_mstate_global->__pyx_int_164
#define __pyx_int_165 __pyx_mstate_global->__pyx_int_165
#define __pyx_int_181 __pyx_mstate_global->__pyx_int_181
#define __pyx_int_182 __pyx_mstate_global->__pyx_int_182
#define __pyx_int_183 __pyx_mstate_global->__pyx_int_183
#define __pyx_int_184 __pyx_mstate_global->__pyx_int_184
#define __pyx_int_185 __pyx_mstate_global->__pyx_int_185
#define __pyx_int_186 __pyx_mstate_global->__pyx_int_186
#define __pyx_int_187 __pyx_mstate_global->__pyx_int_187
#define __pyx_int_188 __pyx_mstate_global->__pyx_int_188
#define __pyx_int_189 __pyx_mstate_global->__pyx_int_189
#define __pyx_int_201 __pyx_mstate_global->__pyx_int_201
#define __pyx_int_202 __pyx_mstate_global->__pyx_int_202
#define __pyx_int_203 __pyx_mstate_global->__pyx_int_203
#define __pyx_int_204 __pyx_mstate_global->__pyx_int_204
#define __pyx_tuple_ __pyx_mstate_global->__pyx_tuple_
#define __pyx_tuple__2 __pyx_mstate_global->__pyx_tuple__2
#define __pyx_tuple__3 __pyx_mstate_global->__pyx_tuple__3
#define __pyx_tuple__4 __pyx_mstate_global->__pyx_tuple__4
#define __pyx_tuple__5 __pyx_mstate_global->__pyx_tuple__5
#define __pyx_tuple__6 __pyx_mstate_global->__pyx_tuple__6
#define __pyx_tuple__7 __pyx_mstate_global->__pyx_tuple__7
#define __pyx_tuple__8 __pyx_mstate_global->__pyx_tuple__8
#define __pyx_tuple__9 __pyx_mstate_global->__pyx_tuple__9
#define __pyx_tuple__10 __pyx_mstate_global->__pyx_tuple__10
#define __pyx_tuple__11 __pyx_mstate_global->__pyx_tuple__11
#define __pyx_tuple__12 __pyx_mstate_global->__pyx_tuple__12
#define __pyx_tuple__13 __pyx_mstate_global->__pyx_tuple__13
#define __pyx_tuple__14 __pyx_mstate_global->__pyx_tuple__14
#define __pyx_tuple__15 __pyx_mstate_global->__pyx_tuple__15
#define __pyx_tuple__16 __pyx_mstate_global->__pyx_tuple__16
#define __pyx_tuple__17 __pyx_mstate_global->__pyx_tuple__17
#define __pyx_tuple__18 __pyx_mstate_global->__pyx_tuple__18
#define __pyx_tuple__19 __pyx_mstate_global->__pyx_tuple__19
#define __pyx_tuple__20 __pyx_mstate_global->__pyx_tuple__20
#define __pyx_tuple__21 __pyx_mstate_global->__pyx_tuple__21
#define __pyx_tuple__22 __pyx_mstate_global->__pyx_tuple__22
#define __pyx_tuple__23 __pyx_mstate_global->__pyx_tuple__23
#define __pyx_tuple__24 __pyx_mstate_global->__pyx_tuple__24
#define __pyx_tuple__25 __pyx_mstate_global->__pyx_tuple__25
#define __pyx_tuple__26 __pyx_mstate_global->__pyx_tuple__26
#define __pyx_tuple__27 __pyx_mstate_global->__pyx_tuple__27
#define __pyx_tuple__28 __pyx_mstate_global->__pyx_tuple__28
#define __pyx_tuple__29 __pyx_mstate_global->__pyx_tuple__29
#define __pyx_tuple__30 __pyx_mstate_global->__pyx_tuple__30
#define __pyx_tuple__31 __pyx_mstate_global->__pyx_tuple__31
#define __pyx_tuple__32 __pyx_mstate_global->__pyx_tuple__32
#define __pyx_tuple__33 __pyx_mstate_global->__pyx_tuple__33
#define __pyx_tuple__34 __pyx_mstate_global->__pyx_tuple__34
#define __pyx_tuple__35 __pyx_mstate_global->__pyx_tuple__35
#define __pyx_tuple__36 __pyx_mstate_global->__pyx_tuple__36
#define __pyx_tuple__37 __pyx_mstate_global->__pyx_tuple__37
#define __pyx_tuple__38 __pyx_mstate_global->__pyx_tuple__38
#define __pyx_tuple__39 __pyx_mstate_global->__pyx_tuple__39
#define __pyx_tuple__40 __pyx_mstate_global->__pyx_tuple__40
#define __pyx_tuple__41 __pyx_mstate_global->__pyx_tuple__41
#define __pyx_tuple__42 __pyx_mstate_global->__pyx_tuple__42
#define __pyx_tuple__43 __pyx_mstate_global->__pyx_tuple__43
#define __pyx_tuple__44 __pyx_mstate_global->__pyx_tuple__44
#define __pyx_tuple__45 __pyx_mstate_global->__pyx_tuple__45
#define __pyx_tuple__46 __pyx_mstate_global->__pyx_tuple__46
#define __pyx_tuple__47 __pyx_mstate_global->__pyx_tuple__47
#define __pyx_tuple__48 __pyx_mstate_global->__pyx_tuple__48
#define __pyx_tuple__49 __pyx_mstate_global->__pyx_tuple__49
#define __pyx_tuple__50 __pyx_mstate_global->__pyx_tuple__50
#define __pyx_tuple__51 __pyx_mstate_global->__pyx_tuple__51
#define __pyx_tuple__52 __pyx_mstate_global->__pyx_tuple__52
#define __pyx_tuple__53 __pyx_mstate_global->__pyx_tuple__53
#define __pyx_tuple__54 __pyx_mstate_global->__pyx_tuple__54
#define __pyx_tuple__55 __pyx_mstate_global->__pyx_tuple__55
#define __pyx_tuple__56 __pyx_mstate_global->__pyx_tuple__56
#define __pyx_tuple__57 __pyx_mstate_global->__pyx_tuple__57
#define __pyx_tuple__58 __pyx_mstate_global->__pyx_tuple__58
#define __pyx_tuple__59 __pyx_mstate_global->__pyx_tuple__59
#define __pyx_tuple__60 __pyx_mstate_global->__pyx_tuple__60
#define __pyx_tuple__61 __pyx_mstate_global->__pyx_tuple__61
#define __pyx_tuple__62 __pyx_mstate_global->__pyx_tuple__62
#define __pyx_tuple__63 __pyx_mstate_global->__pyx_tuple__63
#define __pyx_tuple__64 __pyx_mstate_global->__pyx_tuple__64
#define __pyx_tuple__65 __pyx_mstate_global->__pyx_tuple__65
#define __pyx_tuple__66 __pyx_mstate_global->__pyx_tuple__66
#define __pyx_tuple__67 __pyx_mstate_global->__pyx_tuple__67
#define __pyx_tuple__68 __pyx_mstate_global->__pyx_tuple__68
#define __pyx_tuple__69 __pyx_mstate_global->__pyx_tuple__69
#define __pyx_tuple__70 __pyx_mstate_global->__pyx_tuple__70
#define __pyx_tuple__71 __pyx_mstate_global->__pyx_tuple__71
#define __pyx_tuple__74 __pyx_mstate_global->__pyx_tuple__74
#define __pyx_tuple__77 __pyx_mstate_global->__pyx_tuple__77
#define __pyx_codeobj__72 __pyx_mstate_global->__pyx_codeobj__72
#define __pyx_codeobj__73 __pyx_mstate_global->__pyx_codeobj__73
#define __pyx_codeobj__75 __pyx_mstate_global->__pyx_codeobj__75
#define __pyx_codeobj__76 __pyx_mstate_global->__pyx_codeobj__76
#define __pyx_codeobj__78 __pyx_mstate_global->__pyx_codeobj__78
/* #### Code section: module_code ### */

/* "FromPyStructUtility":12
 * 
 * @cname("__pyx_convert__from_py_struct____pyx_t_12pyretechnics_8cy_types_FuelModel")
 * cdef struct_type __pyx_convert__from_py_struct____pyx_t_12pyretechnics_8cy_types_FuelModel(obj) except *:             # <<<<<<<<<<<<<<
 *     cdef struct_type result
 *     if not PyMapping_Check(obj):
 */

static struct __pyx_t_12pyretechnics_8cy_types_FuelModel __pyx_convert__from_py_struct____pyx_t_12pyretechnics_8cy_types_FuelModel(PyObject *__pyx_v_obj) {
  struct __pyx_t_12pyretechnics_8cy_types_FuelModel __pyx_v_result;
  PyObject *__pyx_v_value = NULL;
  struct __pyx_t_12pyretechnics_8cy_types_FuelModel __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  float __pyx_t_10;
  __pyx_t_12pyretechnics_8cy_types_fclaarr __pyx_t_11;
  __pyx_t_12pyretechnics_8cy_types_fcatarr __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert__from_py_struct____pyx_t_12pyretechnics_8cy_types_FuelModel", 1);

  /* "FromPyStructUtility":14
 * cdef struct_type __pyx_convert__from_py_struct____pyx_t_12pyretechnics_8cy_types_FuelModel(obj) except *:
 *     cdef struct_type result
 *     if not PyMapping_Check(obj):             # <<<<<<<<<<<<<<
 *         __Pyx_RaiseUnexpectedTypeError(b"a mapping", obj)
 * 
 */
  __pyx_t_1 = (!PyMapping_Check(__pyx_v_obj));
  if (__pyx_t_1) {

    /* "FromPyStructUtility":15
 *     cdef struct_type result
 *     if not PyMapping_Check(obj):
 *         __Pyx_RaiseUnexpectedTypeError(b"a mapping", obj)             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
    __pyx_t_2 = __Pyx_RaiseUnexpectedTypeError(((char const *)"a mapping"), __pyx_v_obj); if (unlikely(__pyx_t_2 == ((int)0))) __PYX_ERR(0, 15, __pyx_L1_error)

    /* "FromPyStructUtility":14
 * cdef struct_type __pyx_convert__from_py_struct____pyx_t_12pyretechnics_8cy_types_FuelModel(obj) except *:
 *     cdef struct_type result
 *     if not PyMapping_Check(obj):             # <<<<<<<<<<<<<<
 *         __Pyx_RaiseUnexpectedTypeError(b"a mapping", obj)
 * 
 */
  }

  /* "FromPyStructUtility":17
 *         __Pyx_RaiseUnexpectedTypeError(b"a mapping", obj)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['number']
 *     except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    /*try:*/ {

      /* "FromPyStructUtility":18
 * 
 *     try:
 *         value = obj['number']             # <<<<<<<<<<<<<<
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'number'")
 */
      __pyx_t_6 = __Pyx_PyObject_Dict_GetItem(__pyx_v_obj, __pyx_n_s_number); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 18, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_v_value = __pyx_t_6;
      __pyx_t_6 = 0;

      /* "FromPyStructUtility":17
 *         __Pyx_RaiseUnexpectedTypeError(b"a mapping", obj)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['number']
 *     except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L9_try_end;
    __pyx_L4_error:;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "FromPyStructUtility":19
 *     try:
 *         value = obj['number']
 *     except KeyError:             # <<<<<<<<<<<<<<
 *         raise ValueError("No value specified for struct attribute 'number'")
 *     result.number = value
 */
    __pyx_t_2 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_2) {
      __Pyx_AddTraceback("FromPyStructUtility.__pyx_convert__from_py_struct____pyx_t_12pyretechnics_8cy_types_FuelModel", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8) < 0) __PYX_ERR(0, 19, __pyx_L6_except_error)
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);

      /* "FromPyStructUtility":20
 *         value = obj['number']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'number'")             # <<<<<<<<<<<<<<
 *     result.number = value
 *     try:
 */
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 20, __pyx_L6_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_Raise(__pyx_t_9, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __PYX_ERR(0, 20, __pyx_L6_except_error)
    }
    goto __pyx_L6_except_error;

    /* "FromPyStructUtility":17
 *         __Pyx_RaiseUnexpectedTypeError(b"a mapping", obj)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['number']
 *     except KeyError:
 */
    __pyx_L6_except_error:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    goto __pyx_L1_error;
    __pyx_L9_try_end:;
  }

  /* "FromPyStructUtility":21
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'number'")
 *     result.number = value             # <<<<<<<<<<<<<<
 *     try:
 *         value = obj['delta']
 */
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 21, __pyx_L1_error)
  __pyx_v_result.number = __pyx_t_2;

  /* "FromPyStructUtility":22
 *         raise ValueError("No value specified for struct attribute 'number'")
 *     result.number = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['delta']
 *     except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_4, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "FromPyStructUtility":23
 *     result.number = value
 *     try:
 *         value = obj['delta']             # <<<<<<<<<<<<<<
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'delta'")
 */
      __pyx_t_8 = __Pyx_PyObject_Dict_GetItem(__pyx_v_obj, __pyx_n_s_delta); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 23, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "FromPyStructUtility":22
 *         raise ValueError("No value specified for struct attribute 'number'")
 *     result.number = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['delta']
 *     except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L17_try_end;
    __pyx_L12_error:;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "FromPyStructUtility":24
 *     try:
 *         value = obj['delta']
 *     except KeyError:             # <<<<<<<<<<<<<<
 *         raise ValueError("No value specified for struct attribute 'delta'")
 *     result.delta = value
 */
    __pyx_t_2 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_2) {
      __Pyx_AddTraceback("FromPyStructUtility.__pyx_convert__from_py_struct____pyx_t_12pyretechnics_8cy_types_FuelModel", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_7, &__pyx_t_6) < 0) __PYX_ERR(0, 24, __pyx_L14_except_error)
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_6);

      /* "FromPyStructUtility":25
 *         value = obj['delta']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'delta'")             # <<<<<<<<<<<<<<
 *     result.delta = value
 *     try:
 */
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 25, __pyx_L14_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_Raise(__pyx_t_9, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __PYX_ERR(0, 25, __pyx_L14_except_error)
    }
    goto __pyx_L14_except_error;

    /* "FromPyStructUtility":22
 *         raise ValueError("No value specified for struct attribute 'number'")
 *     result.number = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['delta']
 *     except KeyError:
 */
    __pyx_L14_except_error:;
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_4, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L17_try_end:;
  }

  /* "FromPyStructUtility":26
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'delta'")
 *     result.delta = value             # <<<<<<<<<<<<<<
 *     try:
 *         value = obj['M_x']
 */
  __pyx_t_10 = __pyx_PyFloat_AsFloat(__pyx_v_value); if (unlikely((__pyx_t_10 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 26, __pyx_L1_error)
  __pyx_v_result.delta = __pyx_t_10;

  /* "FromPyStructUtility":27
 *         raise ValueError("No value specified for struct attribute 'delta'")
 *     result.delta = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['M_x']
 *     except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    /*try:*/ {

      /* "FromPyStructUtility":28
 *     result.delta = value
 *     try:
 *         value = obj['M_x']             # <<<<<<<<<<<<<<
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'M_x'")
 */
      __pyx_t_6 = __Pyx_PyObject_Dict_GetItem(__pyx_v_obj, __pyx_n_s_M_x); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 28, __pyx_L20_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "FromPyStructUtility":27
 *         raise ValueError("No value specified for struct attribute 'delta'")
 *     result.delta = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['M_x']
 *     except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L25_try_end;
    __pyx_L20_error:;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "FromPyStructUtility":29
 *     try:
 *         value = obj['M_x']
 *     except KeyError:             # <<<<<<<<<<<<<<
 *         raise ValueError("No value specified for struct attribute 'M_x'")
 *     result.M_x = value
 */
    __pyx_t_2 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_2) {
      __Pyx_AddTraceback("FromPyStructUtility.__pyx_convert__from_py_struct____pyx_t_12pyretechnics_8cy_types_FuelModel", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8) < 0) __PYX_ERR(0, 29, __pyx_L22_except_error)
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);

      /* "FromPyStructUtility":30
 *         value = obj['M_x']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'M_x'")             # <<<<<<<<<<<<<<
 *     result.M_x = value
 *     try:
 */
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 30, __pyx_L22_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_Raise(__pyx_t_9, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __PYX_ERR(0, 30, __pyx_L22_except_error)
    }
    goto __pyx_L22_except_error;

    /* "FromPyStructUtility":27
 *         raise ValueError("No value specified for struct attribute 'delta'")
 *     result.delta = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['M_x']
 *     except KeyError:
 */
    __pyx_L22_except_error:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    goto __pyx_L1_error;
    __pyx_L25_try_end:;
  }

  /* "FromPyStructUtility":31
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'M_x'")
 *     result.M_x = value             # <<<<<<<<<<<<<<
 *     try:
 *         value = obj['M_f']
 */
  __pyx_t_11 = __pyx_convert__from_py___pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float(__pyx_v_value); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 31, __pyx_L1_error)
  __pyx_v_result.M_x = __pyx_t_11;

  /* "FromPyStructUtility":32
 *         raise ValueError("No value specified for struct attribute 'M_x'")
 *     result.M_x = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['M_f']
 *     except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_4, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "FromPyStructUtility":33
 *     result.M_x = value
 *     try:
 *         value = obj['M_f']             # <<<<<<<<<<<<<<
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'M_f'")
 */
      __pyx_t_8 = __Pyx_PyObject_Dict_GetItem(__pyx_v_obj, __pyx_n_s_M_f); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 33, __pyx_L28_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "FromPyStructUtility":32
 *         raise ValueError("No value specified for struct attribute 'M_x'")
 *     result.M_x = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['M_f']
 *     except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L33_try_end;
    __pyx_L28_error:;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "FromPyStructUtility":34
 *     try:
 *         value = obj['M_f']
 *     except KeyError:             # <<<<<<<<<<<<<<
 *         raise ValueError("No value specified for struct attribute 'M_f'")
 *     result.M_f = value
 */
    __pyx_t_2 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_2) {
      __Pyx_AddTraceback("FromPyStructUtility.__pyx_convert__from_py_struct____pyx_t_12pyretechnics_8cy_types_FuelModel", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_7, &__pyx_t_6) < 0) __PYX_ERR(0, 34, __pyx_L30_except_error)
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_6);

      /* "FromPyStructUtility":35
 *         value = obj['M_f']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'M_f'")             # <<<<<<<<<<<<<<
 *     result.M_f = value
 *     try:
 */
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 35, __pyx_L30_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_Raise(__pyx_t_9, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __PYX_ERR(0, 35, __pyx_L30_except_error)
    }
    goto __pyx_L30_except_error;

    /* "FromPyStructUtility":32
 *         raise ValueError("No value specified for struct attribute 'M_x'")
 *     result.M_x = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['M_f']
 *     except KeyError:
 */
    __pyx_L30_except_error:;
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_4, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L33_try_end:;
  }

  /* "FromPyStructUtility":36
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'M_f'")
 *     result.M_f = value             # <<<<<<<<<<<<<<
 *     try:
 *         value = obj['w_o']
 */
  __pyx_t_11 = __pyx_convert__from_py___pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float(__pyx_v_value); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 36, __pyx_L1_error)
  __pyx_v_result.M_f = __pyx_t_11;

  /* "FromPyStructUtility":37
 *         raise ValueError("No value specified for struct attribute 'M_f'")
 *     result.M_f = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['w_o']
 *     except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    /*try:*/ {

      /* "FromPyStructUtility":38
 *     result.M_f = value
 *     try:
 *         value = obj['w_o']             # <<<<<<<<<<<<<<
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'w_o'")
 */
      __pyx_t_6 = __Pyx_PyObject_Dict_GetItem(__pyx_v_obj, __pyx_n_s_w_o); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 38, __pyx_L36_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "FromPyStructUtility":37
 *         raise ValueError("No value specified for struct attribute 'M_f'")
 *     result.M_f = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['w_o']
 *     except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L41_try_end;
    __pyx_L36_error:;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "FromPyStructUtility":39
 *     try:
 *         value = obj['w_o']
 *     except KeyError:             # <<<<<<<<<<<<<<
 *         raise ValueError("No value specified for struct attribute 'w_o'")
 *     result.w_o = value
 */
    __pyx_t_2 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_2) {
      __Pyx_AddTraceback("FromPyStructUtility.__pyx_convert__from_py_struct____pyx_t_12pyretechnics_8cy_types_FuelModel", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8) < 0) __PYX_ERR(0, 39, __pyx_L38_except_error)
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);

      /* "FromPyStructUtility":40
 *         value = obj['w_o']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'w_o'")             # <<<<<<<<<<<<<<
 *     result.w_o = value
 *     try:
 */
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 40, __pyx_L38_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_Raise(__pyx_t_9, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __PYX_ERR(0, 40, __pyx_L38_except_error)
    }
    goto __pyx_L38_except_error;

    /* "FromPyStructUtility":37
 *         raise ValueError("No value specified for struct attribute 'M_f'")
 *     result.M_f = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['w_o']
 *     except KeyError:
 */
    __pyx_L38_except_error:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    goto __pyx_L1_error;
    __pyx_L41_try_end:;
  }

  /* "FromPyStructUtility":41
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'w_o'")
 *     result.w_o = value             # <<<<<<<<<<<<<<
 *     try:
 *         value = obj['sigma']
 */
  __pyx_t_11 = __pyx_convert__from_py___pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float(__pyx_v_value); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 41, __pyx_L1_error)
  __pyx_v_result.w_o = __pyx_t_11;

  /* "FromPyStructUtility":42
 *         raise ValueError("No value specified for struct attribute 'w_o'")
 *     result.w_o = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['sigma']
 *     except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_4, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "FromPyStructUtility":43
 *     result.w_o = value
 *     try:
 *         value = obj['sigma']             # <<<<<<<<<<<<<<
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'sigma'")
 */
      __pyx_t_8 = __Pyx_PyObject_Dict_GetItem(__pyx_v_obj, __pyx_n_s_sigma); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 43, __pyx_L44_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "FromPyStructUtility":42
 *         raise ValueError("No value specified for struct attribute 'w_o'")
 *     result.w_o = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['sigma']
 *     except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L49_try_end;
    __pyx_L44_error:;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "FromPyStructUtility":44
 *     try:
 *         value = obj['sigma']
 *     except KeyError:             # <<<<<<<<<<<<<<
 *         raise ValueError("No value specified for struct attribute 'sigma'")
 *     result.sigma = value
 */
    __pyx_t_2 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_2) {
      __Pyx_AddTraceback("FromPyStructUtility.__pyx_convert__from_py_struct____pyx_t_12pyretechnics_8cy_types_FuelModel", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_7, &__pyx_t_6) < 0) __PYX_ERR(0, 44, __pyx_L46_except_error)
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_6);

      /* "FromPyStructUtility":45
 *         value = obj['sigma']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'sigma'")             # <<<<<<<<<<<<<<
 *     result.sigma = value
 *     try:
 */
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 45, __pyx_L46_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_Raise(__pyx_t_9, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __PYX_ERR(0, 45, __pyx_L46_except_error)
    }
    goto __pyx_L46_except_error;

    /* "FromPyStructUtility":42
 *         raise ValueError("No value specified for struct attribute 'w_o'")
 *     result.w_o = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['sigma']
 *     except KeyError:
 */
    __pyx_L46_except_error:;
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_4, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L49_try_end:;
  }

  /* "FromPyStructUtility":46
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'sigma'")
 *     result.sigma = value             # <<<<<<<<<<<<<<
 *     try:
 *         value = obj['h']
 */
  __pyx_t_11 = __pyx_convert__from_py___pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float(__pyx_v_value); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 46, __pyx_L1_error)
  __pyx_v_result.sigma = __pyx_t_11;

  /* "FromPyStructUtility":47
 *         raise ValueError("No value specified for struct attribute 'sigma'")
 *     result.sigma = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['h']
 *     except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    /*try:*/ {

      /* "FromPyStructUtility":48
 *     result.sigma = value
 *     try:
 *         value = obj['h']             # <<<<<<<<<<<<<<
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'h'")
 */
      __pyx_t_6 = __Pyx_PyObject_Dict_GetItem(__pyx_v_obj, __pyx_n_s_h); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 48, __pyx_L52_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "FromPyStructUtility":47
 *         raise ValueError("No value specified for struct attribute 'sigma'")
 *     result.sigma = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['h']
 *     except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L57_try_end;
    __pyx_L52_error:;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "FromPyStructUtility":49
 *     try:
 *         value = obj['h']
 *     except KeyError:             # <<<<<<<<<<<<<<
 *         raise ValueError("No value specified for struct attribute 'h'")
 *     result.h = value
 */
    __pyx_t_2 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_2) {
      __Pyx_AddTraceback("FromPyStructUtility.__pyx_convert__from_py_struct____pyx_t_12pyretechnics_8cy_types_FuelModel", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8) < 0) __PYX_ERR(0, 49, __pyx_L54_except_error)
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);

      /* "FromPyStructUtility":50
 *         value = obj['h']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'h'")             # <<<<<<<<<<<<<<
 *     result.h = value
 *     try:
 */
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 50, __pyx_L54_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_Raise(__pyx_t_9, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __PYX_ERR(0, 50, __pyx_L54_except_error)
    }
    goto __pyx_L54_except_error;

    /* "FromPyStructUtility":47
 *         raise ValueError("No value specified for struct attribute 'sigma'")
 *     result.sigma = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['h']
 *     except KeyError:
 */
    __pyx_L54_except_error:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    goto __pyx_L1_error;
    __pyx_L57_try_end:;
  }

  /* "FromPyStructUtility":51
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'h'")
 *     result.h = value             # <<<<<<<<<<<<<<
 *     try:
 *         value = obj['rho_p']
 */
  __pyx_t_11 = __pyx_convert__from_py___pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float(__pyx_v_value); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 51, __pyx_L1_error)
  __pyx_v_result.h = __pyx_t_11;

  /* "FromPyStructUtility":52
 *         raise ValueError("No value specified for struct attribute 'h'")
 *     result.h = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['rho_p']
 *     except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_4, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "FromPyStructUtility":53
 *     result.h = value
 *     try:
 *         value = obj['rho_p']             # <<<<<<<<<<<<<<
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'rho_p'")
 */
      __pyx_t_8 = __Pyx_PyObject_Dict_GetItem(__pyx_v_obj, __pyx_n_s_rho_p); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 53, __pyx_L60_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "FromPyStructUtility":52
 *         raise ValueError("No value specified for struct attribute 'h'")
 *     result.h = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['rho_p']
 *     except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L65_try_end;
    __pyx_L60_error:;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "FromPyStructUtility":54
 *     try:
 *         value = obj['rho_p']
 *     except KeyError:             # <<<<<<<<<<<<<<
 *         raise ValueError("No value specified for struct attribute 'rho_p'")
 *     result.rho_p = value
 */
    __pyx_t_2 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_2) {
      __Pyx_AddTraceback("FromPyStructUtility.__pyx_convert__from_py_struct____pyx_t_12pyretechnics_8cy_types_FuelModel", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_7, &__pyx_t_6) < 0) __PYX_ERR(0, 54, __pyx_L62_except_error)
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_6);

      /* "FromPyStructUtility":55
 *         value = obj['rho_p']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'rho_p'")             # <<<<<<<<<<<<<<
 *     result.rho_p = value
 *     try:
 */
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 55, __pyx_L62_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_Raise(__pyx_t_9, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __PYX_ERR(0, 55, __pyx_L62_except_error)
    }
    goto __pyx_L62_except_error;

    /* "FromPyStructUtility":52
 *         raise ValueError("No value specified for struct attribute 'h'")
 *     result.h = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['rho_p']
 *     except KeyError:
 */
    __pyx_L62_except_error:;
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_4, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L65_try_end:;
  }

  /* "FromPyStructUtility":56
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'rho_p'")
 *     result.rho_p = value             # <<<<<<<<<<<<<<
 *     try:
 *         value = obj['S_T']
 */
  __pyx_t_11 = __pyx_convert__from_py___pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float(__pyx_v_value); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 56, __pyx_L1_error)
  __pyx_v_result.rho_p = __pyx_t_11;

  /* "FromPyStructUtility":57
 *         raise ValueError("No value specified for struct attribute 'rho_p'")
 *     result.rho_p = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['S_T']
 *     except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    /*try:*/ {

      /* "FromPyStructUtility":58
 *     result.rho_p = value
 *     try:
 *         value = obj['S_T']             # <<<<<<<<<<<<<<
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'S_T'")
 */
      __pyx_t_6 = __Pyx_PyObject_Dict_GetItem(__pyx_v_obj, __pyx_n_s_S_T); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 58, __pyx_L68_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "FromPyStructUtility":57
 *         raise ValueError("No value specified for struct attribute 'rho_p'")
 *     result.rho_p = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['S_T']
 *     except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L73_try_end;
    __pyx_L68_error:;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "FromPyStructUtility":59
 *     try:
 *         value = obj['S_T']
 *     except KeyError:             # <<<<<<<<<<<<<<
 *         raise ValueError("No value specified for struct attribute 'S_T'")
 *     result.S_T = value
 */
    __pyx_t_2 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_2) {
      __Pyx_AddTraceback("FromPyStructUtility.__pyx_convert__from_py_struct____pyx_t_12pyretechnics_8cy_types_FuelModel", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8) < 0) __PYX_ERR(0, 59, __pyx_L70_except_error)
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);

      /* "FromPyStructUtility":60
 *         value = obj['S_T']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'S_T'")             # <<<<<<<<<<<<<<
 *     result.S_T = value
 *     try:
 */
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 60, __pyx_L70_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_Raise(__pyx_t_9, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __PYX_ERR(0, 60, __pyx_L70_except_error)
    }
    goto __pyx_L70_except_error;

    /* "FromPyStructUtility":57
 *         raise ValueError("No value specified for struct attribute 'rho_p'")
 *     result.rho_p = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['S_T']
 *     except KeyError:
 */
    __pyx_L70_except_error:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    goto __pyx_L1_error;
    __pyx_L73_try_end:;
  }

  /* "FromPyStructUtility":61
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'S_T'")
 *     result.S_T = value             # <<<<<<<<<<<<<<
 *     try:
 *         value = obj['S_e']
 */
  __pyx_t_11 = __pyx_convert__from_py___pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float(__pyx_v_value); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 61, __pyx_L1_error)
  __pyx_v_result.S_T = __pyx_t_11;

  /* "FromPyStructUtility":62
 *         raise ValueError("No value specified for struct attribute 'S_T'")
 *     result.S_T = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['S_e']
 *     except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_4, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "FromPyStructUtility":63
 *     result.S_T = value
 *     try:
 *         value = obj['S_e']             # <<<<<<<<<<<<<<
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'S_e'")
 */
      __pyx_t_8 = __Pyx_PyObject_Dict_GetItem(__pyx_v_obj, __pyx_n_s_S_e); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 63, __pyx_L76_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "FromPyStructUtility":62
 *         raise ValueError("No value specified for struct attribute 'S_T'")
 *     result.S_T = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['S_e']
 *     except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L81_try_end;
    __pyx_L76_error:;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "FromPyStructUtility":64
 *     try:
 *         value = obj['S_e']
 *     except KeyError:             # <<<<<<<<<<<<<<
 *         raise ValueError("No value specified for struct attribute 'S_e'")
 *     result.S_e = value
 */
    __pyx_t_2 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_2) {
      __Pyx_AddTraceback("FromPyStructUtility.__pyx_convert__from_py_struct____pyx_t_12pyretechnics_8cy_types_FuelModel", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_7, &__pyx_t_6) < 0) __PYX_ERR(0, 64, __pyx_L78_except_error)
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_6);

      /* "FromPyStructUtility":65
 *         value = obj['S_e']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'S_e'")             # <<<<<<<<<<<<<<
 *     result.S_e = value
 *     try:
 */
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 65, __pyx_L78_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_Raise(__pyx_t_9, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __PYX_ERR(0, 65, __pyx_L78_except_error)
    }
    goto __pyx_L78_except_error;

    /* "FromPyStructUtility":62
 *         raise ValueError("No value specified for struct attribute 'S_T'")
 *     result.S_T = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['S_e']
 *     except KeyError:
 */
    __pyx_L78_except_error:;
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_4, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L81_try_end:;
  }

  /* "FromPyStructUtility":66
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'S_e'")
 *     result.S_e = value             # <<<<<<<<<<<<<<
 *     try:
 *         value = obj['dynamic']
 */
  __pyx_t_11 = __pyx_convert__from_py___pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float(__pyx_v_value); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 66, __pyx_L1_error)
  __pyx_v_result.S_e = __pyx_t_11;

  /* "FromPyStructUtility":67
 *         raise ValueError("No value specified for struct attribute 'S_e'")
 *     result.S_e = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['dynamic']
 *     except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    /*try:*/ {

      /* "FromPyStructUtility":68
 *     result.S_e = value
 *     try:
 *         value = obj['dynamic']             # <<<<<<<<<<<<<<
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'dynamic'")
 */
      __pyx_t_6 = __Pyx_PyObject_Dict_GetItem(__pyx_v_obj, __pyx_n_s_dynamic); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 68, __pyx_L84_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "FromPyStructUtility":67
 *         raise ValueError("No value specified for struct attribute 'S_e'")
 *     result.S_e = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['dynamic']
 *     except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L89_try_end;
    __pyx_L84_error:;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "FromPyStructUtility":69
 *     try:
 *         value = obj['dynamic']
 *     except KeyError:             # <<<<<<<<<<<<<<
 *         raise ValueError("No value specified for struct attribute 'dynamic'")
 *     result.dynamic = value
 */
    __pyx_t_2 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_2) {
      __Pyx_AddTraceback("FromPyStructUtility.__pyx_convert__from_py_struct____pyx_t_12pyretechnics_8cy_types_FuelModel", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8) < 0) __PYX_ERR(0, 69, __pyx_L86_except_error)
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);

      /* "FromPyStructUtility":70
 *         value = obj['dynamic']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'dynamic'")             # <<<<<<<<<<<<<<
 *     result.dynamic = value
 *     try:
 */
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 70, __pyx_L86_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_Raise(__pyx_t_9, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __PYX_ERR(0, 70, __pyx_L86_except_error)
    }
    goto __pyx_L86_except_error;

    /* "FromPyStructUtility":67
 *         raise ValueError("No value specified for struct attribute 'S_e'")
 *     result.S_e = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['dynamic']
 *     except KeyError:
 */
    __pyx_L86_except_error:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    goto __pyx_L1_error;
    __pyx_L89_try_end:;
  }

  /* "FromPyStructUtility":71
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'dynamic'")
 *     result.dynamic = value             # <<<<<<<<<<<<<<
 *     try:
 *         value = obj['burnable']
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 71, __pyx_L1_error)
  __pyx_v_result.dynamic = __pyx_t_1;

  /* "FromPyStructUtility":72
 *         raise ValueError("No value specified for struct attribute 'dynamic'")
 *     result.dynamic = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['burnable']
 *     except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_4, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "FromPyStructUtility":73
 *     result.dynamic = value
 *     try:
 *         value = obj['burnable']             # <<<<<<<<<<<<<<
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'burnable'")
 */
      __pyx_t_8 = __Pyx_PyObject_Dict_GetItem(__pyx_v_obj, __pyx_n_s_burnable); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 73, __pyx_L92_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "FromPyStructUtility":72
 *         raise ValueError("No value specified for struct attribute 'dynamic'")
 *     result.dynamic = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['burnable']
 *     except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L97_try_end;
    __pyx_L92_error:;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "FromPyStructUtility":74
 *     try:
 *         value = obj['burnable']
 *     except KeyError:             # <<<<<<<<<<<<<<
 *         raise ValueError("No value specified for struct attribute 'burnable'")
 *     result.burnable = value
 */
    __pyx_t_2 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_2) {
      __Pyx_AddTraceback("FromPyStructUtility.__pyx_convert__from_py_struct____pyx_t_12pyretechnics_8cy_types_FuelModel", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_7, &__pyx_t_6) < 0) __PYX_ERR(0, 74, __pyx_L94_except_error)
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_6);

      /* "FromPyStructUtility":75
 *         value = obj['burnable']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'burnable'")             # <<<<<<<<<<<<<<
 *     result.burnable = value
 *     try:
 */
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 75, __pyx_L94_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_Raise(__pyx_t_9, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __PYX_ERR(0, 75, __pyx_L94_except_error)
    }
    goto __pyx_L94_except_error;

    /* "FromPyStructUtility":72
 *         raise ValueError("No value specified for struct attribute 'dynamic'")
 *     result.dynamic = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['burnable']
 *     except KeyError:
 */
    __pyx_L94_except_error:;
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_4, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L97_try_end:;
  }

  /* "FromPyStructUtility":76
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'burnable'")
 *     result.burnable = value             # <<<<<<<<<<<<<<
 *     try:
 *         value = obj['exp_A_sigma']
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 76, __pyx_L1_error)
  __pyx_v_result.burnable = __pyx_t_1;

  /* "FromPyStructUtility":77
 *         raise ValueError("No value specified for struct attribute 'burnable'")
 *     result.burnable = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['exp_A_sigma']
 *     except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    /*try:*/ {

      /* "FromPyStructUtility":78
 *     result.burnable = value
 *     try:
 *         value = obj['exp_A_sigma']             # <<<<<<<<<<<<<<
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'exp_A_sigma'")
 */
      __pyx_t_6 = __Pyx_PyObject_Dict_GetItem(__pyx_v_obj, __pyx_n_s_exp_A_sigma); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 78, __pyx_L100_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "FromPyStructUtility":77
 *         raise ValueError("No value specified for struct attribute 'burnable'")
 *     result.burnable = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['exp_A_sigma']
 *     except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L105_try_end;
    __pyx_L100_error:;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "FromPyStructUtility":79
 *     try:
 *         value = obj['exp_A_sigma']
 *     except KeyError:             # <<<<<<<<<<<<<<
 *         raise ValueError("No value specified for struct attribute 'exp_A_sigma'")
 *     result.exp_A_sigma = value
 */
    __pyx_t_2 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_2) {
      __Pyx_AddTraceback("FromPyStructUtility.__pyx_convert__from_py_struct____pyx_t_12pyretechnics_8cy_types_FuelModel", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8) < 0) __PYX_ERR(0, 79, __pyx_L102_except_error)
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);

      /* "FromPyStructUtility":80
 *         value = obj['exp_A_sigma']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'exp_A_sigma'")             # <<<<<<<<<<<<<<
 *     result.exp_A_sigma = value
 *     try:
 */
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 80, __pyx_L102_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_Raise(__pyx_t_9, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __PYX_ERR(0, 80, __pyx_L102_except_error)
    }
    goto __pyx_L102_except_error;

    /* "FromPyStructUtility":77
 *         raise ValueError("No value specified for struct attribute 'burnable'")
 *     result.burnable = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['exp_A_sigma']
 *     except KeyError:
 */
    __pyx_L102_except_error:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    goto __pyx_L1_error;
    __pyx_L105_try_end:;
  }

  /* "FromPyStructUtility":81
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'exp_A_sigma'")
 *     result.exp_A_sigma = value             # <<<<<<<<<<<<<<
 *     try:
 *         value = obj['firemod_size_classes']
 */
  __pyx_t_11 = __pyx_convert__from_py___pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float(__pyx_v_value); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 81, __pyx_L1_error)
  __pyx_v_result.exp_A_sigma = __pyx_t_11;

  /* "FromPyStructUtility":82
 *         raise ValueError("No value specified for struct attribute 'exp_A_sigma'")
 *     result.exp_A_sigma = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['firemod_size_classes']
 *     except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_4, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "FromPyStructUtility":83
 *     result.exp_A_sigma = value
 *     try:
 *         value = obj['firemod_size_classes']             # <<<<<<<<<<<<<<
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'firemod_size_classes'")
 */
      __pyx_t_8 = __Pyx_PyObject_Dict_GetItem(__pyx_v_obj, __pyx_n_s_firemod_size_classes); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 83, __pyx_L108_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "FromPyStructUtility":82
 *         raise ValueError("No value specified for struct attribute 'exp_A_sigma'")
 *     result.exp_A_sigma = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['firemod_size_classes']
 *     except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L113_try_end;
    __pyx_L108_error:;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "FromPyStructUtility":84
 *     try:
 *         value = obj['firemod_size_classes']
 *     except KeyError:             # <<<<<<<<<<<<<<
 *         raise ValueError("No value specified for struct attribute 'firemod_size_classes'")
 *     result.firemod_size_classes = value
 */
    __pyx_t_2 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_2) {
      __Pyx_AddTraceback("FromPyStructUtility.__pyx_convert__from_py_struct____pyx_t_12pyretechnics_8cy_types_FuelModel", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_7, &__pyx_t_6) < 0) __PYX_ERR(0, 84, __pyx_L110_except_error)
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_6);

      /* "FromPyStructUtility":85
 *         value = obj['firemod_size_classes']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'firemod_size_classes'")             # <<<<<<<<<<<<<<
 *     result.firemod_size_classes = value
 *     try:
 */
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 85, __pyx_L110_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_Raise(__pyx_t_9, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __PYX_ERR(0, 85, __pyx_L110_except_error)
    }
    goto __pyx_L110_except_error;

    /* "FromPyStructUtility":82
 *         raise ValueError("No value specified for struct attribute 'exp_A_sigma'")
 *     result.exp_A_sigma = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['firemod_size_classes']
 *     except KeyError:
 */
    __pyx_L110_except_error:;
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_4, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L113_try_end:;
  }

  /* "FromPyStructUtility":86
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'firemod_size_classes'")
 *     result.firemod_size_classes = value             # <<<<<<<<<<<<<<
 *     try:
 *         value = obj['f_ij']
 */
  __pyx_t_11 = __pyx_convert__from_py___pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float(__pyx_v_value); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 86, __pyx_L1_error)
  __pyx_v_result.firemod_size_classes = __pyx_t_11;

  /* "FromPyStructUtility":87
 *         raise ValueError("No value specified for struct attribute 'firemod_size_classes'")
 *     result.firemod_size_classes = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['f_ij']
 *     except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    /*try:*/ {

      /* "FromPyStructUtility":88
 *     result.firemod_size_classes = value
 *     try:
 *         value = obj['f_ij']             # <<<<<<<<<<<<<<
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'f_ij'")
 */
      __pyx_t_6 = __Pyx_PyObject_Dict_GetItem(__pyx_v_obj, __pyx_n_s_f_ij); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 88, __pyx_L116_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "FromPyStructUtility":87
 *         raise ValueError("No value specified for struct attribute 'firemod_size_classes'")
 *     result.firemod_size_classes = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['f_ij']
 *     except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L121_try_end;
    __pyx_L116_error:;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "FromPyStructUtility":89
 *     try:
 *         value = obj['f_ij']
 *     except KeyError:             # <<<<<<<<<<<<<<
 *         raise ValueError("No value specified for struct attribute 'f_ij'")
 *     result.f_ij = value
 */
    __pyx_t_2 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_2) {
      __Pyx_AddTraceback("FromPyStructUtility.__pyx_convert__from_py_struct____pyx_t_12pyretechnics_8cy_types_FuelModel", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8) < 0) __PYX_ERR(0, 89, __pyx_L118_except_error)
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);

      /* "FromPyStructUtility":90
 *         value = obj['f_ij']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'f_ij'")             # <<<<<<<<<<<<<<
 *     result.f_ij = value
 *     try:
 */
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 90, __pyx_L118_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_Raise(__pyx_t_9, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __PYX_ERR(0, 90, __pyx_L118_except_error)
    }
    goto __pyx_L118_except_error;

    /* "FromPyStructUtility":87
 *         raise ValueError("No value specified for struct attribute 'firemod_size_classes'")
 *     result.firemod_size_classes = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['f_ij']
 *     except KeyError:
 */
    __pyx_L118_except_error:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    goto __pyx_L1_error;
    __pyx_L121_try_end:;
  }

  /* "FromPyStructUtility":91
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'f_ij'")
 *     result.f_ij = value             # <<<<<<<<<<<<<<
 *     try:
 *         value = obj['f_i']
 */
  __pyx_t_11 = __pyx_convert__from_py___pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float(__pyx_v_value); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 91, __pyx_L1_error)
  __pyx_v_result.f_ij = __pyx_t_11;

  /* "FromPyStructUtility":92
 *         raise ValueError("No value specified for struct attribute 'f_ij'")
 *     result.f_ij = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['f_i']
 *     except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_4, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "FromPyStructUtility":93
 *     result.f_ij = value
 *     try:
 *         value = obj['f_i']             # <<<<<<<<<<<<<<
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'f_i'")
 */
      __pyx_t_8 = __Pyx_PyObject_Dict_GetItem(__pyx_v_obj, __pyx_n_s_f_i); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 93, __pyx_L124_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "FromPyStructUtility":92
 *         raise ValueError("No value specified for struct attribute 'f_ij'")
 *     result.f_ij = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['f_i']
 *     except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L129_try_end;
    __pyx_L124_error:;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "FromPyStructUtility":94
 *     try:
 *         value = obj['f_i']
 *     except KeyError:             # <<<<<<<<<<<<<<
 *         raise ValueError("No value specified for struct attribute 'f_i'")
 *     result.f_i = value
 */
    __pyx_t_2 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_2) {
      __Pyx_AddTraceback("FromPyStructUtility.__pyx_convert__from_py_struct____pyx_t_12pyretechnics_8cy_types_FuelModel", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_7, &__pyx_t_6) < 0) __PYX_ERR(0, 94, __pyx_L126_except_error)
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_6);

      /* "FromPyStructUtility":95
 *         value = obj['f_i']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'f_i'")             # <<<<<<<<<<<<<<
 *     result.f_i = value
 *     try:
 */
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 95, __pyx_L126_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_Raise(__pyx_t_9, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __PYX_ERR(0, 95, __pyx_L126_except_error)
    }
    goto __pyx_L126_except_error;

    /* "FromPyStructUtility":92
 *         raise ValueError("No value specified for struct attribute 'f_ij'")
 *     result.f_ij = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['f_i']
 *     except KeyError:
 */
    __pyx_L126_except_error:;
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_4, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L129_try_end:;
  }

  /* "FromPyStructUtility":96
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'f_i'")
 *     result.f_i = value             # <<<<<<<<<<<<<<
 *     try:
 *         value = obj['g_ij']
 */
  __pyx_t_12 = __pyx_convert__from_py___pyx_ctuple_float__and_float(__pyx_v_value); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 96, __pyx_L1_error)
  __pyx_v_result.f_i = __pyx_t_12;

  /* "FromPyStructUtility":97
 *         raise ValueError("No value specified for struct attribute 'f_i'")
 *     result.f_i = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['g_ij']
 *     except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    /*try:*/ {

      /* "FromPyStructUtility":98
 *     result.f_i = value
 *     try:
 *         value = obj['g_ij']             # <<<<<<<<<<<<<<
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'g_ij'")
 */
      __pyx_t_6 = __Pyx_PyObject_Dict_GetItem(__pyx_v_obj, __pyx_n_s_g_ij); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 98, __pyx_L132_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "FromPyStructUtility":97
 *         raise ValueError("No value specified for struct attribute 'f_i'")
 *     result.f_i = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['g_ij']
 *     except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L137_try_end;
    __pyx_L132_error:;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "FromPyStructUtility":99
 *     try:
 *         value = obj['g_ij']
 *     except KeyError:             # <<<<<<<<<<<<<<
 *         raise ValueError("No value specified for struct attribute 'g_ij'")
 *     result.g_ij = value
 */
    __pyx_t_2 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_2) {
      __Pyx_AddTraceback("FromPyStructUtility.__pyx_convert__from_py_struct____pyx_t_12pyretechnics_8cy_types_FuelModel", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8) < 0) __PYX_ERR(0, 99, __pyx_L134_except_error)
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);

      /* "FromPyStructUtility":100
 *         value = obj['g_ij']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'g_ij'")             # <<<<<<<<<<<<<<
 *     result.g_ij = value
 *     return result
 */
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__17, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 100, __pyx_L134_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_Raise(__pyx_t_9, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __PYX_ERR(0, 100, __pyx_L134_except_error)
    }
    goto __pyx_L134_except_error;

    /* "FromPyStructUtility":97
 *         raise ValueError("No value specified for struct attribute 'f_i'")
 *     result.f_i = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['g_ij']
 *     except KeyError:
 */
    __pyx_L134_except_error:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    goto __pyx_L1_error;
    __pyx_L137_try_end:;
  }

  /* "FromPyStructUtility":101
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'g_ij'")
 *     result.g_ij = value             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
  __pyx_t_11 = __pyx_convert__from_py___pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float(__pyx_v_value); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 101, __pyx_L1_error)
  __pyx_v_result.g_ij = __pyx_t_11;

  /* "FromPyStructUtility":102
 *         raise ValueError("No value specified for struct attribute 'g_ij'")
 *     result.g_ij = value
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "FromPyStructUtility":12
 * 
 * @cname("__pyx_convert__from_py_struct____pyx_t_12pyretechnics_8cy_types_FuelModel")
 * cdef struct_type __pyx_convert__from_py_struct____pyx_t_12pyretechnics_8cy_types_FuelModel(obj) except *:             # <<<<<<<<<<<<<<
 *     cdef struct_type result
 *     if not PyMapping_Check(obj):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("FromPyStructUtility.__pyx_convert__from_py_struct____pyx_t_12pyretechnics_8cy_types_FuelModel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyretechnics/fuel_models.py":110
 * 
 * 
 * @cy.cfunc             # <<<<<<<<<<<<<<
 * @cy.inline
 * @cy.exceptval(check=False)
 */

static CYTHON_INLINE int __pyx_f_12pyretechnics_11fuel_models_is_burnable_fuel_model_number(int __pyx_v_fuel_model_number) {
  int __pyx_r;
  int __pyx_t_1;

  /* "pyretechnics/fuel_models.py":114
 * @cy.exceptval(check=False)
 * def is_burnable_fuel_model_number(fuel_model_number: cy.int) -> cy.bint:
 *     return not (91 <= fuel_model_number <= 99)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = (91 <= __pyx_v_fuel_model_number);
  if (__pyx_t_1) {
    __pyx_t_1 = (__pyx_v_fuel_model_number <= 99);
  }
  __pyx_r = (!__pyx_t_1);
  goto __pyx_L0;

  /* "pyretechnics/fuel_models.py":110
 * 
 * 
 * @cy.cfunc             # <<<<<<<<<<<<<<
 * @cy.inline
 * @cy.exceptval(check=False)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyretechnics/fuel_models.py":117
 * 
 * 
 * @cy.cfunc             # <<<<<<<<<<<<<<
 * @cy.exceptval(check=False)
 * def compute_exp_A_sigma(A: cy.float, sigma_ij: cy.float) -> cy.float:
 */

static float __pyx_f_12pyretechnics_11fuel_models_compute_exp_A_sigma(float __pyx_v_A, float __pyx_v_sigma_ij) {
  float __pyx_r;
  int __pyx_t_1;

  /* "pyretechnics/fuel_models.py":120
 * @cy.exceptval(check=False)
 * def compute_exp_A_sigma(A: cy.float, sigma_ij: cy.float) -> cy.float:
 *     if sigma_ij > almost_zero:             # <<<<<<<<<<<<<<
 *         return exp(A / sigma_ij)
 *     else:
 */
  __pyx_t_1 = (__pyx_v_sigma_ij > __pyx_v_12pyretechnics_11fuel_models_almost_zero);
  if (__pyx_t_1) {

    /* "pyretechnics/fuel_models.py":121
 * def compute_exp_A_sigma(A: cy.float, sigma_ij: cy.float) -> cy.float:
 *     if sigma_ij > almost_zero:
 *         return exp(A / sigma_ij)             # <<<<<<<<<<<<<<
 *     else:
 *         return 0.0
 */
    __pyx_r = exp((__pyx_v_A / __pyx_v_sigma_ij));
    goto __pyx_L0;

    /* "pyretechnics/fuel_models.py":120
 * @cy.exceptval(check=False)
 * def compute_exp_A_sigma(A: cy.float, sigma_ij: cy.float) -> cy.float:
 *     if sigma_ij > almost_zero:             # <<<<<<<<<<<<<<
 *         return exp(A / sigma_ij)
 *     else:
 */
  }

  /* "pyretechnics/fuel_models.py":123
 *         return exp(A / sigma_ij)
 *     else:
 *         return 0.0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __pyx_r = 0.0;
    goto __pyx_L0;
  }

  /* "pyretechnics/fuel_models.py":117
 * 
 * 
 * @cy.cfunc             # <<<<<<<<<<<<<<
 * @cy.exceptval(check=False)
 * def compute_exp_A_sigma(A: cy.float, sigma_ij: cy.float) -> cy.float:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyretechnics/fuel_models.py":126
 * 
 * 
 * @cy.cfunc             # <<<<<<<<<<<<<<
 * @cy.inline
 * @cy.exceptval(check=False)
 */

static CYTHON_INLINE float __pyx_f_12pyretechnics_11fuel_models_compute_firemod_size_class(float __pyx_v_sigma_i) {
  float __pyx_r;
  double __pyx_t_1;
  int __pyx_t_2;
  double __pyx_t_3;
  int __pyx_t_4;
  double __pyx_t_5;
  int __pyx_t_6;
  double __pyx_t_7;
  int __pyx_t_8;
  double __pyx_t_9;
  int __pyx_t_10;

  /* "pyretechnics/fuel_models.py":131
 * def compute_firemod_size_class(sigma_i: cy.float) -> cy.float:
 *     return (
 *         1.0 if (sigma_i >= 1200.0)             # <<<<<<<<<<<<<<
 *         else 2.0 if (sigma_i >= 192.0)
 *         else 3.0 if (sigma_i >= 96.0)
 */
  __pyx_t_2 = (__pyx_v_sigma_i >= 1200.0);
  if (__pyx_t_2) {
    __pyx_t_1 = 1.0;
  } else {

    /* "pyretechnics/fuel_models.py":132
 *     return (
 *         1.0 if (sigma_i >= 1200.0)
 *         else 2.0 if (sigma_i >= 192.0)             # <<<<<<<<<<<<<<
 *         else 3.0 if (sigma_i >= 96.0)
 *         else 4.0 if (sigma_i >= 48.0)
 */
    __pyx_t_4 = (__pyx_v_sigma_i >= 192.0);
    if (__pyx_t_4) {
      __pyx_t_3 = 2.0;
    } else {

      /* "pyretechnics/fuel_models.py":133
 *         1.0 if (sigma_i >= 1200.0)
 *         else 2.0 if (sigma_i >= 192.0)
 *         else 3.0 if (sigma_i >= 96.0)             # <<<<<<<<<<<<<<
 *         else 4.0 if (sigma_i >= 48.0)
 *         else 5.0 if (sigma_i >= 16.0)
 */
      __pyx_t_6 = (__pyx_v_sigma_i >= 96.0);
      if (__pyx_t_6) {
        __pyx_t_5 = 3.0;
      } else {

        /* "pyretechnics/fuel_models.py":134
 *         else 2.0 if (sigma_i >= 192.0)
 *         else 3.0 if (sigma_i >= 96.0)
 *         else 4.0 if (sigma_i >= 48.0)             # <<<<<<<<<<<<<<
 *         else 5.0 if (sigma_i >= 16.0)
 *         else 6.0
 */
        __pyx_t_8 = (__pyx_v_sigma_i >= 48.0);
        if (__pyx_t_8) {
          __pyx_t_7 = 4.0;
        } else {

          /* "pyretechnics/fuel_models.py":135
 *         else 3.0 if (sigma_i >= 96.0)
 *         else 4.0 if (sigma_i >= 48.0)
 *         else 5.0 if (sigma_i >= 16.0)             # <<<<<<<<<<<<<<
 *         else 6.0
 *     )
 */
          __pyx_t_10 = (__pyx_v_sigma_i >= 16.0);
          if (__pyx_t_10) {
            __pyx_t_9 = 5.0;
          } else {
            __pyx_t_9 = 6.0;
          }
          __pyx_t_7 = __pyx_t_9;
        }
        __pyx_t_5 = __pyx_t_7;
      }
      __pyx_t_3 = __pyx_t_5;
    }
    __pyx_t_1 = __pyx_t_3;
  }
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "pyretechnics/fuel_models.py":126
 * 
 * 
 * @cy.cfunc             # <<<<<<<<<<<<<<
 * @cy.inline
 * @cy.exceptval(check=False)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyretechnics/fuel_models.py":140
 * 
 * 
 * @cy.cfunc             # <<<<<<<<<<<<<<
 * @cy.exceptval(check=False)
 * def expand_compact_fuel_model(fuel_model_number: cy.int) -> FuelModel:
 */

static struct __pyx_t_12pyretechnics_8cy_types_FuelModel __pyx_f_12pyretechnics_11fuel_models_expand_compact_fuel_model(int __pyx_v_fuel_model_number) {
  __pyx_t_12pyretechnics_8cy_types_CompactFuelModel __pyx_v_cfm;
  float __pyx_v_delta;
  float __pyx_v_M_x_dead;
  float __pyx_v_h;
  float __pyx_v_w_o_dead_1hr;
  float __pyx_v_w_o_dead_10hr;
  float __pyx_v_w_o_dead_100hr;
  float __pyx_v_w_o_live_herbaceous;
  float __pyx_v_w_o_live_woody;
  float __pyx_v_sigma_dead_1hr;
  float __pyx_v_sigma_dead_10hr;
  float __pyx_v_sigma_dead_100hr;
  float __pyx_v_sigma_live_herbaceous;
  float __pyx_v_sigma_live_woody;
  int __pyx_v_dynamic;
  float __pyx_v_M_x_dead_herbaceous;
  float __pyx_v_sigma_dead_herbaceous;
  struct __pyx_t_12pyretechnics_8cy_types_FuelModel __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __pyx_t_12pyretechnics_8cy_types_CompactFuelModel __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  double __pyx_t_6;
  struct __pyx_t_12pyretechnics_8cy_types_FuelModel __pyx_t_7;
  __pyx_t_12pyretechnics_8cy_types_fclaarr __pyx_t_8;
  __pyx_t_12pyretechnics_8cy_types_fcatarr __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("expand_compact_fuel_model", 1);

  /* "pyretechnics/fuel_models.py":144
 * def expand_compact_fuel_model(fuel_model_number: cy.int) -> FuelModel:
 *     # Look up the CompactFuelModel by fuel_model_number
 *     cfm: CompactFuelModel = compact_fuel_model_table[fuel_model_number]             # <<<<<<<<<<<<<<
 *     # Unpack the CompactFuelModel values
 *     delta                : cy.float = cfm[0]
 */
  if (unlikely(__pyx_v_12pyretechnics_11fuel_models_compact_fuel_model_table == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 144, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_fuel_model_number); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_GetItem(__pyx_v_12pyretechnics_11fuel_models_compact_fuel_model_table, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __pyx_convert__from_py___pyx_ctuple_a8db9c__float__and_float__and_float__and_float__and_fl__etc(__pyx_t_2); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 144, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_cfm = __pyx_t_3;

  /* "pyretechnics/fuel_models.py":146
 *     cfm: CompactFuelModel = compact_fuel_model_table[fuel_model_number]
 *     # Unpack the CompactFuelModel values
 *     delta                : cy.float = cfm[0]             # <<<<<<<<<<<<<<
 *     M_x_dead             : cy.float = cfm[1]
 *     h                    : cy.float = cfm[2]
 */
  __pyx_v_delta = __pyx_v_cfm.f0;

  /* "pyretechnics/fuel_models.py":147
 *     # Unpack the CompactFuelModel values
 *     delta                : cy.float = cfm[0]
 *     M_x_dead             : cy.float = cfm[1]             # <<<<<<<<<<<<<<
 *     h                    : cy.float = cfm[2]
 *     w_o_dead_1hr         : cy.float = cfm[3]
 */
  __pyx_v_M_x_dead = __pyx_v_cfm.f1;

  /* "pyretechnics/fuel_models.py":148
 *     delta                : cy.float = cfm[0]
 *     M_x_dead             : cy.float = cfm[1]
 *     h                    : cy.float = cfm[2]             # <<<<<<<<<<<<<<
 *     w_o_dead_1hr         : cy.float = cfm[3]
 *     w_o_dead_10hr        : cy.float = cfm[4]
 */
  __pyx_v_h = __pyx_v_cfm.f2;

  /* "pyretechnics/fuel_models.py":149
 *     M_x_dead             : cy.float = cfm[1]
 *     h                    : cy.float = cfm[2]
 *     w_o_dead_1hr         : cy.float = cfm[3]             # <<<<<<<<<<<<<<
 *     w_o_dead_10hr        : cy.float = cfm[4]
 *     w_o_dead_100hr       : cy.float = cfm[5]
 */
  __pyx_v_w_o_dead_1hr = __pyx_v_cfm.f3;

  /* "pyretechnics/fuel_models.py":150
 *     h                    : cy.float = cfm[2]
 *     w_o_dead_1hr         : cy.float = cfm[3]
 *     w_o_dead_10hr        : cy.float = cfm[4]             # <<<<<<<<<<<<<<
 *     w_o_dead_100hr       : cy.float = cfm[5]
 *     w_o_live_herbaceous  : cy.float = cfm[6]
 */
  __pyx_v_w_o_dead_10hr = __pyx_v_cfm.f4;

  /* "pyretechnics/fuel_models.py":151
 *     w_o_dead_1hr         : cy.float = cfm[3]
 *     w_o_dead_10hr        : cy.float = cfm[4]
 *     w_o_dead_100hr       : cy.float = cfm[5]             # <<<<<<<<<<<<<<
 *     w_o_live_herbaceous  : cy.float = cfm[6]
 *     w_o_live_woody       : cy.float = cfm[7]
 */
  __pyx_v_w_o_dead_100hr = __pyx_v_cfm.f5;

  /* "pyretechnics/fuel_models.py":152
 *     w_o_dead_10hr        : cy.float = cfm[4]
 *     w_o_dead_100hr       : cy.float = cfm[5]
 *     w_o_live_herbaceous  : cy.float = cfm[6]             # <<<<<<<<<<<<<<
 *     w_o_live_woody       : cy.float = cfm[7]
 *     sigma_dead_1hr       : cy.float = cfm[8]
 */
  __pyx_v_w_o_live_herbaceous = __pyx_v_cfm.f6;

  /* "pyretechnics/fuel_models.py":153
 *     w_o_dead_100hr       : cy.float = cfm[5]
 *     w_o_live_herbaceous  : cy.float = cfm[6]
 *     w_o_live_woody       : cy.float = cfm[7]             # <<<<<<<<<<<<<<
 *     sigma_dead_1hr       : cy.float = cfm[8]
 *     sigma_dead_10hr      : cy.float = cfm[9]
 */
  __pyx_v_w_o_live_woody = __pyx_v_cfm.f7;

  /* "pyretechnics/fuel_models.py":154
 *     w_o_live_herbaceous  : cy.float = cfm[6]
 *     w_o_live_woody       : cy.float = cfm[7]
 *     sigma_dead_1hr       : cy.float = cfm[8]             # <<<<<<<<<<<<<<
 *     sigma_dead_10hr      : cy.float = cfm[9]
 *     sigma_dead_100hr     : cy.float = cfm[10]
 */
  __pyx_v_sigma_dead_1hr = __pyx_v_cfm.f8;

  /* "pyretechnics/fuel_models.py":155
 *     w_o_live_woody       : cy.float = cfm[7]
 *     sigma_dead_1hr       : cy.float = cfm[8]
 *     sigma_dead_10hr      : cy.float = cfm[9]             # <<<<<<<<<<<<<<
 *     sigma_dead_100hr     : cy.float = cfm[10]
 *     sigma_live_herbaceous: cy.float = cfm[11]
 */
  __pyx_v_sigma_dead_10hr = __pyx_v_cfm.f9;

  /* "pyretechnics/fuel_models.py":156
 *     sigma_dead_1hr       : cy.float = cfm[8]
 *     sigma_dead_10hr      : cy.float = cfm[9]
 *     sigma_dead_100hr     : cy.float = cfm[10]             # <<<<<<<<<<<<<<
 *     sigma_live_herbaceous: cy.float = cfm[11]
 *     sigma_live_woody     : cy.float = cfm[12]
 */
  __pyx_v_sigma_dead_100hr = __pyx_v_cfm.f10;

  /* "pyretechnics/fuel_models.py":157
 *     sigma_dead_10hr      : cy.float = cfm[9]
 *     sigma_dead_100hr     : cy.float = cfm[10]
 *     sigma_live_herbaceous: cy.float = cfm[11]             # <<<<<<<<<<<<<<
 *     sigma_live_woody     : cy.float = cfm[12]
 *     # Expand compressed values
 */
  __pyx_v_sigma_live_herbaceous = __pyx_v_cfm.f11;

  /* "pyretechnics/fuel_models.py":158
 *     sigma_dead_100hr     : cy.float = cfm[10]
 *     sigma_live_herbaceous: cy.float = cfm[11]
 *     sigma_live_woody     : cy.float = cfm[12]             # <<<<<<<<<<<<<<
 *     # Expand compressed values
 *     M_x_dead: cy.float = M_x_dead * 0.01
 */
  __pyx_v_sigma_live_woody = __pyx_v_cfm.f12;

  /* "pyretechnics/fuel_models.py":160
 *     sigma_live_woody     : cy.float = cfm[12]
 *     # Expand compressed values
 *     M_x_dead: cy.float = M_x_dead * 0.01             # <<<<<<<<<<<<<<
 *     h       : cy.float = h * 1000.0
 *     # Pre-compute some dynamic fuel model values
 */
  __pyx_v_M_x_dead = (__pyx_v_M_x_dead * 0.01);

  /* "pyretechnics/fuel_models.py":161
 *     # Expand compressed values
 *     M_x_dead: cy.float = M_x_dead * 0.01
 *     h       : cy.float = h * 1000.0             # <<<<<<<<<<<<<<
 *     # Pre-compute some dynamic fuel model values
 *     dynamic              : cy.bint  = fuel_model_number > 100 and w_o_live_herbaceous > almost_zero
 */
  __pyx_v_h = (__pyx_v_h * 1000.0);

  /* "pyretechnics/fuel_models.py":163
 *     h       : cy.float = h * 1000.0
 *     # Pre-compute some dynamic fuel model values
 *     dynamic              : cy.bint  = fuel_model_number > 100 and w_o_live_herbaceous > almost_zero             # <<<<<<<<<<<<<<
 *     M_x_dead_herbaceous  : cy.float = M_x_dead              if dynamic else 0.0
 *     sigma_dead_herbaceous: cy.float = sigma_live_herbaceous if dynamic else 0.0
 */
  __pyx_t_5 = (__pyx_v_fuel_model_number > 0x64);
  if (__pyx_t_5) {
  } else {
    __pyx_t_4 = __pyx_t_5;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_5 = (__pyx_v_w_o_live_herbaceous > __pyx_v_12pyretechnics_11fuel_models_almost_zero);
  __pyx_t_4 = __pyx_t_5;
  __pyx_L3_bool_binop_done:;
  __pyx_v_dynamic = __pyx_t_4;

  /* "pyretechnics/fuel_models.py":164
 *     # Pre-compute some dynamic fuel model values
 *     dynamic              : cy.bint  = fuel_model_number > 100 and w_o_live_herbaceous > almost_zero
 *     M_x_dead_herbaceous  : cy.float = M_x_dead              if dynamic else 0.0             # <<<<<<<<<<<<<<
 *     sigma_dead_herbaceous: cy.float = sigma_live_herbaceous if dynamic else 0.0
 *     # Re-pack everything into a FuelModel struct
 */
  if (__pyx_v_dynamic) {
    __pyx_t_6 = __pyx_v_M_x_dead;
  } else {
    __pyx_t_6 = 0.0;
  }
  __pyx_v_M_x_dead_herbaceous = __pyx_t_6;

  /* "pyretechnics/fuel_models.py":165
 *     dynamic              : cy.bint  = fuel_model_number > 100 and w_o_live_herbaceous > almost_zero
 *     M_x_dead_herbaceous  : cy.float = M_x_dead              if dynamic else 0.0
 *     sigma_dead_herbaceous: cy.float = sigma_live_herbaceous if dynamic else 0.0             # <<<<<<<<<<<<<<
 *     # Re-pack everything into a FuelModel struct
 *     return FuelModel(
 */
  if (__pyx_v_dynamic) {
    __pyx_t_6 = __pyx_v_sigma_live_herbaceous;
  } else {
    __pyx_t_6 = 0.0;
  }
  __pyx_v_sigma_dead_herbaceous = __pyx_t_6;

  /* "pyretechnics/fuel_models.py":168
 *     # Re-pack everything into a FuelModel struct
 *     return FuelModel(
 *         number               = fuel_model_number,             # <<<<<<<<<<<<<<
 *         delta                = delta,
 *         M_x                  = (M_x_dead,
 */
  __pyx_t_7.number = __pyx_v_fuel_model_number;

  /* "pyretechnics/fuel_models.py":169
 *     return FuelModel(
 *         number               = fuel_model_number,
 *         delta                = delta,             # <<<<<<<<<<<<<<
 *         M_x                  = (M_x_dead,
 *                                 M_x_dead,
 */
  __pyx_t_7.delta = __pyx_v_delta;

  /* "pyretechnics/fuel_models.py":170
 *         number               = fuel_model_number,
 *         delta                = delta,
 *         M_x                  = (M_x_dead,             # <<<<<<<<<<<<<<
 *                                 M_x_dead,
 *                                 M_x_dead,
 */
  __pyx_t_8.f0 = __pyx_v_M_x_dead;
  __pyx_t_8.f1 = __pyx_v_M_x_dead;
  __pyx_t_8.f2 = __pyx_v_M_x_dead;
  __pyx_t_8.f3 = __pyx_v_M_x_dead_herbaceous;
  __pyx_t_8.f4 = 0.0;
  __pyx_t_8.f5 = 0.0;
  __pyx_t_7.M_x = __pyx_t_8;

  /* "pyretechnics/fuel_models.py":176
 *                                 0.0,
 *                                 0.0),
 *         M_f                  = (0.0, 0.0, 0.0, 0.0, 0.0, 0.0),             # <<<<<<<<<<<<<<
 *         w_o                  = (w_o_dead_1hr,
 *                                 w_o_dead_10hr,
 */
  __pyx_t_8.f0 = 0.0;
  __pyx_t_8.f1 = 0.0;
  __pyx_t_8.f2 = 0.0;
  __pyx_t_8.f3 = 0.0;
  __pyx_t_8.f4 = 0.0;
  __pyx_t_8.f5 = 0.0;
  __pyx_t_7.M_f = __pyx_t_8;

  /* "pyretechnics/fuel_models.py":177
 *                                 0.0),
 *         M_f                  = (0.0, 0.0, 0.0, 0.0, 0.0, 0.0),
 *         w_o                  = (w_o_dead_1hr,             # <<<<<<<<<<<<<<
 *                                 w_o_dead_10hr,
 *                                 w_o_dead_100hr,
 */
  __pyx_t_8.f0 = __pyx_v_w_o_dead_1hr;
  __pyx_t_8.f1 = __pyx_v_w_o_dead_10hr;
  __pyx_t_8.f2 = __pyx_v_w_o_dead_100hr;
  __pyx_t_8.f3 = 0.0;
  __pyx_t_8.f4 = __pyx_v_w_o_live_herbaceous;
  __pyx_t_8.f5 = __pyx_v_w_o_live_woody;
  __pyx_t_7.w_o = __pyx_t_8;

  /* "pyretechnics/fuel_models.py":183
 *                                 w_o_live_herbaceous,
 *                                 w_o_live_woody),
 *         sigma                = (sigma_dead_1hr,             # <<<<<<<<<<<<<<
 *                                 sigma_dead_10hr,
 *                                 sigma_dead_100hr,
 */
  __pyx_t_8.f0 = __pyx_v_sigma_dead_1hr;
  __pyx_t_8.f1 = __pyx_v_sigma_dead_10hr;
  __pyx_t_8.f2 = __pyx_v_sigma_dead_100hr;
  __pyx_t_8.f3 = __pyx_v_sigma_dead_herbaceous;
  __pyx_t_8.f4 = __pyx_v_sigma_live_herbaceous;
  __pyx_t_8.f5 = __pyx_v_sigma_live_woody;
  __pyx_t_7.sigma = __pyx_t_8;

  /* "pyretechnics/fuel_models.py":189
 *                                 sigma_live_herbaceous,
 *                                 sigma_live_woody),
 *         h                    = (h, h, h, h, h, h),             # <<<<<<<<<<<<<<
 *         rho_p                = (32.0, 32.0, 32.0, 32.0, 32.0, 32.0),
 *         S_T                  = (0.0555, 0.0555, 0.0555, 0.0555, 0.0555, 0.0555),
 */
  __pyx_t_8.f0 = __pyx_v_h;
  __pyx_t_8.f1 = __pyx_v_h;
  __pyx_t_8.f2 = __pyx_v_h;
  __pyx_t_8.f3 = __pyx_v_h;
  __pyx_t_8.f4 = __pyx_v_h;
  __pyx_t_8.f5 = __pyx_v_h;
  __pyx_t_7.h = __pyx_t_8;

  /* "pyretechnics/fuel_models.py":190
 *                                 sigma_live_woody),
 *         h                    = (h, h, h, h, h, h),
 *         rho_p                = (32.0, 32.0, 32.0, 32.0, 32.0, 32.0),             # <<<<<<<<<<<<<<
 *         S_T                  = (0.0555, 0.0555, 0.0555, 0.0555, 0.0555, 0.0555),
 *         S_e                  = (0.01, 0.01, 0.01, 0.01, 0.01, 0.01),
 */
  __pyx_t_8.f0 = 32.0;
  __pyx_t_8.f1 = 32.0;
  __pyx_t_8.f2 = 32.0;
  __pyx_t_8.f3 = 32.0;
  __pyx_t_8.f4 = 32.0;
  __pyx_t_8.f5 = 32.0;
  __pyx_t_7.rho_p = __pyx_t_8;

  /* "pyretechnics/fuel_models.py":191
 *         h                    = (h, h, h, h, h, h),
 *         rho_p                = (32.0, 32.0, 32.0, 32.0, 32.0, 32.0),
 *         S_T                  = (0.0555, 0.0555, 0.0555, 0.0555, 0.0555, 0.0555),             # <<<<<<<<<<<<<<
 *         S_e                  = (0.01, 0.01, 0.01, 0.01, 0.01, 0.01),
 *         dynamic              = dynamic,
 */
  __pyx_t_8.f0 = 0.0555;
  __pyx_t_8.f1 = 0.0555;
  __pyx_t_8.f2 = 0.0555;
  __pyx_t_8.f3 = 0.0555;
  __pyx_t_8.f4 = 0.0555;
  __pyx_t_8.f5 = 0.0555;
  __pyx_t_7.S_T = __pyx_t_8;

  /* "pyretechnics/fuel_models.py":192
 *         rho_p                = (32.0, 32.0, 32.0, 32.0, 32.0, 32.0),
 *         S_T                  = (0.0555, 0.0555, 0.0555, 0.0555, 0.0555, 0.0555),
 *         S_e                  = (0.01, 0.01, 0.01, 0.01, 0.01, 0.01),             # <<<<<<<<<<<<<<
 *         dynamic              = dynamic,
 *         burnable             = is_burnable_fuel_model_number(fuel_model_number),
 */
  __pyx_t_8.f0 = 0.01;
  __pyx_t_8.f1 = 0.01;
  __pyx_t_8.f2 = 0.01;
  __pyx_t_8.f3 = 0.01;
  __pyx_t_8.f4 = 0.01;
  __pyx_t_8.f5 = 0.01;
  __pyx_t_7.S_e = __pyx_t_8;

  /* "pyretechnics/fuel_models.py":193
 *         S_T                  = (0.0555, 0.0555, 0.0555, 0.0555, 0.0555, 0.0555),
 *         S_e                  = (0.01, 0.01, 0.01, 0.01, 0.01, 0.01),
 *         dynamic              = dynamic,             # <<<<<<<<<<<<<<
 *         burnable             = is_burnable_fuel_model_number(fuel_model_number),
 *         exp_A_sigma          = (compute_exp_A_sigma(-138.0, sigma_dead_1hr),
 */
  __pyx_t_7.dynamic = __pyx_v_dynamic;

  /* "pyretechnics/fuel_models.py":194
 *         S_e                  = (0.01, 0.01, 0.01, 0.01, 0.01, 0.01),
 *         dynamic              = dynamic,
 *         burnable             = is_burnable_fuel_model_number(fuel_model_number),             # <<<<<<<<<<<<<<
 *         exp_A_sigma          = (compute_exp_A_sigma(-138.0, sigma_dead_1hr),
 *                                 compute_exp_A_sigma(-138.0, sigma_dead_10hr),
 */
  __pyx_t_7.burnable = __pyx_f_12pyretechnics_11fuel_models_is_burnable_fuel_model_number(__pyx_v_fuel_model_number);

  /* "pyretechnics/fuel_models.py":195
 *         dynamic              = dynamic,
 *         burnable             = is_burnable_fuel_model_number(fuel_model_number),
 *         exp_A_sigma          = (compute_exp_A_sigma(-138.0, sigma_dead_1hr),             # <<<<<<<<<<<<<<
 *                                 compute_exp_A_sigma(-138.0, sigma_dead_10hr),
 *                                 compute_exp_A_sigma(-138.0, sigma_dead_100hr),
 */
  __pyx_t_8.f0 = __pyx_f_12pyretechnics_11fuel_models_compute_exp_A_sigma(-138.0, __pyx_v_sigma_dead_1hr);
  __pyx_t_8.f1 = __pyx_f_12pyretechnics_11fuel_models_compute_exp_A_sigma(-138.0, __pyx_v_sigma_dead_10hr);
  __pyx_t_8.f2 = __pyx_f_12pyretechnics_11fuel_models_compute_exp_A_sigma(-138.0, __pyx_v_sigma_dead_100hr);
  __pyx_t_8.f3 = __pyx_f_12pyretechnics_11fuel_models_compute_exp_A_sigma(-138.0, __pyx_v_sigma_dead_herbaceous);
  __pyx_t_8.f4 = __pyx_f_12pyretechnics_11fuel_models_compute_exp_A_sigma(-500.0, __pyx_v_sigma_live_herbaceous);
  __pyx_t_8.f5 = __pyx_f_12pyretechnics_11fuel_models_compute_exp_A_sigma(-500.0, __pyx_v_sigma_live_woody);
  __pyx_t_7.exp_A_sigma = __pyx_t_8;

  /* "pyretechnics/fuel_models.py":201
 *                                 compute_exp_A_sigma(-500.0, sigma_live_herbaceous),
 *                                 compute_exp_A_sigma(-500.0, sigma_live_woody)),
 *         firemod_size_classes = (compute_firemod_size_class(sigma_dead_1hr),             # <<<<<<<<<<<<<<
 *                                 compute_firemod_size_class(sigma_dead_10hr),
 *                                 compute_firemod_size_class(sigma_dead_100hr),
 */
  __pyx_t_8.f0 = __pyx_f_12pyretechnics_11fuel_models_compute_firemod_size_class(__pyx_v_sigma_dead_1hr);
  __pyx_t_8.f1 = __pyx_f_12pyretechnics_11fuel_models_compute_firemod_size_class(__pyx_v_sigma_dead_10hr);
  __pyx_t_8.f2 = __pyx_f_12pyretechnics_11fuel_models_compute_firemod_size_class(__pyx_v_sigma_dead_100hr);
  __pyx_t_8.f3 = __pyx_f_12pyretechnics_11fuel_models_compute_firemod_size_class(__pyx_v_sigma_dead_herbaceous);
  __pyx_t_8.f4 = __pyx_f_12pyretechnics_11fuel_models_compute_firemod_size_class(__pyx_v_sigma_live_herbaceous);
  __pyx_t_8.f5 = __pyx_f_12pyretechnics_11fuel_models_compute_firemod_size_class(__pyx_v_sigma_live_woody);
  __pyx_t_7.firemod_size_classes = __pyx_t_8;

  /* "pyretechnics/fuel_models.py":207
 *                                 compute_firemod_size_class(sigma_live_herbaceous),
 *                                 compute_firemod_size_class(sigma_live_woody)),
 *         f_ij                 = (0.0, 0.0, 0.0, 0.0, 0.0, 0.0),             # <<<<<<<<<<<<<<
 *         f_i                  = (0.0, 0.0),
 *         g_ij                 = (0.0, 0.0, 0.0, 0.0, 0.0, 0.0),
 */
  __pyx_t_8.f0 = 0.0;
  __pyx_t_8.f1 = 0.0;
  __pyx_t_8.f2 = 0.0;
  __pyx_t_8.f3 = 0.0;
  __pyx_t_8.f4 = 0.0;
  __pyx_t_8.f5 = 0.0;
  __pyx_t_7.f_ij = __pyx_t_8;

  /* "pyretechnics/fuel_models.py":208
 *                                 compute_firemod_size_class(sigma_live_woody)),
 *         f_ij                 = (0.0, 0.0, 0.0, 0.0, 0.0, 0.0),
 *         f_i                  = (0.0, 0.0),             # <<<<<<<<<<<<<<
 *         g_ij                 = (0.0, 0.0, 0.0, 0.0, 0.0, 0.0),
 *     )
 */
  __pyx_t_9.f0 = 0.0;
  __pyx_t_9.f1 = 0.0;
  __pyx_t_7.f_i = __pyx_t_9;

  /* "pyretechnics/fuel_models.py":209
 *         f_ij                 = (0.0, 0.0, 0.0, 0.0, 0.0, 0.0),
 *         f_i                  = (0.0, 0.0),
 *         g_ij                 = (0.0, 0.0, 0.0, 0.0, 0.0, 0.0),             # <<<<<<<<<<<<<<
 *     )
 * 
 */
  __pyx_t_8.f0 = 0.0;
  __pyx_t_8.f1 = 0.0;
  __pyx_t_8.f2 = 0.0;
  __pyx_t_8.f3 = 0.0;
  __pyx_t_8.f4 = 0.0;
  __pyx_t_8.f5 = 0.0;
  __pyx_t_7.g_ij = __pyx_t_8;
  __pyx_r = __pyx_t_7;
  goto __pyx_L0;

  /* "pyretechnics/fuel_models.py":140
 * 
 * 
 * @cy.cfunc             # <<<<<<<<<<<<<<
 * @cy.exceptval(check=False)
 * def expand_compact_fuel_model(fuel_model_number: cy.int) -> FuelModel:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_WriteUnraisable("pyretechnics.fuel_models.expand_compact_fuel_model", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyretechnics/fuel_models.py":219
 * 
 * 
 * @cy.ccall             # <<<<<<<<<<<<<<
 * @cy.inline
 * def list_fuel_model_numbers() -> list[int]:
 */

static PyObject *__pyx_pw_12pyretechnics_11fuel_models_1list_fuel_model_numbers(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_12pyretechnics_11fuel_models_list_fuel_model_numbers(CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("list_fuel_model_numbers", 1);

  /* "pyretechnics/fuel_models.py":222
 * @cy.inline
 * def list_fuel_model_numbers() -> list[int]:
 *     return list(fuel_model_table.keys())             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(__pyx_v_12pyretechnics_11fuel_models_fuel_model_table == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "keys");
    __PYX_ERR(1, 222, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyDict_Keys(__pyx_v_12pyretechnics_11fuel_models_fuel_model_table); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PySequence_ListKeepNew(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pyretechnics/fuel_models.py":219
 * 
 * 
 * @cy.ccall             # <<<<<<<<<<<<<<
 * @cy.inline
 * def list_fuel_model_numbers() -> list[int]:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyretechnics.fuel_models.list_fuel_model_numbers", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyretechnics_11fuel_models_1list_fuel_model_numbers(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_12pyretechnics_11fuel_models_1list_fuel_model_numbers = {"list_fuel_model_numbers", (PyCFunction)__pyx_pw_12pyretechnics_11fuel_models_1list_fuel_model_numbers, METH_NOARGS, 0};
static PyObject *__pyx_pw_12pyretechnics_11fuel_models_1list_fuel_model_numbers(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("list_fuel_model_numbers (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12pyretechnics_11fuel_models_list_fuel_model_numbers(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyretechnics_11fuel_models_list_fuel_model_numbers(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("list_fuel_model_numbers", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyretechnics_11fuel_models_list_fuel_model_numbers(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyretechnics.fuel_models.list_fuel_model_numbers", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyretechnics/fuel_models.py":225
 * 
 * 
 * @cy.ccall             # <<<<<<<<<<<<<<
 * @cy.inline
 * def list_fuel_models() -> list[FuelModel]:
 */

static PyObject *__pyx_pw_12pyretechnics_11fuel_models_3list_fuel_models(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_12pyretechnics_11fuel_models_list_fuel_models(CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("list_fuel_models", 1);

  /* "pyretechnics/fuel_models.py":228
 * @cy.inline
 * def list_fuel_models() -> list[FuelModel]:
 *     return list(fuel_model_table.values())             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(__pyx_v_12pyretechnics_11fuel_models_fuel_model_table == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "values");
    __PYX_ERR(1, 228, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyDict_Values(__pyx_v_12pyretechnics_11fuel_models_fuel_model_table); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PySequence_ListKeepNew(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pyretechnics/fuel_models.py":225
 * 
 * 
 * @cy.ccall             # <<<<<<<<<<<<<<
 * @cy.inline
 * def list_fuel_models() -> list[FuelModel]:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyretechnics.fuel_models.list_fuel_models", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyretechnics_11fuel_models_3list_fuel_models(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_12pyretechnics_11fuel_models_3list_fuel_models = {"list_fuel_models", (PyCFunction)__pyx_pw_12pyretechnics_11fuel_models_3list_fuel_models, METH_NOARGS, 0};
static PyObject *__pyx_pw_12pyretechnics_11fuel_models_3list_fuel_models(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("list_fuel_models (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12pyretechnics_11fuel_models_2list_fuel_models(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyretechnics_11fuel_models_2list_fuel_models(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("list_fuel_models", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyretechnics_11fuel_models_list_fuel_models(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyretechnics.fuel_models.list_fuel_models", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyretechnics/fuel_models.py":231
 * 
 * 
 * @cy.ccall             # <<<<<<<<<<<<<<
 * @cy.inline
 * @cy.exceptval(check=False)
 */

static PyObject *__pyx_pw_12pyretechnics_11fuel_models_5fuel_model_exists(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static CYTHON_INLINE int __pyx_f_12pyretechnics_11fuel_models_fuel_model_exists(int __pyx_v_fuel_model_number, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("fuel_model_exists", 1);

  /* "pyretechnics/fuel_models.py":235
 * @cy.exceptval(check=False)
 * def fuel_model_exists(fuel_model_number: cy.int) -> cy.bint:
 *     return fuel_model_number in fuel_model_table             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_fuel_model_number); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(__pyx_v_12pyretechnics_11fuel_models_fuel_model_table == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 235, __pyx_L1_error)
  }
  __pyx_t_2 = (__Pyx_PyDict_ContainsTF(__pyx_t_1, __pyx_v_12pyretechnics_11fuel_models_fuel_model_table, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 235, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  goto __pyx_L0;

  /* "pyretechnics/fuel_models.py":231
 * 
 * 
 * @cy.ccall             # <<<<<<<<<<<<<<
 * @cy.inline
 * @cy.exceptval(check=False)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("pyretechnics.fuel_models.fuel_model_exists", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyretechnics_11fuel_models_5fuel_model_exists(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_12pyretechnics_11fuel_models_5fuel_model_exists = {"fuel_model_exists", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12pyretechnics_11fuel_models_5fuel_model_exists, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_12pyretechnics_11fuel_models_5fuel_model_exists(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int __pyx_v_fuel_model_number;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("fuel_model_exists (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fuel_model_number,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_fuel_model_number)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 231, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "fuel_model_exists") < 0)) __PYX_ERR(1, 231, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_fuel_model_number = __Pyx_PyInt_As_int(values[0]); if (unlikely((__pyx_v_fuel_model_number == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 234, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("fuel_model_exists", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 231, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("pyretechnics.fuel_models.fuel_model_exists", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyretechnics_11fuel_models_4fuel_model_exists(__pyx_self, __pyx_v_fuel_model_number);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyretechnics_11fuel_models_4fuel_model_exists(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_fuel_model_number) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("fuel_model_exists", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12pyretechnics_11fuel_models_fuel_model_exists(__pyx_v_fuel_model_number, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyretechnics.fuel_models.fuel_model_exists", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyretechnics/fuel_models.py":238
 * 
 * 
 * @cy.ccall             # <<<<<<<<<<<<<<
 * @cy.inline
 * @cy.exceptval(check=False)
 */

static PyObject *__pyx_pw_12pyretechnics_11fuel_models_7get_fuel_model(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static CYTHON_INLINE struct __pyx_t_12pyretechnics_8cy_types_FuelModel __pyx_f_12pyretechnics_11fuel_models_get_fuel_model(int __pyx_v_fuel_model_number, CYTHON_UNUSED int __pyx_skip_dispatch) {
  struct __pyx_t_12pyretechnics_8cy_types_FuelModel __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  struct __pyx_t_12pyretechnics_8cy_types_FuelModel __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_fuel_model", 1);

  /* "pyretechnics/fuel_models.py":242
 * @cy.exceptval(check=False)
 * def get_fuel_model(fuel_model_number: cy.int) -> FuelModel:
 *     return cy.cast(FuelModel, fuel_model_table[fuel_model_number])             # <<<<<<<<<<<<<<
 * # expand-compact-fuel-model-table ends here
 * # [[file:../../org/pyretechnics.org::add-dynamic-fuel-loading][add-dynamic-fuel-loading]]
 */
  if (unlikely(__pyx_v_12pyretechnics_11fuel_models_fuel_model_table == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 242, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_fuel_model_number); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_GetItem(__pyx_v_12pyretechnics_11fuel_models_fuel_model_table, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __pyx_convert__from_py_struct____pyx_t_12pyretechnics_8cy_types_FuelModel(__pyx_t_2); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 242, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  goto __pyx_L0;

  /* "pyretechnics/fuel_models.py":238
 * 
 * 
 * @cy.ccall             # <<<<<<<<<<<<<<
 * @cy.inline
 * @cy.exceptval(check=False)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_WriteUnraisable("pyretechnics.fuel_models.get_fuel_model", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyretechnics_11fuel_models_7get_fuel_model(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_12pyretechnics_11fuel_models_7get_fuel_model = {"get_fuel_model", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12pyretechnics_11fuel_models_7get_fuel_model, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_12pyretechnics_11fuel_models_7get_fuel_model(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int __pyx_v_fuel_model_number;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_fuel_model (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fuel_model_number,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_fuel_model_number)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 238, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "get_fuel_model") < 0)) __PYX_ERR(1, 238, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_fuel_model_number = __Pyx_PyInt_As_int(values[0]); if (unlikely((__pyx_v_fuel_model_number == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 241, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_fuel_model", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 238, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("pyretechnics.fuel_models.get_fuel_model", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyretechnics_11fuel_models_6get_fuel_model(__pyx_self, __pyx_v_fuel_model_number);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyretechnics_11fuel_models_6get_fuel_model(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_fuel_model_number) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_fuel_model", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_convert__to_py_struct____pyx_t_12pyretechnics_8cy_types_FuelModel(__pyx_f_12pyretechnics_11fuel_models_get_fuel_model(__pyx_v_fuel_model_number, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyretechnics.fuel_models.get_fuel_model", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyretechnics/fuel_models.py":245
 * # expand-compact-fuel-model-table ends here
 * # [[file:../../org/pyretechnics.org::add-dynamic-fuel-loading][add-dynamic-fuel-loading]]
 * @cy.cfunc             # <<<<<<<<<<<<<<
 * @cy.exceptval(check=False)
 * def add_dynamic_fuel_loading(fuel_model: FuelModel, M_f: fclaarr) -> FuelModel:
 */

static struct __pyx_t_12pyretechnics_8cy_types_FuelModel __pyx_f_12pyretechnics_11fuel_models_add_dynamic_fuel_loading(struct __pyx_t_12pyretechnics_8cy_types_FuelModel __pyx_v_fuel_model, __pyx_t_12pyretechnics_8cy_types_fclaarr __pyx_v_M_f) {
  __pyx_t_12pyretechnics_8cy_types_fclaarr __pyx_v_w_o;
  float __pyx_v_live_herbaceous_load;
  float __pyx_v_live_herbaceous_moisture;
  float __pyx_v_fraction_green;
  float __pyx_v_fraction_cured;
  __pyx_t_12pyretechnics_8cy_types_fclaarr __pyx_v_M_f_dynamic;
  __pyx_t_12pyretechnics_8cy_types_fclaarr __pyx_v_w_o_dynamic;
  struct __pyx_t_12pyretechnics_8cy_types_FuelModel __pyx_r;
  __pyx_t_12pyretechnics_8cy_types_fclaarr __pyx_t_1;
  double __pyx_t_2;
  double __pyx_t_3;
  double __pyx_t_4;
  int __pyx_t_5;

  /* "pyretechnics/fuel_models.py":251
 *     Updates M_f and w_o. Mutates and returns fuel_model.
 *     """
 *     if fuel_model.dynamic:             # <<<<<<<<<<<<<<
 *         # === Dynamic Fuel Model ===
 *         # Calculate fraction_green and fraction_cured
 */
  if (__pyx_v_fuel_model.dynamic) {

    /* "pyretechnics/fuel_models.py":254
 *         # === Dynamic Fuel Model ===
 *         # Calculate fraction_green and fraction_cured
 *         w_o                     : fclaarr  = fuel_model.w_o             # <<<<<<<<<<<<<<
 *         live_herbaceous_load    : cy.float = w_o[4]
 *         live_herbaceous_moisture: cy.float = M_f[4]
 */
    __pyx_t_1 = __pyx_v_fuel_model.w_o;
    __pyx_v_w_o = __pyx_t_1;

    /* "pyretechnics/fuel_models.py":255
 *         # Calculate fraction_green and fraction_cured
 *         w_o                     : fclaarr  = fuel_model.w_o
 *         live_herbaceous_load    : cy.float = w_o[4]             # <<<<<<<<<<<<<<
 *         live_herbaceous_moisture: cy.float = M_f[4]
 *         fraction_green          : cy.float = max(0.0, min(1.0, (live_herbaceous_moisture / 0.9) - 0.3333333333333333))
 */
    __pyx_v_live_herbaceous_load = __pyx_v_w_o.f4;

    /* "pyretechnics/fuel_models.py":256
 *         w_o                     : fclaarr  = fuel_model.w_o
 *         live_herbaceous_load    : cy.float = w_o[4]
 *         live_herbaceous_moisture: cy.float = M_f[4]             # <<<<<<<<<<<<<<
 *         fraction_green          : cy.float = max(0.0, min(1.0, (live_herbaceous_moisture / 0.9) - 0.3333333333333333))
 *         fraction_cured          : cy.float = 1.0 - fraction_green
 */
    __pyx_v_live_herbaceous_moisture = __pyx_v_M_f.f4;

    /* "pyretechnics/fuel_models.py":257
 *         live_herbaceous_load    : cy.float = w_o[4]
 *         live_herbaceous_moisture: cy.float = M_f[4]
 *         fraction_green          : cy.float = max(0.0, min(1.0, (live_herbaceous_moisture / 0.9) - 0.3333333333333333))             # <<<<<<<<<<<<<<
 *         fraction_cured          : cy.float = 1.0 - fraction_green
 *         # Set M_f[dead_herbaceous] equal to M_f[dead_1hr]
 */
    __pyx_t_2 = ((((double)__pyx_v_live_herbaceous_moisture) / 0.9) - 0.3333333333333333);
    __pyx_t_3 = 1.0;
    __pyx_t_5 = (__pyx_t_2 < __pyx_t_3);
    if (__pyx_t_5) {
      __pyx_t_4 = __pyx_t_2;
    } else {
      __pyx_t_4 = __pyx_t_3;
    }
    __pyx_t_2 = __pyx_t_4;
    __pyx_t_4 = 0.0;
    __pyx_t_5 = (__pyx_t_2 > __pyx_t_4);
    if (__pyx_t_5) {
      __pyx_t_3 = __pyx_t_2;
    } else {
      __pyx_t_3 = __pyx_t_4;
    }
    __pyx_v_fraction_green = __pyx_t_3;

    /* "pyretechnics/fuel_models.py":258
 *         live_herbaceous_moisture: cy.float = M_f[4]
 *         fraction_green          : cy.float = max(0.0, min(1.0, (live_herbaceous_moisture / 0.9) - 0.3333333333333333))
 *         fraction_cured          : cy.float = 1.0 - fraction_green             # <<<<<<<<<<<<<<
 *         # Set M_f[dead_herbaceous] equal to M_f[dead_1hr]
 *         M_f_dynamic: fclaarr = (M_f[0], # dead_1hr
 */
    __pyx_v_fraction_cured = (1.0 - __pyx_v_fraction_green);

    /* "pyretechnics/fuel_models.py":260
 *         fraction_cured          : cy.float = 1.0 - fraction_green
 *         # Set M_f[dead_herbaceous] equal to M_f[dead_1hr]
 *         M_f_dynamic: fclaarr = (M_f[0], # dead_1hr             # <<<<<<<<<<<<<<
 *                                 M_f[1],
 *                                 M_f[2],
 */
    __pyx_t_1.f0 = __pyx_v_M_f.f0;
    __pyx_t_1.f1 = __pyx_v_M_f.f1;
    __pyx_t_1.f2 = __pyx_v_M_f.f2;
    __pyx_t_1.f3 = __pyx_v_M_f.f0;
    __pyx_t_1.f4 = __pyx_v_M_f.f4;
    __pyx_t_1.f5 = __pyx_v_M_f.f5;
    __pyx_v_M_f_dynamic = __pyx_t_1;

    /* "pyretechnics/fuel_models.py":267
 *                                 M_f[5])
 *         # Set w_o[dead_herbaceous] and w_o[live_herbaceous] based on fraction_cured and fraction_green
 *         w_o_dynamic: fclaarr = (w_o[0],             # <<<<<<<<<<<<<<
 *                                 w_o[1],
 *                                 w_o[2],
 */
    __pyx_t_1.f0 = __pyx_v_w_o.f0;
    __pyx_t_1.f1 = __pyx_v_w_o.f1;
    __pyx_t_1.f2 = __pyx_v_w_o.f2;
    __pyx_t_1.f3 = (__pyx_v_live_herbaceous_load * __pyx_v_fraction_cured);
    __pyx_t_1.f4 = (__pyx_v_live_herbaceous_load * __pyx_v_fraction_green);
    __pyx_t_1.f5 = __pyx_v_w_o.f5;
    __pyx_v_w_o_dynamic = __pyx_t_1;

    /* "pyretechnics/fuel_models.py":274
 *                                 w_o[5])
 *         # Update the passed in fuel_model to use M_f_dynamic and w_o_dynamic
 *         fuel_model.M_f = M_f_dynamic             # <<<<<<<<<<<<<<
 *         fuel_model.w_o = w_o_dynamic
 *         return fuel_model
 */
    __pyx_v_fuel_model.M_f = __pyx_v_M_f_dynamic;

    /* "pyretechnics/fuel_models.py":275
 *         # Update the passed in fuel_model to use M_f_dynamic and w_o_dynamic
 *         fuel_model.M_f = M_f_dynamic
 *         fuel_model.w_o = w_o_dynamic             # <<<<<<<<<<<<<<
 *         return fuel_model
 *     else:
 */
    __pyx_v_fuel_model.w_o = __pyx_v_w_o_dynamic;

    /* "pyretechnics/fuel_models.py":276
 *         fuel_model.M_f = M_f_dynamic
 *         fuel_model.w_o = w_o_dynamic
 *         return fuel_model             # <<<<<<<<<<<<<<
 *     else:
 *         # === Static Fuel Model ===
 */
    __pyx_r = __pyx_v_fuel_model;
    goto __pyx_L0;

    /* "pyretechnics/fuel_models.py":251
 *     Updates M_f and w_o. Mutates and returns fuel_model.
 *     """
 *     if fuel_model.dynamic:             # <<<<<<<<<<<<<<
 *         # === Dynamic Fuel Model ===
 *         # Calculate fraction_green and fraction_cured
 */
  }

  /* "pyretechnics/fuel_models.py":280
 *         # === Static Fuel Model ===
 *         # Update the passed in fuel_model to use the passed in M_f
 *         fuel_model.M_f = M_f             # <<<<<<<<<<<<<<
 *         return fuel_model
 * # add-dynamic-fuel-loading ends here
 */
  /*else*/ {
    __pyx_v_fuel_model.M_f = __pyx_v_M_f;

    /* "pyretechnics/fuel_models.py":281
 *         # Update the passed in fuel_model to use the passed in M_f
 *         fuel_model.M_f = M_f
 *         return fuel_model             # <<<<<<<<<<<<<<
 * # add-dynamic-fuel-loading ends here
 * # [[file:../../org/pyretechnics.org::add-weighting-factors][add-weighting-factors]]
 */
    __pyx_r = __pyx_v_fuel_model;
    goto __pyx_L0;
  }

  /* "pyretechnics/fuel_models.py":245
 * # expand-compact-fuel-model-table ends here
 * # [[file:../../org/pyretechnics.org::add-dynamic-fuel-loading][add-dynamic-fuel-loading]]
 * @cy.cfunc             # <<<<<<<<<<<<<<
 * @cy.exceptval(check=False)
 * def add_dynamic_fuel_loading(fuel_model: FuelModel, M_f: fclaarr) -> FuelModel:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyretechnics/fuel_models.py":286
 * # TODO: OPTIM pre-compute this conditional branching since it's fully determined by sigma.
 * #       This information might be represented efficiently in bit flags.
 * @cy.cfunc             # <<<<<<<<<<<<<<
 * @cy.exceptval(check=False)
 * def compute_gij(firemod_size_classes : fclaarr,
 */

static float __pyx_f_12pyretechnics_11fuel_models_compute_gij(__pyx_t_12pyretechnics_8cy_types_fclaarr __pyx_v_firemod_size_classes, __pyx_t_12pyretechnics_8cy_types_fclaarr __pyx_v_f_ij, float __pyx_v_firemod_size_class_ij, int __pyx_v_is_dead) {
  float __pyx_v_f_ij_0;
  float __pyx_v_f_ij_1;
  float __pyx_v_f_ij_2;
  float __pyx_v_f_ij_3;
  float __pyx_v_f_ij_4;
  float __pyx_v_f_ij_5;
  float __pyx_r;
  double __pyx_t_1;
  int __pyx_t_2;

  /* "pyretechnics/fuel_models.py":298
 *           trivially equal to f_ij[i].
 *     """
 *     if is_dead:             # <<<<<<<<<<<<<<
 *         f_ij_0: cy.float = (f_ij[0] if (firemod_size_class_ij == firemod_size_classes[0]) else 0.0)
 *         f_ij_1: cy.float = (f_ij[1] if (firemod_size_class_ij == firemod_size_classes[1]) else 0.0)
 */
  if (__pyx_v_is_dead) {

    /* "pyretechnics/fuel_models.py":299
 *     """
 *     if is_dead:
 *         f_ij_0: cy.float = (f_ij[0] if (firemod_size_class_ij == firemod_size_classes[0]) else 0.0)             # <<<<<<<<<<<<<<
 *         f_ij_1: cy.float = (f_ij[1] if (firemod_size_class_ij == firemod_size_classes[1]) else 0.0)
 *         f_ij_2: cy.float = (f_ij[2] if (firemod_size_class_ij == firemod_size_classes[2]) else 0.0)
 */
    __pyx_t_2 = (__pyx_v_firemod_size_class_ij == __pyx_v_firemod_size_classes.f0);
    if (__pyx_t_2) {
      __pyx_t_1 = __pyx_v_f_ij.f0;
    } else {
      __pyx_t_1 = 0.0;
    }
    __pyx_v_f_ij_0 = __pyx_t_1;

    /* "pyretechnics/fuel_models.py":300
 *     if is_dead:
 *         f_ij_0: cy.float = (f_ij[0] if (firemod_size_class_ij == firemod_size_classes[0]) else 0.0)
 *         f_ij_1: cy.float = (f_ij[1] if (firemod_size_class_ij == firemod_size_classes[1]) else 0.0)             # <<<<<<<<<<<<<<
 *         f_ij_2: cy.float = (f_ij[2] if (firemod_size_class_ij == firemod_size_classes[2]) else 0.0)
 *         f_ij_3: cy.float = (f_ij[3] if (firemod_size_class_ij == firemod_size_classes[3]) else 0.0)
 */
    __pyx_t_2 = (__pyx_v_firemod_size_class_ij == __pyx_v_firemod_size_classes.f1);
    if (__pyx_t_2) {
      __pyx_t_1 = __pyx_v_f_ij.f1;
    } else {
      __pyx_t_1 = 0.0;
    }
    __pyx_v_f_ij_1 = __pyx_t_1;

    /* "pyretechnics/fuel_models.py":301
 *         f_ij_0: cy.float = (f_ij[0] if (firemod_size_class_ij == firemod_size_classes[0]) else 0.0)
 *         f_ij_1: cy.float = (f_ij[1] if (firemod_size_class_ij == firemod_size_classes[1]) else 0.0)
 *         f_ij_2: cy.float = (f_ij[2] if (firemod_size_class_ij == firemod_size_classes[2]) else 0.0)             # <<<<<<<<<<<<<<
 *         f_ij_3: cy.float = (f_ij[3] if (firemod_size_class_ij == firemod_size_classes[3]) else 0.0)
 *         return f_ij_0 + f_ij_1 + f_ij_2 + f_ij_3
 */
    __pyx_t_2 = (__pyx_v_firemod_size_class_ij == __pyx_v_firemod_size_classes.f2);
    if (__pyx_t_2) {
      __pyx_t_1 = __pyx_v_f_ij.f2;
    } else {
      __pyx_t_1 = 0.0;
    }
    __pyx_v_f_ij_2 = __pyx_t_1;

    /* "pyretechnics/fuel_models.py":302
 *         f_ij_1: cy.float = (f_ij[1] if (firemod_size_class_ij == firemod_size_classes[1]) else 0.0)
 *         f_ij_2: cy.float = (f_ij[2] if (firemod_size_class_ij == firemod_size_classes[2]) else 0.0)
 *         f_ij_3: cy.float = (f_ij[3] if (firemod_size_class_ij == firemod_size_classes[3]) else 0.0)             # <<<<<<<<<<<<<<
 *         return f_ij_0 + f_ij_1 + f_ij_2 + f_ij_3
 *     else:
 */
    __pyx_t_2 = (__pyx_v_firemod_size_class_ij == __pyx_v_firemod_size_classes.f3);
    if (__pyx_t_2) {
      __pyx_t_1 = __pyx_v_f_ij.f3;
    } else {
      __pyx_t_1 = 0.0;
    }
    __pyx_v_f_ij_3 = __pyx_t_1;

    /* "pyretechnics/fuel_models.py":303
 *         f_ij_2: cy.float = (f_ij[2] if (firemod_size_class_ij == firemod_size_classes[2]) else 0.0)
 *         f_ij_3: cy.float = (f_ij[3] if (firemod_size_class_ij == firemod_size_classes[3]) else 0.0)
 *         return f_ij_0 + f_ij_1 + f_ij_2 + f_ij_3             # <<<<<<<<<<<<<<
 *     else:
 *         f_ij_4: cy.float = (f_ij[4] if (firemod_size_class_ij == firemod_size_classes[4]) else 0.0)
 */
    __pyx_r = (((__pyx_v_f_ij_0 + __pyx_v_f_ij_1) + __pyx_v_f_ij_2) + __pyx_v_f_ij_3);
    goto __pyx_L0;

    /* "pyretechnics/fuel_models.py":298
 *           trivially equal to f_ij[i].
 *     """
 *     if is_dead:             # <<<<<<<<<<<<<<
 *         f_ij_0: cy.float = (f_ij[0] if (firemod_size_class_ij == firemod_size_classes[0]) else 0.0)
 *         f_ij_1: cy.float = (f_ij[1] if (firemod_size_class_ij == firemod_size_classes[1]) else 0.0)
 */
  }

  /* "pyretechnics/fuel_models.py":305
 *         return f_ij_0 + f_ij_1 + f_ij_2 + f_ij_3
 *     else:
 *         f_ij_4: cy.float = (f_ij[4] if (firemod_size_class_ij == firemod_size_classes[4]) else 0.0)             # <<<<<<<<<<<<<<
 *         f_ij_5: cy.float = (f_ij[5] if (firemod_size_class_ij == firemod_size_classes[5]) else 0.0)
 *         return f_ij_4 + f_ij_5
 */
  /*else*/ {
    __pyx_t_2 = (__pyx_v_firemod_size_class_ij == __pyx_v_firemod_size_classes.f4);
    if (__pyx_t_2) {
      __pyx_t_1 = __pyx_v_f_ij.f4;
    } else {
      __pyx_t_1 = 0.0;
    }
    __pyx_v_f_ij_4 = __pyx_t_1;

    /* "pyretechnics/fuel_models.py":306
 *     else:
 *         f_ij_4: cy.float = (f_ij[4] if (firemod_size_class_ij == firemod_size_classes[4]) else 0.0)
 *         f_ij_5: cy.float = (f_ij[5] if (firemod_size_class_ij == firemod_size_classes[5]) else 0.0)             # <<<<<<<<<<<<<<
 *         return f_ij_4 + f_ij_5
 * 
 */
    __pyx_t_2 = (__pyx_v_firemod_size_class_ij == __pyx_v_firemod_size_classes.f5);
    if (__pyx_t_2) {
      __pyx_t_1 = __pyx_v_f_ij.f5;
    } else {
      __pyx_t_1 = 0.0;
    }
    __pyx_v_f_ij_5 = __pyx_t_1;

    /* "pyretechnics/fuel_models.py":307
 *         f_ij_4: cy.float = (f_ij[4] if (firemod_size_class_ij == firemod_size_classes[4]) else 0.0)
 *         f_ij_5: cy.float = (f_ij[5] if (firemod_size_class_ij == firemod_size_classes[5]) else 0.0)
 *         return f_ij_4 + f_ij_5             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_r = (__pyx_v_f_ij_4 + __pyx_v_f_ij_5);
    goto __pyx_L0;
  }

  /* "pyretechnics/fuel_models.py":286
 * # TODO: OPTIM pre-compute this conditional branching since it's fully determined by sigma.
 * #       This information might be represented efficiently in bit flags.
 * @cy.cfunc             # <<<<<<<<<<<<<<
 * @cy.exceptval(check=False)
 * def compute_gij(firemod_size_classes : fclaarr,
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyretechnics/fuel_models.py":310
 * 
 * 
 * @cy.cfunc             # <<<<<<<<<<<<<<
 * @cy.exceptval(check=False)
 * def add_weighting_factors(fuel_model: FuelModel) -> FuelModel:
 */

static struct __pyx_t_12pyretechnics_8cy_types_FuelModel __pyx_f_12pyretechnics_11fuel_models_add_weighting_factors(struct __pyx_t_12pyretechnics_8cy_types_FuelModel __pyx_v_fuel_model) {
  __pyx_t_12pyretechnics_8cy_types_fclaarr __pyx_v_w_o;
  __pyx_t_12pyretechnics_8cy_types_fclaarr __pyx_v_sigma;
  __pyx_t_12pyretechnics_8cy_types_fclaarr __pyx_v_rho_p;
  __pyx_t_12pyretechnics_8cy_types_fclaarr __pyx_v_A_ij;
  float __pyx_v_A_i_0;
  float __pyx_v_A_i_1;
  CYTHON_UNUSED __pyx_t_12pyretechnics_8cy_types_fcatarr __pyx_v_A_i;
  float __pyx_v_A_T;
  float __pyx_v_f_ij_0;
  float __pyx_v_f_ij_1;
  float __pyx_v_f_ij_2;
  float __pyx_v_f_ij_3;
  float __pyx_v_f_ij_4;
  float __pyx_v_f_ij_5;
  float __pyx_v_A_i_0_inv;
  float __pyx_v_A_i_1_inv;
  __pyx_t_12pyretechnics_8cy_types_fclaarr __pyx_v_f_ij;
  float __pyx_v_f_i_0;
  float __pyx_v_f_i_1;
  float __pyx_v_A_T_inv;
  __pyx_t_12pyretechnics_8cy_types_fcatarr __pyx_v_f_i;
  __pyx_t_12pyretechnics_8cy_types_fclaarr __pyx_v_firemod_size_classes;
  __pyx_t_12pyretechnics_8cy_types_fclaarr __pyx_v_g_ij;
  struct __pyx_t_12pyretechnics_8cy_types_FuelModel __pyx_r;
  __pyx_t_12pyretechnics_8cy_types_fclaarr __pyx_t_1;
  __pyx_t_12pyretechnics_8cy_types_fcatarr __pyx_t_2;
  int __pyx_t_3;

  /* "pyretechnics/fuel_models.py":317
 *     """
 *     # Calculate A_ij, A_i, and A_T
 *     w_o  : fclaarr  = fuel_model.w_o             # <<<<<<<<<<<<<<
 *     sigma: fclaarr  = fuel_model.sigma
 *     rho_p: fclaarr  = fuel_model.rho_p
 */
  __pyx_t_1 = __pyx_v_fuel_model.w_o;
  __pyx_v_w_o = __pyx_t_1;

  /* "pyretechnics/fuel_models.py":318
 *     # Calculate A_ij, A_i, and A_T
 *     w_o  : fclaarr  = fuel_model.w_o
 *     sigma: fclaarr  = fuel_model.sigma             # <<<<<<<<<<<<<<
 *     rho_p: fclaarr  = fuel_model.rho_p
 *     A_ij : fclaarr  = ((sigma[0] * w_o[0]) / rho_p[0],
 */
  __pyx_t_1 = __pyx_v_fuel_model.sigma;
  __pyx_v_sigma = __pyx_t_1;

  /* "pyretechnics/fuel_models.py":319
 *     w_o  : fclaarr  = fuel_model.w_o
 *     sigma: fclaarr  = fuel_model.sigma
 *     rho_p: fclaarr  = fuel_model.rho_p             # <<<<<<<<<<<<<<
 *     A_ij : fclaarr  = ((sigma[0] * w_o[0]) / rho_p[0],
 *                        (sigma[1] * w_o[1]) / rho_p[1],
 */
  __pyx_t_1 = __pyx_v_fuel_model.rho_p;
  __pyx_v_rho_p = __pyx_t_1;

  /* "pyretechnics/fuel_models.py":320
 *     sigma: fclaarr  = fuel_model.sigma
 *     rho_p: fclaarr  = fuel_model.rho_p
 *     A_ij : fclaarr  = ((sigma[0] * w_o[0]) / rho_p[0],             # <<<<<<<<<<<<<<
 *                        (sigma[1] * w_o[1]) / rho_p[1],
 *                        (sigma[2] * w_o[2]) / rho_p[2],
 */
  __pyx_t_1.f0 = ((__pyx_v_sigma.f0 * __pyx_v_w_o.f0) / __pyx_v_rho_p.f0);
  __pyx_t_1.f1 = ((__pyx_v_sigma.f1 * __pyx_v_w_o.f1) / __pyx_v_rho_p.f1);
  __pyx_t_1.f2 = ((__pyx_v_sigma.f2 * __pyx_v_w_o.f2) / __pyx_v_rho_p.f2);
  __pyx_t_1.f3 = ((__pyx_v_sigma.f3 * __pyx_v_w_o.f3) / __pyx_v_rho_p.f3);
  __pyx_t_1.f4 = ((__pyx_v_sigma.f4 * __pyx_v_w_o.f4) / __pyx_v_rho_p.f4);
  __pyx_t_1.f5 = ((__pyx_v_sigma.f5 * __pyx_v_w_o.f5) / __pyx_v_rho_p.f5);
  __pyx_v_A_ij = __pyx_t_1;

  /* "pyretechnics/fuel_models.py":326
 *                        (sigma[4] * w_o[4]) / rho_p[4],
 *                        (sigma[5] * w_o[5]) / rho_p[5]) # TODO: OPTIM pre-compute sigma/rho_p
 *     A_i_0: cy.float = A_ij[0] + A_ij[1] + A_ij[2] + A_ij[3]             # <<<<<<<<<<<<<<
 *     A_i_1: cy.float = A_ij[4] + A_ij[5]
 *     A_i  : fcatarr  = (A_i_0, A_i_1)
 */
  __pyx_v_A_i_0 = (((__pyx_v_A_ij.f0 + __pyx_v_A_ij.f1) + __pyx_v_A_ij.f2) + __pyx_v_A_ij.f3);

  /* "pyretechnics/fuel_models.py":327
 *                        (sigma[5] * w_o[5]) / rho_p[5]) # TODO: OPTIM pre-compute sigma/rho_p
 *     A_i_0: cy.float = A_ij[0] + A_ij[1] + A_ij[2] + A_ij[3]
 *     A_i_1: cy.float = A_ij[4] + A_ij[5]             # <<<<<<<<<<<<<<
 *     A_i  : fcatarr  = (A_i_0, A_i_1)
 *     A_T  : cy.float = A_i_0 + A_i_1
 */
  __pyx_v_A_i_1 = (__pyx_v_A_ij.f4 + __pyx_v_A_ij.f5);

  /* "pyretechnics/fuel_models.py":328
 *     A_i_0: cy.float = A_ij[0] + A_ij[1] + A_ij[2] + A_ij[3]
 *     A_i_1: cy.float = A_ij[4] + A_ij[5]
 *     A_i  : fcatarr  = (A_i_0, A_i_1)             # <<<<<<<<<<<<<<
 *     A_T  : cy.float = A_i_0 + A_i_1
 *     # Calculate f_ij
 */
  __pyx_t_2.f0 = __pyx_v_A_i_0;
  __pyx_t_2.f1 = __pyx_v_A_i_1;
  __pyx_v_A_i = __pyx_t_2;

  /* "pyretechnics/fuel_models.py":329
 *     A_i_1: cy.float = A_ij[4] + A_ij[5]
 *     A_i  : fcatarr  = (A_i_0, A_i_1)
 *     A_T  : cy.float = A_i_0 + A_i_1             # <<<<<<<<<<<<<<
 *     # Calculate f_ij
 *     f_ij_0   : cy.float = 0.0
 */
  __pyx_v_A_T = (__pyx_v_A_i_0 + __pyx_v_A_i_1);

  /* "pyretechnics/fuel_models.py":331
 *     A_T  : cy.float = A_i_0 + A_i_1
 *     # Calculate f_ij
 *     f_ij_0   : cy.float = 0.0             # <<<<<<<<<<<<<<
 *     f_ij_1   : cy.float = 0.0
 *     f_ij_2   : cy.float = 0.0
 */
  __pyx_v_f_ij_0 = 0.0;

  /* "pyretechnics/fuel_models.py":332
 *     # Calculate f_ij
 *     f_ij_0   : cy.float = 0.0
 *     f_ij_1   : cy.float = 0.0             # <<<<<<<<<<<<<<
 *     f_ij_2   : cy.float = 0.0
 *     f_ij_3   : cy.float = 0.0
 */
  __pyx_v_f_ij_1 = 0.0;

  /* "pyretechnics/fuel_models.py":333
 *     f_ij_0   : cy.float = 0.0
 *     f_ij_1   : cy.float = 0.0
 *     f_ij_2   : cy.float = 0.0             # <<<<<<<<<<<<<<
 *     f_ij_3   : cy.float = 0.0
 *     f_ij_4   : cy.float = 0.0
 */
  __pyx_v_f_ij_2 = 0.0;

  /* "pyretechnics/fuel_models.py":334
 *     f_ij_1   : cy.float = 0.0
 *     f_ij_2   : cy.float = 0.0
 *     f_ij_3   : cy.float = 0.0             # <<<<<<<<<<<<<<
 *     f_ij_4   : cy.float = 0.0
 *     f_ij_5   : cy.float = 0.0
 */
  __pyx_v_f_ij_3 = 0.0;

  /* "pyretechnics/fuel_models.py":335
 *     f_ij_2   : cy.float = 0.0
 *     f_ij_3   : cy.float = 0.0
 *     f_ij_4   : cy.float = 0.0             # <<<<<<<<<<<<<<
 *     f_ij_5   : cy.float = 0.0
 *     A_i_0_inv: cy.float
 */
  __pyx_v_f_ij_4 = 0.0;

  /* "pyretechnics/fuel_models.py":336
 *     f_ij_3   : cy.float = 0.0
 *     f_ij_4   : cy.float = 0.0
 *     f_ij_5   : cy.float = 0.0             # <<<<<<<<<<<<<<
 *     A_i_0_inv: cy.float
 *     A_i_1_inv: cy.float
 */
  __pyx_v_f_ij_5 = 0.0;

  /* "pyretechnics/fuel_models.py":339
 *     A_i_0_inv: cy.float
 *     A_i_1_inv: cy.float
 *     if A_i_0 > almost_zero:             # <<<<<<<<<<<<<<
 *         A_i_0_inv = 1.0 / A_i_0
 *         f_ij_0    = A_ij[0] * A_i_0_inv
 */
  __pyx_t_3 = (__pyx_v_A_i_0 > __pyx_v_12pyretechnics_11fuel_models_almost_zero);
  if (__pyx_t_3) {

    /* "pyretechnics/fuel_models.py":340
 *     A_i_1_inv: cy.float
 *     if A_i_0 > almost_zero:
 *         A_i_0_inv = 1.0 / A_i_0             # <<<<<<<<<<<<<<
 *         f_ij_0    = A_ij[0] * A_i_0_inv
 *         f_ij_1    = A_ij[1] * A_i_0_inv
 */
    __pyx_v_A_i_0_inv = (1.0 / ((double)__pyx_v_A_i_0));

    /* "pyretechnics/fuel_models.py":341
 *     if A_i_0 > almost_zero:
 *         A_i_0_inv = 1.0 / A_i_0
 *         f_ij_0    = A_ij[0] * A_i_0_inv             # <<<<<<<<<<<<<<
 *         f_ij_1    = A_ij[1] * A_i_0_inv
 *         f_ij_2    = A_ij[2] * A_i_0_inv
 */
    __pyx_v_f_ij_0 = (__pyx_v_A_ij.f0 * __pyx_v_A_i_0_inv);

    /* "pyretechnics/fuel_models.py":342
 *         A_i_0_inv = 1.0 / A_i_0
 *         f_ij_0    = A_ij[0] * A_i_0_inv
 *         f_ij_1    = A_ij[1] * A_i_0_inv             # <<<<<<<<<<<<<<
 *         f_ij_2    = A_ij[2] * A_i_0_inv
 *         f_ij_3    = A_ij[3] * A_i_0_inv
 */
    __pyx_v_f_ij_1 = (__pyx_v_A_ij.f1 * __pyx_v_A_i_0_inv);

    /* "pyretechnics/fuel_models.py":343
 *         f_ij_0    = A_ij[0] * A_i_0_inv
 *         f_ij_1    = A_ij[1] * A_i_0_inv
 *         f_ij_2    = A_ij[2] * A_i_0_inv             # <<<<<<<<<<<<<<
 *         f_ij_3    = A_ij[3] * A_i_0_inv
 *     if A_i_1 > almost_zero:
 */
    __pyx_v_f_ij_2 = (__pyx_v_A_ij.f2 * __pyx_v_A_i_0_inv);

    /* "pyretechnics/fuel_models.py":344
 *         f_ij_1    = A_ij[1] * A_i_0_inv
 *         f_ij_2    = A_ij[2] * A_i_0_inv
 *         f_ij_3    = A_ij[3] * A_i_0_inv             # <<<<<<<<<<<<<<
 *     if A_i_1 > almost_zero:
 *         A_i_1_inv = 1.0 / A_i_1
 */
    __pyx_v_f_ij_3 = (__pyx_v_A_ij.f3 * __pyx_v_A_i_0_inv);

    /* "pyretechnics/fuel_models.py":339
 *     A_i_0_inv: cy.float
 *     A_i_1_inv: cy.float
 *     if A_i_0 > almost_zero:             # <<<<<<<<<<<<<<
 *         A_i_0_inv = 1.0 / A_i_0
 *         f_ij_0    = A_ij[0] * A_i_0_inv
 */
  }

  /* "pyretechnics/fuel_models.py":345
 *         f_ij_2    = A_ij[2] * A_i_0_inv
 *         f_ij_3    = A_ij[3] * A_i_0_inv
 *     if A_i_1 > almost_zero:             # <<<<<<<<<<<<<<
 *         A_i_1_inv = 1.0 / A_i_1
 *         f_ij_4    = A_ij[4] * A_i_1_inv
 */
  __pyx_t_3 = (__pyx_v_A_i_1 > __pyx_v_12pyretechnics_11fuel_models_almost_zero);
  if (__pyx_t_3) {

    /* "pyretechnics/fuel_models.py":346
 *         f_ij_3    = A_ij[3] * A_i_0_inv
 *     if A_i_1 > almost_zero:
 *         A_i_1_inv = 1.0 / A_i_1             # <<<<<<<<<<<<<<
 *         f_ij_4    = A_ij[4] * A_i_1_inv
 *         f_ij_5    = A_ij[5] * A_i_1_inv
 */
    __pyx_v_A_i_1_inv = (1.0 / ((double)__pyx_v_A_i_1));

    /* "pyretechnics/fuel_models.py":347
 *     if A_i_1 > almost_zero:
 *         A_i_1_inv = 1.0 / A_i_1
 *         f_ij_4    = A_ij[4] * A_i_1_inv             # <<<<<<<<<<<<<<
 *         f_ij_5    = A_ij[5] * A_i_1_inv
 *     f_ij: fclaarr = (f_ij_0, f_ij_1, f_ij_2, f_ij_3, f_ij_4, f_ij_5)
 */
    __pyx_v_f_ij_4 = (__pyx_v_A_ij.f4 * __pyx_v_A_i_1_inv);

    /* "pyretechnics/fuel_models.py":348
 *         A_i_1_inv = 1.0 / A_i_1
 *         f_ij_4    = A_ij[4] * A_i_1_inv
 *         f_ij_5    = A_ij[5] * A_i_1_inv             # <<<<<<<<<<<<<<
 *     f_ij: fclaarr = (f_ij_0, f_ij_1, f_ij_2, f_ij_3, f_ij_4, f_ij_5)
 *     # Calculate f_i
 */
    __pyx_v_f_ij_5 = (__pyx_v_A_ij.f5 * __pyx_v_A_i_1_inv);

    /* "pyretechnics/fuel_models.py":345
 *         f_ij_2    = A_ij[2] * A_i_0_inv
 *         f_ij_3    = A_ij[3] * A_i_0_inv
 *     if A_i_1 > almost_zero:             # <<<<<<<<<<<<<<
 *         A_i_1_inv = 1.0 / A_i_1
 *         f_ij_4    = A_ij[4] * A_i_1_inv
 */
  }

  /* "pyretechnics/fuel_models.py":349
 *         f_ij_4    = A_ij[4] * A_i_1_inv
 *         f_ij_5    = A_ij[5] * A_i_1_inv
 *     f_ij: fclaarr = (f_ij_0, f_ij_1, f_ij_2, f_ij_3, f_ij_4, f_ij_5)             # <<<<<<<<<<<<<<
 *     # Calculate f_i
 *     f_i_0  : cy.float = 0.0
 */
  __pyx_t_1.f0 = __pyx_v_f_ij_0;
  __pyx_t_1.f1 = __pyx_v_f_ij_1;
  __pyx_t_1.f2 = __pyx_v_f_ij_2;
  __pyx_t_1.f3 = __pyx_v_f_ij_3;
  __pyx_t_1.f4 = __pyx_v_f_ij_4;
  __pyx_t_1.f5 = __pyx_v_f_ij_5;
  __pyx_v_f_ij = __pyx_t_1;

  /* "pyretechnics/fuel_models.py":351
 *     f_ij: fclaarr = (f_ij_0, f_ij_1, f_ij_2, f_ij_3, f_ij_4, f_ij_5)
 *     # Calculate f_i
 *     f_i_0  : cy.float = 0.0             # <<<<<<<<<<<<<<
 *     f_i_1  : cy.float = 0.0
 *     A_T_inv: cy.float
 */
  __pyx_v_f_i_0 = 0.0;

  /* "pyretechnics/fuel_models.py":352
 *     # Calculate f_i
 *     f_i_0  : cy.float = 0.0
 *     f_i_1  : cy.float = 0.0             # <<<<<<<<<<<<<<
 *     A_T_inv: cy.float
 *     if A_T > almost_zero:
 */
  __pyx_v_f_i_1 = 0.0;

  /* "pyretechnics/fuel_models.py":354
 *     f_i_1  : cy.float = 0.0
 *     A_T_inv: cy.float
 *     if A_T > almost_zero:             # <<<<<<<<<<<<<<
 *         A_T_inv = 1.0 / A_T
 *         f_i_0   = A_i_0 * A_T_inv
 */
  __pyx_t_3 = (__pyx_v_A_T > __pyx_v_12pyretechnics_11fuel_models_almost_zero);
  if (__pyx_t_3) {

    /* "pyretechnics/fuel_models.py":355
 *     A_T_inv: cy.float
 *     if A_T > almost_zero:
 *         A_T_inv = 1.0 / A_T             # <<<<<<<<<<<<<<
 *         f_i_0   = A_i_0 * A_T_inv
 *         f_i_1   = A_i_1 * A_T_inv
 */
    __pyx_v_A_T_inv = (1.0 / ((double)__pyx_v_A_T));

    /* "pyretechnics/fuel_models.py":356
 *     if A_T > almost_zero:
 *         A_T_inv = 1.0 / A_T
 *         f_i_0   = A_i_0 * A_T_inv             # <<<<<<<<<<<<<<
 *         f_i_1   = A_i_1 * A_T_inv
 *     f_i: fcatarr = (f_i_0, f_i_1)
 */
    __pyx_v_f_i_0 = (__pyx_v_A_i_0 * __pyx_v_A_T_inv);

    /* "pyretechnics/fuel_models.py":357
 *         A_T_inv = 1.0 / A_T
 *         f_i_0   = A_i_0 * A_T_inv
 *         f_i_1   = A_i_1 * A_T_inv             # <<<<<<<<<<<<<<
 *     f_i: fcatarr = (f_i_0, f_i_1)
 *     # Calculate g_ij
 */
    __pyx_v_f_i_1 = (__pyx_v_A_i_1 * __pyx_v_A_T_inv);

    /* "pyretechnics/fuel_models.py":354
 *     f_i_1  : cy.float = 0.0
 *     A_T_inv: cy.float
 *     if A_T > almost_zero:             # <<<<<<<<<<<<<<
 *         A_T_inv = 1.0 / A_T
 *         f_i_0   = A_i_0 * A_T_inv
 */
  }

  /* "pyretechnics/fuel_models.py":358
 *         f_i_0   = A_i_0 * A_T_inv
 *         f_i_1   = A_i_1 * A_T_inv
 *     f_i: fcatarr = (f_i_0, f_i_1)             # <<<<<<<<<<<<<<
 *     # Calculate g_ij
 *     firemod_size_classes: fclaarr = fuel_model.firemod_size_classes
 */
  __pyx_t_2.f0 = __pyx_v_f_i_0;
  __pyx_t_2.f1 = __pyx_v_f_i_1;
  __pyx_v_f_i = __pyx_t_2;

  /* "pyretechnics/fuel_models.py":360
 *     f_i: fcatarr = (f_i_0, f_i_1)
 *     # Calculate g_ij
 *     firemod_size_classes: fclaarr = fuel_model.firemod_size_classes             # <<<<<<<<<<<<<<
 *     g_ij                : fclaarr = (compute_gij(firemod_size_classes, f_ij, firemod_size_classes[0], True),
 *                                      compute_gij(firemod_size_classes, f_ij, firemod_size_classes[1], True),
 */
  __pyx_t_1 = __pyx_v_fuel_model.firemod_size_classes;
  __pyx_v_firemod_size_classes = __pyx_t_1;

  /* "pyretechnics/fuel_models.py":361
 *     # Calculate g_ij
 *     firemod_size_classes: fclaarr = fuel_model.firemod_size_classes
 *     g_ij                : fclaarr = (compute_gij(firemod_size_classes, f_ij, firemod_size_classes[0], True),             # <<<<<<<<<<<<<<
 *                                      compute_gij(firemod_size_classes, f_ij, firemod_size_classes[1], True),
 *                                      compute_gij(firemod_size_classes, f_ij, firemod_size_classes[2], True),
 */
  __pyx_t_1.f0 = __pyx_f_12pyretechnics_11fuel_models_compute_gij(__pyx_v_firemod_size_classes, __pyx_v_f_ij, __pyx_v_firemod_size_classes.f0, 1);
  __pyx_t_1.f1 = __pyx_f_12pyretechnics_11fuel_models_compute_gij(__pyx_v_firemod_size_classes, __pyx_v_f_ij, __pyx_v_firemod_size_classes.f1, 1);
  __pyx_t_1.f2 = __pyx_f_12pyretechnics_11fuel_models_compute_gij(__pyx_v_firemod_size_classes, __pyx_v_f_ij, __pyx_v_firemod_size_classes.f2, 1);
  __pyx_t_1.f3 = __pyx_f_12pyretechnics_11fuel_models_compute_gij(__pyx_v_firemod_size_classes, __pyx_v_f_ij, __pyx_v_firemod_size_classes.f3, 1);
  __pyx_t_1.f4 = __pyx_f_12pyretechnics_11fuel_models_compute_gij(__pyx_v_firemod_size_classes, __pyx_v_f_ij, __pyx_v_firemod_size_classes.f4, 0);
  __pyx_t_1.f5 = __pyx_f_12pyretechnics_11fuel_models_compute_gij(__pyx_v_firemod_size_classes, __pyx_v_f_ij, __pyx_v_firemod_size_classes.f5, 0);
  __pyx_v_g_ij = __pyx_t_1;

  /* "pyretechnics/fuel_models.py":368
 *                                      compute_gij(firemod_size_classes, f_ij, firemod_size_classes[5], False))
 *     # Update the passed in fuel_model to use f_ij, f_i, and g_ij
 *     fuel_model.f_ij = f_ij             # <<<<<<<<<<<<<<
 *     fuel_model.f_i  = f_i
 *     fuel_model.g_ij = g_ij
 */
  __pyx_v_fuel_model.f_ij = __pyx_v_f_ij;

  /* "pyretechnics/fuel_models.py":369
 *     # Update the passed in fuel_model to use f_ij, f_i, and g_ij
 *     fuel_model.f_ij = f_ij
 *     fuel_model.f_i  = f_i             # <<<<<<<<<<<<<<
 *     fuel_model.g_ij = g_ij
 *     return fuel_model
 */
  __pyx_v_fuel_model.f_i = __pyx_v_f_i;

  /* "pyretechnics/fuel_models.py":370
 *     fuel_model.f_ij = f_ij
 *     fuel_model.f_i  = f_i
 *     fuel_model.g_ij = g_ij             # <<<<<<<<<<<<<<
 *     return fuel_model
 * # add-weighting-factors ends here
 */
  __pyx_v_fuel_model.g_ij = __pyx_v_g_ij;

  /* "pyretechnics/fuel_models.py":371
 *     fuel_model.f_i  = f_i
 *     fuel_model.g_ij = g_ij
 *     return fuel_model             # <<<<<<<<<<<<<<
 * # add-weighting-factors ends here
 * # [[file:../../org/pyretechnics.org::add-live-moisture-of-extinction][add-live-moisture-of-extinction]]
 */
  __pyx_r = __pyx_v_fuel_model;
  goto __pyx_L0;

  /* "pyretechnics/fuel_models.py":310
 * 
 * 
 * @cy.cfunc             # <<<<<<<<<<<<<<
 * @cy.exceptval(check=False)
 * def add_weighting_factors(fuel_model: FuelModel) -> FuelModel:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyretechnics/fuel_models.py":374
 * # add-weighting-factors ends here
 * # [[file:../../org/pyretechnics.org::add-live-moisture-of-extinction][add-live-moisture-of-extinction]]
 * @cy.cfunc             # <<<<<<<<<<<<<<
 * @cy.exceptval(check=False)
 * def add_live_moisture_of_extinction(fuel_model: FuelModel) -> FuelModel:
 */

static struct __pyx_t_12pyretechnics_8cy_types_FuelModel __pyx_f_12pyretechnics_11fuel_models_add_live_moisture_of_extinction(struct __pyx_t_12pyretechnics_8cy_types_FuelModel __pyx_v_fuel_model) {
  __pyx_t_12pyretechnics_8cy_types_fclaarr __pyx_v_w_o;
  __pyx_t_12pyretechnics_8cy_types_fclaarr __pyx_v_exp_A_sigma;
  __pyx_t_12pyretechnics_8cy_types_fclaarr __pyx_v_M_f;
  __pyx_t_12pyretechnics_8cy_types_fclaarr __pyx_v_M_x;
  float __pyx_v_M_x_dead;
  __pyx_t_12pyretechnics_8cy_types_fclaarr __pyx_v_loading_factors;
  float __pyx_v_dead_moisture_factor;
  float __pyx_v_dead_loading_factor;
  float __pyx_v_live_loading_factor;
  float __pyx_v_dead_fuel_moisture;
  float __pyx_v_dead_to_live_ratio;
  float __pyx_v_M_x_live;
  __pyx_t_12pyretechnics_8cy_types_fclaarr __pyx_v_M_x_new;
  struct __pyx_t_12pyretechnics_8cy_types_FuelModel __pyx_r;
  __pyx_t_12pyretechnics_8cy_types_fclaarr __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  double __pyx_t_4;
  float __pyx_t_5;
  double __pyx_t_6;

  /* "pyretechnics/fuel_models.py":383
 *     """
 *     # Calculate dead_moisture_factor, dead_loading_factor, and live_loading_factor
 *     w_o                 : fclaarr  = fuel_model.w_o             # <<<<<<<<<<<<<<
 *     exp_A_sigma         : fclaarr  = fuel_model.exp_A_sigma
 *     M_f                 : fclaarr  = fuel_model.M_f
 */
  __pyx_t_1 = __pyx_v_fuel_model.w_o;
  __pyx_v_w_o = __pyx_t_1;

  /* "pyretechnics/fuel_models.py":384
 *     # Calculate dead_moisture_factor, dead_loading_factor, and live_loading_factor
 *     w_o                 : fclaarr  = fuel_model.w_o
 *     exp_A_sigma         : fclaarr  = fuel_model.exp_A_sigma             # <<<<<<<<<<<<<<
 *     M_f                 : fclaarr  = fuel_model.M_f
 *     M_x                 : fclaarr  = fuel_model.M_x
 */
  __pyx_t_1 = __pyx_v_fuel_model.exp_A_sigma;
  __pyx_v_exp_A_sigma = __pyx_t_1;

  /* "pyretechnics/fuel_models.py":385
 *     w_o                 : fclaarr  = fuel_model.w_o
 *     exp_A_sigma         : fclaarr  = fuel_model.exp_A_sigma
 *     M_f                 : fclaarr  = fuel_model.M_f             # <<<<<<<<<<<<<<
 *     M_x                 : fclaarr  = fuel_model.M_x
 *     M_x_dead            : cy.float = M_x[0]
 */
  __pyx_t_1 = __pyx_v_fuel_model.M_f;
  __pyx_v_M_f = __pyx_t_1;

  /* "pyretechnics/fuel_models.py":386
 *     exp_A_sigma         : fclaarr  = fuel_model.exp_A_sigma
 *     M_f                 : fclaarr  = fuel_model.M_f
 *     M_x                 : fclaarr  = fuel_model.M_x             # <<<<<<<<<<<<<<
 *     M_x_dead            : cy.float = M_x[0]
 *     loading_factors     : fclaarr  = (w_o[0] * exp_A_sigma[0],
 */
  __pyx_t_1 = __pyx_v_fuel_model.M_x;
  __pyx_v_M_x = __pyx_t_1;

  /* "pyretechnics/fuel_models.py":387
 *     M_f                 : fclaarr  = fuel_model.M_f
 *     M_x                 : fclaarr  = fuel_model.M_x
 *     M_x_dead            : cy.float = M_x[0]             # <<<<<<<<<<<<<<
 *     loading_factors     : fclaarr  = (w_o[0] * exp_A_sigma[0],
 *                                       w_o[1] * exp_A_sigma[1],
 */
  __pyx_v_M_x_dead = __pyx_v_M_x.f0;

  /* "pyretechnics/fuel_models.py":388
 *     M_x                 : fclaarr  = fuel_model.M_x
 *     M_x_dead            : cy.float = M_x[0]
 *     loading_factors     : fclaarr  = (w_o[0] * exp_A_sigma[0],             # <<<<<<<<<<<<<<
 *                                       w_o[1] * exp_A_sigma[1],
 *                                       w_o[2] * exp_A_sigma[2],
 */
  __pyx_t_1.f0 = (__pyx_v_w_o.f0 * __pyx_v_exp_A_sigma.f0);
  __pyx_t_1.f1 = (__pyx_v_w_o.f1 * __pyx_v_exp_A_sigma.f1);
  __pyx_t_1.f2 = (__pyx_v_w_o.f2 * __pyx_v_exp_A_sigma.f2);
  __pyx_t_1.f3 = (__pyx_v_w_o.f3 * __pyx_v_exp_A_sigma.f3);
  __pyx_t_1.f4 = (__pyx_v_w_o.f4 * __pyx_v_exp_A_sigma.f4);
  __pyx_t_1.f5 = (__pyx_v_w_o.f5 * __pyx_v_exp_A_sigma.f5);
  __pyx_v_loading_factors = __pyx_t_1;

  /* "pyretechnics/fuel_models.py":396
 *     dead_moisture_factor: cy.float = (M_f[0] * loading_factors[0] +
 *                                       M_f[1] * loading_factors[1] +
 *                                       M_f[2] * loading_factors[2] +             # <<<<<<<<<<<<<<
 *                                       M_f[3] * loading_factors[3])
 *     dead_loading_factor : cy.float = loading_factors[0] + loading_factors[1] + loading_factors[2] + loading_factors[3]
 */
  __pyx_v_dead_moisture_factor = ((((__pyx_v_M_f.f0 * __pyx_v_loading_factors.f0) + (__pyx_v_M_f.f1 * __pyx_v_loading_factors.f1)) + (__pyx_v_M_f.f2 * __pyx_v_loading_factors.f2)) + (__pyx_v_M_f.f3 * __pyx_v_loading_factors.f3));

  /* "pyretechnics/fuel_models.py":398
 *                                       M_f[2] * loading_factors[2] +
 *                                       M_f[3] * loading_factors[3])
 *     dead_loading_factor : cy.float = loading_factors[0] + loading_factors[1] + loading_factors[2] + loading_factors[3]             # <<<<<<<<<<<<<<
 *     live_loading_factor : cy.float = loading_factors[4] + loading_factors[5]
 *     # Calculate M_x_live
 */
  __pyx_v_dead_loading_factor = (((__pyx_v_loading_factors.f0 + __pyx_v_loading_factors.f1) + __pyx_v_loading_factors.f2) + __pyx_v_loading_factors.f3);

  /* "pyretechnics/fuel_models.py":399
 *                                       M_f[3] * loading_factors[3])
 *     dead_loading_factor : cy.float = loading_factors[0] + loading_factors[1] + loading_factors[2] + loading_factors[3]
 *     live_loading_factor : cy.float = loading_factors[4] + loading_factors[5]             # <<<<<<<<<<<<<<
 *     # Calculate M_x_live
 *     dead_fuel_moisture: cy.float
 */
  __pyx_v_live_loading_factor = (__pyx_v_loading_factors.f4 + __pyx_v_loading_factors.f5);

  /* "pyretechnics/fuel_models.py":404
 *     dead_to_live_ratio: cy.float
 *     M_x_live          : cy.float
 *     if (dead_loading_factor > almost_zero and live_loading_factor > almost_zero):             # <<<<<<<<<<<<<<
 *         dead_fuel_moisture = dead_moisture_factor / dead_loading_factor
 *         dead_to_live_ratio = dead_loading_factor / live_loading_factor
 */
  __pyx_t_3 = (__pyx_v_dead_loading_factor > __pyx_v_12pyretechnics_11fuel_models_almost_zero);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (__pyx_v_live_loading_factor > __pyx_v_12pyretechnics_11fuel_models_almost_zero);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "pyretechnics/fuel_models.py":405
 *     M_x_live          : cy.float
 *     if (dead_loading_factor > almost_zero and live_loading_factor > almost_zero):
 *         dead_fuel_moisture = dead_moisture_factor / dead_loading_factor             # <<<<<<<<<<<<<<
 *         dead_to_live_ratio = dead_loading_factor / live_loading_factor
 *         M_x_live           = max(M_x_dead,
 */
    __pyx_v_dead_fuel_moisture = (__pyx_v_dead_moisture_factor / __pyx_v_dead_loading_factor);

    /* "pyretechnics/fuel_models.py":406
 *     if (dead_loading_factor > almost_zero and live_loading_factor > almost_zero):
 *         dead_fuel_moisture = dead_moisture_factor / dead_loading_factor
 *         dead_to_live_ratio = dead_loading_factor / live_loading_factor             # <<<<<<<<<<<<<<
 *         M_x_live           = max(M_x_dead,
 *                                  (2.9 * dead_to_live_ratio * (1.0 - (dead_fuel_moisture / M_x_dead))) - 0.226)
 */
    __pyx_v_dead_to_live_ratio = (__pyx_v_dead_loading_factor / __pyx_v_live_loading_factor);

    /* "pyretechnics/fuel_models.py":408
 *         dead_to_live_ratio = dead_loading_factor / live_loading_factor
 *         M_x_live           = max(M_x_dead,
 *                                  (2.9 * dead_to_live_ratio * (1.0 - (dead_fuel_moisture / M_x_dead))) - 0.226)             # <<<<<<<<<<<<<<
 *     else:
 *         M_x_live = M_x_dead
 */
    __pyx_t_4 = (((2.9 * __pyx_v_dead_to_live_ratio) * (1.0 - (__pyx_v_dead_fuel_moisture / __pyx_v_M_x_dead))) - 0.226);

    /* "pyretechnics/fuel_models.py":407
 *         dead_fuel_moisture = dead_moisture_factor / dead_loading_factor
 *         dead_to_live_ratio = dead_loading_factor / live_loading_factor
 *         M_x_live           = max(M_x_dead,             # <<<<<<<<<<<<<<
 *                                  (2.9 * dead_to_live_ratio * (1.0 - (dead_fuel_moisture / M_x_dead))) - 0.226)
 *     else:
 */
    __pyx_t_5 = __pyx_v_M_x_dead;

    /* "pyretechnics/fuel_models.py":408
 *         dead_to_live_ratio = dead_loading_factor / live_loading_factor
 *         M_x_live           = max(M_x_dead,
 *                                  (2.9 * dead_to_live_ratio * (1.0 - (dead_fuel_moisture / M_x_dead))) - 0.226)             # <<<<<<<<<<<<<<
 *     else:
 *         M_x_live = M_x_dead
 */
    __pyx_t_2 = (__pyx_t_4 > __pyx_t_5);
    if (__pyx_t_2) {
      __pyx_t_6 = __pyx_t_4;
    } else {
      __pyx_t_6 = __pyx_t_5;
    }
    __pyx_v_M_x_live = __pyx_t_6;

    /* "pyretechnics/fuel_models.py":404
 *     dead_to_live_ratio: cy.float
 *     M_x_live          : cy.float
 *     if (dead_loading_factor > almost_zero and live_loading_factor > almost_zero):             # <<<<<<<<<<<<<<
 *         dead_fuel_moisture = dead_moisture_factor / dead_loading_factor
 *         dead_to_live_ratio = dead_loading_factor / live_loading_factor
 */
    goto __pyx_L3;
  }

  /* "pyretechnics/fuel_models.py":410
 *                                  (2.9 * dead_to_live_ratio * (1.0 - (dead_fuel_moisture / M_x_dead))) - 0.226)
 *     else:
 *         M_x_live = M_x_dead             # <<<<<<<<<<<<<<
 *     # Calculate M_x_new
 *     M_x_new: fclaarr = (M_x[0],
 */
  /*else*/ {
    __pyx_v_M_x_live = __pyx_v_M_x_dead;
  }
  __pyx_L3:;

  /* "pyretechnics/fuel_models.py":412
 *         M_x_live = M_x_dead
 *     # Calculate M_x_new
 *     M_x_new: fclaarr = (M_x[0],             # <<<<<<<<<<<<<<
 *                         M_x[1],
 *                         M_x[2],
 */
  __pyx_t_1.f0 = __pyx_v_M_x.f0;
  __pyx_t_1.f1 = __pyx_v_M_x.f1;
  __pyx_t_1.f2 = __pyx_v_M_x.f2;
  __pyx_t_1.f3 = __pyx_v_M_x.f3;
  __pyx_t_1.f4 = __pyx_v_M_x_live;
  __pyx_t_1.f5 = __pyx_v_M_x_live;
  __pyx_v_M_x_new = __pyx_t_1;

  /* "pyretechnics/fuel_models.py":419
 *                         M_x_live)
 *     # Update the passed in fuel_model to use M_x_new
 *     fuel_model.M_x = M_x_new             # <<<<<<<<<<<<<<
 *     return fuel_model
 * # add-live-moisture-of-extinction ends here
 */
  __pyx_v_fuel_model.M_x = __pyx_v_M_x_new;

  /* "pyretechnics/fuel_models.py":420
 *     # Update the passed in fuel_model to use M_x_new
 *     fuel_model.M_x = M_x_new
 *     return fuel_model             # <<<<<<<<<<<<<<
 * # add-live-moisture-of-extinction ends here
 * # [[file:../../org/pyretechnics.org::moisturize][moisturize]]
 */
  __pyx_r = __pyx_v_fuel_model;
  goto __pyx_L0;

  /* "pyretechnics/fuel_models.py":374
 * # add-weighting-factors ends here
 * # [[file:../../org/pyretechnics.org::add-live-moisture-of-extinction][add-live-moisture-of-extinction]]
 * @cy.cfunc             # <<<<<<<<<<<<<<
 * @cy.exceptval(check=False)
 * def add_live_moisture_of_extinction(fuel_model: FuelModel) -> FuelModel:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyretechnics/fuel_models.py":423
 * # add-live-moisture-of-extinction ends here
 * # [[file:../../org/pyretechnics.org::moisturize][moisturize]]
 * @cy.ccall             # <<<<<<<<<<<<<<
 * @cy.exceptval(check=False)
 * def moisturize(fuel_model: FuelModel, fuel_moisture: fclaarr) -> FuelModel:
 */

static PyObject *__pyx_pw_12pyretechnics_11fuel_models_9moisturize(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_t_12pyretechnics_8cy_types_FuelModel __pyx_f_12pyretechnics_11fuel_models_moisturize(struct __pyx_t_12pyretechnics_8cy_types_FuelModel __pyx_v_fuel_model, __pyx_t_12pyretechnics_8cy_types_fclaarr __pyx_v_fuel_moisture, CYTHON_UNUSED int __pyx_skip_dispatch) {
  struct __pyx_t_12pyretechnics_8cy_types_FuelModel __pyx_v_fuel_model_copy;
  struct __pyx_t_12pyretechnics_8cy_types_FuelModel __pyx_v_dynamic_fuel_model;
  struct __pyx_t_12pyretechnics_8cy_types_FuelModel __pyx_v_weighted_fuel_model;
  struct __pyx_t_12pyretechnics_8cy_types_FuelModel __pyx_v_moisturized_fuel_model;
  struct __pyx_t_12pyretechnics_8cy_types_FuelModel __pyx_r;
  struct __pyx_t_12pyretechnics_8cy_types_FuelModel __pyx_t_1;

  /* "pyretechnics/fuel_models.py":432
 *     # Create a copy of fuel_model
 *     fuel_model_copy: FuelModel = FuelModel(
 *         number               = fuel_model.number,             # <<<<<<<<<<<<<<
 *         delta                = fuel_model.delta,
 *         M_x                  = fuel_model.M_x,
 */
  __pyx_t_1.number = __pyx_v_fuel_model.number;

  /* "pyretechnics/fuel_models.py":433
 *     fuel_model_copy: FuelModel = FuelModel(
 *         number               = fuel_model.number,
 *         delta                = fuel_model.delta,             # <<<<<<<<<<<<<<
 *         M_x                  = fuel_model.M_x,
 *         M_f                  = fuel_model.M_f,
 */
  __pyx_t_1.delta = __pyx_v_fuel_model.delta;

  /* "pyretechnics/fuel_models.py":434
 *         number               = fuel_model.number,
 *         delta                = fuel_model.delta,
 *         M_x                  = fuel_model.M_x,             # <<<<<<<<<<<<<<
 *         M_f                  = fuel_model.M_f,
 *         w_o                  = fuel_model.w_o,
 */
  __pyx_t_1.M_x = __pyx_v_fuel_model.M_x;

  /* "pyretechnics/fuel_models.py":435
 *         delta                = fuel_model.delta,
 *         M_x                  = fuel_model.M_x,
 *         M_f                  = fuel_model.M_f,             # <<<<<<<<<<<<<<
 *         w_o                  = fuel_model.w_o,
 *         sigma                = fuel_model.sigma,
 */
  __pyx_t_1.M_f = __pyx_v_fuel_model.M_f;

  /* "pyretechnics/fuel_models.py":436
 *         M_x                  = fuel_model.M_x,
 *         M_f                  = fuel_model.M_f,
 *         w_o                  = fuel_model.w_o,             # <<<<<<<<<<<<<<
 *         sigma                = fuel_model.sigma,
 *         h                    = fuel_model.h,
 */
  __pyx_t_1.w_o = __pyx_v_fuel_model.w_o;

  /* "pyretechnics/fuel_models.py":437
 *         M_f                  = fuel_model.M_f,
 *         w_o                  = fuel_model.w_o,
 *         sigma                = fuel_model.sigma,             # <<<<<<<<<<<<<<
 *         h                    = fuel_model.h,
 *         rho_p                = fuel_model.rho_p,
 */
  __pyx_t_1.sigma = __pyx_v_fuel_model.sigma;

  /* "pyretechnics/fuel_models.py":438
 *         w_o                  = fuel_model.w_o,
 *         sigma                = fuel_model.sigma,
 *         h                    = fuel_model.h,             # <<<<<<<<<<<<<<
 *         rho_p                = fuel_model.rho_p,
 *         S_T                  = fuel_model.S_T,
 */
  __pyx_t_1.h = __pyx_v_fuel_model.h;

  /* "pyretechnics/fuel_models.py":439
 *         sigma                = fuel_model.sigma,
 *         h                    = fuel_model.h,
 *         rho_p                = fuel_model.rho_p,             # <<<<<<<<<<<<<<
 *         S_T                  = fuel_model.S_T,
 *         S_e                  = fuel_model.S_e,
 */
  __pyx_t_1.rho_p = __pyx_v_fuel_model.rho_p;

  /* "pyretechnics/fuel_models.py":440
 *         h                    = fuel_model.h,
 *         rho_p                = fuel_model.rho_p,
 *         S_T                  = fuel_model.S_T,             # <<<<<<<<<<<<<<
 *         S_e                  = fuel_model.S_e,
 *         dynamic              = fuel_model.dynamic,
 */
  __pyx_t_1.S_T = __pyx_v_fuel_model.S_T;

  /* "pyretechnics/fuel_models.py":441
 *         rho_p                = fuel_model.rho_p,
 *         S_T                  = fuel_model.S_T,
 *         S_e                  = fuel_model.S_e,             # <<<<<<<<<<<<<<
 *         dynamic              = fuel_model.dynamic,
 *         burnable             = fuel_model.burnable,
 */
  __pyx_t_1.S_e = __pyx_v_fuel_model.S_e;

  /* "pyretechnics/fuel_models.py":442
 *         S_T                  = fuel_model.S_T,
 *         S_e                  = fuel_model.S_e,
 *         dynamic              = fuel_model.dynamic,             # <<<<<<<<<<<<<<
 *         burnable             = fuel_model.burnable,
 *         exp_A_sigma          = fuel_model.exp_A_sigma,
 */
  __pyx_t_1.dynamic = __pyx_v_fuel_model.dynamic;

  /* "pyretechnics/fuel_models.py":443
 *         S_e                  = fuel_model.S_e,
 *         dynamic              = fuel_model.dynamic,
 *         burnable             = fuel_model.burnable,             # <<<<<<<<<<<<<<
 *         exp_A_sigma          = fuel_model.exp_A_sigma,
 *         firemod_size_classes = fuel_model.firemod_size_classes,
 */
  __pyx_t_1.burnable = __pyx_v_fuel_model.burnable;

  /* "pyretechnics/fuel_models.py":444
 *         dynamic              = fuel_model.dynamic,
 *         burnable             = fuel_model.burnable,
 *         exp_A_sigma          = fuel_model.exp_A_sigma,             # <<<<<<<<<<<<<<
 *         firemod_size_classes = fuel_model.firemod_size_classes,
 *         f_ij                 = fuel_model.f_ij,
 */
  __pyx_t_1.exp_A_sigma = __pyx_v_fuel_model.exp_A_sigma;

  /* "pyretechnics/fuel_models.py":445
 *         burnable             = fuel_model.burnable,
 *         exp_A_sigma          = fuel_model.exp_A_sigma,
 *         firemod_size_classes = fuel_model.firemod_size_classes,             # <<<<<<<<<<<<<<
 *         f_ij                 = fuel_model.f_ij,
 *         f_i                  = fuel_model.f_i,
 */
  __pyx_t_1.firemod_size_classes = __pyx_v_fuel_model.firemod_size_classes;

  /* "pyretechnics/fuel_models.py":446
 *         exp_A_sigma          = fuel_model.exp_A_sigma,
 *         firemod_size_classes = fuel_model.firemod_size_classes,
 *         f_ij                 = fuel_model.f_ij,             # <<<<<<<<<<<<<<
 *         f_i                  = fuel_model.f_i,
 *         g_ij                 = fuel_model.g_ij,
 */
  __pyx_t_1.f_ij = __pyx_v_fuel_model.f_ij;

  /* "pyretechnics/fuel_models.py":447
 *         firemod_size_classes = fuel_model.firemod_size_classes,
 *         f_ij                 = fuel_model.f_ij,
 *         f_i                  = fuel_model.f_i,             # <<<<<<<<<<<<<<
 *         g_ij                 = fuel_model.g_ij,
 *     )
 */
  __pyx_t_1.f_i = __pyx_v_fuel_model.f_i;

  /* "pyretechnics/fuel_models.py":448
 *         f_ij                 = fuel_model.f_ij,
 *         f_i                  = fuel_model.f_i,
 *         g_ij                 = fuel_model.g_ij,             # <<<<<<<<<<<<<<
 *     )
 *     # Mutate and return the copy
 */
  __pyx_t_1.g_ij = __pyx_v_fuel_model.g_ij;
  __pyx_v_fuel_model_copy = __pyx_t_1;

  /* "pyretechnics/fuel_models.py":451
 *     )
 *     # Mutate and return the copy
 *     dynamic_fuel_model    : FuelModel = add_dynamic_fuel_loading(fuel_model_copy, fuel_moisture)             # <<<<<<<<<<<<<<
 *     weighted_fuel_model   : FuelModel = add_weighting_factors(dynamic_fuel_model)
 *     moisturized_fuel_model: FuelModel = add_live_moisture_of_extinction(weighted_fuel_model)
 */
  __pyx_v_dynamic_fuel_model = __pyx_f_12pyretechnics_11fuel_models_add_dynamic_fuel_loading(__pyx_v_fuel_model_copy, __pyx_v_fuel_moisture);

  /* "pyretechnics/fuel_models.py":452
 *     # Mutate and return the copy
 *     dynamic_fuel_model    : FuelModel = add_dynamic_fuel_loading(fuel_model_copy, fuel_moisture)
 *     weighted_fuel_model   : FuelModel = add_weighting_factors(dynamic_fuel_model)             # <<<<<<<<<<<<<<
 *     moisturized_fuel_model: FuelModel = add_live_moisture_of_extinction(weighted_fuel_model)
 *     return moisturized_fuel_model
 */
  __pyx_v_weighted_fuel_model = __pyx_f_12pyretechnics_11fuel_models_add_weighting_factors(__pyx_v_dynamic_fuel_model);

  /* "pyretechnics/fuel_models.py":453
 *     dynamic_fuel_model    : FuelModel = add_dynamic_fuel_loading(fuel_model_copy, fuel_moisture)
 *     weighted_fuel_model   : FuelModel = add_weighting_factors(dynamic_fuel_model)
 *     moisturized_fuel_model: FuelModel = add_live_moisture_of_extinction(weighted_fuel_model)             # <<<<<<<<<<<<<<
 *     return moisturized_fuel_model
 * # moisturize ends here
 */
  __pyx_v_moisturized_fuel_model = __pyx_f_12pyretechnics_11fuel_models_add_live_moisture_of_extinction(__pyx_v_weighted_fuel_model);

  /* "pyretechnics/fuel_models.py":454
 *     weighted_fuel_model   : FuelModel = add_weighting_factors(dynamic_fuel_model)
 *     moisturized_fuel_model: FuelModel = add_live_moisture_of_extinction(weighted_fuel_model)
 *     return moisturized_fuel_model             # <<<<<<<<<<<<<<
 * # moisturize ends here
 */
  __pyx_r = __pyx_v_moisturized_fuel_model;
  goto __pyx_L0;

  /* "pyretechnics/fuel_models.py":423
 * # add-live-moisture-of-extinction ends here
 * # [[file:../../org/pyretechnics.org::moisturize][moisturize]]
 * @cy.ccall             # <<<<<<<<<<<<<<
 * @cy.exceptval(check=False)
 * def moisturize(fuel_model: FuelModel, fuel_moisture: fclaarr) -> FuelModel:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyretechnics_11fuel_models_9moisturize(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_12pyretechnics_11fuel_models_8moisturize, "\n    Updates w_o, M_f, and M_x and assigns f_ij, f_i, and g_ij.\n    Returns a new FuelModel struct.\n    ");
static PyMethodDef __pyx_mdef_12pyretechnics_11fuel_models_9moisturize = {"moisturize", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12pyretechnics_11fuel_models_9moisturize, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12pyretechnics_11fuel_models_8moisturize};
static PyObject *__pyx_pw_12pyretechnics_11fuel_models_9moisturize(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_t_12pyretechnics_8cy_types_FuelModel __pyx_v_fuel_model;
  __pyx_t_12pyretechnics_8cy_types_fclaarr __pyx_v_fuel_moisture;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("moisturize (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fuel_model,&__pyx_n_s_fuel_moisture,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_fuel_model)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 423, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_fuel_moisture)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 423, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("moisturize", 1, 2, 2, 1); __PYX_ERR(1, 423, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "moisturize") < 0)) __PYX_ERR(1, 423, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_fuel_model = __pyx_convert__from_py_struct____pyx_t_12pyretechnics_8cy_types_FuelModel(values[0]); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 425, __pyx_L3_error)
    __pyx_v_fuel_moisture = __pyx_convert__from_py___pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float(values[1]); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 425, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("moisturize", 1, 2, 2, __pyx_nargs); __PYX_ERR(1, 423, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("pyretechnics.fuel_models.moisturize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyretechnics_11fuel_models_8moisturize(__pyx_self, __pyx_v_fuel_model, __pyx_v_fuel_moisture);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyretechnics_11fuel_models_8moisturize(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_t_12pyretechnics_8cy_types_FuelModel __pyx_v_fuel_model, __pyx_t_12pyretechnics_8cy_types_fclaarr __pyx_v_fuel_moisture) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("moisturize", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_convert__to_py_struct____pyx_t_12pyretechnics_8cy_types_FuelModel(__pyx_f_12pyretechnics_11fuel_models_moisturize(__pyx_v_fuel_model, __pyx_v_fuel_moisture, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyretechnics.fuel_models.moisturize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif
/* #### Code section: pystring_table ### */

static int __Pyx_CreateStringTabAndInitStrings(void) {
  __Pyx_StringTabEntry __pyx_string_tab[] = {
    {&__pyx_n_s_FuelModel, __pyx_k_FuelModel, sizeof(__pyx_k_FuelModel), 0, 0, 1, 1},
    {&__pyx_n_s_KeyError, __pyx_k_KeyError, sizeof(__pyx_k_KeyError), 0, 0, 1, 1},
    {&__pyx_n_s_M_f, __pyx_k_M_f, sizeof(__pyx_k_M_f), 0, 0, 1, 1},
    {&__pyx_n_s_M_x, __pyx_k_M_x, sizeof(__pyx_k_M_x), 0, 0, 1, 1},
    {&__pyx_kp_s_No_value_specified_for_struct_at, __pyx_k_No_value_specified_for_struct_at, sizeof(__pyx_k_No_value_specified_for_struct_at), 0, 0, 1, 0},
    {&__pyx_kp_s_No_value_specified_for_struct_at_10, __pyx_k_No_value_specified_for_struct_at_10, sizeof(__pyx_k_No_value_specified_for_struct_at_10), 0, 0, 1, 0},
    {&__pyx_kp_s_No_value_specified_for_struct_at_11, __pyx_k_No_value_specified_for_struct_at_11, sizeof(__pyx_k_No_value_specified_for_struct_at_11), 0, 0, 1, 0},
    {&__pyx_kp_s_No_value_specified_for_struct_at_12, __pyx_k_No_value_specified_for_struct_at_12, sizeof(__pyx_k_No_value_specified_for_struct_at_12), 0, 0, 1, 0},
    {&__pyx_kp_s_No_value_specified_for_struct_at_13, __pyx_k_No_value_specified_for_struct_at_13, sizeof(__pyx_k_No_value_specified_for_struct_at_13), 0, 0, 1, 0},
    {&__pyx_kp_s_No_value_specified_for_struct_at_14, __pyx_k_No_value_specified_for_struct_at_14, sizeof(__pyx_k_No_value_specified_for_struct_at_14), 0, 0, 1, 0},
    {&__pyx_kp_s_No_value_specified_for_struct_at_15, __pyx_k_No_value_specified_for_struct_at_15, sizeof(__pyx_k_No_value_specified_for_struct_at_15), 0, 0, 1, 0},
    {&__pyx_kp_s_No_value_specified_for_struct_at_16, __pyx_k_No_value_specified_for_struct_at_16, sizeof(__pyx_k_No_value_specified_for_struct_at_16), 0, 0, 1, 0},
    {&__pyx_kp_s_No_value_specified_for_struct_at_17, __pyx_k_No_value_specified_for_struct_at_17, sizeof(__pyx_k_No_value_specified_for_struct_at_17), 0, 0, 1, 0},
    {&__pyx_kp_s_No_value_specified_for_struct_at_2, __pyx_k_No_value_specified_for_struct_at_2, sizeof(__pyx_k_No_value_specified_for_struct_at_2), 0, 0, 1, 0},
    {&__pyx_kp_s_No_value_specified_for_struct_at_3, __pyx_k_No_value_specified_for_struct_at_3, sizeof(__pyx_k_No_value_specified_for_struct_at_3), 0, 0, 1, 0},
    {&__pyx_kp_s_No_value_specified_for_struct_at_4, __pyx_k_No_value_specified_for_struct_at_4, sizeof(__pyx_k_No_value_specified_for_struct_at_4), 0, 0, 1, 0},
    {&__pyx_kp_s_No_value_specified_for_struct_at_5, __pyx_k_No_value_specified_for_struct_at_5, sizeof(__pyx_k_No_value_specified_for_struct_at_5), 0, 0, 1, 0},
    {&__pyx_kp_s_No_value_specified_for_struct_at_6, __pyx_k_No_value_specified_for_struct_at_6, sizeof(__pyx_k_No_value_specified_for_struct_at_6), 0, 0, 1, 0},
    {&__pyx_kp_s_No_value_specified_for_struct_at_7, __pyx_k_No_value_specified_for_struct_at_7, sizeof(__pyx_k_No_value_specified_for_struct_at_7), 0, 0, 1, 0},
    {&__pyx_kp_s_No_value_specified_for_struct_at_8, __pyx_k_No_value_specified_for_struct_at_8, sizeof(__pyx_k_No_value_specified_for_struct_at_8), 0, 0, 1, 0},
    {&__pyx_kp_s_No_value_specified_for_struct_at_9, __pyx_k_No_value_specified_for_struct_at_9, sizeof(__pyx_k_No_value_specified_for_struct_at_9), 0, 0, 1, 0},
    {&__pyx_n_s_S_T, __pyx_k_S_T, sizeof(__pyx_k_S_T), 0, 0, 1, 1},
    {&__pyx_n_s_S_e, __pyx_k_S_e, sizeof(__pyx_k_S_e), 0, 0, 1, 1},
    {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
    {&__pyx_n_s__79, __pyx_k__79, sizeof(__pyx_k__79), 0, 0, 1, 1},
    {&__pyx_n_s_almost_zero, __pyx_k_almost_zero, sizeof(__pyx_k_almost_zero), 0, 0, 1, 1},
    {&__pyx_n_s_asyncio_coroutines, __pyx_k_asyncio_coroutines, sizeof(__pyx_k_asyncio_coroutines), 0, 0, 1, 1},
    {&__pyx_n_s_burnable, __pyx_k_burnable, sizeof(__pyx_k_burnable), 0, 0, 1, 1},
    {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
    {&__pyx_n_s_compact_fuel_model_table, __pyx_k_compact_fuel_model_table, sizeof(__pyx_k_compact_fuel_model_table), 0, 0, 1, 1},
    {&__pyx_kp_s_cy_int, __pyx_k_cy_int, sizeof(__pyx_k_cy_int), 0, 0, 1, 0},
    {&__pyx_n_s_delta, __pyx_k_delta, sizeof(__pyx_k_delta), 0, 0, 1, 1},
    {&__pyx_n_s_dynamic, __pyx_k_dynamic, sizeof(__pyx_k_dynamic), 0, 0, 1, 1},
    {&__pyx_n_s_exp_A_sigma, __pyx_k_exp_A_sigma, sizeof(__pyx_k_exp_A_sigma), 0, 0, 1, 1},
    {&__pyx_n_s_f_i, __pyx_k_f_i, sizeof(__pyx_k_f_i), 0, 0, 1, 1},
    {&__pyx_n_s_f_ij, __pyx_k_f_ij, sizeof(__pyx_k_f_ij), 0, 0, 1, 1},
    {&__pyx_n_s_fclaarr, __pyx_k_fclaarr, sizeof(__pyx_k_fclaarr), 0, 0, 1, 1},
    {&__pyx_n_s_firemod_size_classes, __pyx_k_firemod_size_classes, sizeof(__pyx_k_firemod_size_classes), 0, 0, 1, 1},
    {&__pyx_n_s_fuel_model, __pyx_k_fuel_model, sizeof(__pyx_k_fuel_model), 0, 0, 1, 1},
    {&__pyx_n_s_fuel_model_exists, __pyx_k_fuel_model_exists, sizeof(__pyx_k_fuel_model_exists), 0, 0, 1, 1},
    {&__pyx_n_s_fuel_model_number, __pyx_k_fuel_model_number, sizeof(__pyx_k_fuel_model_number), 0, 0, 1, 1},
    {&__pyx_n_s_fuel_model_table, __pyx_k_fuel_model_table, sizeof(__pyx_k_fuel_model_table), 0, 0, 1, 1},
    {&__pyx_n_s_fuel_moisture, __pyx_k_fuel_moisture, sizeof(__pyx_k_fuel_moisture), 0, 0, 1, 1},
    {&__pyx_n_s_g_ij, __pyx_k_g_ij, sizeof(__pyx_k_g_ij), 0, 0, 1, 1},
    {&__pyx_n_s_get_fuel_model, __pyx_k_get_fuel_model, sizeof(__pyx_k_get_fuel_model), 0, 0, 1, 1},
    {&__pyx_n_s_h, __pyx_k_h, sizeof(__pyx_k_h), 0, 0, 1, 1},
    {&__pyx_n_s_is_coroutine, __pyx_k_is_coroutine, sizeof(__pyx_k_is_coroutine), 0, 0, 1, 1},
    {&__pyx_n_s_keys, __pyx_k_keys, sizeof(__pyx_k_keys), 0, 0, 1, 1},
    {&__pyx_n_s_list_fuel_model_numbers, __pyx_k_list_fuel_model_numbers, sizeof(__pyx_k_list_fuel_model_numbers), 0, 0, 1, 1},
    {&__pyx_n_s_list_fuel_models, __pyx_k_list_fuel_models, sizeof(__pyx_k_list_fuel_models), 0, 0, 1, 1},
    {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
    {&__pyx_n_s_moisturize, __pyx_k_moisturize, sizeof(__pyx_k_moisturize), 0, 0, 1, 1},
    {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
    {&__pyx_n_s_number, __pyx_k_number, sizeof(__pyx_k_number), 0, 0, 1, 1},
    {&__pyx_n_s_pyretechnics_fuel_models, __pyx_k_pyretechnics_fuel_models, sizeof(__pyx_k_pyretechnics_fuel_models), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_capi, __pyx_k_pyx_capi, sizeof(__pyx_k_pyx_capi), 0, 0, 1, 1},
    {&__pyx_n_s_rho_p, __pyx_k_rho_p, sizeof(__pyx_k_rho_p), 0, 0, 1, 1},
    {&__pyx_n_s_sigma, __pyx_k_sigma, sizeof(__pyx_k_sigma), 0, 0, 1, 1},
    {&__pyx_kp_s_src_pyretechnics_fuel_models_py, __pyx_k_src_pyretechnics_fuel_models_py, sizeof(__pyx_k_src_pyretechnics_fuel_models_py), 0, 0, 1, 0},
    {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
    {&__pyx_n_s_values, __pyx_k_values, sizeof(__pyx_k_values), 0, 0, 1, 1},
    {&__pyx_n_s_w_o, __pyx_k_w_o, sizeof(__pyx_k_w_o), 0, 0, 1, 1},
    {0, 0, 0, 0, 0, 0, 0}
  };
  return __Pyx_InitStrings(__pyx_string_tab);
}
/* #### Code section: cached_builtins ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_KeyError = __Pyx_GetBuiltinName(__pyx_n_s_KeyError); if (!__pyx_builtin_KeyError) __PYX_ERR(0, 19, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 20, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cached_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "FromPyStructUtility":20
 *         value = obj['number']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'number'")             # <<<<<<<<<<<<<<
 *     result.number = value
 *     try:
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_s_No_value_specified_for_struct_at); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "FromPyStructUtility":25
 *         value = obj['delta']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'delta'")             # <<<<<<<<<<<<<<
 *     result.delta = value
 *     try:
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_No_value_specified_for_struct_at_2); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "FromPyStructUtility":30
 *         value = obj['M_x']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'M_x'")             # <<<<<<<<<<<<<<
 *     result.M_x = value
 *     try:
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_kp_s_No_value_specified_for_struct_at_3); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "FromPyStructUtility":35
 *         value = obj['M_f']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'M_f'")             # <<<<<<<<<<<<<<
 *     result.M_f = value
 *     try:
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_s_No_value_specified_for_struct_at_4); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "FromPyStructUtility":40
 *         value = obj['w_o']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'w_o'")             # <<<<<<<<<<<<<<
 *     result.w_o = value
 *     try:
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_s_No_value_specified_for_struct_at_5); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "FromPyStructUtility":45
 *         value = obj['sigma']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'sigma'")             # <<<<<<<<<<<<<<
 *     result.sigma = value
 *     try:
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_s_No_value_specified_for_struct_at_6); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "FromPyStructUtility":50
 *         value = obj['h']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'h'")             # <<<<<<<<<<<<<<
 *     result.h = value
 *     try:
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_kp_s_No_value_specified_for_struct_at_7); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "FromPyStructUtility":55
 *         value = obj['rho_p']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'rho_p'")             # <<<<<<<<<<<<<<
 *     result.rho_p = value
 *     try:
 */
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_kp_s_No_value_specified_for_struct_at_8); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "FromPyStructUtility":60
 *         value = obj['S_T']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'S_T'")             # <<<<<<<<<<<<<<
 *     result.S_T = value
 *     try:
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_kp_s_No_value_specified_for_struct_at_9); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "FromPyStructUtility":65
 *         value = obj['S_e']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'S_e'")             # <<<<<<<<<<<<<<
 *     result.S_e = value
 *     try:
 */
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_kp_s_No_value_specified_for_struct_at_10); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "FromPyStructUtility":70
 *         value = obj['dynamic']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'dynamic'")             # <<<<<<<<<<<<<<
 *     result.dynamic = value
 *     try:
 */
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_kp_s_No_value_specified_for_struct_at_11); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "FromPyStructUtility":75
 *         value = obj['burnable']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'burnable'")             # <<<<<<<<<<<<<<
 *     result.burnable = value
 *     try:
 */
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_kp_s_No_value_specified_for_struct_at_12); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(0, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "FromPyStructUtility":80
 *         value = obj['exp_A_sigma']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'exp_A_sigma'")             # <<<<<<<<<<<<<<
 *     result.exp_A_sigma = value
 *     try:
 */
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_kp_s_No_value_specified_for_struct_at_13); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "FromPyStructUtility":85
 *         value = obj['firemod_size_classes']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'firemod_size_classes'")             # <<<<<<<<<<<<<<
 *     result.firemod_size_classes = value
 *     try:
 */
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_kp_s_No_value_specified_for_struct_at_14); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(0, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "FromPyStructUtility":90
 *         value = obj['f_ij']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'f_ij'")             # <<<<<<<<<<<<<<
 *     result.f_ij = value
 *     try:
 */
  __pyx_tuple__15 = PyTuple_Pack(1, __pyx_kp_s_No_value_specified_for_struct_at_15); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "FromPyStructUtility":95
 *         value = obj['f_i']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'f_i'")             # <<<<<<<<<<<<<<
 *     result.f_i = value
 *     try:
 */
  __pyx_tuple__16 = PyTuple_Pack(1, __pyx_kp_s_No_value_specified_for_struct_at_16); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "FromPyStructUtility":100
 *         value = obj['g_ij']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'g_ij'")             # <<<<<<<<<<<<<<
 *     result.g_ij = value
 *     return result
 */
  __pyx_tuple__17 = PyTuple_Pack(1, __pyx_kp_s_No_value_specified_for_struct_at_17); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "pyretechnics/fuel_models.py":34
 *     # Anderson 13:
 *     # Grass and Grass-dominated (short-grass,timber-grass-and-understory,tall-grass)
 *     1  : (1.0, 12.0, 8.0, 0.0340, 0.0000, 0.0000, 0.0000, 0.0000, 3500.0,   0.0,  0.0,    0.0,    0.0), # R01             # <<<<<<<<<<<<<<
 *     2  : (1.0, 15.0, 8.0, 0.0920, 0.0460, 0.0230, 0.0230, 0.0000, 3000.0, 109.0, 30.0, 1500.0,    0.0), # R02
 *     3  : (2.5, 25.0, 8.0, 0.1380, 0.0000, 0.0000, 0.0000, 0.0000, 1500.0,   0.0,  0.0,    0.0,    0.0), # R03
 */
  __pyx_tuple__18 = PyTuple_Pack(13, __pyx_float_1_0, __pyx_float_12_0, __pyx_float_8_0, __pyx_float_0_0340, __pyx_float_0_0000, __pyx_float_0_0000, __pyx_float_0_0000, __pyx_float_0_0000, __pyx_float_3500_0, __pyx_float_0_0, __pyx_float_0_0, __pyx_float_0_0, __pyx_float_0_0); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(1, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);

  /* "pyretechnics/fuel_models.py":35
 *     # Grass and Grass-dominated (short-grass,timber-grass-and-understory,tall-grass)
 *     1  : (1.0, 12.0, 8.0, 0.0340, 0.0000, 0.0000, 0.0000, 0.0000, 3500.0,   0.0,  0.0,    0.0,    0.0), # R01
 *     2  : (1.0, 15.0, 8.0, 0.0920, 0.0460, 0.0230, 0.0230, 0.0000, 3000.0, 109.0, 30.0, 1500.0,    0.0), # R02             # <<<<<<<<<<<<<<
 *     3  : (2.5, 25.0, 8.0, 0.1380, 0.0000, 0.0000, 0.0000, 0.0000, 1500.0,   0.0,  0.0,    0.0,    0.0), # R03
 *     # Chaparral and Shrubfields (chaparral,brush,dormant-brush-hardwood-slash,southern-rough)
 */
  __pyx_tuple__19 = PyTuple_Pack(13, __pyx_float_1_0, __pyx_float_15_0, __pyx_float_8_0, __pyx_float_0_0920, __pyx_float_0_0460, __pyx_float_0_0230, __pyx_float_0_0230, __pyx_float_0_0000, __pyx_float_3000_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_1500_0, __pyx_float_0_0); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(1, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);

  /* "pyretechnics/fuel_models.py":36
 *     1  : (1.0, 12.0, 8.0, 0.0340, 0.0000, 0.0000, 0.0000, 0.0000, 3500.0,   0.0,  0.0,    0.0,    0.0), # R01
 *     2  : (1.0, 15.0, 8.0, 0.0920, 0.0460, 0.0230, 0.0230, 0.0000, 3000.0, 109.0, 30.0, 1500.0,    0.0), # R02
 *     3  : (2.5, 25.0, 8.0, 0.1380, 0.0000, 0.0000, 0.0000, 0.0000, 1500.0,   0.0,  0.0,    0.0,    0.0), # R03             # <<<<<<<<<<<<<<
 *     # Chaparral and Shrubfields (chaparral,brush,dormant-brush-hardwood-slash,southern-rough)
 *     4  : (6.0, 20.0, 8.0, 0.2300, 0.1840, 0.0920, 0.2300, 0.0000, 2000.0, 109.0, 30.0, 1500.0,    0.0), # R04
 */
  __pyx_tuple__20 = PyTuple_Pack(13, __pyx_float_2_5, __pyx_float_25_0, __pyx_float_8_0, __pyx_float_0_1380, __pyx_float_0_0000, __pyx_float_0_0000, __pyx_float_0_0000, __pyx_float_0_0000, __pyx_float_1500_0, __pyx_float_0_0, __pyx_float_0_0, __pyx_float_0_0, __pyx_float_0_0); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(1, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);

  /* "pyretechnics/fuel_models.py":38
 *     3  : (2.5, 25.0, 8.0, 0.1380, 0.0000, 0.0000, 0.0000, 0.0000, 1500.0,   0.0,  0.0,    0.0,    0.0), # R03
 *     # Chaparral and Shrubfields (chaparral,brush,dormant-brush-hardwood-slash,southern-rough)
 *     4  : (6.0, 20.0, 8.0, 0.2300, 0.1840, 0.0920, 0.2300, 0.0000, 2000.0, 109.0, 30.0, 1500.0,    0.0), # R04             # <<<<<<<<<<<<<<
 *     5  : (2.0, 20.0, 8.0, 0.0460, 0.0230, 0.0000, 0.0920, 0.0000, 2000.0, 109.0,  0.0, 1500.0,    0.0), # R05
 *     6  : (2.5, 25.0, 8.0, 0.0690, 0.1150, 0.0920, 0.0000, 0.0000, 1750.0, 109.0, 30.0,    0.0,    0.0), # R06
 */
  __pyx_tuple__21 = PyTuple_Pack(13, __pyx_float_6_0, __pyx_float_20_0, __pyx_float_8_0, __pyx_float_0_2300, __pyx_float_0_1840, __pyx_float_0_0920, __pyx_float_0_2300, __pyx_float_0_0000, __pyx_float_2000_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_1500_0, __pyx_float_0_0); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(1, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);

  /* "pyretechnics/fuel_models.py":39
 *     # Chaparral and Shrubfields (chaparral,brush,dormant-brush-hardwood-slash,southern-rough)
 *     4  : (6.0, 20.0, 8.0, 0.2300, 0.1840, 0.0920, 0.2300, 0.0000, 2000.0, 109.0, 30.0, 1500.0,    0.0), # R04
 *     5  : (2.0, 20.0, 8.0, 0.0460, 0.0230, 0.0000, 0.0920, 0.0000, 2000.0, 109.0,  0.0, 1500.0,    0.0), # R05             # <<<<<<<<<<<<<<
 *     6  : (2.5, 25.0, 8.0, 0.0690, 0.1150, 0.0920, 0.0000, 0.0000, 1750.0, 109.0, 30.0,    0.0,    0.0), # R06
 *     7  : (2.5, 40.0, 8.0, 0.0520, 0.0860, 0.0690, 0.0170, 0.0000, 1750.0, 109.0, 30.0, 1550.0,    0.0), # R07
 */
  __pyx_tuple__22 = PyTuple_Pack(13, __pyx_float_2_0, __pyx_float_20_0, __pyx_float_8_0, __pyx_float_0_0460, __pyx_float_0_0230, __pyx_float_0_0000, __pyx_float_0_0920, __pyx_float_0_0000, __pyx_float_2000_0, __pyx_float_109_0, __pyx_float_0_0, __pyx_float_1500_0, __pyx_float_0_0); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(1, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);

  /* "pyretechnics/fuel_models.py":40
 *     4  : (6.0, 20.0, 8.0, 0.2300, 0.1840, 0.0920, 0.2300, 0.0000, 2000.0, 109.0, 30.0, 1500.0,    0.0), # R04
 *     5  : (2.0, 20.0, 8.0, 0.0460, 0.0230, 0.0000, 0.0920, 0.0000, 2000.0, 109.0,  0.0, 1500.0,    0.0), # R05
 *     6  : (2.5, 25.0, 8.0, 0.0690, 0.1150, 0.0920, 0.0000, 0.0000, 1750.0, 109.0, 30.0,    0.0,    0.0), # R06             # <<<<<<<<<<<<<<
 *     7  : (2.5, 40.0, 8.0, 0.0520, 0.0860, 0.0690, 0.0170, 0.0000, 1750.0, 109.0, 30.0, 1550.0,    0.0), # R07
 *     # Timber Litter (closed-timber-litter,hardwood-litter,timber-litter-and-understory)
 */
  __pyx_tuple__23 = PyTuple_Pack(13, __pyx_float_2_5, __pyx_float_25_0, __pyx_float_8_0, __pyx_float_0_0690, __pyx_float_0_1150, __pyx_float_0_0920, __pyx_float_0_0000, __pyx_float_0_0000, __pyx_float_1750_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_0_0, __pyx_float_0_0); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);

  /* "pyretechnics/fuel_models.py":41
 *     5  : (2.0, 20.0, 8.0, 0.0460, 0.0230, 0.0000, 0.0920, 0.0000, 2000.0, 109.0,  0.0, 1500.0,    0.0), # R05
 *     6  : (2.5, 25.0, 8.0, 0.0690, 0.1150, 0.0920, 0.0000, 0.0000, 1750.0, 109.0, 30.0,    0.0,    0.0), # R06
 *     7  : (2.5, 40.0, 8.0, 0.0520, 0.0860, 0.0690, 0.0170, 0.0000, 1750.0, 109.0, 30.0, 1550.0,    0.0), # R07             # <<<<<<<<<<<<<<
 *     # Timber Litter (closed-timber-litter,hardwood-litter,timber-litter-and-understory)
 *     8  : (0.2, 30.0, 8.0, 0.0690, 0.0460, 0.1150, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # R08
 */
  __pyx_tuple__24 = PyTuple_Pack(13, __pyx_float_2_5, __pyx_float_40_0, __pyx_float_8_0, __pyx_float_0_0520, __pyx_float_0_0860, __pyx_float_0_0690, __pyx_float_0_0170, __pyx_float_0_0000, __pyx_float_1750_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_1550_0, __pyx_float_0_0); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(1, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);

  /* "pyretechnics/fuel_models.py":43
 *     7  : (2.5, 40.0, 8.0, 0.0520, 0.0860, 0.0690, 0.0170, 0.0000, 1750.0, 109.0, 30.0, 1550.0,    0.0), # R07
 *     # Timber Litter (closed-timber-litter,hardwood-litter,timber-litter-and-understory)
 *     8  : (0.2, 30.0, 8.0, 0.0690, 0.0460, 0.1150, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # R08             # <<<<<<<<<<<<<<
 *     9  : (0.2, 25.0, 8.0, 0.1340, 0.0190, 0.0070, 0.0000, 0.0000, 2500.0, 109.0, 30.0,    0.0,    0.0), # R09
 *     10 : (1.0, 25.0, 8.0, 0.1380, 0.0920, 0.2300, 0.0920, 0.0000, 2000.0, 109.0, 30.0, 1500.0,    0.0), # R10
 */
  __pyx_tuple__25 = PyTuple_Pack(13, __pyx_float_0_2, __pyx_float_30_0, __pyx_float_8_0, __pyx_float_0_0690, __pyx_float_0_0460, __pyx_float_0_1150, __pyx_float_0_0000, __pyx_float_0_0000, __pyx_float_2000_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_0_0, __pyx_float_0_0); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(1, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);

  /* "pyretechnics/fuel_models.py":44
 *     # Timber Litter (closed-timber-litter,hardwood-litter,timber-litter-and-understory)
 *     8  : (0.2, 30.0, 8.0, 0.0690, 0.0460, 0.1150, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # R08
 *     9  : (0.2, 25.0, 8.0, 0.1340, 0.0190, 0.0070, 0.0000, 0.0000, 2500.0, 109.0, 30.0,    0.0,    0.0), # R09             # <<<<<<<<<<<<<<
 *     10 : (1.0, 25.0, 8.0, 0.1380, 0.0920, 0.2300, 0.0920, 0.0000, 2000.0, 109.0, 30.0, 1500.0,    0.0), # R10
 *     # Logging Slash (light-logging-slash,medium-logging-slash,heavy-logging-slash)
 */
  __pyx_tuple__26 = PyTuple_Pack(13, __pyx_float_0_2, __pyx_float_25_0, __pyx_float_8_0, __pyx_float_0_1340, __pyx_float_0_0190, __pyx_float_0_0070, __pyx_float_0_0000, __pyx_float_0_0000, __pyx_float_2500_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_0_0, __pyx_float_0_0); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(1, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);

  /* "pyretechnics/fuel_models.py":45
 *     8  : (0.2, 30.0, 8.0, 0.0690, 0.0460, 0.1150, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # R08
 *     9  : (0.2, 25.0, 8.0, 0.1340, 0.0190, 0.0070, 0.0000, 0.0000, 2500.0, 109.0, 30.0,    0.0,    0.0), # R09
 *     10 : (1.0, 25.0, 8.0, 0.1380, 0.0920, 0.2300, 0.0920, 0.0000, 2000.0, 109.0, 30.0, 1500.0,    0.0), # R10             # <<<<<<<<<<<<<<
 *     # Logging Slash (light-logging-slash,medium-logging-slash,heavy-logging-slash)
 *     11 : (1.0, 15.0, 8.0, 0.0690, 0.2070, 0.2530, 0.0000, 0.0000, 1500.0, 109.0, 30.0,    0.0,    0.0), # R11
 */
  __pyx_tuple__27 = PyTuple_Pack(13, __pyx_float_1_0, __pyx_float_25_0, __pyx_float_8_0, __pyx_float_0_1380, __pyx_float_0_0920, __pyx_float_0_2300, __pyx_float_0_0920, __pyx_float_0_0000, __pyx_float_2000_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_1500_0, __pyx_float_0_0); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(1, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);

  /* "pyretechnics/fuel_models.py":47
 *     10 : (1.0, 25.0, 8.0, 0.1380, 0.0920, 0.2300, 0.0920, 0.0000, 2000.0, 109.0, 30.0, 1500.0,    0.0), # R10
 *     # Logging Slash (light-logging-slash,medium-logging-slash,heavy-logging-slash)
 *     11 : (1.0, 15.0, 8.0, 0.0690, 0.2070, 0.2530, 0.0000, 0.0000, 1500.0, 109.0, 30.0,    0.0,    0.0), # R11             # <<<<<<<<<<<<<<
 *     12 : (2.3, 20.0, 8.0, 0.1840, 0.6440, 0.7590, 0.0000, 0.0000, 1500.0, 109.0, 30.0,    0.0,    0.0), # R12
 *     13 : (3.0, 25.0, 8.0, 0.3220, 1.0580, 1.2880, 0.0000, 0.0000, 1500.0, 109.0, 30.0,    0.0,    0.0), # R13
 */
  __pyx_tuple__28 = PyTuple_Pack(13, __pyx_float_1_0, __pyx_float_15_0, __pyx_float_8_0, __pyx_float_0_0690, __pyx_float_0_2070, __pyx_float_0_2530, __pyx_float_0_0000, __pyx_float_0_0000, __pyx_float_1500_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_0_0, __pyx_float_0_0); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(1, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);

  /* "pyretechnics/fuel_models.py":48
 *     # Logging Slash (light-logging-slash,medium-logging-slash,heavy-logging-slash)
 *     11 : (1.0, 15.0, 8.0, 0.0690, 0.2070, 0.2530, 0.0000, 0.0000, 1500.0, 109.0, 30.0,    0.0,    0.0), # R11
 *     12 : (2.3, 20.0, 8.0, 0.1840, 0.6440, 0.7590, 0.0000, 0.0000, 1500.0, 109.0, 30.0,    0.0,    0.0), # R12             # <<<<<<<<<<<<<<
 *     13 : (3.0, 25.0, 8.0, 0.3220, 1.0580, 1.2880, 0.0000, 0.0000, 1500.0, 109.0, 30.0,    0.0,    0.0), # R13
 *     # Nonburnable (NB)
 */
  __pyx_tuple__29 = PyTuple_Pack(13, __pyx_float_2_3, __pyx_float_20_0, __pyx_float_8_0, __pyx_float_0_1840, __pyx_float_0_6440, __pyx_float_0_7590, __pyx_float_0_0000, __pyx_float_0_0000, __pyx_float_1500_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_0_0, __pyx_float_0_0); if (unlikely(!__pyx_tuple__29)) __PYX_ERR(1, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__29);
  __Pyx_GIVEREF(__pyx_tuple__29);

  /* "pyretechnics/fuel_models.py":49
 *     11 : (1.0, 15.0, 8.0, 0.0690, 0.2070, 0.2530, 0.0000, 0.0000, 1500.0, 109.0, 30.0,    0.0,    0.0), # R11
 *     12 : (2.3, 20.0, 8.0, 0.1840, 0.6440, 0.7590, 0.0000, 0.0000, 1500.0, 109.0, 30.0,    0.0,    0.0), # R12
 *     13 : (3.0, 25.0, 8.0, 0.3220, 1.0580, 1.2880, 0.0000, 0.0000, 1500.0, 109.0, 30.0,    0.0,    0.0), # R13             # <<<<<<<<<<<<<<
 *     # Nonburnable (NB)
 *     91 : (0.0,  0.0, 0.0, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000,    0.0,   0.0,  0.0,    0.0,    0.0), # NB1
 */
  __pyx_tuple__30 = PyTuple_Pack(13, __pyx_float_3_0, __pyx_float_25_0, __pyx_float_8_0, __pyx_float_0_3220, __pyx_float_1_0580, __pyx_float_1_2880, __pyx_float_0_0000, __pyx_float_0_0000, __pyx_float_1500_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_0_0, __pyx_float_0_0); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(1, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);

  /* "pyretechnics/fuel_models.py":51
 *     13 : (3.0, 25.0, 8.0, 0.3220, 1.0580, 1.2880, 0.0000, 0.0000, 1500.0, 109.0, 30.0,    0.0,    0.0), # R13
 *     # Nonburnable (NB)
 *     91 : (0.0,  0.0, 0.0, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000,    0.0,   0.0,  0.0,    0.0,    0.0), # NB1             # <<<<<<<<<<<<<<
 *     92 : (0.0,  0.0, 0.0, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000,    0.0,   0.0,  0.0,    0.0,    0.0), # NB2
 *     93 : (0.0,  0.0, 0.0, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000,    0.0,   0.0,  0.0,    0.0,    0.0), # NB3
 */
  __pyx_tuple__31 = PyTuple_Pack(13, __pyx_float_0_0, __pyx_float_0_0, __pyx_float_0_0, __pyx_float_0_0000, __pyx_float_0_0000, __pyx_float_0_0000, __pyx_float_0_0000, __pyx_float_0_0000, __pyx_float_0_0, __pyx_float_0_0, __pyx_float_0_0, __pyx_float_0_0, __pyx_float_0_0); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(1, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);

  /* "pyretechnics/fuel_models.py":58
 *     # Scott & Burgan 40:
 *     # Grass (GR)
 *     101: (0.4, 15.0, 8.0, 0.0046, 0.0000, 0.0000, 0.0138, 0.0000, 2200.0, 109.0, 30.0, 2000.0,    0.0), # GR1             # <<<<<<<<<<<<<<
 *     102: (1.0, 15.0, 8.0, 0.0046, 0.0000, 0.0000, 0.0459, 0.0000, 2000.0, 109.0, 30.0, 1800.0,    0.0), # GR2
 *     103: (2.0, 30.0, 8.0, 0.0046, 0.0184, 0.0000, 0.0689, 0.0000, 1500.0, 109.0, 30.0, 1300.0,    0.0), # GR3
 */
  __pyx_tuple__32 = PyTuple_Pack(13, __pyx_float_0_4, __pyx_float_15_0, __pyx_float_8_0, __pyx_float_0_0046, __pyx_float_0_0000, __pyx_float_0_0000, __pyx_float_0_0138, __pyx_float_0_0000, __pyx_float_2200_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_2000_0, __pyx_float_0_0); if (unlikely(!__pyx_tuple__32)) __PYX_ERR(1, 58, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__32);
  __Pyx_GIVEREF(__pyx_tuple__32);

  /* "pyretechnics/fuel_models.py":59
 *     # Grass (GR)
 *     101: (0.4, 15.0, 8.0, 0.0046, 0.0000, 0.0000, 0.0138, 0.0000, 2200.0, 109.0, 30.0, 2000.0,    0.0), # GR1
 *     102: (1.0, 15.0, 8.0, 0.0046, 0.0000, 0.0000, 0.0459, 0.0000, 2000.0, 109.0, 30.0, 1800.0,    0.0), # GR2             # <<<<<<<<<<<<<<
 *     103: (2.0, 30.0, 8.0, 0.0046, 0.0184, 0.0000, 0.0689, 0.0000, 1500.0, 109.0, 30.0, 1300.0,    0.0), # GR3
 *     104: (2.0, 15.0, 8.0, 0.0115, 0.0000, 0.0000, 0.0872, 0.0000, 2000.0, 109.0, 30.0, 1800.0,    0.0), # GR4
 */
  __pyx_tuple__33 = PyTuple_Pack(13, __pyx_float_1_0, __pyx_float_15_0, __pyx_float_8_0, __pyx_float_0_0046, __pyx_float_0_0000, __pyx_float_0_0000, __pyx_float_0_0459, __pyx_float_0_0000, __pyx_float_2000_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_1800_0, __pyx_float_0_0); if (unlikely(!__pyx_tuple__33)) __PYX_ERR(1, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__33);
  __Pyx_GIVEREF(__pyx_tuple__33);

  /* "pyretechnics/fuel_models.py":60
 *     101: (0.4, 15.0, 8.0, 0.0046, 0.0000, 0.0000, 0.0138, 0.0000, 2200.0, 109.0, 30.0, 2000.0,    0.0), # GR1
 *     102: (1.0, 15.0, 8.0, 0.0046, 0.0000, 0.0000, 0.0459, 0.0000, 2000.0, 109.0, 30.0, 1800.0,    0.0), # GR2
 *     103: (2.0, 30.0, 8.0, 0.0046, 0.0184, 0.0000, 0.0689, 0.0000, 1500.0, 109.0, 30.0, 1300.0,    0.0), # GR3             # <<<<<<<<<<<<<<
 *     104: (2.0, 15.0, 8.0, 0.0115, 0.0000, 0.0000, 0.0872, 0.0000, 2000.0, 109.0, 30.0, 1800.0,    0.0), # GR4
 *     105: (1.5, 40.0, 8.0, 0.0184, 0.0000, 0.0000, 0.1148, 0.0000, 1800.0, 109.0, 30.0, 1600.0,    0.0), # GR5
 */
  __pyx_tuple__34 = PyTuple_Pack(13, __pyx_float_2_0, __pyx_float_30_0, __pyx_float_8_0, __pyx_float_0_0046, __pyx_float_0_0184, __pyx_float_0_0000, __pyx_float_0_0689, __pyx_float_0_0000, __pyx_float_1500_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_1300_0, __pyx_float_0_0); if (unlikely(!__pyx_tuple__34)) __PYX_ERR(1, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__34);
  __Pyx_GIVEREF(__pyx_tuple__34);

  /* "pyretechnics/fuel_models.py":61
 *     102: (1.0, 15.0, 8.0, 0.0046, 0.0000, 0.0000, 0.0459, 0.0000, 2000.0, 109.0, 30.0, 1800.0,    0.0), # GR2
 *     103: (2.0, 30.0, 8.0, 0.0046, 0.0184, 0.0000, 0.0689, 0.0000, 1500.0, 109.0, 30.0, 1300.0,    0.0), # GR3
 *     104: (2.0, 15.0, 8.0, 0.0115, 0.0000, 0.0000, 0.0872, 0.0000, 2000.0, 109.0, 30.0, 1800.0,    0.0), # GR4             # <<<<<<<<<<<<<<
 *     105: (1.5, 40.0, 8.0, 0.0184, 0.0000, 0.0000, 0.1148, 0.0000, 1800.0, 109.0, 30.0, 1600.0,    0.0), # GR5
 *     106: (1.5, 40.0, 9.0, 0.0046, 0.0000, 0.0000, 0.1561, 0.0000, 2200.0, 109.0, 30.0, 2000.0,    0.0), # GR6
 */
  __pyx_tuple__35 = PyTuple_Pack(13, __pyx_float_2_0, __pyx_float_15_0, __pyx_float_8_0, __pyx_float_0_0115, __pyx_float_0_0000, __pyx_float_0_0000, __pyx_float_0_0872, __pyx_float_0_0000, __pyx_float_2000_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_1800_0, __pyx_float_0_0); if (unlikely(!__pyx_tuple__35)) __PYX_ERR(1, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__35);
  __Pyx_GIVEREF(__pyx_tuple__35);

  /* "pyretechnics/fuel_models.py":62
 *     103: (2.0, 30.0, 8.0, 0.0046, 0.0184, 0.0000, 0.0689, 0.0000, 1500.0, 109.0, 30.0, 1300.0,    0.0), # GR3
 *     104: (2.0, 15.0, 8.0, 0.0115, 0.0000, 0.0000, 0.0872, 0.0000, 2000.0, 109.0, 30.0, 1800.0,    0.0), # GR4
 *     105: (1.5, 40.0, 8.0, 0.0184, 0.0000, 0.0000, 0.1148, 0.0000, 1800.0, 109.0, 30.0, 1600.0,    0.0), # GR5             # <<<<<<<<<<<<<<
 *     106: (1.5, 40.0, 9.0, 0.0046, 0.0000, 0.0000, 0.1561, 0.0000, 2200.0, 109.0, 30.0, 2000.0,    0.0), # GR6
 *     107: (3.0, 15.0, 8.0, 0.0459, 0.0000, 0.0000, 0.2479, 0.0000, 2000.0, 109.0, 30.0, 1800.0,    0.0), # GR7
 */
  __pyx_tuple__36 = PyTuple_Pack(13, __pyx_float_1_5, __pyx_float_40_0, __pyx_float_8_0, __pyx_float_0_0184, __pyx_float_0_0000, __pyx_float_0_0000, __pyx_float_0_1148, __pyx_float_0_0000, __pyx_float_1800_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_1600_0, __pyx_float_0_0); if (unlikely(!__pyx_tuple__36)) __PYX_ERR(1, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__36);
  __Pyx_GIVEREF(__pyx_tuple__36);

  /* "pyretechnics/fuel_models.py":63
 *     104: (2.0, 15.0, 8.0, 0.0115, 0.0000, 0.0000, 0.0872, 0.0000, 2000.0, 109.0, 30.0, 1800.0,    0.0), # GR4
 *     105: (1.5, 40.0, 8.0, 0.0184, 0.0000, 0.0000, 0.1148, 0.0000, 1800.0, 109.0, 30.0, 1600.0,    0.0), # GR5
 *     106: (1.5, 40.0, 9.0, 0.0046, 0.0000, 0.0000, 0.1561, 0.0000, 2200.0, 109.0, 30.0, 2000.0,    0.0), # GR6             # <<<<<<<<<<<<<<
 *     107: (3.0, 15.0, 8.0, 0.0459, 0.0000, 0.0000, 0.2479, 0.0000, 2000.0, 109.0, 30.0, 1800.0,    0.0), # GR7
 *     108: (4.0, 30.0, 8.0, 0.0230, 0.0459, 0.0000, 0.3352, 0.0000, 1500.0, 109.0, 30.0, 1300.0,    0.0), # GR8
 */
  __pyx_tuple__37 = PyTuple_Pack(13, __pyx_float_1_5, __pyx_float_40_0, __pyx_float_9_0, __pyx_float_0_0046, __pyx_float_0_0000, __pyx_float_0_0000, __pyx_float_0_1561, __pyx_float_0_0000, __pyx_float_2200_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_2000_0, __pyx_float_0_0); if (unlikely(!__pyx_tuple__37)) __PYX_ERR(1, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__37);
  __Pyx_GIVEREF(__pyx_tuple__37);

  /* "pyretechnics/fuel_models.py":64
 *     105: (1.5, 40.0, 8.0, 0.0184, 0.0000, 0.0000, 0.1148, 0.0000, 1800.0, 109.0, 30.0, 1600.0,    0.0), # GR5
 *     106: (1.5, 40.0, 9.0, 0.0046, 0.0000, 0.0000, 0.1561, 0.0000, 2200.0, 109.0, 30.0, 2000.0,    0.0), # GR6
 *     107: (3.0, 15.0, 8.0, 0.0459, 0.0000, 0.0000, 0.2479, 0.0000, 2000.0, 109.0, 30.0, 1800.0,    0.0), # GR7             # <<<<<<<<<<<<<<
 *     108: (4.0, 30.0, 8.0, 0.0230, 0.0459, 0.0000, 0.3352, 0.0000, 1500.0, 109.0, 30.0, 1300.0,    0.0), # GR8
 *     109: (5.0, 40.0, 8.0, 0.0459, 0.0459, 0.0000, 0.4132, 0.0000, 1800.0, 109.0, 30.0, 1600.0,    0.0), # GR9
 */
  __pyx_tuple__38 = PyTuple_Pack(13, __pyx_float_3_0, __pyx_float_15_0, __pyx_float_8_0, __pyx_float_0_0459, __pyx_float_0_0000, __pyx_float_0_0000, __pyx_float_0_2479, __pyx_float_0_0000, __pyx_float_2000_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_1800_0, __pyx_float_0_0); if (unlikely(!__pyx_tuple__38)) __PYX_ERR(1, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__38);
  __Pyx_GIVEREF(__pyx_tuple__38);

  /* "pyretechnics/fuel_models.py":65
 *     106: (1.5, 40.0, 9.0, 0.0046, 0.0000, 0.0000, 0.1561, 0.0000, 2200.0, 109.0, 30.0, 2000.0,    0.0), # GR6
 *     107: (3.0, 15.0, 8.0, 0.0459, 0.0000, 0.0000, 0.2479, 0.0000, 2000.0, 109.0, 30.0, 1800.0,    0.0), # GR7
 *     108: (4.0, 30.0, 8.0, 0.0230, 0.0459, 0.0000, 0.3352, 0.0000, 1500.0, 109.0, 30.0, 1300.0,    0.0), # GR8             # <<<<<<<<<<<<<<
 *     109: (5.0, 40.0, 8.0, 0.0459, 0.0459, 0.0000, 0.4132, 0.0000, 1800.0, 109.0, 30.0, 1600.0,    0.0), # GR9
 *     # Grass-Shrub (GS)
 */
  __pyx_tuple__39 = PyTuple_Pack(13, __pyx_float_4_0, __pyx_float_30_0, __pyx_float_8_0, __pyx_float_0_0230, __pyx_float_0_0459, __pyx_float_0_0000, __pyx_float_0_3352, __pyx_float_0_0000, __pyx_float_1500_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_1300_0, __pyx_float_0_0); if (unlikely(!__pyx_tuple__39)) __PYX_ERR(1, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__39);
  __Pyx_GIVEREF(__pyx_tuple__39);

  /* "pyretechnics/fuel_models.py":66
 *     107: (3.0, 15.0, 8.0, 0.0459, 0.0000, 0.0000, 0.2479, 0.0000, 2000.0, 109.0, 30.0, 1800.0,    0.0), # GR7
 *     108: (4.0, 30.0, 8.0, 0.0230, 0.0459, 0.0000, 0.3352, 0.0000, 1500.0, 109.0, 30.0, 1300.0,    0.0), # GR8
 *     109: (5.0, 40.0, 8.0, 0.0459, 0.0459, 0.0000, 0.4132, 0.0000, 1800.0, 109.0, 30.0, 1600.0,    0.0), # GR9             # <<<<<<<<<<<<<<
 *     # Grass-Shrub (GS)
 *     121: (0.9, 15.0, 8.0, 0.0092, 0.0000, 0.0000, 0.0230, 0.0298, 2000.0, 109.0, 30.0, 1800.0, 1800.0), # GS1
 */
  __pyx_tuple__40 = PyTuple_Pack(13, __pyx_float_5_0, __pyx_float_40_0, __pyx_float_8_0, __pyx_float_0_0459, __pyx_float_0_0459, __pyx_float_0_0000, __pyx_float_0_4132, __pyx_float_0_0000, __pyx_float_1800_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_1600_0, __pyx_float_0_0); if (unlikely(!__pyx_tuple__40)) __PYX_ERR(1, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__40);
  __Pyx_GIVEREF(__pyx_tuple__40);

  /* "pyretechnics/fuel_models.py":68
 *     109: (5.0, 40.0, 8.0, 0.0459, 0.0459, 0.0000, 0.4132, 0.0000, 1800.0, 109.0, 30.0, 1600.0,    0.0), # GR9
 *     # Grass-Shrub (GS)
 *     121: (0.9, 15.0, 8.0, 0.0092, 0.0000, 0.0000, 0.0230, 0.0298, 2000.0, 109.0, 30.0, 1800.0, 1800.0), # GS1             # <<<<<<<<<<<<<<
 *     122: (1.5, 15.0, 8.0, 0.0230, 0.0230, 0.0000, 0.0275, 0.0459, 2000.0, 109.0, 30.0, 1800.0, 1800.0), # GS2
 *     123: (1.8, 40.0, 8.0, 0.0138, 0.0115, 0.0000, 0.0666, 0.0574, 1800.0, 109.0, 30.0, 1600.0, 1600.0), # GS3
 */
  __pyx_tuple__41 = PyTuple_Pack(13, __pyx_float_0_9, __pyx_float_15_0, __pyx_float_8_0, __pyx_float_0_0092, __pyx_float_0_0000, __pyx_float_0_0000, __pyx_float_0_0230, __pyx_float_0_0298, __pyx_float_2000_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_1800_0, __pyx_float_1800_0); if (unlikely(!__pyx_tuple__41)) __PYX_ERR(1, 68, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__41);
  __Pyx_GIVEREF(__pyx_tuple__41);

  /* "pyretechnics/fuel_models.py":69
 *     # Grass-Shrub (GS)
 *     121: (0.9, 15.0, 8.0, 0.0092, 0.0000, 0.0000, 0.0230, 0.0298, 2000.0, 109.0, 30.0, 1800.0, 1800.0), # GS1
 *     122: (1.5, 15.0, 8.0, 0.0230, 0.0230, 0.0000, 0.0275, 0.0459, 2000.0, 109.0, 30.0, 1800.0, 1800.0), # GS2             # <<<<<<<<<<<<<<
 *     123: (1.8, 40.0, 8.0, 0.0138, 0.0115, 0.0000, 0.0666, 0.0574, 1800.0, 109.0, 30.0, 1600.0, 1600.0), # GS3
 *     124: (2.1, 40.0, 8.0, 0.0872, 0.0138, 0.0046, 0.1561, 0.3260, 1800.0, 109.0, 30.0, 1600.0, 1600.0), # GS4
 */
  __pyx_tuple__42 = PyTuple_Pack(13, __pyx_float_1_5, __pyx_float_15_0, __pyx_float_8_0, __pyx_float_0_0230, __pyx_float_0_0230, __pyx_float_0_0000, __pyx_float_0_0275, __pyx_float_0_0459, __pyx_float_2000_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_1800_0, __pyx_float_1800_0); if (unlikely(!__pyx_tuple__42)) __PYX_ERR(1, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__42);
  __Pyx_GIVEREF(__pyx_tuple__42);

  /* "pyretechnics/fuel_models.py":70
 *     121: (0.9, 15.0, 8.0, 0.0092, 0.0000, 0.0000, 0.0230, 0.0298, 2000.0, 109.0, 30.0, 1800.0, 1800.0), # GS1
 *     122: (1.5, 15.0, 8.0, 0.0230, 0.0230, 0.0000, 0.0275, 0.0459, 2000.0, 109.0, 30.0, 1800.0, 1800.0), # GS2
 *     123: (1.8, 40.0, 8.0, 0.0138, 0.0115, 0.0000, 0.0666, 0.0574, 1800.0, 109.0, 30.0, 1600.0, 1600.0), # GS3             # <<<<<<<<<<<<<<
 *     124: (2.1, 40.0, 8.0, 0.0872, 0.0138, 0.0046, 0.1561, 0.3260, 1800.0, 109.0, 30.0, 1600.0, 1600.0), # GS4
 *     # Shrub (SH)
 */
  __pyx_tuple__43 = PyTuple_Pack(13, __pyx_float_1_8, __pyx_float_40_0, __pyx_float_8_0, __pyx_float_0_0138, __pyx_float_0_0115, __pyx_float_0_0000, __pyx_float_0_0666, __pyx_float_0_0574, __pyx_float_1800_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_1600_0, __pyx_float_1600_0); if (unlikely(!__pyx_tuple__43)) __PYX_ERR(1, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__43);
  __Pyx_GIVEREF(__pyx_tuple__43);

  /* "pyretechnics/fuel_models.py":71
 *     122: (1.5, 15.0, 8.0, 0.0230, 0.0230, 0.0000, 0.0275, 0.0459, 2000.0, 109.0, 30.0, 1800.0, 1800.0), # GS2
 *     123: (1.8, 40.0, 8.0, 0.0138, 0.0115, 0.0000, 0.0666, 0.0574, 1800.0, 109.0, 30.0, 1600.0, 1600.0), # GS3
 *     124: (2.1, 40.0, 8.0, 0.0872, 0.0138, 0.0046, 0.1561, 0.3260, 1800.0, 109.0, 30.0, 1600.0, 1600.0), # GS4             # <<<<<<<<<<<<<<
 *     # Shrub (SH)
 *     141: (1.0, 15.0, 8.0, 0.0115, 0.0115, 0.0000, 0.0069, 0.0597, 2000.0, 109.0, 30.0, 1800.0, 1600.0), # SH1
 */
  __pyx_tuple__44 = PyTuple_Pack(13, __pyx_float_2_1, __pyx_float_40_0, __pyx_float_8_0, __pyx_float_0_0872, __pyx_float_0_0138, __pyx_float_0_0046, __pyx_float_0_1561, __pyx_float_0_3260, __pyx_float_1800_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_1600_0, __pyx_float_1600_0); if (unlikely(!__pyx_tuple__44)) __PYX_ERR(1, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__44);
  __Pyx_GIVEREF(__pyx_tuple__44);

  /* "pyretechnics/fuel_models.py":73
 *     124: (2.1, 40.0, 8.0, 0.0872, 0.0138, 0.0046, 0.1561, 0.3260, 1800.0, 109.0, 30.0, 1600.0, 1600.0), # GS4
 *     # Shrub (SH)
 *     141: (1.0, 15.0, 8.0, 0.0115, 0.0115, 0.0000, 0.0069, 0.0597, 2000.0, 109.0, 30.0, 1800.0, 1600.0), # SH1             # <<<<<<<<<<<<<<
 *     142: (1.0, 15.0, 8.0, 0.0620, 0.1102, 0.0344, 0.0000, 0.1768, 2000.0, 109.0, 30.0,    0.0, 1600.0), # SH2
 *     143: (2.4, 40.0, 8.0, 0.0207, 0.1377, 0.0000, 0.0000, 0.2847, 1600.0, 109.0, 30.0,    0.0, 1400.0), # SH3
 */
  __pyx_tuple__45 = PyTuple_Pack(13, __pyx_float_1_0, __pyx_float_15_0, __pyx_float_8_0, __pyx_float_0_0115, __pyx_float_0_0115, __pyx_float_0_0000, __pyx_float_0_0069, __pyx_float_0_0597, __pyx_float_2000_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_1800_0, __pyx_float_1600_0); if (unlikely(!__pyx_tuple__45)) __PYX_ERR(1, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__45);
  __Pyx_GIVEREF(__pyx_tuple__45);

  /* "pyretechnics/fuel_models.py":74
 *     # Shrub (SH)
 *     141: (1.0, 15.0, 8.0, 0.0115, 0.0115, 0.0000, 0.0069, 0.0597, 2000.0, 109.0, 30.0, 1800.0, 1600.0), # SH1
 *     142: (1.0, 15.0, 8.0, 0.0620, 0.1102, 0.0344, 0.0000, 0.1768, 2000.0, 109.0, 30.0,    0.0, 1600.0), # SH2             # <<<<<<<<<<<<<<
 *     143: (2.4, 40.0, 8.0, 0.0207, 0.1377, 0.0000, 0.0000, 0.2847, 1600.0, 109.0, 30.0,    0.0, 1400.0), # SH3
 *     144: (3.0, 30.0, 8.0, 0.0390, 0.0528, 0.0092, 0.0000, 0.1171, 2000.0, 109.0, 30.0, 1800.0, 1600.0), # SH4
 */
  __pyx_tuple__46 = PyTuple_Pack(13, __pyx_float_1_0, __pyx_float_15_0, __pyx_float_8_0, __pyx_float_0_0620, __pyx_float_0_1102, __pyx_float_0_0344, __pyx_float_0_0000, __pyx_float_0_1768, __pyx_float_2000_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_0_0, __pyx_float_1600_0); if (unlikely(!__pyx_tuple__46)) __PYX_ERR(1, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__46);
  __Pyx_GIVEREF(__pyx_tuple__46);

  /* "pyretechnics/fuel_models.py":75
 *     141: (1.0, 15.0, 8.0, 0.0115, 0.0115, 0.0000, 0.0069, 0.0597, 2000.0, 109.0, 30.0, 1800.0, 1600.0), # SH1
 *     142: (1.0, 15.0, 8.0, 0.0620, 0.1102, 0.0344, 0.0000, 0.1768, 2000.0, 109.0, 30.0,    0.0, 1600.0), # SH2
 *     143: (2.4, 40.0, 8.0, 0.0207, 0.1377, 0.0000, 0.0000, 0.2847, 1600.0, 109.0, 30.0,    0.0, 1400.0), # SH3             # <<<<<<<<<<<<<<
 *     144: (3.0, 30.0, 8.0, 0.0390, 0.0528, 0.0092, 0.0000, 0.1171, 2000.0, 109.0, 30.0, 1800.0, 1600.0), # SH4
 *     145: (6.0, 15.0, 8.0, 0.1653, 0.0964, 0.0000, 0.0000, 0.1331,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH5
 */
  __pyx_tuple__47 = PyTuple_Pack(13, __pyx_float_2_4, __pyx_float_40_0, __pyx_float_8_0, __pyx_float_0_0207, __pyx_float_0_1377, __pyx_float_0_0000, __pyx_float_0_0000, __pyx_float_0_2847, __pyx_float_1600_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_0_0, __pyx_float_1400_0); if (unlikely(!__pyx_tuple__47)) __PYX_ERR(1, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__47);
  __Pyx_GIVEREF(__pyx_tuple__47);

  /* "pyretechnics/fuel_models.py":76
 *     142: (1.0, 15.0, 8.0, 0.0620, 0.1102, 0.0344, 0.0000, 0.1768, 2000.0, 109.0, 30.0,    0.0, 1600.0), # SH2
 *     143: (2.4, 40.0, 8.0, 0.0207, 0.1377, 0.0000, 0.0000, 0.2847, 1600.0, 109.0, 30.0,    0.0, 1400.0), # SH3
 *     144: (3.0, 30.0, 8.0, 0.0390, 0.0528, 0.0092, 0.0000, 0.1171, 2000.0, 109.0, 30.0, 1800.0, 1600.0), # SH4             # <<<<<<<<<<<<<<
 *     145: (6.0, 15.0, 8.0, 0.1653, 0.0964, 0.0000, 0.0000, 0.1331,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH5
 *     146: (2.0, 30.0, 8.0, 0.1331, 0.0666, 0.0000, 0.0000, 0.0643,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH6
 */
  __pyx_tuple__48 = PyTuple_Pack(13, __pyx_float_3_0, __pyx_float_30_0, __pyx_float_8_0, __pyx_float_0_0390, __pyx_float_0_0528, __pyx_float_0_0092, __pyx_float_0_0000, __pyx_float_0_1171, __pyx_float_2000_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_1800_0, __pyx_float_1600_0); if (unlikely(!__pyx_tuple__48)) __PYX_ERR(1, 76, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__48);
  __Pyx_GIVEREF(__pyx_tuple__48);

  /* "pyretechnics/fuel_models.py":77
 *     143: (2.4, 40.0, 8.0, 0.0207, 0.1377, 0.0000, 0.0000, 0.2847, 1600.0, 109.0, 30.0,    0.0, 1400.0), # SH3
 *     144: (3.0, 30.0, 8.0, 0.0390, 0.0528, 0.0092, 0.0000, 0.1171, 2000.0, 109.0, 30.0, 1800.0, 1600.0), # SH4
 *     145: (6.0, 15.0, 8.0, 0.1653, 0.0964, 0.0000, 0.0000, 0.1331,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH5             # <<<<<<<<<<<<<<
 *     146: (2.0, 30.0, 8.0, 0.1331, 0.0666, 0.0000, 0.0000, 0.0643,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH6
 *     147: (6.0, 15.0, 8.0, 0.1607, 0.2433, 0.1010, 0.0000, 0.1561,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH7
 */
  __pyx_tuple__49 = PyTuple_Pack(13, __pyx_float_6_0, __pyx_float_15_0, __pyx_float_8_0, __pyx_float_0_1653, __pyx_float_0_0964, __pyx_float_0_0000, __pyx_float_0_0000, __pyx_float_0_1331, __pyx_float_750_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_0_0, __pyx_float_1600_0); if (unlikely(!__pyx_tuple__49)) __PYX_ERR(1, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__49);
  __Pyx_GIVEREF(__pyx_tuple__49);

  /* "pyretechnics/fuel_models.py":78
 *     144: (3.0, 30.0, 8.0, 0.0390, 0.0528, 0.0092, 0.0000, 0.1171, 2000.0, 109.0, 30.0, 1800.0, 1600.0), # SH4
 *     145: (6.0, 15.0, 8.0, 0.1653, 0.0964, 0.0000, 0.0000, 0.1331,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH5
 *     146: (2.0, 30.0, 8.0, 0.1331, 0.0666, 0.0000, 0.0000, 0.0643,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH6             # <<<<<<<<<<<<<<
 *     147: (6.0, 15.0, 8.0, 0.1607, 0.2433, 0.1010, 0.0000, 0.1561,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH7
 *     148: (3.0, 40.0, 8.0, 0.0941, 0.1561, 0.0390, 0.0000, 0.1997,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH8
 */
  __pyx_tuple__50 = PyTuple_Pack(13, __pyx_float_2_0, __pyx_float_30_0, __pyx_float_8_0, __pyx_float_0_1331, __pyx_float_0_0666, __pyx_float_0_0000, __pyx_float_0_0000, __pyx_float_0_0643, __pyx_float_750_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_0_0, __pyx_float_1600_0); if (unlikely(!__pyx_tuple__50)) __PYX_ERR(1, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__50);
  __Pyx_GIVEREF(__pyx_tuple__50);

  /* "pyretechnics/fuel_models.py":79
 *     145: (6.0, 15.0, 8.0, 0.1653, 0.0964, 0.0000, 0.0000, 0.1331,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH5
 *     146: (2.0, 30.0, 8.0, 0.1331, 0.0666, 0.0000, 0.0000, 0.0643,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH6
 *     147: (6.0, 15.0, 8.0, 0.1607, 0.2433, 0.1010, 0.0000, 0.1561,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH7             # <<<<<<<<<<<<<<
 *     148: (3.0, 40.0, 8.0, 0.0941, 0.1561, 0.0390, 0.0000, 0.1997,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH8
 *     149: (4.4, 40.0, 8.0, 0.2066, 0.1125, 0.0000, 0.0712, 0.3214,  750.0, 109.0, 30.0, 1800.0, 1500.0), # SH9
 */
  __pyx_tuple__51 = PyTuple_Pack(13, __pyx_float_6_0, __pyx_float_15_0, __pyx_float_8_0, __pyx_float_0_1607, __pyx_float_0_2433, __pyx_float_0_1010, __pyx_float_0_0000, __pyx_float_0_1561, __pyx_float_750_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_0_0, __pyx_float_1600_0); if (unlikely(!__pyx_tuple__51)) __PYX_ERR(1, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__51);
  __Pyx_GIVEREF(__pyx_tuple__51);

  /* "pyretechnics/fuel_models.py":80
 *     146: (2.0, 30.0, 8.0, 0.1331, 0.0666, 0.0000, 0.0000, 0.0643,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH6
 *     147: (6.0, 15.0, 8.0, 0.1607, 0.2433, 0.1010, 0.0000, 0.1561,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH7
 *     148: (3.0, 40.0, 8.0, 0.0941, 0.1561, 0.0390, 0.0000, 0.1997,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH8             # <<<<<<<<<<<<<<
 *     149: (4.4, 40.0, 8.0, 0.2066, 0.1125, 0.0000, 0.0712, 0.3214,  750.0, 109.0, 30.0, 1800.0, 1500.0), # SH9
 *     # Timber-Understory (TU)
 */
  __pyx_tuple__52 = PyTuple_Pack(13, __pyx_float_3_0, __pyx_float_40_0, __pyx_float_8_0, __pyx_float_0_0941, __pyx_float_0_1561, __pyx_float_0_0390, __pyx_float_0_0000, __pyx_float_0_1997, __pyx_float_750_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_0_0, __pyx_float_1600_0); if (unlikely(!__pyx_tuple__52)) __PYX_ERR(1, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__52);
  __Pyx_GIVEREF(__pyx_tuple__52);

  /* "pyretechnics/fuel_models.py":81
 *     147: (6.0, 15.0, 8.0, 0.1607, 0.2433, 0.1010, 0.0000, 0.1561,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH7
 *     148: (3.0, 40.0, 8.0, 0.0941, 0.1561, 0.0390, 0.0000, 0.1997,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH8
 *     149: (4.4, 40.0, 8.0, 0.2066, 0.1125, 0.0000, 0.0712, 0.3214,  750.0, 109.0, 30.0, 1800.0, 1500.0), # SH9             # <<<<<<<<<<<<<<
 *     # Timber-Understory (TU)
 *     161: (0.6, 20.0, 8.0, 0.0092, 0.0413, 0.0689, 0.0092, 0.0413, 2000.0, 109.0, 30.0, 1800.0, 1600.0), # TU1
 */
  __pyx_tuple__53 = PyTuple_Pack(13, __pyx_float_4_4, __pyx_float_40_0, __pyx_float_8_0, __pyx_float_0_2066, __pyx_float_0_1125, __pyx_float_0_0000, __pyx_float_0_0712, __pyx_float_0_3214, __pyx_float_750_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_1800_0, __pyx_float_1500_0); if (unlikely(!__pyx_tuple__53)) __PYX_ERR(1, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__53);
  __Pyx_GIVEREF(__pyx_tuple__53);

  /* "pyretechnics/fuel_models.py":83
 *     149: (4.4, 40.0, 8.0, 0.2066, 0.1125, 0.0000, 0.0712, 0.3214,  750.0, 109.0, 30.0, 1800.0, 1500.0), # SH9
 *     # Timber-Understory (TU)
 *     161: (0.6, 20.0, 8.0, 0.0092, 0.0413, 0.0689, 0.0092, 0.0413, 2000.0, 109.0, 30.0, 1800.0, 1600.0), # TU1             # <<<<<<<<<<<<<<
 *     162: (1.0, 30.0, 8.0, 0.0436, 0.0826, 0.0574, 0.0000, 0.0092, 2000.0, 109.0, 30.0,    0.0, 1600.0), # TU2
 *     163: (1.3, 30.0, 8.0, 0.0505, 0.0069, 0.0115, 0.0298, 0.0505, 1800.0, 109.0, 30.0, 1600.0, 1400.0), # TU3
 */
  __pyx_tuple__54 = PyTuple_Pack(13, __pyx_float_0_6, __pyx_float_20_0, __pyx_float_8_0, __pyx_float_0_0092, __pyx_float_0_0413, __pyx_float_0_0689, __pyx_float_0_0092, __pyx_float_0_0413, __pyx_float_2000_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_1800_0, __pyx_float_1600_0); if (unlikely(!__pyx_tuple__54)) __PYX_ERR(1, 83, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__54);
  __Pyx_GIVEREF(__pyx_tuple__54);

  /* "pyretechnics/fuel_models.py":84
 *     # Timber-Understory (TU)
 *     161: (0.6, 20.0, 8.0, 0.0092, 0.0413, 0.0689, 0.0092, 0.0413, 2000.0, 109.0, 30.0, 1800.0, 1600.0), # TU1
 *     162: (1.0, 30.0, 8.0, 0.0436, 0.0826, 0.0574, 0.0000, 0.0092, 2000.0, 109.0, 30.0,    0.0, 1600.0), # TU2             # <<<<<<<<<<<<<<
 *     163: (1.3, 30.0, 8.0, 0.0505, 0.0069, 0.0115, 0.0298, 0.0505, 1800.0, 109.0, 30.0, 1600.0, 1400.0), # TU3
 *     164: (0.5, 12.0, 8.0, 0.2066, 0.0000, 0.0000, 0.0000, 0.0918, 2300.0, 109.0, 30.0,    0.0, 2000.0), # TU4
 */
  __pyx_tuple__55 = PyTuple_Pack(13, __pyx_float_1_0, __pyx_float_30_0, __pyx_float_8_0, __pyx_float_0_0436, __pyx_float_0_0826, __pyx_float_0_0574, __pyx_float_0_0000, __pyx_float_0_0092, __pyx_float_2000_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_0_0, __pyx_float_1600_0); if (unlikely(!__pyx_tuple__55)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__55);
  __Pyx_GIVEREF(__pyx_tuple__55);

  /* "pyretechnics/fuel_models.py":85
 *     161: (0.6, 20.0, 8.0, 0.0092, 0.0413, 0.0689, 0.0092, 0.0413, 2000.0, 109.0, 30.0, 1800.0, 1600.0), # TU1
 *     162: (1.0, 30.0, 8.0, 0.0436, 0.0826, 0.0574, 0.0000, 0.0092, 2000.0, 109.0, 30.0,    0.0, 1600.0), # TU2
 *     163: (1.3, 30.0, 8.0, 0.0505, 0.0069, 0.0115, 0.0298, 0.0505, 1800.0, 109.0, 30.0, 1600.0, 1400.0), # TU3             # <<<<<<<<<<<<<<
 *     164: (0.5, 12.0, 8.0, 0.2066, 0.0000, 0.0000, 0.0000, 0.0918, 2300.0, 109.0, 30.0,    0.0, 2000.0), # TU4
 *     165: (1.0, 25.0, 8.0, 0.1837, 0.1837, 0.1377, 0.0000, 0.1377, 1500.0, 109.0, 30.0,    0.0,  750.0), # TU5
 */
  __pyx_tuple__56 = PyTuple_Pack(13, __pyx_float_1_3, __pyx_float_30_0, __pyx_float_8_0, __pyx_float_0_0505, __pyx_float_0_0069, __pyx_float_0_0115, __pyx_float_0_0298, __pyx_float_0_0505, __pyx_float_1800_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_1600_0, __pyx_float_1400_0); if (unlikely(!__pyx_tuple__56)) __PYX_ERR(1, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__56);
  __Pyx_GIVEREF(__pyx_tuple__56);

  /* "pyretechnics/fuel_models.py":86
 *     162: (1.0, 30.0, 8.0, 0.0436, 0.0826, 0.0574, 0.0000, 0.0092, 2000.0, 109.0, 30.0,    0.0, 1600.0), # TU2
 *     163: (1.3, 30.0, 8.0, 0.0505, 0.0069, 0.0115, 0.0298, 0.0505, 1800.0, 109.0, 30.0, 1600.0, 1400.0), # TU3
 *     164: (0.5, 12.0, 8.0, 0.2066, 0.0000, 0.0000, 0.0000, 0.0918, 2300.0, 109.0, 30.0,    0.0, 2000.0), # TU4             # <<<<<<<<<<<<<<
 *     165: (1.0, 25.0, 8.0, 0.1837, 0.1837, 0.1377, 0.0000, 0.1377, 1500.0, 109.0, 30.0,    0.0,  750.0), # TU5
 *     # Timber Litter (TL)
 */
  __pyx_tuple__57 = PyTuple_Pack(13, __pyx_float_0_5, __pyx_float_12_0, __pyx_float_8_0, __pyx_float_0_2066, __pyx_float_0_0000, __pyx_float_0_0000, __pyx_float_0_0000, __pyx_float_0_0918, __pyx_float_2300_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_0_0, __pyx_float_2000_0); if (unlikely(!__pyx_tuple__57)) __PYX_ERR(1, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__57);
  __Pyx_GIVEREF(__pyx_tuple__57);

  /* "pyretechnics/fuel_models.py":87
 *     163: (1.3, 30.0, 8.0, 0.0505, 0.0069, 0.0115, 0.0298, 0.0505, 1800.0, 109.0, 30.0, 1600.0, 1400.0), # TU3
 *     164: (0.5, 12.0, 8.0, 0.2066, 0.0000, 0.0000, 0.0000, 0.0918, 2300.0, 109.0, 30.0,    0.0, 2000.0), # TU4
 *     165: (1.0, 25.0, 8.0, 0.1837, 0.1837, 0.1377, 0.0000, 0.1377, 1500.0, 109.0, 30.0,    0.0,  750.0), # TU5             # <<<<<<<<<<<<<<
 *     # Timber Litter (TL)
 *     181: (0.2, 30.0, 8.0, 0.0459, 0.1010, 0.1653, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL1
 */
  __pyx_tuple__58 = PyTuple_Pack(13, __pyx_float_1_0, __pyx_float_25_0, __pyx_float_8_0, __pyx_float_0_1837, __pyx_float_0_1837, __pyx_float_0_1377, __pyx_float_0_0000, __pyx_float_0_1377, __pyx_float_1500_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_0_0, __pyx_float_750_0); if (unlikely(!__pyx_tuple__58)) __PYX_ERR(1, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__58);
  __Pyx_GIVEREF(__pyx_tuple__58);

  /* "pyretechnics/fuel_models.py":89
 *     165: (1.0, 25.0, 8.0, 0.1837, 0.1837, 0.1377, 0.0000, 0.1377, 1500.0, 109.0, 30.0,    0.0,  750.0), # TU5
 *     # Timber Litter (TL)
 *     181: (0.2, 30.0, 8.0, 0.0459, 0.1010, 0.1653, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL1             # <<<<<<<<<<<<<<
 *     182: (0.2, 25.0, 8.0, 0.0643, 0.1056, 0.1010, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL2
 *     183: (0.3, 20.0, 8.0, 0.0230, 0.1010, 0.1286, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL3
 */
  __pyx_tuple__59 = PyTuple_Pack(13, __pyx_float_0_2, __pyx_float_30_0, __pyx_float_8_0, __pyx_float_0_0459, __pyx_float_0_1010, __pyx_float_0_1653, __pyx_float_0_0000, __pyx_float_0_0000, __pyx_float_2000_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_0_0, __pyx_float_0_0); if (unlikely(!__pyx_tuple__59)) __PYX_ERR(1, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__59);
  __Pyx_GIVEREF(__pyx_tuple__59);

  /* "pyretechnics/fuel_models.py":90
 *     # Timber Litter (TL)
 *     181: (0.2, 30.0, 8.0, 0.0459, 0.1010, 0.1653, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL1
 *     182: (0.2, 25.0, 8.0, 0.0643, 0.1056, 0.1010, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL2             # <<<<<<<<<<<<<<
 *     183: (0.3, 20.0, 8.0, 0.0230, 0.1010, 0.1286, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL3
 *     184: (0.4, 25.0, 8.0, 0.0230, 0.0689, 0.1928, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL4
 */
  __pyx_tuple__60 = PyTuple_Pack(13, __pyx_float_0_2, __pyx_float_25_0, __pyx_float_8_0, __pyx_float_0_0643, __pyx_float_0_1056, __pyx_float_0_1010, __pyx_float_0_0000, __pyx_float_0_0000, __pyx_float_2000_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_0_0, __pyx_float_0_0); if (unlikely(!__pyx_tuple__60)) __PYX_ERR(1, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__60);
  __Pyx_GIVEREF(__pyx_tuple__60);

  /* "pyretechnics/fuel_models.py":91
 *     181: (0.2, 30.0, 8.0, 0.0459, 0.1010, 0.1653, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL1
 *     182: (0.2, 25.0, 8.0, 0.0643, 0.1056, 0.1010, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL2
 *     183: (0.3, 20.0, 8.0, 0.0230, 0.1010, 0.1286, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL3             # <<<<<<<<<<<<<<
 *     184: (0.4, 25.0, 8.0, 0.0230, 0.0689, 0.1928, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL4
 *     185: (0.6, 25.0, 8.0, 0.0528, 0.1148, 0.2020, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0, 1600.0), # TL5
 */
  __pyx_tuple__61 = PyTuple_Pack(13, __pyx_float_0_3, __pyx_float_20_0, __pyx_float_8_0, __pyx_float_0_0230, __pyx_float_0_1010, __pyx_float_0_1286, __pyx_float_0_0000, __pyx_float_0_0000, __pyx_float_2000_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_0_0, __pyx_float_0_0); if (unlikely(!__pyx_tuple__61)) __PYX_ERR(1, 91, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__61);
  __Pyx_GIVEREF(__pyx_tuple__61);

  /* "pyretechnics/fuel_models.py":92
 *     182: (0.2, 25.0, 8.0, 0.0643, 0.1056, 0.1010, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL2
 *     183: (0.3, 20.0, 8.0, 0.0230, 0.1010, 0.1286, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL3
 *     184: (0.4, 25.0, 8.0, 0.0230, 0.0689, 0.1928, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL4             # <<<<<<<<<<<<<<
 *     185: (0.6, 25.0, 8.0, 0.0528, 0.1148, 0.2020, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0, 1600.0), # TL5
 *     186: (0.3, 25.0, 8.0, 0.1102, 0.0551, 0.0551, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL6
 */
  __pyx_tuple__62 = PyTuple_Pack(13, __pyx_float_0_4, __pyx_float_25_0, __pyx_float_8_0, __pyx_float_0_0230, __pyx_float_0_0689, __pyx_float_0_1928, __pyx_float_0_0000, __pyx_float_0_0000, __pyx_float_2000_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_0_0, __pyx_float_0_0); if (unlikely(!__pyx_tuple__62)) __PYX_ERR(1, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__62);
  __Pyx_GIVEREF(__pyx_tuple__62);

  /* "pyretechnics/fuel_models.py":93
 *     183: (0.3, 20.0, 8.0, 0.0230, 0.1010, 0.1286, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL3
 *     184: (0.4, 25.0, 8.0, 0.0230, 0.0689, 0.1928, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL4
 *     185: (0.6, 25.0, 8.0, 0.0528, 0.1148, 0.2020, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0, 1600.0), # TL5             # <<<<<<<<<<<<<<
 *     186: (0.3, 25.0, 8.0, 0.1102, 0.0551, 0.0551, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL6
 *     187: (0.4, 25.0, 8.0, 0.0138, 0.0643, 0.3719, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL7
 */
  __pyx_tuple__63 = PyTuple_Pack(13, __pyx_float_0_6, __pyx_float_25_0, __pyx_float_8_0, __pyx_float_0_0528, __pyx_float_0_1148, __pyx_float_0_2020, __pyx_float_0_0000, __pyx_float_0_0000, __pyx_float_2000_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_0_0, __pyx_float_1600_0); if (unlikely(!__pyx_tuple__63)) __PYX_ERR(1, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__63);
  __Pyx_GIVEREF(__pyx_tuple__63);

  /* "pyretechnics/fuel_models.py":94
 *     184: (0.4, 25.0, 8.0, 0.0230, 0.0689, 0.1928, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL4
 *     185: (0.6, 25.0, 8.0, 0.0528, 0.1148, 0.2020, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0, 1600.0), # TL5
 *     186: (0.3, 25.0, 8.0, 0.1102, 0.0551, 0.0551, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL6             # <<<<<<<<<<<<<<
 *     187: (0.4, 25.0, 8.0, 0.0138, 0.0643, 0.3719, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL7
 *     188: (0.3, 35.0, 8.0, 0.2663, 0.0643, 0.0505, 0.0000, 0.0000, 1800.0, 109.0, 30.0,    0.0,    0.0), # TL8
 */
  __pyx_tuple__64 = PyTuple_Pack(13, __pyx_float_0_3, __pyx_float_25_0, __pyx_float_8_0, __pyx_float_0_1102, __pyx_float_0_0551, __pyx_float_0_0551, __pyx_float_0_0000, __pyx_float_0_0000, __pyx_float_2000_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_0_0, __pyx_float_0_0); if (unlikely(!__pyx_tuple__64)) __PYX_ERR(1, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__64);
  __Pyx_GIVEREF(__pyx_tuple__64);

  /* "pyretechnics/fuel_models.py":95
 *     185: (0.6, 25.0, 8.0, 0.0528, 0.1148, 0.2020, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0, 1600.0), # TL5
 *     186: (0.3, 25.0, 8.0, 0.1102, 0.0551, 0.0551, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL6
 *     187: (0.4, 25.0, 8.0, 0.0138, 0.0643, 0.3719, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL7             # <<<<<<<<<<<<<<
 *     188: (0.3, 35.0, 8.0, 0.2663, 0.0643, 0.0505, 0.0000, 0.0000, 1800.0, 109.0, 30.0,    0.0,    0.0), # TL8
 *     189: (0.6, 35.0, 8.0, 0.3053, 0.1515, 0.1905, 0.0000, 0.0000, 1800.0, 109.0, 30.0,    0.0, 1600.0), # TL9
 */
  __pyx_tuple__65 = PyTuple_Pack(13, __pyx_float_0_4, __pyx_float_25_0, __pyx_float_8_0, __pyx_float_0_0138, __pyx_float_0_0643, __pyx_float_0_3719, __pyx_float_0_0000, __pyx_float_0_0000, __pyx_float_2000_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_0_0, __pyx_float_0_0); if (unlikely(!__pyx_tuple__65)) __PYX_ERR(1, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__65);
  __Pyx_GIVEREF(__pyx_tuple__65);

  /* "pyretechnics/fuel_models.py":96
 *     186: (0.3, 25.0, 8.0, 0.1102, 0.0551, 0.0551, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL6
 *     187: (0.4, 25.0, 8.0, 0.0138, 0.0643, 0.3719, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL7
 *     188: (0.3, 35.0, 8.0, 0.2663, 0.0643, 0.0505, 0.0000, 0.0000, 1800.0, 109.0, 30.0,    0.0,    0.0), # TL8             # <<<<<<<<<<<<<<
 *     189: (0.6, 35.0, 8.0, 0.3053, 0.1515, 0.1905, 0.0000, 0.0000, 1800.0, 109.0, 30.0,    0.0, 1600.0), # TL9
 *     # Slash-Blowdown (SB)
 */
  __pyx_tuple__66 = PyTuple_Pack(13, __pyx_float_0_3, __pyx_float_35_0, __pyx_float_8_0, __pyx_float_0_2663, __pyx_float_0_0643, __pyx_float_0_0505, __pyx_float_0_0000, __pyx_float_0_0000, __pyx_float_1800_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_0_0, __pyx_float_0_0); if (unlikely(!__pyx_tuple__66)) __PYX_ERR(1, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__66);
  __Pyx_GIVEREF(__pyx_tuple__66);

  /* "pyretechnics/fuel_models.py":97
 *     187: (0.4, 25.0, 8.0, 0.0138, 0.0643, 0.3719, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL7
 *     188: (0.3, 35.0, 8.0, 0.2663, 0.0643, 0.0505, 0.0000, 0.0000, 1800.0, 109.0, 30.0,    0.0,    0.0), # TL8
 *     189: (0.6, 35.0, 8.0, 0.3053, 0.1515, 0.1905, 0.0000, 0.0000, 1800.0, 109.0, 30.0,    0.0, 1600.0), # TL9             # <<<<<<<<<<<<<<
 *     # Slash-Blowdown (SB)
 *     201: (1.0, 25.0, 8.0, 0.0689, 0.1377, 0.5051, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # SB1
 */
  __pyx_tuple__67 = PyTuple_Pack(13, __pyx_float_0_6, __pyx_float_35_0, __pyx_float_8_0, __pyx_float_0_3053, __pyx_float_0_1515, __pyx_float_0_1905, __pyx_float_0_0000, __pyx_float_0_0000, __pyx_float_1800_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_0_0, __pyx_float_1600_0); if (unlikely(!__pyx_tuple__67)) __PYX_ERR(1, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__67);
  __Pyx_GIVEREF(__pyx_tuple__67);

  /* "pyretechnics/fuel_models.py":99
 *     189: (0.6, 35.0, 8.0, 0.3053, 0.1515, 0.1905, 0.0000, 0.0000, 1800.0, 109.0, 30.0,    0.0, 1600.0), # TL9
 *     # Slash-Blowdown (SB)
 *     201: (1.0, 25.0, 8.0, 0.0689, 0.1377, 0.5051, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # SB1             # <<<<<<<<<<<<<<
 *     202: (1.0, 25.0, 8.0, 0.2066, 0.1951, 0.1837, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # SB2
 *     203: (1.2, 25.0, 8.0, 0.2525, 0.1263, 0.1377, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # SB3
 */
  __pyx_tuple__68 = PyTuple_Pack(13, __pyx_float_1_0, __pyx_float_25_0, __pyx_float_8_0, __pyx_float_0_0689, __pyx_float_0_1377, __pyx_float_0_5051, __pyx_float_0_0000, __pyx_float_0_0000, __pyx_float_2000_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_0_0, __pyx_float_0_0); if (unlikely(!__pyx_tuple__68)) __PYX_ERR(1, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__68);
  __Pyx_GIVEREF(__pyx_tuple__68);

  /* "pyretechnics/fuel_models.py":100
 *     # Slash-Blowdown (SB)
 *     201: (1.0, 25.0, 8.0, 0.0689, 0.1377, 0.5051, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # SB1
 *     202: (1.0, 25.0, 8.0, 0.2066, 0.1951, 0.1837, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # SB2             # <<<<<<<<<<<<<<
 *     203: (1.2, 25.0, 8.0, 0.2525, 0.1263, 0.1377, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # SB3
 *     204: (2.7, 25.0, 8.0, 0.2410, 0.1607, 0.2410, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # SB4
 */
  __pyx_tuple__69 = PyTuple_Pack(13, __pyx_float_1_0, __pyx_float_25_0, __pyx_float_8_0, __pyx_float_0_2066, __pyx_float_0_1951, __pyx_float_0_1837, __pyx_float_0_0000, __pyx_float_0_0000, __pyx_float_2000_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_0_0, __pyx_float_0_0); if (unlikely(!__pyx_tuple__69)) __PYX_ERR(1, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__69);
  __Pyx_GIVEREF(__pyx_tuple__69);

  /* "pyretechnics/fuel_models.py":101
 *     201: (1.0, 25.0, 8.0, 0.0689, 0.1377, 0.5051, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # SB1
 *     202: (1.0, 25.0, 8.0, 0.2066, 0.1951, 0.1837, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # SB2
 *     203: (1.2, 25.0, 8.0, 0.2525, 0.1263, 0.1377, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # SB3             # <<<<<<<<<<<<<<
 *     204: (2.7, 25.0, 8.0, 0.2410, 0.1607, 0.2410, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # SB4
 * })
 */
  __pyx_tuple__70 = PyTuple_Pack(13, __pyx_float_1_2, __pyx_float_25_0, __pyx_float_8_0, __pyx_float_0_2525, __pyx_float_0_1263, __pyx_float_0_1377, __pyx_float_0_0000, __pyx_float_0_0000, __pyx_float_2000_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_0_0, __pyx_float_0_0); if (unlikely(!__pyx_tuple__70)) __PYX_ERR(1, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__70);
  __Pyx_GIVEREF(__pyx_tuple__70);

  /* "pyretechnics/fuel_models.py":102
 *     202: (1.0, 25.0, 8.0, 0.2066, 0.1951, 0.1837, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # SB2
 *     203: (1.2, 25.0, 8.0, 0.2525, 0.1263, 0.1377, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # SB3
 *     204: (2.7, 25.0, 8.0, 0.2410, 0.1607, 0.2410, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # SB4             # <<<<<<<<<<<<<<
 * })
 * # fuel-model-compact-table ends here
 */
  __pyx_tuple__71 = PyTuple_Pack(13, __pyx_float_2_7, __pyx_float_25_0, __pyx_float_8_0, __pyx_float_0_2410, __pyx_float_0_1607, __pyx_float_0_2410, __pyx_float_0_0000, __pyx_float_0_0000, __pyx_float_2000_0, __pyx_float_109_0, __pyx_float_30_0, __pyx_float_0_0, __pyx_float_0_0); if (unlikely(!__pyx_tuple__71)) __PYX_ERR(1, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__71);
  __Pyx_GIVEREF(__pyx_tuple__71);

  /* "pyretechnics/fuel_models.py":219
 * 
 * 
 * @cy.ccall             # <<<<<<<<<<<<<<
 * @cy.inline
 * def list_fuel_model_numbers() -> list[int]:
 */
  __pyx_codeobj__72 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_pyretechnics_fuel_models_py, __pyx_n_s_list_fuel_model_numbers, 219, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__72)) __PYX_ERR(1, 219, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":225
 * 
 * 
 * @cy.ccall             # <<<<<<<<<<<<<<
 * @cy.inline
 * def list_fuel_models() -> list[FuelModel]:
 */
  __pyx_codeobj__73 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_pyretechnics_fuel_models_py, __pyx_n_s_list_fuel_models, 225, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__73)) __PYX_ERR(1, 225, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":231
 * 
 * 
 * @cy.ccall             # <<<<<<<<<<<<<<
 * @cy.inline
 * @cy.exceptval(check=False)
 */
  __pyx_tuple__74 = PyTuple_Pack(1, __pyx_n_s_fuel_model_number); if (unlikely(!__pyx_tuple__74)) __PYX_ERR(1, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__74);
  __Pyx_GIVEREF(__pyx_tuple__74);
  __pyx_codeobj__75 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__74, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_pyretechnics_fuel_models_py, __pyx_n_s_fuel_model_exists, 231, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__75)) __PYX_ERR(1, 231, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":238
 * 
 * 
 * @cy.ccall             # <<<<<<<<<<<<<<
 * @cy.inline
 * @cy.exceptval(check=False)
 */
  __pyx_codeobj__76 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__74, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_pyretechnics_fuel_models_py, __pyx_n_s_get_fuel_model, 238, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__76)) __PYX_ERR(1, 238, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":423
 * # add-live-moisture-of-extinction ends here
 * # [[file:../../org/pyretechnics.org::moisturize][moisturize]]
 * @cy.ccall             # <<<<<<<<<<<<<<
 * @cy.exceptval(check=False)
 * def moisturize(fuel_model: FuelModel, fuel_moisture: fclaarr) -> FuelModel:
 */
  __pyx_tuple__77 = PyTuple_Pack(2, __pyx_n_s_fuel_model, __pyx_n_s_fuel_moisture); if (unlikely(!__pyx_tuple__77)) __PYX_ERR(1, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__77);
  __Pyx_GIVEREF(__pyx_tuple__77);
  __pyx_codeobj__78 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__77, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_pyretechnics_fuel_models_py, __pyx_n_s_moisturize, 423, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__78)) __PYX_ERR(1, 423, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}
/* #### Code section: init_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitConstants(void) {
  __pyx_umethod_PyDict_Type_keys.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PyDict_Type_keys.method_name = &__pyx_n_s_keys;
  __pyx_umethod_PyDict_Type_values.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PyDict_Type_values.method_name = &__pyx_n_s_values;
  if (__Pyx_CreateStringTabAndInitStrings() < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  __pyx_float_0_0 = PyFloat_FromDouble(0.0); if (unlikely(!__pyx_float_0_0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_2 = PyFloat_FromDouble(0.2); if (unlikely(!__pyx_float_0_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_3 = PyFloat_FromDouble(0.3); if (unlikely(!__pyx_float_0_3)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_4 = PyFloat_FromDouble(0.4); if (unlikely(!__pyx_float_0_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_5 = PyFloat_FromDouble(0.5); if (unlikely(!__pyx_float_0_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_6 = PyFloat_FromDouble(0.6); if (unlikely(!__pyx_float_0_6)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_9 = PyFloat_FromDouble(0.9); if (unlikely(!__pyx_float_0_9)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_1_0 = PyFloat_FromDouble(1.0); if (unlikely(!__pyx_float_1_0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_1_2 = PyFloat_FromDouble(1.2); if (unlikely(!__pyx_float_1_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_1_3 = PyFloat_FromDouble(1.3); if (unlikely(!__pyx_float_1_3)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_1_5 = PyFloat_FromDouble(1.5); if (unlikely(!__pyx_float_1_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_1_8 = PyFloat_FromDouble(1.8); if (unlikely(!__pyx_float_1_8)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_2_0 = PyFloat_FromDouble(2.0); if (unlikely(!__pyx_float_2_0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_2_1 = PyFloat_FromDouble(2.1); if (unlikely(!__pyx_float_2_1)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_2_3 = PyFloat_FromDouble(2.3); if (unlikely(!__pyx_float_2_3)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_2_4 = PyFloat_FromDouble(2.4); if (unlikely(!__pyx_float_2_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_2_5 = PyFloat_FromDouble(2.5); if (unlikely(!__pyx_float_2_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_2_7 = PyFloat_FromDouble(2.7); if (unlikely(!__pyx_float_2_7)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_3_0 = PyFloat_FromDouble(3.0); if (unlikely(!__pyx_float_3_0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_4_0 = PyFloat_FromDouble(4.0); if (unlikely(!__pyx_float_4_0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_4_4 = PyFloat_FromDouble(4.4); if (unlikely(!__pyx_float_4_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_5_0 = PyFloat_FromDouble(5.0); if (unlikely(!__pyx_float_5_0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_6_0 = PyFloat_FromDouble(6.0); if (unlikely(!__pyx_float_6_0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_8_0 = PyFloat_FromDouble(8.0); if (unlikely(!__pyx_float_8_0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_9_0 = PyFloat_FromDouble(9.0); if (unlikely(!__pyx_float_9_0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_12_0 = PyFloat_FromDouble(12.0); if (unlikely(!__pyx_float_12_0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_15_0 = PyFloat_FromDouble(15.0); if (unlikely(!__pyx_float_15_0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_20_0 = PyFloat_FromDouble(20.0); if (unlikely(!__pyx_float_20_0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_25_0 = PyFloat_FromDouble(25.0); if (unlikely(!__pyx_float_25_0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_30_0 = PyFloat_FromDouble(30.0); if (unlikely(!__pyx_float_30_0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_35_0 = PyFloat_FromDouble(35.0); if (unlikely(!__pyx_float_35_0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_40_0 = PyFloat_FromDouble(40.0); if (unlikely(!__pyx_float_40_0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_109_0 = PyFloat_FromDouble(109.0); if (unlikely(!__pyx_float_109_0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_750_0 = PyFloat_FromDouble(750.0); if (unlikely(!__pyx_float_750_0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_0000 = PyFloat_FromDouble(0.0000); if (unlikely(!__pyx_float_0_0000)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_0046 = PyFloat_FromDouble(0.0046); if (unlikely(!__pyx_float_0_0046)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_0069 = PyFloat_FromDouble(0.0069); if (unlikely(!__pyx_float_0_0069)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_0070 = PyFloat_FromDouble(0.0070); if (unlikely(!__pyx_float_0_0070)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_0092 = PyFloat_FromDouble(0.0092); if (unlikely(!__pyx_float_0_0092)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_0115 = PyFloat_FromDouble(0.0115); if (unlikely(!__pyx_float_0_0115)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_0138 = PyFloat_FromDouble(0.0138); if (unlikely(!__pyx_float_0_0138)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_0170 = PyFloat_FromDouble(0.0170); if (unlikely(!__pyx_float_0_0170)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_0184 = PyFloat_FromDouble(0.0184); if (unlikely(!__pyx_float_0_0184)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_0190 = PyFloat_FromDouble(0.0190); if (unlikely(!__pyx_float_0_0190)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_0207 = PyFloat_FromDouble(0.0207); if (unlikely(!__pyx_float_0_0207)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_0230 = PyFloat_FromDouble(0.0230); if (unlikely(!__pyx_float_0_0230)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_0275 = PyFloat_FromDouble(0.0275); if (unlikely(!__pyx_float_0_0275)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_0298 = PyFloat_FromDouble(0.0298); if (unlikely(!__pyx_float_0_0298)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_0340 = PyFloat_FromDouble(0.0340); if (unlikely(!__pyx_float_0_0340)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_0344 = PyFloat_FromDouble(0.0344); if (unlikely(!__pyx_float_0_0344)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_0390 = PyFloat_FromDouble(0.0390); if (unlikely(!__pyx_float_0_0390)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_0413 = PyFloat_FromDouble(0.0413); if (unlikely(!__pyx_float_0_0413)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_0436 = PyFloat_FromDouble(0.0436); if (unlikely(!__pyx_float_0_0436)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_0459 = PyFloat_FromDouble(0.0459); if (unlikely(!__pyx_float_0_0459)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_0460 = PyFloat_FromDouble(0.0460); if (unlikely(!__pyx_float_0_0460)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_0505 = PyFloat_FromDouble(0.0505); if (unlikely(!__pyx_float_0_0505)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_0520 = PyFloat_FromDouble(0.0520); if (unlikely(!__pyx_float_0_0520)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_0528 = PyFloat_FromDouble(0.0528); if (unlikely(!__pyx_float_0_0528)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_0551 = PyFloat_FromDouble(0.0551); if (unlikely(!__pyx_float_0_0551)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_0574 = PyFloat_FromDouble(0.0574); if (unlikely(!__pyx_float_0_0574)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_0597 = PyFloat_FromDouble(0.0597); if (unlikely(!__pyx_float_0_0597)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_0620 = PyFloat_FromDouble(0.0620); if (unlikely(!__pyx_float_0_0620)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_0643 = PyFloat_FromDouble(0.0643); if (unlikely(!__pyx_float_0_0643)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_0666 = PyFloat_FromDouble(0.0666); if (unlikely(!__pyx_float_0_0666)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_0689 = PyFloat_FromDouble(0.0689); if (unlikely(!__pyx_float_0_0689)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_0690 = PyFloat_FromDouble(0.0690); if (unlikely(!__pyx_float_0_0690)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_0712 = PyFloat_FromDouble(0.0712); if (unlikely(!__pyx_float_0_0712)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_0826 = PyFloat_FromDouble(0.0826); if (unlikely(!__pyx_float_0_0826)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_0860 = PyFloat_FromDouble(0.0860); if (unlikely(!__pyx_float_0_0860)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_0872 = PyFloat_FromDouble(0.0872); if (unlikely(!__pyx_float_0_0872)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_0918 = PyFloat_FromDouble(0.0918); if (unlikely(!__pyx_float_0_0918)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_0920 = PyFloat_FromDouble(0.0920); if (unlikely(!__pyx_float_0_0920)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_0941 = PyFloat_FromDouble(0.0941); if (unlikely(!__pyx_float_0_0941)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_0964 = PyFloat_FromDouble(0.0964); if (unlikely(!__pyx_float_0_0964)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_1010 = PyFloat_FromDouble(0.1010); if (unlikely(!__pyx_float_0_1010)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_1056 = PyFloat_FromDouble(0.1056); if (unlikely(!__pyx_float_0_1056)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_1102 = PyFloat_FromDouble(0.1102); if (unlikely(!__pyx_float_0_1102)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_1125 = PyFloat_FromDouble(0.1125); if (unlikely(!__pyx_float_0_1125)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_1148 = PyFloat_FromDouble(0.1148); if (unlikely(!__pyx_float_0_1148)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_1150 = PyFloat_FromDouble(0.1150); if (unlikely(!__pyx_float_0_1150)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_1171 = PyFloat_FromDouble(0.1171); if (unlikely(!__pyx_float_0_1171)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_1263 = PyFloat_FromDouble(0.1263); if (unlikely(!__pyx_float_0_1263)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_1286 = PyFloat_FromDouble(0.1286); if (unlikely(!__pyx_float_0_1286)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_1331 = PyFloat_FromDouble(0.1331); if (unlikely(!__pyx_float_0_1331)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_1340 = PyFloat_FromDouble(0.1340); if (unlikely(!__pyx_float_0_1340)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_1377 = PyFloat_FromDouble(0.1377); if (unlikely(!__pyx_float_0_1377)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_1380 = PyFloat_FromDouble(0.1380); if (unlikely(!__pyx_float_0_1380)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_1515 = PyFloat_FromDouble(0.1515); if (unlikely(!__pyx_float_0_1515)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_1561 = PyFloat_FromDouble(0.1561); if (unlikely(!__pyx_float_0_1561)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_1607 = PyFloat_FromDouble(0.1607); if (unlikely(!__pyx_float_0_1607)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_1653 = PyFloat_FromDouble(0.1653); if (unlikely(!__pyx_float_0_1653)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_1768 = PyFloat_FromDouble(0.1768); if (unlikely(!__pyx_float_0_1768)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_1837 = PyFloat_FromDouble(0.1837); if (unlikely(!__pyx_float_0_1837)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_1840 = PyFloat_FromDouble(0.1840); if (unlikely(!__pyx_float_0_1840)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_1905 = PyFloat_FromDouble(0.1905); if (unlikely(!__pyx_float_0_1905)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_1928 = PyFloat_FromDouble(0.1928); if (unlikely(!__pyx_float_0_1928)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_1951 = PyFloat_FromDouble(0.1951); if (unlikely(!__pyx_float_0_1951)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_1997 = PyFloat_FromDouble(0.1997); if (unlikely(!__pyx_float_0_1997)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_2020 = PyFloat_FromDouble(0.2020); if (unlikely(!__pyx_float_0_2020)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_2066 = PyFloat_FromDouble(0.2066); if (unlikely(!__pyx_float_0_2066)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_2070 = PyFloat_FromDouble(0.2070); if (unlikely(!__pyx_float_0_2070)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_2300 = PyFloat_FromDouble(0.2300); if (unlikely(!__pyx_float_0_2300)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_2410 = PyFloat_FromDouble(0.2410); if (unlikely(!__pyx_float_0_2410)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_2433 = PyFloat_FromDouble(0.2433); if (unlikely(!__pyx_float_0_2433)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_2479 = PyFloat_FromDouble(0.2479); if (unlikely(!__pyx_float_0_2479)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_2525 = PyFloat_FromDouble(0.2525); if (unlikely(!__pyx_float_0_2525)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_2530 = PyFloat_FromDouble(0.2530); if (unlikely(!__pyx_float_0_2530)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_2663 = PyFloat_FromDouble(0.2663); if (unlikely(!__pyx_float_0_2663)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_2847 = PyFloat_FromDouble(0.2847); if (unlikely(!__pyx_float_0_2847)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_3053 = PyFloat_FromDouble(0.3053); if (unlikely(!__pyx_float_0_3053)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_3214 = PyFloat_FromDouble(0.3214); if (unlikely(!__pyx_float_0_3214)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_3220 = PyFloat_FromDouble(0.3220); if (unlikely(!__pyx_float_0_3220)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_3260 = PyFloat_FromDouble(0.3260); if (unlikely(!__pyx_float_0_3260)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_3352 = PyFloat_FromDouble(0.3352); if (unlikely(!__pyx_float_0_3352)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_3719 = PyFloat_FromDouble(0.3719); if (unlikely(!__pyx_float_0_3719)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_4132 = PyFloat_FromDouble(0.4132); if (unlikely(!__pyx_float_0_4132)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_5051 = PyFloat_FromDouble(0.5051); if (unlikely(!__pyx_float_0_5051)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_6440 = PyFloat_FromDouble(0.6440); if (unlikely(!__pyx_float_0_6440)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_0_7590 = PyFloat_FromDouble(0.7590); if (unlikely(!__pyx_float_0_7590)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_1_0580 = PyFloat_FromDouble(1.0580); if (unlikely(!__pyx_float_1_0580)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_1_2880 = PyFloat_FromDouble(1.2880); if (unlikely(!__pyx_float_1_2880)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_1300_0 = PyFloat_FromDouble(1300.0); if (unlikely(!__pyx_float_1300_0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_1400_0 = PyFloat_FromDouble(1400.0); if (unlikely(!__pyx_float_1400_0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_1500_0 = PyFloat_FromDouble(1500.0); if (unlikely(!__pyx_float_1500_0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_1550_0 = PyFloat_FromDouble(1550.0); if (unlikely(!__pyx_float_1550_0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_1600_0 = PyFloat_FromDouble(1600.0); if (unlikely(!__pyx_float_1600_0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_1750_0 = PyFloat_FromDouble(1750.0); if (unlikely(!__pyx_float_1750_0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_1800_0 = PyFloat_FromDouble(1800.0); if (unlikely(!__pyx_float_1800_0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_2000_0 = PyFloat_FromDouble(2000.0); if (unlikely(!__pyx_float_2000_0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_2200_0 = PyFloat_FromDouble(2200.0); if (unlikely(!__pyx_float_2200_0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_2300_0 = PyFloat_FromDouble(2300.0); if (unlikely(!__pyx_float_2300_0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_2500_0 = PyFloat_FromDouble(2500.0); if (unlikely(!__pyx_float_2500_0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_3000_0 = PyFloat_FromDouble(3000.0); if (unlikely(!__pyx_float_3000_0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_float_3500_0 = PyFloat_FromDouble(3500.0); if (unlikely(!__pyx_float_3500_0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_4 = PyInt_FromLong(4); if (unlikely(!__pyx_int_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_5 = PyInt_FromLong(5); if (unlikely(!__pyx_int_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_6 = PyInt_FromLong(6); if (unlikely(!__pyx_int_6)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_7 = PyInt_FromLong(7); if (unlikely(!__pyx_int_7)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_8 = PyInt_FromLong(8); if (unlikely(!__pyx_int_8)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_9 = PyInt_FromLong(9); if (unlikely(!__pyx_int_9)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_10 = PyInt_FromLong(10); if (unlikely(!__pyx_int_10)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_11 = PyInt_FromLong(11); if (unlikely(!__pyx_int_11)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_12 = PyInt_FromLong(12); if (unlikely(!__pyx_int_12)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_13 = PyInt_FromLong(13); if (unlikely(!__pyx_int_13)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_91 = PyInt_FromLong(91); if (unlikely(!__pyx_int_91)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_92 = PyInt_FromLong(92); if (unlikely(!__pyx_int_92)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_93 = PyInt_FromLong(93); if (unlikely(!__pyx_int_93)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_98 = PyInt_FromLong(98); if (unlikely(!__pyx_int_98)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_99 = PyInt_FromLong(99); if (unlikely(!__pyx_int_99)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_101 = PyInt_FromLong(101); if (unlikely(!__pyx_int_101)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_102 = PyInt_FromLong(102); if (unlikely(!__pyx_int_102)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_103 = PyInt_FromLong(103); if (unlikely(!__pyx_int_103)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_104 = PyInt_FromLong(104); if (unlikely(!__pyx_int_104)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_105 = PyInt_FromLong(105); if (unlikely(!__pyx_int_105)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_106 = PyInt_FromLong(106); if (unlikely(!__pyx_int_106)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_107 = PyInt_FromLong(107); if (unlikely(!__pyx_int_107)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_108 = PyInt_FromLong(108); if (unlikely(!__pyx_int_108)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_109 = PyInt_FromLong(109); if (unlikely(!__pyx_int_109)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_121 = PyInt_FromLong(121); if (unlikely(!__pyx_int_121)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_122 = PyInt_FromLong(122); if (unlikely(!__pyx_int_122)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_123 = PyInt_FromLong(123); if (unlikely(!__pyx_int_123)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_124 = PyInt_FromLong(124); if (unlikely(!__pyx_int_124)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_141 = PyInt_FromLong(141); if (unlikely(!__pyx_int_141)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_142 = PyInt_FromLong(142); if (unlikely(!__pyx_int_142)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_143 = PyInt_FromLong(143); if (unlikely(!__pyx_int_143)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_144 = PyInt_FromLong(144); if (unlikely(!__pyx_int_144)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_145 = PyInt_FromLong(145); if (unlikely(!__pyx_int_145)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_146 = PyInt_FromLong(146); if (unlikely(!__pyx_int_146)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_147 = PyInt_FromLong(147); if (unlikely(!__pyx_int_147)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_148 = PyInt_FromLong(148); if (unlikely(!__pyx_int_148)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_149 = PyInt_FromLong(149); if (unlikely(!__pyx_int_149)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_161 = PyInt_FromLong(161); if (unlikely(!__pyx_int_161)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_162 = PyInt_FromLong(162); if (unlikely(!__pyx_int_162)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_163 = PyInt_FromLong(163); if (unlikely(!__pyx_int_163)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_164 = PyInt_FromLong(164); if (unlikely(!__pyx_int_164)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_165 = PyInt_FromLong(165); if (unlikely(!__pyx_int_165)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_181 = PyInt_FromLong(181); if (unlikely(!__pyx_int_181)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_182 = PyInt_FromLong(182); if (unlikely(!__pyx_int_182)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_183 = PyInt_FromLong(183); if (unlikely(!__pyx_int_183)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_184 = PyInt_FromLong(184); if (unlikely(!__pyx_int_184)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_185 = PyInt_FromLong(185); if (unlikely(!__pyx_int_185)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_186 = PyInt_FromLong(186); if (unlikely(!__pyx_int_186)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_187 = PyInt_FromLong(187); if (unlikely(!__pyx_int_187)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_188 = PyInt_FromLong(188); if (unlikely(!__pyx_int_188)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_189 = PyInt_FromLong(189); if (unlikely(!__pyx_int_189)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_201 = PyInt_FromLong(201); if (unlikely(!__pyx_int_201)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_202 = PyInt_FromLong(202); if (unlikely(!__pyx_int_202)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_203 = PyInt_FromLong(203); if (unlikely(!__pyx_int_203)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_204 = PyInt_FromLong(204); if (unlikely(!__pyx_int_204)) __PYX_ERR(1, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_globals ### */

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  return 0;
}
/* #### Code section: init_module ### */

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __pyx_v_12pyretechnics_11fuel_models_compact_fuel_model_table = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_12pyretechnics_11fuel_models_fuel_model_table = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_7genexpr__pyx_v_12pyretechnics_11fuel_models_k = Py_None; Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  if (__Pyx_ExportVoidPtr(__pyx_n_s_compact_fuel_model_table, (void *)&__pyx_v_12pyretechnics_11fuel_models_compact_fuel_model_table, "PyObject *") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_almost_zero, (void *)&__pyx_v_12pyretechnics_11fuel_models_almost_zero, "float") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_fuel_model_table, (void *)&__pyx_v_12pyretechnics_11fuel_models_fuel_model_table, "PyObject *") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  if (__Pyx_ExportFunction("is_burnable_fuel_model_number", (void (*)(void))__pyx_f_12pyretechnics_11fuel_models_is_burnable_fuel_model_number, "int (int)") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("compute_exp_A_sigma", (void (*)(void))__pyx_f_12pyretechnics_11fuel_models_compute_exp_A_sigma, "float (float, float)") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("compute_firemod_size_class", (void (*)(void))__pyx_f_12pyretechnics_11fuel_models_compute_firemod_size_class, "float (float)") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("expand_compact_fuel_model", (void (*)(void))__pyx_f_12pyretechnics_11fuel_models_expand_compact_fuel_model, "struct __pyx_t_12pyretechnics_8cy_types_FuelModel (int)") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("list_fuel_model_numbers", (void (*)(void))__pyx_f_12pyretechnics_11fuel_models_list_fuel_model_numbers, "PyObject *(int __pyx_skip_dispatch)") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("list_fuel_models", (void (*)(void))__pyx_f_12pyretechnics_11fuel_models_list_fuel_models, "PyObject *(int __pyx_skip_dispatch)") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("fuel_model_exists", (void (*)(void))__pyx_f_12pyretechnics_11fuel_models_fuel_model_exists, "int (int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("get_fuel_model", (void (*)(void))__pyx_f_12pyretechnics_11fuel_models_get_fuel_model, "struct __pyx_t_12pyretechnics_8cy_types_FuelModel (int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("add_dynamic_fuel_loading", (void (*)(void))__pyx_f_12pyretechnics_11fuel_models_add_dynamic_fuel_loading, "struct __pyx_t_12pyretechnics_8cy_types_FuelModel (struct __pyx_t_12pyretechnics_8cy_types_FuelModel, __pyx_t_12pyretechnics_8cy_types_fclaarr)") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("compute_gij", (void (*)(void))__pyx_f_12pyretechnics_11fuel_models_compute_gij, "float (__pyx_t_12pyretechnics_8cy_types_fclaarr, __pyx_t_12pyretechnics_8cy_types_fclaarr, float, int)") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("add_weighting_factors", (void (*)(void))__pyx_f_12pyretechnics_11fuel_models_add_weighting_factors, "struct __pyx_t_12pyretechnics_8cy_types_FuelModel (struct __pyx_t_12pyretechnics_8cy_types_FuelModel)") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("add_live_moisture_of_extinction", (void (*)(void))__pyx_f_12pyretechnics_11fuel_models_add_live_moisture_of_extinction, "struct __pyx_t_12pyretechnics_8cy_types_FuelModel (struct __pyx_t_12pyretechnics_8cy_types_FuelModel)") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("moisturize", (void (*)(void))__pyx_f_12pyretechnics_11fuel_models_moisturize, "struct __pyx_t_12pyretechnics_8cy_types_FuelModel (struct __pyx_t_12pyretechnics_8cy_types_FuelModel, __pyx_t_12pyretechnics_8cy_types_fclaarr, int __pyx_skip_dispatch)") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_fuel_models(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_fuel_models},
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "fuel_models",
      0, /* m_doc */
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      0, /* m_size */
    #elif CYTHON_USE_MODULE_STATE
      sizeof(__pyx_mstate), /* m_size */
    #else
      -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_USE_MODULE_STATE
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif
#endif

#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initfuel_models(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initfuel_models(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_fuel_models(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_fuel_models(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *module, const char* from_name, const char* to_name, int allow_none)
#else
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none)
#endif
{
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
#if CYTHON_COMPILING_IN_LIMITED_API
            result = PyModule_AddObject(module, to_name, value);
#else
            result = PyDict_SetItemString(moddict, to_name, value);
#endif
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    CYTHON_UNUSED_VAR(def);
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    moddict = module;
#else
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
#endif
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_fuel_models(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  int stringtab_initialized = 0;
  #if CYTHON_USE_MODULE_STATE
  int pystate_addmodule_run = 0;
  #endif
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'fuel_models' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("fuel_models", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  if (unlikely(!__pyx_m)) __PYX_ERR(1, 1, __pyx_L1_error)
  #elif CYTHON_USE_MODULE_STATE
  __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1, __pyx_L1_error)
  {
    int add_module_result = PyState_AddModule(__pyx_t_1, &__pyx_moduledef);
    __pyx_t_1 = 0; /* transfer ownership from __pyx_t_1 to "fuel_models" pseudovariable */
    if (unlikely((add_module_result < 0))) __PYX_ERR(1, 1, __pyx_L1_error)
    pystate_addmodule_run = 1;
  }
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  if (unlikely(!__pyx_m)) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #endif
  CYTHON_UNUSED_VAR(__pyx_t_1);
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(1, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = __Pyx_PyImport_AddModuleRef(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_cython_runtime = __Pyx_PyImport_AddModuleRef((const char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(1, 1, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_fuel_models(void)", 0);
  if (__Pyx_check_binary_version(__PYX_LIMITED_VERSION_HEX, __Pyx_get_runtime_version(), CYTHON_COMPILING_IN_LIMITED_API) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(1, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init(__pyx_m) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init(__pyx_m) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init(__pyx_m) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init(__pyx_m) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init(__pyx_m) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init(__pyx_m) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitConstants() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  stringtab_initialized = 1;
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_pyretechnics__fuel_models) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(1, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "pyretechnics.fuel_models")) {
      if (unlikely((PyDict_SetItemString(modules, "pyretechnics.fuel_models", __pyx_m) < 0))) __PYX_ERR(1, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  if (unlikely((__Pyx_modinit_variable_export_code() < 0))) __PYX_ERR(1, 1, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_function_export_code() < 0))) __PYX_ERR(1, 1, __pyx_L1_error)
  (void)__Pyx_modinit_type_init_code();
  (void)__Pyx_modinit_type_import_code();
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif

  /* "pyretechnics/fuel_models.py":34
 *     # Anderson 13:
 *     # Grass and Grass-dominated (short-grass,timber-grass-and-understory,tall-grass)
 *     1  : (1.0, 12.0, 8.0, 0.0340, 0.0000, 0.0000, 0.0000, 0.0000, 3500.0,   0.0,  0.0,    0.0,    0.0), # R01             # <<<<<<<<<<<<<<
 *     2  : (1.0, 15.0, 8.0, 0.0920, 0.0460, 0.0230, 0.0230, 0.0000, 3000.0, 109.0, 30.0, 1500.0,    0.0), # R02
 *     3  : (2.5, 25.0, 8.0, 0.1380, 0.0000, 0.0000, 0.0000, 0.0000, 1500.0,   0.0,  0.0,    0.0,    0.0), # R03
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(58); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_1, __pyx_tuple__18) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":35
 *     # Grass and Grass-dominated (short-grass,timber-grass-and-understory,tall-grass)
 *     1  : (1.0, 12.0, 8.0, 0.0340, 0.0000, 0.0000, 0.0000, 0.0000, 3500.0,   0.0,  0.0,    0.0,    0.0), # R01
 *     2  : (1.0, 15.0, 8.0, 0.0920, 0.0460, 0.0230, 0.0230, 0.0000, 3000.0, 109.0, 30.0, 1500.0,    0.0), # R02             # <<<<<<<<<<<<<<
 *     3  : (2.5, 25.0, 8.0, 0.1380, 0.0000, 0.0000, 0.0000, 0.0000, 1500.0,   0.0,  0.0,    0.0,    0.0), # R03
 *     # Chaparral and Shrubfields (chaparral,brush,dormant-brush-hardwood-slash,southern-rough)
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_2, __pyx_tuple__19) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":36
 *     1  : (1.0, 12.0, 8.0, 0.0340, 0.0000, 0.0000, 0.0000, 0.0000, 3500.0,   0.0,  0.0,    0.0,    0.0), # R01
 *     2  : (1.0, 15.0, 8.0, 0.0920, 0.0460, 0.0230, 0.0230, 0.0000, 3000.0, 109.0, 30.0, 1500.0,    0.0), # R02
 *     3  : (2.5, 25.0, 8.0, 0.1380, 0.0000, 0.0000, 0.0000, 0.0000, 1500.0,   0.0,  0.0,    0.0,    0.0), # R03             # <<<<<<<<<<<<<<
 *     # Chaparral and Shrubfields (chaparral,brush,dormant-brush-hardwood-slash,southern-rough)
 *     4  : (6.0, 20.0, 8.0, 0.2300, 0.1840, 0.0920, 0.2300, 0.0000, 2000.0, 109.0, 30.0, 1500.0,    0.0), # R04
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_3, __pyx_tuple__20) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":38
 *     3  : (2.5, 25.0, 8.0, 0.1380, 0.0000, 0.0000, 0.0000, 0.0000, 1500.0,   0.0,  0.0,    0.0,    0.0), # R03
 *     # Chaparral and Shrubfields (chaparral,brush,dormant-brush-hardwood-slash,southern-rough)
 *     4  : (6.0, 20.0, 8.0, 0.2300, 0.1840, 0.0920, 0.2300, 0.0000, 2000.0, 109.0, 30.0, 1500.0,    0.0), # R04             # <<<<<<<<<<<<<<
 *     5  : (2.0, 20.0, 8.0, 0.0460, 0.0230, 0.0000, 0.0920, 0.0000, 2000.0, 109.0,  0.0, 1500.0,    0.0), # R05
 *     6  : (2.5, 25.0, 8.0, 0.0690, 0.1150, 0.0920, 0.0000, 0.0000, 1750.0, 109.0, 30.0,    0.0,    0.0), # R06
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_4, __pyx_tuple__21) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":39
 *     # Chaparral and Shrubfields (chaparral,brush,dormant-brush-hardwood-slash,southern-rough)
 *     4  : (6.0, 20.0, 8.0, 0.2300, 0.1840, 0.0920, 0.2300, 0.0000, 2000.0, 109.0, 30.0, 1500.0,    0.0), # R04
 *     5  : (2.0, 20.0, 8.0, 0.0460, 0.0230, 0.0000, 0.0920, 0.0000, 2000.0, 109.0,  0.0, 1500.0,    0.0), # R05             # <<<<<<<<<<<<<<
 *     6  : (2.5, 25.0, 8.0, 0.0690, 0.1150, 0.0920, 0.0000, 0.0000, 1750.0, 109.0, 30.0,    0.0,    0.0), # R06
 *     7  : (2.5, 40.0, 8.0, 0.0520, 0.0860, 0.0690, 0.0170, 0.0000, 1750.0, 109.0, 30.0, 1550.0,    0.0), # R07
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_5, __pyx_tuple__22) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":40
 *     4  : (6.0, 20.0, 8.0, 0.2300, 0.1840, 0.0920, 0.2300, 0.0000, 2000.0, 109.0, 30.0, 1500.0,    0.0), # R04
 *     5  : (2.0, 20.0, 8.0, 0.0460, 0.0230, 0.0000, 0.0920, 0.0000, 2000.0, 109.0,  0.0, 1500.0,    0.0), # R05
 *     6  : (2.5, 25.0, 8.0, 0.0690, 0.1150, 0.0920, 0.0000, 0.0000, 1750.0, 109.0, 30.0,    0.0,    0.0), # R06             # <<<<<<<<<<<<<<
 *     7  : (2.5, 40.0, 8.0, 0.0520, 0.0860, 0.0690, 0.0170, 0.0000, 1750.0, 109.0, 30.0, 1550.0,    0.0), # R07
 *     # Timber Litter (closed-timber-litter,hardwood-litter,timber-litter-and-understory)
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_6, __pyx_tuple__23) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":41
 *     5  : (2.0, 20.0, 8.0, 0.0460, 0.0230, 0.0000, 0.0920, 0.0000, 2000.0, 109.0,  0.0, 1500.0,    0.0), # R05
 *     6  : (2.5, 25.0, 8.0, 0.0690, 0.1150, 0.0920, 0.0000, 0.0000, 1750.0, 109.0, 30.0,    0.0,    0.0), # R06
 *     7  : (2.5, 40.0, 8.0, 0.0520, 0.0860, 0.0690, 0.0170, 0.0000, 1750.0, 109.0, 30.0, 1550.0,    0.0), # R07             # <<<<<<<<<<<<<<
 *     # Timber Litter (closed-timber-litter,hardwood-litter,timber-litter-and-understory)
 *     8  : (0.2, 30.0, 8.0, 0.0690, 0.0460, 0.1150, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # R08
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_7, __pyx_tuple__24) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":43
 *     7  : (2.5, 40.0, 8.0, 0.0520, 0.0860, 0.0690, 0.0170, 0.0000, 1750.0, 109.0, 30.0, 1550.0,    0.0), # R07
 *     # Timber Litter (closed-timber-litter,hardwood-litter,timber-litter-and-understory)
 *     8  : (0.2, 30.0, 8.0, 0.0690, 0.0460, 0.1150, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # R08             # <<<<<<<<<<<<<<
 *     9  : (0.2, 25.0, 8.0, 0.1340, 0.0190, 0.0070, 0.0000, 0.0000, 2500.0, 109.0, 30.0,    0.0,    0.0), # R09
 *     10 : (1.0, 25.0, 8.0, 0.1380, 0.0920, 0.2300, 0.0920, 0.0000, 2000.0, 109.0, 30.0, 1500.0,    0.0), # R10
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_8, __pyx_tuple__25) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":44
 *     # Timber Litter (closed-timber-litter,hardwood-litter,timber-litter-and-understory)
 *     8  : (0.2, 30.0, 8.0, 0.0690, 0.0460, 0.1150, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # R08
 *     9  : (0.2, 25.0, 8.0, 0.1340, 0.0190, 0.0070, 0.0000, 0.0000, 2500.0, 109.0, 30.0,    0.0,    0.0), # R09             # <<<<<<<<<<<<<<
 *     10 : (1.0, 25.0, 8.0, 0.1380, 0.0920, 0.2300, 0.0920, 0.0000, 2000.0, 109.0, 30.0, 1500.0,    0.0), # R10
 *     # Logging Slash (light-logging-slash,medium-logging-slash,heavy-logging-slash)
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_9, __pyx_tuple__26) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":45
 *     8  : (0.2, 30.0, 8.0, 0.0690, 0.0460, 0.1150, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # R08
 *     9  : (0.2, 25.0, 8.0, 0.1340, 0.0190, 0.0070, 0.0000, 0.0000, 2500.0, 109.0, 30.0,    0.0,    0.0), # R09
 *     10 : (1.0, 25.0, 8.0, 0.1380, 0.0920, 0.2300, 0.0920, 0.0000, 2000.0, 109.0, 30.0, 1500.0,    0.0), # R10             # <<<<<<<<<<<<<<
 *     # Logging Slash (light-logging-slash,medium-logging-slash,heavy-logging-slash)
 *     11 : (1.0, 15.0, 8.0, 0.0690, 0.2070, 0.2530, 0.0000, 0.0000, 1500.0, 109.0, 30.0,    0.0,    0.0), # R11
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_10, __pyx_tuple__27) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":47
 *     10 : (1.0, 25.0, 8.0, 0.1380, 0.0920, 0.2300, 0.0920, 0.0000, 2000.0, 109.0, 30.0, 1500.0,    0.0), # R10
 *     # Logging Slash (light-logging-slash,medium-logging-slash,heavy-logging-slash)
 *     11 : (1.0, 15.0, 8.0, 0.0690, 0.2070, 0.2530, 0.0000, 0.0000, 1500.0, 109.0, 30.0,    0.0,    0.0), # R11             # <<<<<<<<<<<<<<
 *     12 : (2.3, 20.0, 8.0, 0.1840, 0.6440, 0.7590, 0.0000, 0.0000, 1500.0, 109.0, 30.0,    0.0,    0.0), # R12
 *     13 : (3.0, 25.0, 8.0, 0.3220, 1.0580, 1.2880, 0.0000, 0.0000, 1500.0, 109.0, 30.0,    0.0,    0.0), # R13
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_11, __pyx_tuple__28) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":48
 *     # Logging Slash (light-logging-slash,medium-logging-slash,heavy-logging-slash)
 *     11 : (1.0, 15.0, 8.0, 0.0690, 0.2070, 0.2530, 0.0000, 0.0000, 1500.0, 109.0, 30.0,    0.0,    0.0), # R11
 *     12 : (2.3, 20.0, 8.0, 0.1840, 0.6440, 0.7590, 0.0000, 0.0000, 1500.0, 109.0, 30.0,    0.0,    0.0), # R12             # <<<<<<<<<<<<<<
 *     13 : (3.0, 25.0, 8.0, 0.3220, 1.0580, 1.2880, 0.0000, 0.0000, 1500.0, 109.0, 30.0,    0.0,    0.0), # R13
 *     # Nonburnable (NB)
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_12, __pyx_tuple__29) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":49
 *     11 : (1.0, 15.0, 8.0, 0.0690, 0.2070, 0.2530, 0.0000, 0.0000, 1500.0, 109.0, 30.0,    0.0,    0.0), # R11
 *     12 : (2.3, 20.0, 8.0, 0.1840, 0.6440, 0.7590, 0.0000, 0.0000, 1500.0, 109.0, 30.0,    0.0,    0.0), # R12
 *     13 : (3.0, 25.0, 8.0, 0.3220, 1.0580, 1.2880, 0.0000, 0.0000, 1500.0, 109.0, 30.0,    0.0,    0.0), # R13             # <<<<<<<<<<<<<<
 *     # Nonburnable (NB)
 *     91 : (0.0,  0.0, 0.0, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000,    0.0,   0.0,  0.0,    0.0,    0.0), # NB1
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_13, __pyx_tuple__30) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":51
 *     13 : (3.0, 25.0, 8.0, 0.3220, 1.0580, 1.2880, 0.0000, 0.0000, 1500.0, 109.0, 30.0,    0.0,    0.0), # R13
 *     # Nonburnable (NB)
 *     91 : (0.0,  0.0, 0.0, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000,    0.0,   0.0,  0.0,    0.0,    0.0), # NB1             # <<<<<<<<<<<<<<
 *     92 : (0.0,  0.0, 0.0, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000,    0.0,   0.0,  0.0,    0.0,    0.0), # NB2
 *     93 : (0.0,  0.0, 0.0, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000,    0.0,   0.0,  0.0,    0.0,    0.0), # NB3
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_91, __pyx_tuple__31) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":52
 *     # Nonburnable (NB)
 *     91 : (0.0,  0.0, 0.0, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000,    0.0,   0.0,  0.0,    0.0,    0.0), # NB1
 *     92 : (0.0,  0.0, 0.0, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000,    0.0,   0.0,  0.0,    0.0,    0.0), # NB2             # <<<<<<<<<<<<<<
 *     93 : (0.0,  0.0, 0.0, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000,    0.0,   0.0,  0.0,    0.0,    0.0), # NB3
 *     98 : (0.0,  0.0, 0.0, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000,    0.0,   0.0,  0.0,    0.0,    0.0), # NB4
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_92, __pyx_tuple__31) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":53
 *     91 : (0.0,  0.0, 0.0, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000,    0.0,   0.0,  0.0,    0.0,    0.0), # NB1
 *     92 : (0.0,  0.0, 0.0, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000,    0.0,   0.0,  0.0,    0.0,    0.0), # NB2
 *     93 : (0.0,  0.0, 0.0, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000,    0.0,   0.0,  0.0,    0.0,    0.0), # NB3             # <<<<<<<<<<<<<<
 *     98 : (0.0,  0.0, 0.0, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000,    0.0,   0.0,  0.0,    0.0,    0.0), # NB4
 *     99 : (0.0,  0.0, 0.0, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000,    0.0,   0.0,  0.0,    0.0,    0.0), # NB5
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_93, __pyx_tuple__31) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":54
 *     92 : (0.0,  0.0, 0.0, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000,    0.0,   0.0,  0.0,    0.0,    0.0), # NB2
 *     93 : (0.0,  0.0, 0.0, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000,    0.0,   0.0,  0.0,    0.0,    0.0), # NB3
 *     98 : (0.0,  0.0, 0.0, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000,    0.0,   0.0,  0.0,    0.0,    0.0), # NB4             # <<<<<<<<<<<<<<
 *     99 : (0.0,  0.0, 0.0, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000,    0.0,   0.0,  0.0,    0.0,    0.0), # NB5
 *     # Scott & Burgan 40:
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_98, __pyx_tuple__31) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":55
 *     93 : (0.0,  0.0, 0.0, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000,    0.0,   0.0,  0.0,    0.0,    0.0), # NB3
 *     98 : (0.0,  0.0, 0.0, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000,    0.0,   0.0,  0.0,    0.0,    0.0), # NB4
 *     99 : (0.0,  0.0, 0.0, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000,    0.0,   0.0,  0.0,    0.0,    0.0), # NB5             # <<<<<<<<<<<<<<
 *     # Scott & Burgan 40:
 *     # Grass (GR)
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_99, __pyx_tuple__31) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":58
 *     # Scott & Burgan 40:
 *     # Grass (GR)
 *     101: (0.4, 15.0, 8.0, 0.0046, 0.0000, 0.0000, 0.0138, 0.0000, 2200.0, 109.0, 30.0, 2000.0,    0.0), # GR1             # <<<<<<<<<<<<<<
 *     102: (1.0, 15.0, 8.0, 0.0046, 0.0000, 0.0000, 0.0459, 0.0000, 2000.0, 109.0, 30.0, 1800.0,    0.0), # GR2
 *     103: (2.0, 30.0, 8.0, 0.0046, 0.0184, 0.0000, 0.0689, 0.0000, 1500.0, 109.0, 30.0, 1300.0,    0.0), # GR3
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_101, __pyx_tuple__32) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":59
 *     # Grass (GR)
 *     101: (0.4, 15.0, 8.0, 0.0046, 0.0000, 0.0000, 0.0138, 0.0000, 2200.0, 109.0, 30.0, 2000.0,    0.0), # GR1
 *     102: (1.0, 15.0, 8.0, 0.0046, 0.0000, 0.0000, 0.0459, 0.0000, 2000.0, 109.0, 30.0, 1800.0,    0.0), # GR2             # <<<<<<<<<<<<<<
 *     103: (2.0, 30.0, 8.0, 0.0046, 0.0184, 0.0000, 0.0689, 0.0000, 1500.0, 109.0, 30.0, 1300.0,    0.0), # GR3
 *     104: (2.0, 15.0, 8.0, 0.0115, 0.0000, 0.0000, 0.0872, 0.0000, 2000.0, 109.0, 30.0, 1800.0,    0.0), # GR4
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_102, __pyx_tuple__33) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":60
 *     101: (0.4, 15.0, 8.0, 0.0046, 0.0000, 0.0000, 0.0138, 0.0000, 2200.0, 109.0, 30.0, 2000.0,    0.0), # GR1
 *     102: (1.0, 15.0, 8.0, 0.0046, 0.0000, 0.0000, 0.0459, 0.0000, 2000.0, 109.0, 30.0, 1800.0,    0.0), # GR2
 *     103: (2.0, 30.0, 8.0, 0.0046, 0.0184, 0.0000, 0.0689, 0.0000, 1500.0, 109.0, 30.0, 1300.0,    0.0), # GR3             # <<<<<<<<<<<<<<
 *     104: (2.0, 15.0, 8.0, 0.0115, 0.0000, 0.0000, 0.0872, 0.0000, 2000.0, 109.0, 30.0, 1800.0,    0.0), # GR4
 *     105: (1.5, 40.0, 8.0, 0.0184, 0.0000, 0.0000, 0.1148, 0.0000, 1800.0, 109.0, 30.0, 1600.0,    0.0), # GR5
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_103, __pyx_tuple__34) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":61
 *     102: (1.0, 15.0, 8.0, 0.0046, 0.0000, 0.0000, 0.0459, 0.0000, 2000.0, 109.0, 30.0, 1800.0,    0.0), # GR2
 *     103: (2.0, 30.0, 8.0, 0.0046, 0.0184, 0.0000, 0.0689, 0.0000, 1500.0, 109.0, 30.0, 1300.0,    0.0), # GR3
 *     104: (2.0, 15.0, 8.0, 0.0115, 0.0000, 0.0000, 0.0872, 0.0000, 2000.0, 109.0, 30.0, 1800.0,    0.0), # GR4             # <<<<<<<<<<<<<<
 *     105: (1.5, 40.0, 8.0, 0.0184, 0.0000, 0.0000, 0.1148, 0.0000, 1800.0, 109.0, 30.0, 1600.0,    0.0), # GR5
 *     106: (1.5, 40.0, 9.0, 0.0046, 0.0000, 0.0000, 0.1561, 0.0000, 2200.0, 109.0, 30.0, 2000.0,    0.0), # GR6
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_104, __pyx_tuple__35) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":62
 *     103: (2.0, 30.0, 8.0, 0.0046, 0.0184, 0.0000, 0.0689, 0.0000, 1500.0, 109.0, 30.0, 1300.0,    0.0), # GR3
 *     104: (2.0, 15.0, 8.0, 0.0115, 0.0000, 0.0000, 0.0872, 0.0000, 2000.0, 109.0, 30.0, 1800.0,    0.0), # GR4
 *     105: (1.5, 40.0, 8.0, 0.0184, 0.0000, 0.0000, 0.1148, 0.0000, 1800.0, 109.0, 30.0, 1600.0,    0.0), # GR5             # <<<<<<<<<<<<<<
 *     106: (1.5, 40.0, 9.0, 0.0046, 0.0000, 0.0000, 0.1561, 0.0000, 2200.0, 109.0, 30.0, 2000.0,    0.0), # GR6
 *     107: (3.0, 15.0, 8.0, 0.0459, 0.0000, 0.0000, 0.2479, 0.0000, 2000.0, 109.0, 30.0, 1800.0,    0.0), # GR7
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_105, __pyx_tuple__36) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":63
 *     104: (2.0, 15.0, 8.0, 0.0115, 0.0000, 0.0000, 0.0872, 0.0000, 2000.0, 109.0, 30.0, 1800.0,    0.0), # GR4
 *     105: (1.5, 40.0, 8.0, 0.0184, 0.0000, 0.0000, 0.1148, 0.0000, 1800.0, 109.0, 30.0, 1600.0,    0.0), # GR5
 *     106: (1.5, 40.0, 9.0, 0.0046, 0.0000, 0.0000, 0.1561, 0.0000, 2200.0, 109.0, 30.0, 2000.0,    0.0), # GR6             # <<<<<<<<<<<<<<
 *     107: (3.0, 15.0, 8.0, 0.0459, 0.0000, 0.0000, 0.2479, 0.0000, 2000.0, 109.0, 30.0, 1800.0,    0.0), # GR7
 *     108: (4.0, 30.0, 8.0, 0.0230, 0.0459, 0.0000, 0.3352, 0.0000, 1500.0, 109.0, 30.0, 1300.0,    0.0), # GR8
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_106, __pyx_tuple__37) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":64
 *     105: (1.5, 40.0, 8.0, 0.0184, 0.0000, 0.0000, 0.1148, 0.0000, 1800.0, 109.0, 30.0, 1600.0,    0.0), # GR5
 *     106: (1.5, 40.0, 9.0, 0.0046, 0.0000, 0.0000, 0.1561, 0.0000, 2200.0, 109.0, 30.0, 2000.0,    0.0), # GR6
 *     107: (3.0, 15.0, 8.0, 0.0459, 0.0000, 0.0000, 0.2479, 0.0000, 2000.0, 109.0, 30.0, 1800.0,    0.0), # GR7             # <<<<<<<<<<<<<<
 *     108: (4.0, 30.0, 8.0, 0.0230, 0.0459, 0.0000, 0.3352, 0.0000, 1500.0, 109.0, 30.0, 1300.0,    0.0), # GR8
 *     109: (5.0, 40.0, 8.0, 0.0459, 0.0459, 0.0000, 0.4132, 0.0000, 1800.0, 109.0, 30.0, 1600.0,    0.0), # GR9
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_107, __pyx_tuple__38) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":65
 *     106: (1.5, 40.0, 9.0, 0.0046, 0.0000, 0.0000, 0.1561, 0.0000, 2200.0, 109.0, 30.0, 2000.0,    0.0), # GR6
 *     107: (3.0, 15.0, 8.0, 0.0459, 0.0000, 0.0000, 0.2479, 0.0000, 2000.0, 109.0, 30.0, 1800.0,    0.0), # GR7
 *     108: (4.0, 30.0, 8.0, 0.0230, 0.0459, 0.0000, 0.3352, 0.0000, 1500.0, 109.0, 30.0, 1300.0,    0.0), # GR8             # <<<<<<<<<<<<<<
 *     109: (5.0, 40.0, 8.0, 0.0459, 0.0459, 0.0000, 0.4132, 0.0000, 1800.0, 109.0, 30.0, 1600.0,    0.0), # GR9
 *     # Grass-Shrub (GS)
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_108, __pyx_tuple__39) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":66
 *     107: (3.0, 15.0, 8.0, 0.0459, 0.0000, 0.0000, 0.2479, 0.0000, 2000.0, 109.0, 30.0, 1800.0,    0.0), # GR7
 *     108: (4.0, 30.0, 8.0, 0.0230, 0.0459, 0.0000, 0.3352, 0.0000, 1500.0, 109.0, 30.0, 1300.0,    0.0), # GR8
 *     109: (5.0, 40.0, 8.0, 0.0459, 0.0459, 0.0000, 0.4132, 0.0000, 1800.0, 109.0, 30.0, 1600.0,    0.0), # GR9             # <<<<<<<<<<<<<<
 *     # Grass-Shrub (GS)
 *     121: (0.9, 15.0, 8.0, 0.0092, 0.0000, 0.0000, 0.0230, 0.0298, 2000.0, 109.0, 30.0, 1800.0, 1800.0), # GS1
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_109, __pyx_tuple__40) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":68
 *     109: (5.0, 40.0, 8.0, 0.0459, 0.0459, 0.0000, 0.4132, 0.0000, 1800.0, 109.0, 30.0, 1600.0,    0.0), # GR9
 *     # Grass-Shrub (GS)
 *     121: (0.9, 15.0, 8.0, 0.0092, 0.0000, 0.0000, 0.0230, 0.0298, 2000.0, 109.0, 30.0, 1800.0, 1800.0), # GS1             # <<<<<<<<<<<<<<
 *     122: (1.5, 15.0, 8.0, 0.0230, 0.0230, 0.0000, 0.0275, 0.0459, 2000.0, 109.0, 30.0, 1800.0, 1800.0), # GS2
 *     123: (1.8, 40.0, 8.0, 0.0138, 0.0115, 0.0000, 0.0666, 0.0574, 1800.0, 109.0, 30.0, 1600.0, 1600.0), # GS3
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_121, __pyx_tuple__41) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":69
 *     # Grass-Shrub (GS)
 *     121: (0.9, 15.0, 8.0, 0.0092, 0.0000, 0.0000, 0.0230, 0.0298, 2000.0, 109.0, 30.0, 1800.0, 1800.0), # GS1
 *     122: (1.5, 15.0, 8.0, 0.0230, 0.0230, 0.0000, 0.0275, 0.0459, 2000.0, 109.0, 30.0, 1800.0, 1800.0), # GS2             # <<<<<<<<<<<<<<
 *     123: (1.8, 40.0, 8.0, 0.0138, 0.0115, 0.0000, 0.0666, 0.0574, 1800.0, 109.0, 30.0, 1600.0, 1600.0), # GS3
 *     124: (2.1, 40.0, 8.0, 0.0872, 0.0138, 0.0046, 0.1561, 0.3260, 1800.0, 109.0, 30.0, 1600.0, 1600.0), # GS4
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_122, __pyx_tuple__42) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":70
 *     121: (0.9, 15.0, 8.0, 0.0092, 0.0000, 0.0000, 0.0230, 0.0298, 2000.0, 109.0, 30.0, 1800.0, 1800.0), # GS1
 *     122: (1.5, 15.0, 8.0, 0.0230, 0.0230, 0.0000, 0.0275, 0.0459, 2000.0, 109.0, 30.0, 1800.0, 1800.0), # GS2
 *     123: (1.8, 40.0, 8.0, 0.0138, 0.0115, 0.0000, 0.0666, 0.0574, 1800.0, 109.0, 30.0, 1600.0, 1600.0), # GS3             # <<<<<<<<<<<<<<
 *     124: (2.1, 40.0, 8.0, 0.0872, 0.0138, 0.0046, 0.1561, 0.3260, 1800.0, 109.0, 30.0, 1600.0, 1600.0), # GS4
 *     # Shrub (SH)
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_123, __pyx_tuple__43) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":71
 *     122: (1.5, 15.0, 8.0, 0.0230, 0.0230, 0.0000, 0.0275, 0.0459, 2000.0, 109.0, 30.0, 1800.0, 1800.0), # GS2
 *     123: (1.8, 40.0, 8.0, 0.0138, 0.0115, 0.0000, 0.0666, 0.0574, 1800.0, 109.0, 30.0, 1600.0, 1600.0), # GS3
 *     124: (2.1, 40.0, 8.0, 0.0872, 0.0138, 0.0046, 0.1561, 0.3260, 1800.0, 109.0, 30.0, 1600.0, 1600.0), # GS4             # <<<<<<<<<<<<<<
 *     # Shrub (SH)
 *     141: (1.0, 15.0, 8.0, 0.0115, 0.0115, 0.0000, 0.0069, 0.0597, 2000.0, 109.0, 30.0, 1800.0, 1600.0), # SH1
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_124, __pyx_tuple__44) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":73
 *     124: (2.1, 40.0, 8.0, 0.0872, 0.0138, 0.0046, 0.1561, 0.3260, 1800.0, 109.0, 30.0, 1600.0, 1600.0), # GS4
 *     # Shrub (SH)
 *     141: (1.0, 15.0, 8.0, 0.0115, 0.0115, 0.0000, 0.0069, 0.0597, 2000.0, 109.0, 30.0, 1800.0, 1600.0), # SH1             # <<<<<<<<<<<<<<
 *     142: (1.0, 15.0, 8.0, 0.0620, 0.1102, 0.0344, 0.0000, 0.1768, 2000.0, 109.0, 30.0,    0.0, 1600.0), # SH2
 *     143: (2.4, 40.0, 8.0, 0.0207, 0.1377, 0.0000, 0.0000, 0.2847, 1600.0, 109.0, 30.0,    0.0, 1400.0), # SH3
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_141, __pyx_tuple__45) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":74
 *     # Shrub (SH)
 *     141: (1.0, 15.0, 8.0, 0.0115, 0.0115, 0.0000, 0.0069, 0.0597, 2000.0, 109.0, 30.0, 1800.0, 1600.0), # SH1
 *     142: (1.0, 15.0, 8.0, 0.0620, 0.1102, 0.0344, 0.0000, 0.1768, 2000.0, 109.0, 30.0,    0.0, 1600.0), # SH2             # <<<<<<<<<<<<<<
 *     143: (2.4, 40.0, 8.0, 0.0207, 0.1377, 0.0000, 0.0000, 0.2847, 1600.0, 109.0, 30.0,    0.0, 1400.0), # SH3
 *     144: (3.0, 30.0, 8.0, 0.0390, 0.0528, 0.0092, 0.0000, 0.1171, 2000.0, 109.0, 30.0, 1800.0, 1600.0), # SH4
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_142, __pyx_tuple__46) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":75
 *     141: (1.0, 15.0, 8.0, 0.0115, 0.0115, 0.0000, 0.0069, 0.0597, 2000.0, 109.0, 30.0, 1800.0, 1600.0), # SH1
 *     142: (1.0, 15.0, 8.0, 0.0620, 0.1102, 0.0344, 0.0000, 0.1768, 2000.0, 109.0, 30.0,    0.0, 1600.0), # SH2
 *     143: (2.4, 40.0, 8.0, 0.0207, 0.1377, 0.0000, 0.0000, 0.2847, 1600.0, 109.0, 30.0,    0.0, 1400.0), # SH3             # <<<<<<<<<<<<<<
 *     144: (3.0, 30.0, 8.0, 0.0390, 0.0528, 0.0092, 0.0000, 0.1171, 2000.0, 109.0, 30.0, 1800.0, 1600.0), # SH4
 *     145: (6.0, 15.0, 8.0, 0.1653, 0.0964, 0.0000, 0.0000, 0.1331,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH5
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_143, __pyx_tuple__47) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":76
 *     142: (1.0, 15.0, 8.0, 0.0620, 0.1102, 0.0344, 0.0000, 0.1768, 2000.0, 109.0, 30.0,    0.0, 1600.0), # SH2
 *     143: (2.4, 40.0, 8.0, 0.0207, 0.1377, 0.0000, 0.0000, 0.2847, 1600.0, 109.0, 30.0,    0.0, 1400.0), # SH3
 *     144: (3.0, 30.0, 8.0, 0.0390, 0.0528, 0.0092, 0.0000, 0.1171, 2000.0, 109.0, 30.0, 1800.0, 1600.0), # SH4             # <<<<<<<<<<<<<<
 *     145: (6.0, 15.0, 8.0, 0.1653, 0.0964, 0.0000, 0.0000, 0.1331,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH5
 *     146: (2.0, 30.0, 8.0, 0.1331, 0.0666, 0.0000, 0.0000, 0.0643,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH6
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_144, __pyx_tuple__48) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":77
 *     143: (2.4, 40.0, 8.0, 0.0207, 0.1377, 0.0000, 0.0000, 0.2847, 1600.0, 109.0, 30.0,    0.0, 1400.0), # SH3
 *     144: (3.0, 30.0, 8.0, 0.0390, 0.0528, 0.0092, 0.0000, 0.1171, 2000.0, 109.0, 30.0, 1800.0, 1600.0), # SH4
 *     145: (6.0, 15.0, 8.0, 0.1653, 0.0964, 0.0000, 0.0000, 0.1331,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH5             # <<<<<<<<<<<<<<
 *     146: (2.0, 30.0, 8.0, 0.1331, 0.0666, 0.0000, 0.0000, 0.0643,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH6
 *     147: (6.0, 15.0, 8.0, 0.1607, 0.2433, 0.1010, 0.0000, 0.1561,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH7
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_145, __pyx_tuple__49) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":78
 *     144: (3.0, 30.0, 8.0, 0.0390, 0.0528, 0.0092, 0.0000, 0.1171, 2000.0, 109.0, 30.0, 1800.0, 1600.0), # SH4
 *     145: (6.0, 15.0, 8.0, 0.1653, 0.0964, 0.0000, 0.0000, 0.1331,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH5
 *     146: (2.0, 30.0, 8.0, 0.1331, 0.0666, 0.0000, 0.0000, 0.0643,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH6             # <<<<<<<<<<<<<<
 *     147: (6.0, 15.0, 8.0, 0.1607, 0.2433, 0.1010, 0.0000, 0.1561,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH7
 *     148: (3.0, 40.0, 8.0, 0.0941, 0.1561, 0.0390, 0.0000, 0.1997,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH8
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_146, __pyx_tuple__50) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":79
 *     145: (6.0, 15.0, 8.0, 0.1653, 0.0964, 0.0000, 0.0000, 0.1331,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH5
 *     146: (2.0, 30.0, 8.0, 0.1331, 0.0666, 0.0000, 0.0000, 0.0643,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH6
 *     147: (6.0, 15.0, 8.0, 0.1607, 0.2433, 0.1010, 0.0000, 0.1561,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH7             # <<<<<<<<<<<<<<
 *     148: (3.0, 40.0, 8.0, 0.0941, 0.1561, 0.0390, 0.0000, 0.1997,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH8
 *     149: (4.4, 40.0, 8.0, 0.2066, 0.1125, 0.0000, 0.0712, 0.3214,  750.0, 109.0, 30.0, 1800.0, 1500.0), # SH9
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_147, __pyx_tuple__51) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":80
 *     146: (2.0, 30.0, 8.0, 0.1331, 0.0666, 0.0000, 0.0000, 0.0643,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH6
 *     147: (6.0, 15.0, 8.0, 0.1607, 0.2433, 0.1010, 0.0000, 0.1561,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH7
 *     148: (3.0, 40.0, 8.0, 0.0941, 0.1561, 0.0390, 0.0000, 0.1997,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH8             # <<<<<<<<<<<<<<
 *     149: (4.4, 40.0, 8.0, 0.2066, 0.1125, 0.0000, 0.0712, 0.3214,  750.0, 109.0, 30.0, 1800.0, 1500.0), # SH9
 *     # Timber-Understory (TU)
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_148, __pyx_tuple__52) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":81
 *     147: (6.0, 15.0, 8.0, 0.1607, 0.2433, 0.1010, 0.0000, 0.1561,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH7
 *     148: (3.0, 40.0, 8.0, 0.0941, 0.1561, 0.0390, 0.0000, 0.1997,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH8
 *     149: (4.4, 40.0, 8.0, 0.2066, 0.1125, 0.0000, 0.0712, 0.3214,  750.0, 109.0, 30.0, 1800.0, 1500.0), # SH9             # <<<<<<<<<<<<<<
 *     # Timber-Understory (TU)
 *     161: (0.6, 20.0, 8.0, 0.0092, 0.0413, 0.0689, 0.0092, 0.0413, 2000.0, 109.0, 30.0, 1800.0, 1600.0), # TU1
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_149, __pyx_tuple__53) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":83
 *     149: (4.4, 40.0, 8.0, 0.2066, 0.1125, 0.0000, 0.0712, 0.3214,  750.0, 109.0, 30.0, 1800.0, 1500.0), # SH9
 *     # Timber-Understory (TU)
 *     161: (0.6, 20.0, 8.0, 0.0092, 0.0413, 0.0689, 0.0092, 0.0413, 2000.0, 109.0, 30.0, 1800.0, 1600.0), # TU1             # <<<<<<<<<<<<<<
 *     162: (1.0, 30.0, 8.0, 0.0436, 0.0826, 0.0574, 0.0000, 0.0092, 2000.0, 109.0, 30.0,    0.0, 1600.0), # TU2
 *     163: (1.3, 30.0, 8.0, 0.0505, 0.0069, 0.0115, 0.0298, 0.0505, 1800.0, 109.0, 30.0, 1600.0, 1400.0), # TU3
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_161, __pyx_tuple__54) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":84
 *     # Timber-Understory (TU)
 *     161: (0.6, 20.0, 8.0, 0.0092, 0.0413, 0.0689, 0.0092, 0.0413, 2000.0, 109.0, 30.0, 1800.0, 1600.0), # TU1
 *     162: (1.0, 30.0, 8.0, 0.0436, 0.0826, 0.0574, 0.0000, 0.0092, 2000.0, 109.0, 30.0,    0.0, 1600.0), # TU2             # <<<<<<<<<<<<<<
 *     163: (1.3, 30.0, 8.0, 0.0505, 0.0069, 0.0115, 0.0298, 0.0505, 1800.0, 109.0, 30.0, 1600.0, 1400.0), # TU3
 *     164: (0.5, 12.0, 8.0, 0.2066, 0.0000, 0.0000, 0.0000, 0.0918, 2300.0, 109.0, 30.0,    0.0, 2000.0), # TU4
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_162, __pyx_tuple__55) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":85
 *     161: (0.6, 20.0, 8.0, 0.0092, 0.0413, 0.0689, 0.0092, 0.0413, 2000.0, 109.0, 30.0, 1800.0, 1600.0), # TU1
 *     162: (1.0, 30.0, 8.0, 0.0436, 0.0826, 0.0574, 0.0000, 0.0092, 2000.0, 109.0, 30.0,    0.0, 1600.0), # TU2
 *     163: (1.3, 30.0, 8.0, 0.0505, 0.0069, 0.0115, 0.0298, 0.0505, 1800.0, 109.0, 30.0, 1600.0, 1400.0), # TU3             # <<<<<<<<<<<<<<
 *     164: (0.5, 12.0, 8.0, 0.2066, 0.0000, 0.0000, 0.0000, 0.0918, 2300.0, 109.0, 30.0,    0.0, 2000.0), # TU4
 *     165: (1.0, 25.0, 8.0, 0.1837, 0.1837, 0.1377, 0.0000, 0.1377, 1500.0, 109.0, 30.0,    0.0,  750.0), # TU5
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_163, __pyx_tuple__56) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":86
 *     162: (1.0, 30.0, 8.0, 0.0436, 0.0826, 0.0574, 0.0000, 0.0092, 2000.0, 109.0, 30.0,    0.0, 1600.0), # TU2
 *     163: (1.3, 30.0, 8.0, 0.0505, 0.0069, 0.0115, 0.0298, 0.0505, 1800.0, 109.0, 30.0, 1600.0, 1400.0), # TU3
 *     164: (0.5, 12.0, 8.0, 0.2066, 0.0000, 0.0000, 0.0000, 0.0918, 2300.0, 109.0, 30.0,    0.0, 2000.0), # TU4             # <<<<<<<<<<<<<<
 *     165: (1.0, 25.0, 8.0, 0.1837, 0.1837, 0.1377, 0.0000, 0.1377, 1500.0, 109.0, 30.0,    0.0,  750.0), # TU5
 *     # Timber Litter (TL)
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_164, __pyx_tuple__57) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":87
 *     163: (1.3, 30.0, 8.0, 0.0505, 0.0069, 0.0115, 0.0298, 0.0505, 1800.0, 109.0, 30.0, 1600.0, 1400.0), # TU3
 *     164: (0.5, 12.0, 8.0, 0.2066, 0.0000, 0.0000, 0.0000, 0.0918, 2300.0, 109.0, 30.0,    0.0, 2000.0), # TU4
 *     165: (1.0, 25.0, 8.0, 0.1837, 0.1837, 0.1377, 0.0000, 0.1377, 1500.0, 109.0, 30.0,    0.0,  750.0), # TU5             # <<<<<<<<<<<<<<
 *     # Timber Litter (TL)
 *     181: (0.2, 30.0, 8.0, 0.0459, 0.1010, 0.1653, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL1
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_165, __pyx_tuple__58) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":89
 *     165: (1.0, 25.0, 8.0, 0.1837, 0.1837, 0.1377, 0.0000, 0.1377, 1500.0, 109.0, 30.0,    0.0,  750.0), # TU5
 *     # Timber Litter (TL)
 *     181: (0.2, 30.0, 8.0, 0.0459, 0.1010, 0.1653, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL1             # <<<<<<<<<<<<<<
 *     182: (0.2, 25.0, 8.0, 0.0643, 0.1056, 0.1010, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL2
 *     183: (0.3, 20.0, 8.0, 0.0230, 0.1010, 0.1286, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL3
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_181, __pyx_tuple__59) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":90
 *     # Timber Litter (TL)
 *     181: (0.2, 30.0, 8.0, 0.0459, 0.1010, 0.1653, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL1
 *     182: (0.2, 25.0, 8.0, 0.0643, 0.1056, 0.1010, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL2             # <<<<<<<<<<<<<<
 *     183: (0.3, 20.0, 8.0, 0.0230, 0.1010, 0.1286, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL3
 *     184: (0.4, 25.0, 8.0, 0.0230, 0.0689, 0.1928, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL4
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_182, __pyx_tuple__60) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":91
 *     181: (0.2, 30.0, 8.0, 0.0459, 0.1010, 0.1653, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL1
 *     182: (0.2, 25.0, 8.0, 0.0643, 0.1056, 0.1010, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL2
 *     183: (0.3, 20.0, 8.0, 0.0230, 0.1010, 0.1286, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL3             # <<<<<<<<<<<<<<
 *     184: (0.4, 25.0, 8.0, 0.0230, 0.0689, 0.1928, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL4
 *     185: (0.6, 25.0, 8.0, 0.0528, 0.1148, 0.2020, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0, 1600.0), # TL5
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_183, __pyx_tuple__61) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":92
 *     182: (0.2, 25.0, 8.0, 0.0643, 0.1056, 0.1010, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL2
 *     183: (0.3, 20.0, 8.0, 0.0230, 0.1010, 0.1286, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL3
 *     184: (0.4, 25.0, 8.0, 0.0230, 0.0689, 0.1928, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL4             # <<<<<<<<<<<<<<
 *     185: (0.6, 25.0, 8.0, 0.0528, 0.1148, 0.2020, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0, 1600.0), # TL5
 *     186: (0.3, 25.0, 8.0, 0.1102, 0.0551, 0.0551, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL6
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_184, __pyx_tuple__62) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":93
 *     183: (0.3, 20.0, 8.0, 0.0230, 0.1010, 0.1286, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL3
 *     184: (0.4, 25.0, 8.0, 0.0230, 0.0689, 0.1928, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL4
 *     185: (0.6, 25.0, 8.0, 0.0528, 0.1148, 0.2020, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0, 1600.0), # TL5             # <<<<<<<<<<<<<<
 *     186: (0.3, 25.0, 8.0, 0.1102, 0.0551, 0.0551, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL6
 *     187: (0.4, 25.0, 8.0, 0.0138, 0.0643, 0.3719, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL7
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_185, __pyx_tuple__63) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":94
 *     184: (0.4, 25.0, 8.0, 0.0230, 0.0689, 0.1928, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL4
 *     185: (0.6, 25.0, 8.0, 0.0528, 0.1148, 0.2020, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0, 1600.0), # TL5
 *     186: (0.3, 25.0, 8.0, 0.1102, 0.0551, 0.0551, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL6             # <<<<<<<<<<<<<<
 *     187: (0.4, 25.0, 8.0, 0.0138, 0.0643, 0.3719, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL7
 *     188: (0.3, 35.0, 8.0, 0.2663, 0.0643, 0.0505, 0.0000, 0.0000, 1800.0, 109.0, 30.0,    0.0,    0.0), # TL8
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_186, __pyx_tuple__64) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":95
 *     185: (0.6, 25.0, 8.0, 0.0528, 0.1148, 0.2020, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0, 1600.0), # TL5
 *     186: (0.3, 25.0, 8.0, 0.1102, 0.0551, 0.0551, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL6
 *     187: (0.4, 25.0, 8.0, 0.0138, 0.0643, 0.3719, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL7             # <<<<<<<<<<<<<<
 *     188: (0.3, 35.0, 8.0, 0.2663, 0.0643, 0.0505, 0.0000, 0.0000, 1800.0, 109.0, 30.0,    0.0,    0.0), # TL8
 *     189: (0.6, 35.0, 8.0, 0.3053, 0.1515, 0.1905, 0.0000, 0.0000, 1800.0, 109.0, 30.0,    0.0, 1600.0), # TL9
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_187, __pyx_tuple__65) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":96
 *     186: (0.3, 25.0, 8.0, 0.1102, 0.0551, 0.0551, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL6
 *     187: (0.4, 25.0, 8.0, 0.0138, 0.0643, 0.3719, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL7
 *     188: (0.3, 35.0, 8.0, 0.2663, 0.0643, 0.0505, 0.0000, 0.0000, 1800.0, 109.0, 30.0,    0.0,    0.0), # TL8             # <<<<<<<<<<<<<<
 *     189: (0.6, 35.0, 8.0, 0.3053, 0.1515, 0.1905, 0.0000, 0.0000, 1800.0, 109.0, 30.0,    0.0, 1600.0), # TL9
 *     # Slash-Blowdown (SB)
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_188, __pyx_tuple__66) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":97
 *     187: (0.4, 25.0, 8.0, 0.0138, 0.0643, 0.3719, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL7
 *     188: (0.3, 35.0, 8.0, 0.2663, 0.0643, 0.0505, 0.0000, 0.0000, 1800.0, 109.0, 30.0,    0.0,    0.0), # TL8
 *     189: (0.6, 35.0, 8.0, 0.3053, 0.1515, 0.1905, 0.0000, 0.0000, 1800.0, 109.0, 30.0,    0.0, 1600.0), # TL9             # <<<<<<<<<<<<<<
 *     # Slash-Blowdown (SB)
 *     201: (1.0, 25.0, 8.0, 0.0689, 0.1377, 0.5051, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # SB1
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_189, __pyx_tuple__67) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":99
 *     189: (0.6, 35.0, 8.0, 0.3053, 0.1515, 0.1905, 0.0000, 0.0000, 1800.0, 109.0, 30.0,    0.0, 1600.0), # TL9
 *     # Slash-Blowdown (SB)
 *     201: (1.0, 25.0, 8.0, 0.0689, 0.1377, 0.5051, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # SB1             # <<<<<<<<<<<<<<
 *     202: (1.0, 25.0, 8.0, 0.2066, 0.1951, 0.1837, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # SB2
 *     203: (1.2, 25.0, 8.0, 0.2525, 0.1263, 0.1377, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # SB3
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_201, __pyx_tuple__68) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":100
 *     # Slash-Blowdown (SB)
 *     201: (1.0, 25.0, 8.0, 0.0689, 0.1377, 0.5051, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # SB1
 *     202: (1.0, 25.0, 8.0, 0.2066, 0.1951, 0.1837, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # SB2             # <<<<<<<<<<<<<<
 *     203: (1.2, 25.0, 8.0, 0.2525, 0.1263, 0.1377, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # SB3
 *     204: (2.7, 25.0, 8.0, 0.2410, 0.1607, 0.2410, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # SB4
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_202, __pyx_tuple__69) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":101
 *     201: (1.0, 25.0, 8.0, 0.0689, 0.1377, 0.5051, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # SB1
 *     202: (1.0, 25.0, 8.0, 0.2066, 0.1951, 0.1837, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # SB2
 *     203: (1.2, 25.0, 8.0, 0.2525, 0.1263, 0.1377, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # SB3             # <<<<<<<<<<<<<<
 *     204: (2.7, 25.0, 8.0, 0.2410, 0.1607, 0.2410, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # SB4
 * })
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_203, __pyx_tuple__70) < 0) __PYX_ERR(1, 34, __pyx_L1_error)

  /* "pyretechnics/fuel_models.py":102
 *     202: (1.0, 25.0, 8.0, 0.2066, 0.1951, 0.1837, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # SB2
 *     203: (1.2, 25.0, 8.0, 0.2525, 0.1263, 0.1377, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # SB3
 *     204: (2.7, 25.0, 8.0, 0.2410, 0.1607, 0.2410, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # SB4             # <<<<<<<<<<<<<<
 * })
 * # fuel-model-compact-table ends here
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_204, __pyx_tuple__71) < 0) __PYX_ERR(1, 34, __pyx_L1_error)
  __Pyx_XGOTREF(__pyx_v_12pyretechnics_11fuel_models_compact_fuel_model_table);
  __Pyx_DECREF_SET(__pyx_v_12pyretechnics_11fuel_models_compact_fuel_model_table, ((PyObject*)__pyx_t_2));
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "pyretechnics/fuel_models.py":107
 * # [[file:../../org/pyretechnics.org::expand-compact-fuel-model-table][expand-compact-fuel-model-table]]
 * # NOTE: We use this variable for X > 0.0 comparisons to account for floating point precision issues.
 * almost_zero = cy.declare(cy.float, 1e-6)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_12pyretechnics_11fuel_models_almost_zero = 1e-6;

  /* "pyretechnics/fuel_models.py":214
 * 
 * # TODO: OPTIM Replace this dictionary with something more efficient
 * fuel_model_table = cy.declare(dict, { # dict[int, FuelModel]             # <<<<<<<<<<<<<<
 *     k: expand_compact_fuel_model(k) for k in compact_fuel_model_table.keys()
 * })
 */
  { /* enter inner scope */
    __pyx_t_2 = PyDict_New(); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 214, __pyx_L4_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "pyretechnics/fuel_models.py":215
 * # TODO: OPTIM Replace this dictionary with something more efficient
 * fuel_model_table = cy.declare(dict, { # dict[int, FuelModel]
 *     k: expand_compact_fuel_model(k) for k in compact_fuel_model_table.keys()             # <<<<<<<<<<<<<<
 * })
 * 
 */
    __pyx_t_4 = 0;
    if (unlikely(__pyx_v_12pyretechnics_11fuel_models_compact_fuel_model_table == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "keys");
      __PYX_ERR(1, 215, __pyx_L4_error)
    }
    __pyx_t_7 = __Pyx_dict_iterator(__pyx_v_12pyretechnics_11fuel_models_compact_fuel_model_table, 1, __pyx_n_s_keys, (&__pyx_t_5), (&__pyx_t_6)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 215, __pyx_L4_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_3);
    __pyx_t_3 = __pyx_t_7;
    __pyx_t_7 = 0;
    while (1) {
      __pyx_t_8 = __Pyx_dict_iter_next(__pyx_t_3, __pyx_t_5, &__pyx_t_4, &__pyx_t_7, NULL, NULL, __pyx_t_6);
      if (unlikely(__pyx_t_8 == 0)) break;
      if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(1, 215, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_7genexpr__pyx_v_12pyretechnics_11fuel_models_k);
      __Pyx_DECREF_SET(__pyx_7genexpr__pyx_v_12pyretechnics_11fuel_models_k, __pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_7);
      __pyx_t_7 = 0;
      __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_7genexpr__pyx_v_12pyretechnics_11fuel_models_k); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 215, __pyx_L4_error)
      __pyx_t_7 = __pyx_convert__to_py_struct____pyx_t_12pyretechnics_8cy_types_FuelModel(__pyx_f_12pyretechnics_11fuel_models_expand_compact_fuel_model(__pyx_t_8)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 215, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (unlikely(PyDict_SetItem(__pyx_t_2, (PyObject*)__pyx_7genexpr__pyx_v_12pyretechnics_11fuel_models_k, (PyObject*)__pyx_t_7))) __PYX_ERR(1, 215, __pyx_L4_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_7genexpr__pyx_v_12pyretechnics_11fuel_models_k);
    __Pyx_DECREF_SET(__pyx_7genexpr__pyx_v_12pyretechnics_11fuel_models_k, Py_None);
    goto __pyx_L7_exit_scope;
    __pyx_L4_error:;
    __Pyx_GOTREF(__pyx_7genexpr__pyx_v_12pyretechnics_11fuel_models_k);
    __Pyx_DECREF_SET(__pyx_7genexpr__pyx_v_12pyretechnics_11fuel_models_k, Py_None);
    goto __pyx_L1_error;
    __pyx_L7_exit_scope:;
  } /* exit inner scope */
  __Pyx_XGOTREF(__pyx_v_12pyretechnics_11fuel_models_fuel_model_table);
  __Pyx_DECREF_SET(__pyx_v_12pyretechnics_11fuel_models_fuel_model_table, ((PyObject*)__pyx_t_2));
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "pyretechnics/fuel_models.py":219
 * 
 * 
 * @cy.ccall             # <<<<<<<<<<<<<<
 * @cy.inline
 * def list_fuel_model_numbers() -> list[int]:
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_12pyretechnics_11fuel_models_1list_fuel_model_numbers, 0, __pyx_n_s_list_fuel_model_numbers, NULL, __pyx_n_s_pyretechnics_fuel_models, __pyx_d, ((PyObject *)__pyx_codeobj__72)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_list_fuel_model_numbers, __pyx_t_2) < 0) __PYX_ERR(1, 219, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyretechnics/fuel_models.py":225
 * 
 * 
 * @cy.ccall             # <<<<<<<<<<<<<<
 * @cy.inline
 * def list_fuel_models() -> list[FuelModel]:
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_12pyretechnics_11fuel_models_3list_fuel_models, 0, __pyx_n_s_list_fuel_models, NULL, __pyx_n_s_pyretechnics_fuel_models, __pyx_d, ((PyObject *)__pyx_codeobj__73)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_list_fuel_models, __pyx_t_2) < 0) __PYX_ERR(1, 225, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyretechnics/fuel_models.py":231
 * 
 * 
 * @cy.ccall             # <<<<<<<<<<<<<<
 * @cy.inline
 * @cy.exceptval(check=False)
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_fuel_model_number, __pyx_kp_s_cy_int) < 0) __PYX_ERR(1, 231, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_12pyretechnics_11fuel_models_5fuel_model_exists, 0, __pyx_n_s_fuel_model_exists, NULL, __pyx_n_s_pyretechnics_fuel_models, __pyx_d, ((PyObject *)__pyx_codeobj__75)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_fuel_model_exists, __pyx_t_3) < 0) __PYX_ERR(1, 231, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyretechnics/fuel_models.py":238
 * 
 * 
 * @cy.ccall             # <<<<<<<<<<<<<<
 * @cy.inline
 * @cy.exceptval(check=False)
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_fuel_model_number, __pyx_kp_s_cy_int) < 0) __PYX_ERR(1, 238, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_12pyretechnics_11fuel_models_7get_fuel_model, 0, __pyx_n_s_get_fuel_model, NULL, __pyx_n_s_pyretechnics_fuel_models, __pyx_d, ((PyObject *)__pyx_codeobj__76)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_get_fuel_model, __pyx_t_2) < 0) __PYX_ERR(1, 238, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyretechnics/fuel_models.py":423
 * # add-live-moisture-of-extinction ends here
 * # [[file:../../org/pyretechnics.org::moisturize][moisturize]]
 * @cy.ccall             # <<<<<<<<<<<<<<
 * @cy.exceptval(check=False)
 * def moisturize(fuel_model: FuelModel, fuel_moisture: fclaarr) -> FuelModel:
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_fuel_model, __pyx_n_s_FuelModel) < 0) __PYX_ERR(1, 423, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_fuel_moisture, __pyx_n_s_fclaarr) < 0) __PYX_ERR(1, 423, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_12pyretechnics_11fuel_models_9moisturize, 0, __pyx_n_s_moisturize, NULL, __pyx_n_s_pyretechnics_fuel_models, __pyx_d, ((PyObject *)__pyx_codeobj__78)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_moisturize, __pyx_t_3) < 0) __PYX_ERR(1, 423, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyretechnics/fuel_models.py":1
 * # [[file:../../org/pyretechnics.org::fuel-models-imports][fuel-models-imports]]             # <<<<<<<<<<<<<<
 * import cython
 * import cython as cy
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_3) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  if (__pyx_m) {
    if (__pyx_d && stringtab_initialized) {
      __Pyx_AddTraceback("init pyretechnics.fuel_models", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    #if !CYTHON_USE_MODULE_STATE
    Py_CLEAR(__pyx_m);
    #else
    Py_DECREF(__pyx_m);
    if (pystate_addmodule_run) {
      PyObject *tp, *value, *tb;
      PyErr_Fetch(&tp, &value, &tb);
      PyState_RemoveModule(&__pyx_moduledef);
      PyErr_Restore(tp, value, tb);
    }
    #endif
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init pyretechnics.fuel_models");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_pragmas ### */
#ifdef _MSC_VER
#pragma warning( push )
/* Warning 4127: conditional expression is constant
 * Cython uses constant conditional expressions to allow in inline functions to be optimized at
 * compile-time, so this warning is not useful
 */
#pragma warning( disable : 4127 )
#endif



/* #### Code section: utility_code_def ### */

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    int result;
    PyObject *exc_type;
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *current_exception = tstate->current_exception;
    if (unlikely(!current_exception)) return 0;
    exc_type = (PyObject*) Py_TYPE(current_exception);
    if (exc_type == err) return 1;
#else
    exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
#endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(exc_type);
    #endif
    if (unlikely(PyTuple_Check(err))) {
        result = __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    } else {
        result = __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
    }
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(exc_type);
    #endif
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *tmp_value;
    assert(type == NULL || (value != NULL && type == (PyObject*) Py_TYPE(value)));
    if (value) {
        #if CYTHON_COMPILING_IN_CPYTHON
        if (unlikely(((PyBaseExceptionObject*) value)->traceback != tb))
        #endif
            PyException_SetTraceback(value, tb);
    }
    tmp_value = tstate->current_exception;
    tstate->current_exception = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
#else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject* exc_value;
    exc_value = tstate->current_exception;
    tstate->current_exception = 0;
    *value = exc_value;
    *type = NULL;
    *tb = NULL;
    if (exc_value) {
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        #if CYTHON_COMPILING_IN_CPYTHON
        *tb = ((PyBaseExceptionObject*) exc_value)->traceback;
        Py_XINCREF(*tb);
        #else
        *tb = PyException_GetTraceback(exc_value);
        #endif
    }
#else
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#endif
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError */
#if __PYX_LIMITED_VERSION_HEX < 0x030d00A1
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d00A1
    (void) PyObject_GetOptionalAttr(obj, attr_name, &result);
    return result;
#else
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
#endif
}

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStrNoError(__pyx_b, name);
    if (unlikely(!result) && !PyErr_Occurred()) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_value == NULL || exc_info->exc_value == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    PyObject *exc_value = exc_info->exc_value;
    if (exc_value == NULL || exc_value == Py_None) {
        *value = NULL;
        *type = NULL;
        *tb = NULL;
    } else {
        *value = exc_value;
        Py_INCREF(*value);
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        *tb = PyException_GetTraceback(exc_value);
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #endif
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    PyObject *tmp_value = exc_info->exc_value;
    exc_info->exc_value = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
  #else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
  #endif
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type = NULL, *local_value, *local_tb = NULL;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if PY_VERSION_HEX >= 0x030C00A6
    local_value = tstate->current_exception;
    tstate->current_exception = 0;
    if (likely(local_value)) {
        local_type = (PyObject*) Py_TYPE(local_value);
        Py_INCREF(local_type);
        local_tb = PyException_GetTraceback(local_value);
    }
  #else
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
  #endif
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE && PY_VERSION_HEX >= 0x030C00A6
    if (unlikely(tstate->current_exception))
#elif CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
      #if PY_VERSION_HEX >= 0x030B00a4
        tmp_value = exc_info->exc_value;
        exc_info->exc_value = local_value;
        tmp_type = NULL;
        tmp_tb = NULL;
        Py_XDECREF(local_type);
        Py_XDECREF(local_tb);
      #else
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
      #endif
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    #endif
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    __Pyx_PyThreadState_declare
    CYTHON_UNUSED_VAR(cause);
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
      #if PY_VERSION_HEX >= 0x030C00A6
        PyException_SetTraceback(value, tb);
      #elif CYTHON_FAST_THREAD_STATE
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#else
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
    else state = (PyGILState_STATE)0;
#endif
    CYTHON_UNUSED_VAR(clineno);
    CYTHON_UNUSED_VAR(lineno);
    CYTHON_UNUSED_VAR(filename);
    CYTHON_MAYBE_UNUSED_VAR(nogil);
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(0);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* UnpackUnboundCMethod */
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *args, PyObject *kwargs) {
    PyObject *result;
    PyObject *selfless_args = PyTuple_GetSlice(args, 1, PyTuple_Size(args));
    if (unlikely(!selfless_args)) return NULL;
    result = PyObject_Call(method, selfless_args, kwargs);
    Py_DECREF(selfless_args);
    return result;
}
static PyMethodDef __Pyx_UnboundCMethod_Def = {
     "CythonUnboundCMethod",
     __PYX_REINTERPRET_FUNCION(PyCFunction, __Pyx_SelflessCall),
     METH_VARARGS | METH_KEYWORDS,
     NULL
};
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #else
    if (likely(!__Pyx_CyOrPyCFunction_Check(method)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    } else
#endif
#if CYTHON_COMPILING_IN_PYPY
#else
    if (PyCFunction_Check(method))
#endif
    {
        PyObject *self;
        int self_found;
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        self = PyObject_GetAttrString(method, "__self__");
        if (!self) {
            PyErr_Clear();
        }
#else
        self = PyCFunction_GET_SELF(method);
#endif
        self_found = (self && self != Py_None);
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        Py_XDECREF(self);
#endif
        if (self_found) {
            PyObject *unbound_method = PyCFunction_New(&__Pyx_UnboundCMethod_Def, method);
            if (unlikely(!unbound_method)) return -1;
            Py_DECREF(method);
            target->method = unbound_method;
        }
    }
    return 0;
}

/* CallUnboundCMethod0 */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_ASSUME_SAFE_MACROS
    args = PyTuple_New(1);
    if (unlikely(!args)) goto bad;
    Py_INCREF(self);
    PyTuple_SET_ITEM(args, 0, self);
#else
    args = PyTuple_Pack(1, self);
    if (unlikely(!args)) goto bad;
#endif
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    Py_DECREF(args);
bad:
    return result;
}

/* py_dict_keys */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Keys(PyObject* d) {
    if (PY_MAJOR_VERSION >= 3)
        return __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyDict_Type_keys, d);
    else
        return PyDict_Keys(d);
}

/* TupleAndListFromArray */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject *const *CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        Py_INCREF(__pyx_empty_tuple);
        return __pyx_empty_tuple;
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyTupleObject*)res)->ob_item, n);
    return res;
}
static CYTHON_INLINE PyObject *
__Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return PyList_New(0);
    }
    res = PyList_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyListObject*)res)->ob_item, n);
    return res;
}
#endif

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* fastcall */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s)
{
    Py_ssize_t i, n = PyTuple_GET_SIZE(kwnames);
    for (i = 0; i < n; i++)
    {
        if (s == PyTuple_GET_ITEM(kwnames, i)) return kwvalues[i];
    }
    for (i = 0; i < n; i++)
    {
        int eq = __Pyx_PyUnicode_Equals(s, PyTuple_GET_ITEM(kwnames, i), Py_EQ);
        if (unlikely(eq != 0)) {
            if (unlikely(eq < 0)) return NULL;
            return kwvalues[i];
        }
    }
    return NULL;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues) {
    Py_ssize_t i, nkwargs = PyTuple_GET_SIZE(kwnames);
    PyObject *dict;
    dict = PyDict_New();
    if (unlikely(!dict))
        return NULL;
    for (i=0; i<nkwargs; i++) {
        PyObject *key = PyTuple_GET_ITEM(kwnames, i);
        if (unlikely(PyDict_SetItem(dict, key, kwvalues[i]) < 0))
            goto bad;
    }
    return dict;
bad:
    Py_DECREF(dict);
    return NULL;
}
#endif
#endif

/* py_dict_values */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Values(PyObject* d) {
    if (PY_MAJOR_VERSION >= 3)
        return __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyDict_Type_values, d);
    else
        return PyDict_Values(d);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    int kwds_is_tuple = CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds));
    while (1) {
        Py_XDECREF(key); key = NULL;
        Py_XDECREF(value); value = NULL;
        if (kwds_is_tuple) {
            Py_ssize_t size;
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(kwds);
#else
            size = PyTuple_Size(kwds);
            if (size < 0) goto bad;
#endif
            if (pos >= size) break;
#if CYTHON_AVOID_BORROWED_REFS
            key = __Pyx_PySequence_ITEM(kwds, pos);
            if (!key) goto bad;
#elif CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kwds, pos);
#else
            key = PyTuple_GetItem(kwds, pos);
            if (!key) goto bad;
#endif
            value = kwvalues[pos];
            pos++;
        }
        else
        {
            if (!PyDict_Next(kwds, &pos, &key, &value)) break;
#if CYTHON_AVOID_BORROWED_REFS
            Py_INCREF(key);
#endif
        }
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
            Py_INCREF(value);
            Py_DECREF(key);
#endif
            key = NULL;
            value = NULL;
            continue;
        }
#if !CYTHON_AVOID_BORROWED_REFS
        Py_INCREF(key);
#endif
        Py_INCREF(value);
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
                    value = NULL;
#endif
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key)
                );
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
                    value = NULL;
#endif
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    Py_XDECREF(key);
    Py_XDECREF(value);
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    #if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_TypeError,
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    Py_XDECREF(key);
    Py_XDECREF(value);
    return -1;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
    PyObject* exc_type;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    exc_type = __Pyx_PyErr_CurrentExceptionType();
    if (unlikely(exc_type)) {
        if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))
            return -1;
        __Pyx_PyErr_Clear();
        return 0;
    }
    return 0;
}

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL && !CYTHON_VECTORCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object"))) {
        return NULL;
    }
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object"))) {
        return NULL;
    }
    #endif
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = __Pyx_CyOrPyCFunction_GET_FUNCTION(func);
    self = __Pyx_CyOrPyCFunction_GET_SELF(func);
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    #endif
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall */
#if PY_VERSION_HEX < 0x03090000 || CYTHON_COMPILING_IN_LIMITED_API
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject **args, size_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result = 0;
    size_t i;
    argstuple = PyTuple_New((Py_ssize_t)nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        if (__Pyx_PyTuple_SET_ITEM(argstuple, (Py_ssize_t)i, args[i]) < 0) goto bad;
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
  bad:
    Py_DECREF(argstuple);
    return result;
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, size_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_NOARGS))
            return __Pyx_PyObject_CallMethO(func, NULL);
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_O))
            return __Pyx_PyObject_CallMethO(func, args[0]);
    }
#endif
    #if PY_VERSION_HEX < 0x030800B1
    #if CYTHON_FAST_PYCCALL
    if (PyCFunction_Check(func)) {
        if (kwargs) {
            return _PyCFunction_FastCallDict(func, args, nargs, kwargs);
        } else {
            return _PyCFunction_FastCallKeywords(func, args, nargs, NULL);
        }
    }
    #if PY_VERSION_HEX >= 0x030700A1
    if (!kwargs && __Pyx_IS_TYPE(func, &PyMethodDescr_Type)) {
        return _PyMethodDescr_FastCallKeywords(func, args, nargs, NULL);
    }
    #endif
    #endif
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs);
    }
    #endif
    #endif
    if (kwargs == NULL) {
        #if CYTHON_VECTORCALL
        #if PY_VERSION_HEX < 0x03090000
        vectorcallfunc f = _PyVectorcall_Function(func);
        #else
        vectorcallfunc f = PyVectorcall_Function(func);
        #endif
        if (f) {
            return f(func, args, (size_t)nargs, NULL);
        }
        #elif defined(__Pyx_CyFunction_USED) && CYTHON_BACKPORT_VECTORCALL
        if (__Pyx_CyFunction_CheckExact(func)) {
            __pyx_vectorcallfunc f = __Pyx_CyFunction_func_vectorcall(func);
            if (f) return f(func, args, (size_t)nargs, NULL);
        }
        #endif
    }
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_empty_tuple, kwargs);
    }
    #if PY_VERSION_HEX >= 0x03090000 && !CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_VectorcallDict(func, args, (size_t)nargs, kwargs);
    #else
    return __Pyx_PyObject_FastCall_fallback(func, args, (size_t)nargs, kwargs);
    #endif
}

/* PyObjectCallNoArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
    PyObject *arg[2] = {NULL, NULL};
    return __Pyx_PyObject_FastCall(func, arg + 1, 0 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectCallOneArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *args[2] = {NULL, arg};
    return __Pyx_PyObject_FastCall(func, args+1, 1 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    __Pyx_TypeName type_name;
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if defined(Py_TPFLAGS_METHOD_DESCRIPTOR) && Py_TPFLAGS_METHOD_DESCRIPTOR
        if (__Pyx_PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_METHOD_DESCRIPTOR))
#elif PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (likely(descr != NULL)) {
        *method = descr;
        return 0;
    }
    type_name = __Pyx_PyType_GetName(tp);
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, name);
#else
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%.400s'",
                 type_name, PyString_AS_STRING(name));
#endif
    __Pyx_DECREF_TypeName(type_name);
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    }
    return __Pyx_IterFinish();
}

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* UnpackTupleError */
static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else if (PyTuple_GET_SIZE(t) < index) {
      __Pyx_RaiseNeedMoreValuesError(PyTuple_GET_SIZE(t));
    } else {
      __Pyx_RaiseTooManyValuesError(index);
    }
}

/* UnpackTuple2 */
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
        PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2, int decref_tuple) {
    PyObject *value1 = NULL, *value2 = NULL;
#if CYTHON_COMPILING_IN_PYPY
    value1 = PySequence_ITEM(tuple, 0);  if (unlikely(!value1)) goto bad;
    value2 = PySequence_ITEM(tuple, 1);  if (unlikely(!value2)) goto bad;
#else
    value1 = PyTuple_GET_ITEM(tuple, 0);  Py_INCREF(value1);
    value2 = PyTuple_GET_ITEM(tuple, 1);  Py_INCREF(value2);
#endif
    if (decref_tuple) {
        Py_DECREF(tuple);
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
#if CYTHON_COMPILING_IN_PYPY
bad:
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
#endif
}
static int __Pyx_unpack_tuple2_generic(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                       int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    iternextfunc iternext;
    iter = PyObject_GetIter(tuple);
    if (unlikely(!iter)) goto bad;
    if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
    iternext = __Pyx_PyObject_GetIterNextFunc(iter);
    value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
    value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
    if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
    Py_DECREF(iter);
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
#if CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
#include <string.h>
#endif
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
    if (is_dict) {
#if !CYTHON_COMPILING_IN_PYPY
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
#elif PY_MAJOR_VERSION >= 3
        static PyObject *py_items = NULL, *py_keys = NULL, *py_values = NULL;
        PyObject **pp = NULL;
        if (method_name) {
            const char *name = PyUnicode_AsUTF8(method_name);
            if (strcmp(name, "iteritems") == 0) pp = &py_items;
            else if (strcmp(name, "iterkeys") == 0) pp = &py_keys;
            else if (strcmp(name, "itervalues") == 0) pp = &py_values;
            if (pp) {
                if (!*pp) {
                    *pp = PyUnicode_FromString(name + 4);
                    if (!*pp)
                        return NULL;
                }
                method_name = *pp;
            }
        }
#endif
    }
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_COMPILING_IN_PYPY
    if (source_is_dict) {
        PyObject *key, *value;
        if (unlikely(orig_length != PyDict_Size(iter_obj))) {
            PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
            return -1;
        }
        if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
            return 0;
        }
        if (pitem) {
            PyObject* tuple = PyTuple_New(2);
            if (unlikely(!tuple)) {
                return -1;
            }
            Py_INCREF(key);
            Py_INCREF(value);
            PyTuple_SET_ITEM(tuple, 0, key);
            PyTuple_SET_ITEM(tuple, 1, value);
            *pitem = tuple;
        } else {
            if (pkey) {
                Py_INCREF(key);
                *pkey = key;
            }
            if (pvalue) {
                Py_INCREF(value);
                *pvalue = value;
            }
        }
        return 1;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyTuple_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyList_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyList_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* FixUpExtensionType */
#if CYTHON_USE_TYPE_SPECS
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type) {
#if PY_VERSION_HEX > 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    CYTHON_UNUSED_VAR(spec);
    CYTHON_UNUSED_VAR(type);
#else
    const PyType_Slot *slot = spec->slots;
    while (slot && slot->slot && slot->slot != Py_tp_members)
        slot++;
    if (slot && slot->slot == Py_tp_members) {
        int changed = 0;
#if !(PY_VERSION_HEX <= 0x030900b1 && CYTHON_COMPILING_IN_CPYTHON)
        const
#endif
            PyMemberDef *memb = (PyMemberDef*) slot->pfunc;
        while (memb && memb->name) {
            if (memb->name[0] == '_' && memb->name[1] == '_') {
#if PY_VERSION_HEX < 0x030900b1
                if (strcmp(memb->name, "__weaklistoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_weaklistoffset = memb->offset;
                    changed = 1;
                }
                else if (strcmp(memb->name, "__dictoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_dictoffset = memb->offset;
                    changed = 1;
                }
#if CYTHON_METH_FASTCALL
                else if (strcmp(memb->name, "__vectorcalloffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
#if PY_VERSION_HEX >= 0x030800b4
                    type->tp_vectorcall_offset = memb->offset;
#else
                    type->tp_print = (printfunc) memb->offset;
#endif
                    changed = 1;
                }
#endif
#else
                if ((0));
#endif
#if PY_VERSION_HEX <= 0x030900b1 && CYTHON_COMPILING_IN_CPYTHON
                else if (strcmp(memb->name, "__module__") == 0) {
                    PyObject *descr;
                    assert(memb->type == T_OBJECT);
                    assert(memb->flags == 0 || memb->flags == READONLY);
                    descr = PyDescr_NewMember(type, memb);
                    if (unlikely(!descr))
                        return -1;
                    if (unlikely(PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr) < 0)) {
                        Py_DECREF(descr);
                        return -1;
                    }
                    Py_DECREF(descr);
                    changed = 1;
                }
#endif
            }
            memb++;
        }
        if (changed)
            PyType_Modified(type);
    }
#endif
    return 0;
}
#endif

/* FetchSharedCythonModule */
static PyObject *__Pyx_FetchSharedCythonABIModule(void) {
    return __Pyx_PyImport_AddModuleRef((char*) __PYX_ABI_MODULE_NAME);
}

/* FetchCommonType */
static int __Pyx_VerifyCachedType(PyObject *cached_type,
                               const char *name,
                               Py_ssize_t basicsize,
                               Py_ssize_t expected_basicsize) {
    if (!PyType_Check(cached_type)) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s is not a type object", name);
        return -1;
    }
    if (basicsize != expected_basicsize) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s has the wrong size, try recompiling",
            name);
        return -1;
    }
    return 0;
}
#if !CYTHON_USE_TYPE_SPECS
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* abi_module;
    const char* object_name;
    PyTypeObject *cached_type = NULL;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    object_name = strrchr(type->tp_name, '.');
    object_name = object_name ? object_name+1 : type->tp_name;
    cached_type = (PyTypeObject*) PyObject_GetAttrString(abi_module, object_name);
    if (cached_type) {
        if (__Pyx_VerifyCachedType(
              (PyObject *)cached_type,
              object_name,
              cached_type->tp_basicsize,
              type->tp_basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    if (PyType_Ready(type) < 0) goto bad;
    if (PyObject_SetAttrString(abi_module, object_name, (PyObject *)type) < 0)
        goto bad;
    Py_INCREF(type);
    cached_type = type;
done:
    Py_DECREF(abi_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#else
static PyTypeObject *__Pyx_FetchCommonTypeFromSpec(PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *abi_module, *cached_type = NULL;
    const char* object_name = strrchr(spec->name, '.');
    object_name = object_name ? object_name+1 : spec->name;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    cached_type = PyObject_GetAttrString(abi_module, object_name);
    if (cached_type) {
        Py_ssize_t basicsize;
#if CYTHON_COMPILING_IN_LIMITED_API
        PyObject *py_basicsize;
        py_basicsize = PyObject_GetAttrString(cached_type, "__basicsize__");
        if (unlikely(!py_basicsize)) goto bad;
        basicsize = PyLong_AsSsize_t(py_basicsize);
        Py_DECREF(py_basicsize);
        py_basicsize = 0;
        if (unlikely(basicsize == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
#else
        basicsize = likely(PyType_Check(cached_type)) ? ((PyTypeObject*) cached_type)->tp_basicsize : -1;
#endif
        if (__Pyx_VerifyCachedType(
              cached_type,
              object_name,
              basicsize,
              spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    CYTHON_UNUSED_VAR(module);
    cached_type = __Pyx_PyType_FromModuleAndSpec(abi_module, spec, bases);
    if (unlikely(!cached_type)) goto bad;
    if (unlikely(__Pyx_fix_up_extension_type_from_spec(spec, (PyTypeObject *) cached_type) < 0)) goto bad;
    if (PyObject_SetAttrString(abi_module, object_name, cached_type) < 0) goto bad;
done:
    Py_DECREF(abi_module);
    assert(cached_type == NULL || PyType_Check(cached_type));
    return (PyTypeObject *) cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#endif

/* PyVectorcallFastCallDict */
#if CYTHON_METH_FASTCALL
static PyObject *__Pyx_PyVectorcall_FastCallDict_kw(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    PyObject *res = NULL;
    PyObject *kwnames;
    PyObject **newargs;
    PyObject **kwvalues;
    Py_ssize_t i, pos;
    size_t j;
    PyObject *key, *value;
    unsigned long keys_are_strings;
    Py_ssize_t nkw = PyDict_GET_SIZE(kw);
    newargs = (PyObject **)PyMem_Malloc((nargs + (size_t)nkw) * sizeof(args[0]));
    if (unlikely(newargs == NULL)) {
        PyErr_NoMemory();
        return NULL;
    }
    for (j = 0; j < nargs; j++) newargs[j] = args[j];
    kwnames = PyTuple_New(nkw);
    if (unlikely(kwnames == NULL)) {
        PyMem_Free(newargs);
        return NULL;
    }
    kwvalues = newargs + nargs;
    pos = i = 0;
    keys_are_strings = Py_TPFLAGS_UNICODE_SUBCLASS;
    while (PyDict_Next(kw, &pos, &key, &value)) {
        keys_are_strings &= Py_TYPE(key)->tp_flags;
        Py_INCREF(key);
        Py_INCREF(value);
        PyTuple_SET_ITEM(kwnames, i, key);
        kwvalues[i] = value;
        i++;
    }
    if (unlikely(!keys_are_strings)) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        goto cleanup;
    }
    res = vc(func, newargs, nargs, kwnames);
cleanup:
    Py_DECREF(kwnames);
    for (i = 0; i < nkw; i++)
        Py_DECREF(kwvalues[i]);
    PyMem_Free(newargs);
    return res;
}
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    if (likely(kw == NULL) || PyDict_GET_SIZE(kw) == 0) {
        return vc(func, args, nargs, NULL);
    }
    return __Pyx_PyVectorcall_FastCallDict_kw(func, vc, args, nargs, kw);
}
#endif

/* CythonFunctionShared */
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc) {
    if (__Pyx_CyFunction_Check(func)) {
        return PyCFunction_GetFunction(((__pyx_CyFunctionObject*)func)->func) == (PyCFunction) cfunc;
    } else if (PyCFunction_Check(func)) {
        return PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
    }
    return 0;
}
#else
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc) {
    return __Pyx_CyOrPyCFunction_Check(func) && __Pyx_CyOrPyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
}
#endif
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj) {
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    __Pyx_Py_XDECREF_SET(
        __Pyx_CyFunction_GetClassObj(f),
            ((classobj) ? __Pyx_NewRef(classobj) : NULL));
#else
    __Pyx_Py_XDECREF_SET(
        ((PyCMethodObject *) (f))->mm_class,
        (PyTypeObject*)((classobj) ? __Pyx_NewRef(classobj) : NULL));
#endif
}
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, void *closure)
{
    CYTHON_UNUSED_VAR(closure);
    if (unlikely(op->func_doc == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_doc = PyObject_GetAttrString(op->func, "__doc__");
        if (unlikely(!op->func_doc)) return NULL;
#else
        if (((PyCFunctionObject*)op)->m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
#endif
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_doc, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_name == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_name = PyObject_GetAttrString(op->func, "__name__");
#elif PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_name, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_qualname, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_dict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(op);
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = __Pyx_PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = __Pyx_PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyTuple_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__defaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_tuple, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->defaults_tuple;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__kwdefaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_kwdict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->defaults_kwdict;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value || value == Py_None) {
        value = NULL;
    } else if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_annotations, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->func_annotations;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine(__pyx_CyFunctionObject *op, void *context) {
    int is_coroutine;
    CYTHON_UNUSED_VAR(context);
    if (op->func_is_coroutine) {
        return __Pyx_NewRef(op->func_is_coroutine);
    }
    is_coroutine = op->flags & __Pyx_CYFUNCTION_COROUTINE;
#if PY_VERSION_HEX >= 0x03050000
    if (is_coroutine) {
        PyObject *module, *fromlist, *marker = __pyx_n_s_is_coroutine;
        fromlist = PyList_New(1);
        if (unlikely(!fromlist)) return NULL;
        Py_INCREF(marker);
#if CYTHON_ASSUME_SAFE_MACROS
        PyList_SET_ITEM(fromlist, 0, marker);
#else
        if (unlikely(PyList_SetItem(fromlist, 0, marker) < 0)) {
            Py_DECREF(marker);
            Py_DECREF(fromlist);
            return NULL;
        }
#endif
        module = PyImport_ImportModuleLevelObject(__pyx_n_s_asyncio_coroutines, NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
        if (unlikely(!module)) goto ignore;
        op->func_is_coroutine = __Pyx_PyObject_GetAttrStr(module, marker);
        Py_DECREF(module);
        if (likely(op->func_is_coroutine)) {
            return __Pyx_NewRef(op->func_is_coroutine);
        }
ignore:
        PyErr_Clear();
    }
#endif
    op->func_is_coroutine = __Pyx_PyBool_FromLong(is_coroutine);
    return __Pyx_NewRef(op->func_is_coroutine);
}
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *
__Pyx_CyFunction_get_module(__pyx_CyFunctionObject *op, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_GetAttrString(op->func, "__module__");
}
static int
__Pyx_CyFunction_set_module(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_SetAttrString(op->func, "__module__", value);
}
#endif
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {(char *) "_is_coroutine", (getter)__Pyx_CyFunction_get_is_coroutine, 0, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", (getter)__Pyx_CyFunction_get_module, (setter)__Pyx_CyFunction_set_module, 0, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
#if !CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), 0, 0},
#endif
#if CYTHON_USE_TYPE_SPECS
    {(char *) "__dictoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_dict), READONLY, 0},
#if CYTHON_METH_FASTCALL
#if CYTHON_BACKPORT_VECTORCALL
    {(char *) "__vectorcalloffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_vectorcall), READONLY, 0},
#else
#if !CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__vectorcalloffset__", T_PYSSIZET, offsetof(PyCFunctionObject, vectorcall), READONLY, 0},
#endif
#endif
#endif
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_weakreflist), READONLY, 0},
#else
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(PyCFunctionObject, m_weakreflist), READONLY, 0},
#endif
#endif
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, PyObject *args)
{
    CYTHON_UNUSED_VAR(args);
#if PY_MAJOR_VERSION >= 3
    Py_INCREF(m->func_qualname);
    return m->func_qualname;
#else
    return PyString_FromString(((PyCFunctionObject*)m)->m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) (((PyCFunctionObject*)cyfunc)->m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
#if !CYTHON_COMPILING_IN_LIMITED_API
    PyCFunctionObject *cf = (PyCFunctionObject*) op;
#endif
    if (unlikely(op == NULL))
        return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    op->func = PyCFunction_NewEx(ml, (PyObject*)op, module);
    if (unlikely(!op->func)) return NULL;
#endif
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
#if !CYTHON_COMPILING_IN_LIMITED_API
    cf->m_ml = ml;
    cf->m_self = (PyObject *) op;
#endif
    Py_XINCREF(closure);
    op->func_closure = closure;
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_XINCREF(module);
    cf->m_module = module;
#endif
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_classobj = NULL;
#else
    ((PyCMethodObject*)op)->mm_class = NULL;
#endif
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults_size = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    op->func_is_coroutine = NULL;
#if CYTHON_METH_FASTCALL
    switch (ml->ml_flags & (METH_VARARGS | METH_FASTCALL | METH_NOARGS | METH_O | METH_KEYWORDS | METH_METHOD)) {
    case METH_NOARGS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_NOARGS;
        break;
    case METH_O:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_O;
        break;
    case METH_METHOD | METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD;
        break;
    case METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS;
        break;
    case METH_VARARGS | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = NULL;
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        Py_DECREF(op);
        return NULL;
    }
#endif
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func);
#else
    Py_CLEAR(((PyCFunctionObject*)m)->m_module);
#endif
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
#if PY_VERSION_HEX < 0x030900B1
    Py_CLEAR(__Pyx_CyFunction_GetClassObj(m));
#else
    {
        PyObject *cls = (PyObject*) ((PyCMethodObject *) (m))->mm_class;
        ((PyCMethodObject *) (m))->mm_class = NULL;
        Py_XDECREF(cls);
    }
#endif
#endif
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    Py_CLEAR(m->func_is_coroutine);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    __Pyx_PyHeapTypeObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func);
#else
    Py_VISIT(((PyCFunctionObject*)m)->m_module);
#endif
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(__Pyx_CyFunction_GetClassObj(m));
#endif
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    Py_VISIT(m->func_is_coroutine);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *f = ((__pyx_CyFunctionObject*)func)->func;
    PyObject *py_name = NULL;
    PyCFunction meth;
    int flags;
    meth = PyCFunction_GetFunction(f);
    if (unlikely(!meth)) return NULL;
    flags = PyCFunction_GetFlags(f);
    if (unlikely(flags < 0)) return NULL;
#else
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    int flags = f->m_ml->ml_flags;
#endif
    Py_ssize_t size;
    switch (flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 0))
                return (*meth)(self, NULL);
#if CYTHON_COMPILING_IN_LIMITED_API
            py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
            if (!py_name) return NULL;
            PyErr_Format(PyExc_TypeError,
                "%.200S() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                py_name, size);
            Py_DECREF(py_name);
#else
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
#endif
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = __Pyx_PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
#if CYTHON_COMPILING_IN_LIMITED_API
            py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
            if (!py_name) return NULL;
            PyErr_Format(PyExc_TypeError,
                "%.200S() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                py_name, size);
            Py_DECREF(py_name);
#else
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
#endif
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        return NULL;
    }
#if CYTHON_COMPILING_IN_LIMITED_API
    py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
    if (!py_name) return NULL;
    PyErr_Format(PyExc_TypeError, "%.200S() takes no keyword arguments",
                 py_name);
    Py_DECREF(py_name);
#else
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
#endif
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *self, *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)func)->func);
    if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
    self = ((PyCFunctionObject*)func)->m_self;
#endif
    result = __Pyx_CyFunction_CallMethod(func, self, arg, kw);
    return result;
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
#if CYTHON_METH_FASTCALL
     __pyx_vectorcallfunc vc = __Pyx_CyFunction_func_vectorcall(cyfunc);
    if (vc) {
#if CYTHON_ASSUME_SAFE_MACROS
        return __Pyx_PyVectorcall_FastCallDict(func, vc, &PyTuple_GET_ITEM(args, 0), (size_t)PyTuple_GET_SIZE(args), kw);
#else
        (void) &__Pyx_PyVectorcall_FastCallDict;
        return PyVectorcall_Call(func, args, kw);
#endif
    }
#endif
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
#if CYTHON_ASSUME_SAFE_MACROS
        argc = PyTuple_GET_SIZE(args);
#else
        argc = PyTuple_Size(args);
        if (unlikely(!argc) < 0) return NULL;
#endif
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
#if PY_MAJOR_VERSION > 2
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
#else
            PyErr_SetString(PyExc_TypeError,
                            "unbound method needs an argument");
#endif
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE int __Pyx_CyFunction_Vectorcall_CheckArgs(__pyx_CyFunctionObject *cyfunc, Py_ssize_t nargs, PyObject *kwnames)
{
    int ret = 0;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        if (unlikely(nargs < 1)) {
            PyErr_Format(PyExc_TypeError, "%.200s() needs an argument",
                         ((PyCFunctionObject*)cyfunc)->m_ml->ml_name);
            return -1;
        }
        ret = 1;
    }
    if (unlikely(kwnames) && unlikely(PyTuple_GET_SIZE(kwnames))) {
        PyErr_Format(PyExc_TypeError,
                     "%.200s() takes no keyword arguments", ((PyCFunctionObject*)cyfunc)->m_ml->ml_name);
        return -1;
    }
    return ret;
}
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 0)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, NULL);
}
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 1)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, args[0]);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))def->ml_meth)(self, args, nargs, kwnames);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
    PyTypeObject *cls = (PyTypeObject *) __Pyx_CyFunction_GetClassObj(cyfunc);
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCMethod)(void(*)(void))def->ml_meth)(self, cls, args, (size_t)nargs, kwnames);
}
#endif
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_CyFunctionType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_CyFunction_dealloc},
    {Py_tp_repr, (void *)__Pyx_CyFunction_repr},
    {Py_tp_call, (void *)__Pyx_CyFunction_CallAsMethod},
    {Py_tp_traverse, (void *)__Pyx_CyFunction_traverse},
    {Py_tp_clear, (void *)__Pyx_CyFunction_clear},
    {Py_tp_methods, (void *)__pyx_CyFunction_methods},
    {Py_tp_members, (void *)__pyx_CyFunction_members},
    {Py_tp_getset, (void *)__pyx_CyFunction_getsets},
    {Py_tp_descr_get, (void *)__Pyx_PyMethod_New},
    {0, 0},
};
static PyType_Spec __pyx_CyFunctionType_spec = {
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if (defined(_Py_TPFLAGS_HAVE_VECTORCALL) && CYTHON_METH_FASTCALL)
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    __pyx_CyFunctionType_slots
};
#else
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
#if !CYTHON_METH_FASTCALL
    0,
#elif CYTHON_BACKPORT_VECTORCALL
    (printfunc)offsetof(__pyx_CyFunctionObject, func_vectorcall),
#else
    offsetof(PyCFunctionObject, vectorcall),
#endif
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if defined(_Py_TPFLAGS_HAVE_VECTORCALL) && CYTHON_METH_FASTCALL
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_PyMethod_New,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if __PYX_NEED_TP_PRINT_SLOT
    0,
#endif
#if PY_VERSION_HEX >= 0x030C0000
    0,
#endif
#if PY_VERSION_HEX >= 0x030d00A4
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
    0,
#endif
};
#endif
static int __pyx_CyFunction_init(PyObject *module) {
#if CYTHON_USE_TYPE_SPECS
    __pyx_CyFunctionType = __Pyx_FetchCommonTypeFromSpec(module, &__pyx_CyFunctionType_spec, NULL);
#else
    CYTHON_UNUSED_VAR(module);
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
#endif
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    m->defaults_size = size;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    CYTHON_MAYBE_UNUSED_VAR(tstate);
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStrNoError(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}
#endif

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API && !defined(PYPY_VERSION)
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyCode_Replace_For_AddTraceback(PyObject *code, PyObject *scratch_dict,
                                                       PyObject *firstlineno, PyObject *name) {
    PyObject *replace = NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_firstlineno", firstlineno))) return NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_name", name))) return NULL;
    replace = PyObject_GetAttrString(code, "replace");
    if (likely(replace)) {
        PyObject *result;
        result = PyObject_Call(replace, __pyx_empty_tuple, scratch_dict);
        Py_DECREF(replace);
        return result;
    }
    PyErr_Clear();
    #if __PYX_LIMITED_VERSION_HEX < 0x030780000
    {
        PyObject *compiled = NULL, *result = NULL;
        if (unlikely(PyDict_SetItemString(scratch_dict, "code", code))) return NULL;
        if (unlikely(PyDict_SetItemString(scratch_dict, "type", (PyObject*)(&PyType_Type)))) return NULL;
        compiled = Py_CompileString(
            "out = type(code)(\n"
            "  code.co_argcount, code.co_kwonlyargcount, code.co_nlocals, code.co_stacksize,\n"
            "  code.co_flags, code.co_code, code.co_consts, code.co_names,\n"
            "  code.co_varnames, code.co_filename, co_name, co_firstlineno,\n"
            "  code.co_lnotab)\n", "<dummy>", Py_file_input);
        if (!compiled) return NULL;
        result = PyEval_EvalCode(compiled, scratch_dict, scratch_dict);
        Py_DECREF(compiled);
        if (!result) PyErr_Print();
        Py_DECREF(result);
        result = PyDict_GetItemString(scratch_dict, "out");
        if (result) Py_INCREF(result);
        return result;
    }
    #else
    return NULL;
    #endif
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyObject *code_object = NULL, *py_py_line = NULL, *py_funcname = NULL, *dict = NULL;
    PyObject *replace = NULL, *getframe = NULL, *frame = NULL;
    PyObject *exc_type, *exc_value, *exc_traceback;
    int success = 0;
    if (c_line) {
        (void) __pyx_cfilenm;
        (void) __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
    }
    PyErr_Fetch(&exc_type, &exc_value, &exc_traceback);
    code_object = Py_CompileString("_getframe()", filename, Py_eval_input);
    if (unlikely(!code_object)) goto bad;
    py_py_line = PyLong_FromLong(py_line);
    if (unlikely(!py_py_line)) goto bad;
    py_funcname = PyUnicode_FromString(funcname);
    if (unlikely(!py_funcname)) goto bad;
    dict = PyDict_New();
    if (unlikely(!dict)) goto bad;
    {
        PyObject *old_code_object = code_object;
        code_object = __Pyx_PyCode_Replace_For_AddTraceback(code_object, dict, py_py_line, py_funcname);
        Py_DECREF(old_code_object);
    }
    if (unlikely(!code_object)) goto bad;
    getframe = PySys_GetObject("_getframe");
    if (unlikely(!getframe)) goto bad;
    if (unlikely(PyDict_SetItemString(dict, "_getframe", getframe))) goto bad;
    frame = PyEval_EvalCode(code_object, dict, dict);
    if (unlikely(!frame) || frame == Py_None) goto bad;
    success = 1;
  bad:
    PyErr_Restore(exc_type, exc_value, exc_traceback);
    Py_XDECREF(code_object);
    Py_XDECREF(py_py_line);
    Py_XDECREF(py_funcname);
    Py_XDECREF(dict);
    Py_XDECREF(replace);
    if (success) {
        PyTraceBack_Here(
            (struct _frame*)frame);
    }
    Py_XDECREF(frame);
}
#else
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}
#endif

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* FromPyCTupleUtility */
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
static void __Pyx_tuple___pyx_convert__from_py___pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float(PyObject * o, __pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float *result) {
        result->f0 = __pyx_PyFloat_AsFloat(PyTuple_GET_ITEM(o, 0));
        if ((result->f0 == (float)-1) && PyErr_Occurred()) goto bad;
        result->f1 = __pyx_PyFloat_AsFloat(PyTuple_GET_ITEM(o, 1));
        if ((result->f1 == (float)-1) && PyErr_Occurred()) goto bad;
        result->f2 = __pyx_PyFloat_AsFloat(PyTuple_GET_ITEM(o, 2));
        if ((result->f2 == (float)-1) && PyErr_Occurred()) goto bad;
        result->f3 = __pyx_PyFloat_AsFloat(PyTuple_GET_ITEM(o, 3));
        if ((result->f3 == (float)-1) && PyErr_Occurred()) goto bad;
        result->f4 = __pyx_PyFloat_AsFloat(PyTuple_GET_ITEM(o, 4));
        if ((result->f4 == (float)-1) && PyErr_Occurred()) goto bad;
        result->f5 = __pyx_PyFloat_AsFloat(PyTuple_GET_ITEM(o, 5));
        if ((result->f5 == (float)-1) && PyErr_Occurred()) goto bad;
    return;
bad:
    return;
}
static void __Pyx_list___pyx_convert__from_py___pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float(PyObject * o, __pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float *result) {
        result->f0 = __pyx_PyFloat_AsFloat(PyList_GET_ITEM(o, 0));
        if ((result->f0 == (float)-1) && PyErr_Occurred()) goto bad;
        result->f1 = __pyx_PyFloat_AsFloat(PyList_GET_ITEM(o, 1));
        if ((result->f1 == (float)-1) && PyErr_Occurred()) goto bad;
        result->f2 = __pyx_PyFloat_AsFloat(PyList_GET_ITEM(o, 2));
        if ((result->f2 == (float)-1) && PyErr_Occurred()) goto bad;
        result->f3 = __pyx_PyFloat_AsFloat(PyList_GET_ITEM(o, 3));
        if ((result->f3 == (float)-1) && PyErr_Occurred()) goto bad;
        result->f4 = __pyx_PyFloat_AsFloat(PyList_GET_ITEM(o, 4));
        if ((result->f4 == (float)-1) && PyErr_Occurred()) goto bad;
        result->f5 = __pyx_PyFloat_AsFloat(PyList_GET_ITEM(o, 5));
        if ((result->f5 == (float)-1) && PyErr_Occurred()) goto bad;
    return;
bad:
    return;
}
#endif
static void __Pyx_seq___pyx_convert__from_py___pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float(PyObject * o, __pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float *result) {
    if (unlikely(!PySequence_Check(o))) {
        __Pyx_TypeName o_type_name = __Pyx_PyType_GetName(Py_TYPE(o));
        PyErr_Format(PyExc_TypeError,
                     "Expected a sequence of size %zd, got " __Pyx_FMT_TYPENAME, (Py_ssize_t) 6, o_type_name);
        __Pyx_DECREF_TypeName(o_type_name);
        goto bad;
    } else if (unlikely(PySequence_Length(o) != 6)) {
        PyErr_Format(PyExc_TypeError,
                     "Expected a sequence of size %zd, got size %zd", (Py_ssize_t) 6, PySequence_Length(o));
        goto bad;
    }
    {
        PyObject *item;
        item = PySequence_ITEM(o, 0);  if (unlikely(!item)) goto bad;
        result->f0 = __pyx_PyFloat_AsFloat(item);
        Py_DECREF(item);
        if ((result->f0 == (float)-1) && PyErr_Occurred()) goto bad;
        item = PySequence_ITEM(o, 1);  if (unlikely(!item)) goto bad;
        result->f1 = __pyx_PyFloat_AsFloat(item);
        Py_DECREF(item);
        if ((result->f1 == (float)-1) && PyErr_Occurred()) goto bad;
        item = PySequence_ITEM(o, 2);  if (unlikely(!item)) goto bad;
        result->f2 = __pyx_PyFloat_AsFloat(item);
        Py_DECREF(item);
        if ((result->f2 == (float)-1) && PyErr_Occurred()) goto bad;
        item = PySequence_ITEM(o, 3);  if (unlikely(!item)) goto bad;
        result->f3 = __pyx_PyFloat_AsFloat(item);
        Py_DECREF(item);
        if ((result->f3 == (float)-1) && PyErr_Occurred()) goto bad;
        item = PySequence_ITEM(o, 4);  if (unlikely(!item)) goto bad;
        result->f4 = __pyx_PyFloat_AsFloat(item);
        Py_DECREF(item);
        if ((result->f4 == (float)-1) && PyErr_Occurred()) goto bad;
        item = PySequence_ITEM(o, 5);  if (unlikely(!item)) goto bad;
        result->f5 = __pyx_PyFloat_AsFloat(item);
        Py_DECREF(item);
        if ((result->f5 == (float)-1) && PyErr_Occurred()) goto bad;
    }
    return;
bad:
    return;
}
static CYTHON_INLINE __pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float __pyx_convert__from_py___pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float(PyObject * o) {
    __pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float result;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_Check(o) && PyTuple_GET_SIZE(o) == 6)) {
        __Pyx_tuple___pyx_convert__from_py___pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float(o, &result);
    } else if (likely(PyList_Check(o) && PyList_GET_SIZE(o) == 6)) {
        __Pyx_list___pyx_convert__from_py___pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float(o, &result);
    } else
    #endif
    {
        __Pyx_seq___pyx_convert__from_py___pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float(o, &result);
    }
    return result;
}

/* FromPyCTupleUtility */
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
static void __Pyx_tuple___pyx_convert__from_py___pyx_ctuple_float__and_float(PyObject * o, __pyx_ctuple_float__and_float *result) {
        result->f0 = __pyx_PyFloat_AsFloat(PyTuple_GET_ITEM(o, 0));
        if ((result->f0 == (float)-1) && PyErr_Occurred()) goto bad;
        result->f1 = __pyx_PyFloat_AsFloat(PyTuple_GET_ITEM(o, 1));
        if ((result->f1 == (float)-1) && PyErr_Occurred()) goto bad;
    return;
bad:
    return;
}
static void __Pyx_list___pyx_convert__from_py___pyx_ctuple_float__and_float(PyObject * o, __pyx_ctuple_float__and_float *result) {
        result->f0 = __pyx_PyFloat_AsFloat(PyList_GET_ITEM(o, 0));
        if ((result->f0 == (float)-1) && PyErr_Occurred()) goto bad;
        result->f1 = __pyx_PyFloat_AsFloat(PyList_GET_ITEM(o, 1));
        if ((result->f1 == (float)-1) && PyErr_Occurred()) goto bad;
    return;
bad:
    return;
}
#endif
static void __Pyx_seq___pyx_convert__from_py___pyx_ctuple_float__and_float(PyObject * o, __pyx_ctuple_float__and_float *result) {
    if (unlikely(!PySequence_Check(o))) {
        __Pyx_TypeName o_type_name = __Pyx_PyType_GetName(Py_TYPE(o));
        PyErr_Format(PyExc_TypeError,
                     "Expected a sequence of size %zd, got " __Pyx_FMT_TYPENAME, (Py_ssize_t) 2, o_type_name);
        __Pyx_DECREF_TypeName(o_type_name);
        goto bad;
    } else if (unlikely(PySequence_Length(o) != 2)) {
        PyErr_Format(PyExc_TypeError,
                     "Expected a sequence of size %zd, got size %zd", (Py_ssize_t) 2, PySequence_Length(o));
        goto bad;
    }
    {
        PyObject *item;
        item = PySequence_ITEM(o, 0);  if (unlikely(!item)) goto bad;
        result->f0 = __pyx_PyFloat_AsFloat(item);
        Py_DECREF(item);
        if ((result->f0 == (float)-1) && PyErr_Occurred()) goto bad;
        item = PySequence_ITEM(o, 1);  if (unlikely(!item)) goto bad;
        result->f1 = __pyx_PyFloat_AsFloat(item);
        Py_DECREF(item);
        if ((result->f1 == (float)-1) && PyErr_Occurred()) goto bad;
    }
    return;
bad:
    return;
}
static CYTHON_INLINE __pyx_ctuple_float__and_float __pyx_convert__from_py___pyx_ctuple_float__and_float(PyObject * o) {
    __pyx_ctuple_float__and_float result;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_Check(o) && PyTuple_GET_SIZE(o) == 2)) {
        __Pyx_tuple___pyx_convert__from_py___pyx_ctuple_float__and_float(o, &result);
    } else if (likely(PyList_Check(o) && PyList_GET_SIZE(o) == 2)) {
        __Pyx_list___pyx_convert__from_py___pyx_ctuple_float__and_float(o, &result);
    } else
    #endif
    {
        __Pyx_seq___pyx_convert__from_py___pyx_ctuple_float__and_float(o, &result);
    }
    return result;
}

/* RaiseUnexpectedTypeError */
static int
__Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj)
{
    __Pyx_TypeName obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError, "Expected %s, got " __Pyx_FMT_TYPENAME,
                 expected, obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* FromPyCTupleUtility */
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
static void __Pyx_tuple___pyx_convert__from_py___pyx_ctuple_a8db9c__float__and_float__and_float__and_float__and_fl__etc(PyObject * o, __pyx_ctuple_a8db9c__float__and_float__and_float__and_float__and_fl__etc *result) {
        result->f0 = __pyx_PyFloat_AsFloat(PyTuple_GET_ITEM(o, 0));
        if ((result->f0 == (float)-1) && PyErr_Occurred()) goto bad;
        result->f1 = __pyx_PyFloat_AsFloat(PyTuple_GET_ITEM(o, 1));
        if ((result->f1 == (float)-1) && PyErr_Occurred()) goto bad;
        result->f2 = __pyx_PyFloat_AsFloat(PyTuple_GET_ITEM(o, 2));
        if ((result->f2 == (float)-1) && PyErr_Occurred()) goto bad;
        result->f3 = __pyx_PyFloat_AsFloat(PyTuple_GET_ITEM(o, 3));
        if ((result->f3 == (float)-1) && PyErr_Occurred()) goto bad;
        result->f4 = __pyx_PyFloat_AsFloat(PyTuple_GET_ITEM(o, 4));
        if ((result->f4 == (float)-1) && PyErr_Occurred()) goto bad;
        result->f5 = __pyx_PyFloat_AsFloat(PyTuple_GET_ITEM(o, 5));
        if ((result->f5 == (float)-1) && PyErr_Occurred()) goto bad;
        result->f6 = __pyx_PyFloat_AsFloat(PyTuple_GET_ITEM(o, 6));
        if ((result->f6 == (float)-1) && PyErr_Occurred()) goto bad;
        result->f7 = __pyx_PyFloat_AsFloat(PyTuple_GET_ITEM(o, 7));
        if ((result->f7 == (float)-1) && PyErr_Occurred()) goto bad;
        result->f8 = __pyx_PyFloat_AsFloat(PyTuple_GET_ITEM(o, 8));
        if ((result->f8 == (float)-1) && PyErr_Occurred()) goto bad;
        result->f9 = __pyx_PyFloat_AsFloat(PyTuple_GET_ITEM(o, 9));
        if ((result->f9 == (float)-1) && PyErr_Occurred()) goto bad;
        result->f10 = __pyx_PyFloat_AsFloat(PyTuple_GET_ITEM(o, 10));
        if ((result->f10 == (float)-1) && PyErr_Occurred()) goto bad;
        result->f11 = __pyx_PyFloat_AsFloat(PyTuple_GET_ITEM(o, 11));
        if ((result->f11 == (float)-1) && PyErr_Occurred()) goto bad;
        result->f12 = __pyx_PyFloat_AsFloat(PyTuple_GET_ITEM(o, 12));
        if ((result->f12 == (float)-1) && PyErr_Occurred()) goto bad;
    return;
bad:
    return;
}
static void __Pyx_list___pyx_convert__from_py___pyx_ctuple_a8db9c__float__and_float__and_float__and_float__and_fl__etc(PyObject * o, __pyx_ctuple_a8db9c__float__and_float__and_float__and_float__and_fl__etc *result) {
        result->f0 = __pyx_PyFloat_AsFloat(PyList_GET_ITEM(o, 0));
        if ((result->f0 == (float)-1) && PyErr_Occurred()) goto bad;
        result->f1 = __pyx_PyFloat_AsFloat(PyList_GET_ITEM(o, 1));
        if ((result->f1 == (float)-1) && PyErr_Occurred()) goto bad;
        result->f2 = __pyx_PyFloat_AsFloat(PyList_GET_ITEM(o, 2));
        if ((result->f2 == (float)-1) && PyErr_Occurred()) goto bad;
        result->f3 = __pyx_PyFloat_AsFloat(PyList_GET_ITEM(o, 3));
        if ((result->f3 == (float)-1) && PyErr_Occurred()) goto bad;
        result->f4 = __pyx_PyFloat_AsFloat(PyList_GET_ITEM(o, 4));
        if ((result->f4 == (float)-1) && PyErr_Occurred()) goto bad;
        result->f5 = __pyx_PyFloat_AsFloat(PyList_GET_ITEM(o, 5));
        if ((result->f5 == (float)-1) && PyErr_Occurred()) goto bad;
        result->f6 = __pyx_PyFloat_AsFloat(PyList_GET_ITEM(o, 6));
        if ((result->f6 == (float)-1) && PyErr_Occurred()) goto bad;
        result->f7 = __pyx_PyFloat_AsFloat(PyList_GET_ITEM(o, 7));
        if ((result->f7 == (float)-1) && PyErr_Occurred()) goto bad;
        result->f8 = __pyx_PyFloat_AsFloat(PyList_GET_ITEM(o, 8));
        if ((result->f8 == (float)-1) && PyErr_Occurred()) goto bad;
        result->f9 = __pyx_PyFloat_AsFloat(PyList_GET_ITEM(o, 9));
        if ((result->f9 == (float)-1) && PyErr_Occurred()) goto bad;
        result->f10 = __pyx_PyFloat_AsFloat(PyList_GET_ITEM(o, 10));
        if ((result->f10 == (float)-1) && PyErr_Occurred()) goto bad;
        result->f11 = __pyx_PyFloat_AsFloat(PyList_GET_ITEM(o, 11));
        if ((result->f11 == (float)-1) && PyErr_Occurred()) goto bad;
        result->f12 = __pyx_PyFloat_AsFloat(PyList_GET_ITEM(o, 12));
        if ((result->f12 == (float)-1) && PyErr_Occurred()) goto bad;
    return;
bad:
    return;
}
#endif
static void __Pyx_seq___pyx_convert__from_py___pyx_ctuple_a8db9c__float__and_float__and_float__and_float__and_fl__etc(PyObject * o, __pyx_ctuple_a8db9c__float__and_float__and_float__and_float__and_fl__etc *result) {
    if (unlikely(!PySequence_Check(o))) {
        __Pyx_TypeName o_type_name = __Pyx_PyType_GetName(Py_TYPE(o));
        PyErr_Format(PyExc_TypeError,
                     "Expected a sequence of size %zd, got " __Pyx_FMT_TYPENAME, (Py_ssize_t) 13, o_type_name);
        __Pyx_DECREF_TypeName(o_type_name);
        goto bad;
    } else if (unlikely(PySequence_Length(o) != 13)) {
        PyErr_Format(PyExc_TypeError,
                     "Expected a sequence of size %zd, got size %zd", (Py_ssize_t) 13, PySequence_Length(o));
        goto bad;
    }
    {
        PyObject *item;
        item = PySequence_ITEM(o, 0);  if (unlikely(!item)) goto bad;
        result->f0 = __pyx_PyFloat_AsFloat(item);
        Py_DECREF(item);
        if ((result->f0 == (float)-1) && PyErr_Occurred()) goto bad;
        item = PySequence_ITEM(o, 1);  if (unlikely(!item)) goto bad;
        result->f1 = __pyx_PyFloat_AsFloat(item);
        Py_DECREF(item);
        if ((result->f1 == (float)-1) && PyErr_Occurred()) goto bad;
        item = PySequence_ITEM(o, 2);  if (unlikely(!item)) goto bad;
        result->f2 = __pyx_PyFloat_AsFloat(item);
        Py_DECREF(item);
        if ((result->f2 == (float)-1) && PyErr_Occurred()) goto bad;
        item = PySequence_ITEM(o, 3);  if (unlikely(!item)) goto bad;
        result->f3 = __pyx_PyFloat_AsFloat(item);
        Py_DECREF(item);
        if ((result->f3 == (float)-1) && PyErr_Occurred()) goto bad;
        item = PySequence_ITEM(o, 4);  if (unlikely(!item)) goto bad;
        result->f4 = __pyx_PyFloat_AsFloat(item);
        Py_DECREF(item);
        if ((result->f4 == (float)-1) && PyErr_Occurred()) goto bad;
        item = PySequence_ITEM(o, 5);  if (unlikely(!item)) goto bad;
        result->f5 = __pyx_PyFloat_AsFloat(item);
        Py_DECREF(item);
        if ((result->f5 == (float)-1) && PyErr_Occurred()) goto bad;
        item = PySequence_ITEM(o, 6);  if (unlikely(!item)) goto bad;
        result->f6 = __pyx_PyFloat_AsFloat(item);
        Py_DECREF(item);
        if ((result->f6 == (float)-1) && PyErr_Occurred()) goto bad;
        item = PySequence_ITEM(o, 7);  if (unlikely(!item)) goto bad;
        result->f7 = __pyx_PyFloat_AsFloat(item);
        Py_DECREF(item);
        if ((result->f7 == (float)-1) && PyErr_Occurred()) goto bad;
        item = PySequence_ITEM(o, 8);  if (unlikely(!item)) goto bad;
        result->f8 = __pyx_PyFloat_AsFloat(item);
        Py_DECREF(item);
        if ((result->f8 == (float)-1) && PyErr_Occurred()) goto bad;
        item = PySequence_ITEM(o, 9);  if (unlikely(!item)) goto bad;
        result->f9 = __pyx_PyFloat_AsFloat(item);
        Py_DECREF(item);
        if ((result->f9 == (float)-1) && PyErr_Occurred()) goto bad;
        item = PySequence_ITEM(o, 10);  if (unlikely(!item)) goto bad;
        result->f10 = __pyx_PyFloat_AsFloat(item);
        Py_DECREF(item);
        if ((result->f10 == (float)-1) && PyErr_Occurred()) goto bad;
        item = PySequence_ITEM(o, 11);  if (unlikely(!item)) goto bad;
        result->f11 = __pyx_PyFloat_AsFloat(item);
        Py_DECREF(item);
        if ((result->f11 == (float)-1) && PyErr_Occurred()) goto bad;
        item = PySequence_ITEM(o, 12);  if (unlikely(!item)) goto bad;
        result->f12 = __pyx_PyFloat_AsFloat(item);
        Py_DECREF(item);
        if ((result->f12 == (float)-1) && PyErr_Occurred()) goto bad;
    }
    return;
bad:
    return;
}
static CYTHON_INLINE __pyx_ctuple_a8db9c__float__and_float__and_float__and_float__and_fl__etc __pyx_convert__from_py___pyx_ctuple_a8db9c__float__and_float__and_float__and_float__and_fl__etc(PyObject * o) {
    __pyx_ctuple_a8db9c__float__and_float__and_float__and_float__and_fl__etc result;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_Check(o) && PyTuple_GET_SIZE(o) == 13)) {
        __Pyx_tuple___pyx_convert__from_py___pyx_ctuple_a8db9c__float__and_float__and_float__and_float__and_fl__etc(o, &result);
    } else if (likely(PyList_Check(o) && PyList_GET_SIZE(o) == 13)) {
        __Pyx_list___pyx_convert__from_py___pyx_ctuple_a8db9c__float__and_float__and_float__and_float__and_fl__etc(o, &result);
    } else
    #endif
    {
        __Pyx_seq___pyx_convert__from_py___pyx_ctuple_a8db9c__float__and_float__and_float__and_float__and_fl__etc(o, &result);
    }
    return result;
}

/* ToPyCTupleUtility */
static PyObject* __pyx_convert__to_py___pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float(__pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float value) {
    PyObject* item = NULL;
    PyObject* result = PyTuple_New(6);
    if (!result) goto bad;
        item = PyFloat_FromDouble(value.f0);
        if (!item) goto bad;
        PyTuple_SET_ITEM(result, 0, item);
        item = PyFloat_FromDouble(value.f1);
        if (!item) goto bad;
        PyTuple_SET_ITEM(result, 1, item);
        item = PyFloat_FromDouble(value.f2);
        if (!item) goto bad;
        PyTuple_SET_ITEM(result, 2, item);
        item = PyFloat_FromDouble(value.f3);
        if (!item) goto bad;
        PyTuple_SET_ITEM(result, 3, item);
        item = PyFloat_FromDouble(value.f4);
        if (!item) goto bad;
        PyTuple_SET_ITEM(result, 4, item);
        item = PyFloat_FromDouble(value.f5);
        if (!item) goto bad;
        PyTuple_SET_ITEM(result, 5, item);
    return result;
bad:
    Py_XDECREF(item);
    Py_XDECREF(result);
    return NULL;
}

/* ToPyCTupleUtility */
static PyObject* __pyx_convert__to_py___pyx_ctuple_float__and_float(__pyx_ctuple_float__and_float value) {
    PyObject* item = NULL;
    PyObject* result = PyTuple_New(2);
    if (!result) goto bad;
        item = PyFloat_FromDouble(value.f0);
        if (!item) goto bad;
        PyTuple_SET_ITEM(result, 0, item);
        item = PyFloat_FromDouble(value.f1);
        if (!item) goto bad;
        PyTuple_SET_ITEM(result, 1, item);
    return result;
bad:
    Py_XDECREF(item);
    Py_XDECREF(result);
    return NULL;
}

static PyObject* __pyx_convert__to_py_struct____pyx_t_12pyretechnics_8cy_types_FuelModel(struct __pyx_t_12pyretechnics_8cy_types_FuelModel s) {
  PyObject* res;
  PyObject* member;
  res = __Pyx_PyDict_NewPresized(17); if (unlikely(!res)) return NULL;
  member = __Pyx_PyInt_From_int(s.number); if (unlikely(!member)) goto bad;
  if (unlikely(PyDict_SetItem(res, __pyx_n_s_number, member) < 0)) goto bad;
  Py_DECREF(member);
  member = PyFloat_FromDouble(s.delta); if (unlikely(!member)) goto bad;
  if (unlikely(PyDict_SetItem(res, __pyx_n_s_delta, member) < 0)) goto bad;
  Py_DECREF(member);
  member = __pyx_convert__to_py___pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float(s.M_x); if (unlikely(!member)) goto bad;
  if (unlikely(PyDict_SetItem(res, __pyx_n_s_M_x, member) < 0)) goto bad;
  Py_DECREF(member);
  member = __pyx_convert__to_py___pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float(s.M_f); if (unlikely(!member)) goto bad;
  if (unlikely(PyDict_SetItem(res, __pyx_n_s_M_f, member) < 0)) goto bad;
  Py_DECREF(member);
  member = __pyx_convert__to_py___pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float(s.w_o); if (unlikely(!member)) goto bad;
  if (unlikely(PyDict_SetItem(res, __pyx_n_s_w_o, member) < 0)) goto bad;
  Py_DECREF(member);
  member = __pyx_convert__to_py___pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float(s.sigma); if (unlikely(!member)) goto bad;
  if (unlikely(PyDict_SetItem(res, __pyx_n_s_sigma, member) < 0)) goto bad;
  Py_DECREF(member);
  member = __pyx_convert__to_py___pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float(s.h); if (unlikely(!member)) goto bad;
  if (unlikely(PyDict_SetItem(res, __pyx_n_s_h, member) < 0)) goto bad;
  Py_DECREF(member);
  member = __pyx_convert__to_py___pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float(s.rho_p); if (unlikely(!member)) goto bad;
  if (unlikely(PyDict_SetItem(res, __pyx_n_s_rho_p, member) < 0)) goto bad;
  Py_DECREF(member);
  member = __pyx_convert__to_py___pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float(s.S_T); if (unlikely(!member)) goto bad;
  if (unlikely(PyDict_SetItem(res, __pyx_n_s_S_T, member) < 0)) goto bad;
  Py_DECREF(member);
  member = __pyx_convert__to_py___pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float(s.S_e); if (unlikely(!member)) goto bad;
  if (unlikely(PyDict_SetItem(res, __pyx_n_s_S_e, member) < 0)) goto bad;
  Py_DECREF(member);
  member = __Pyx_PyBool_FromLong(s.dynamic); if (unlikely(!member)) goto bad;
  if (unlikely(PyDict_SetItem(res, __pyx_n_s_dynamic, member) < 0)) goto bad;
  Py_DECREF(member);
  member = __Pyx_PyBool_FromLong(s.burnable); if (unlikely(!member)) goto bad;
  if (unlikely(PyDict_SetItem(res, __pyx_n_s_burnable, member) < 0)) goto bad;
  Py_DECREF(member);
  member = __pyx_convert__to_py___pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float(s.exp_A_sigma); if (unlikely(!member)) goto bad;
  if (unlikely(PyDict_SetItem(res, __pyx_n_s_exp_A_sigma, member) < 0)) goto bad;
  Py_DECREF(member);
  member = __pyx_convert__to_py___pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float(s.firemod_size_classes); if (unlikely(!member)) goto bad;
  if (unlikely(PyDict_SetItem(res, __pyx_n_s_firemod_size_classes, member) < 0)) goto bad;
  Py_DECREF(member);
  member = __pyx_convert__to_py___pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float(s.f_ij); if (unlikely(!member)) goto bad;
  if (unlikely(PyDict_SetItem(res, __pyx_n_s_f_ij, member) < 0)) goto bad;
  Py_DECREF(member);
  member = __pyx_convert__to_py___pyx_ctuple_float__and_float(s.f_i); if (unlikely(!member)) goto bad;
  if (unlikely(PyDict_SetItem(res, __pyx_n_s_f_i, member) < 0)) goto bad;
  Py_DECREF(member);
  member = __pyx_convert__to_py___pyx_ctuple_float__and_float__and_float__and_float__and_float__and_float(s.g_ij); if (unlikely(!member)) goto bad;
  if (unlikely(PyDict_SetItem(res, __pyx_n_s_g_ij, member) < 0)) goto bad;
  Py_DECREF(member);
  return res;
  bad:
  Py_XDECREF(member);
  Py_DECREF(res);
  return NULL;
}
/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(int) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 2 * PyLong_SHIFT)) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 3 * PyLong_SHIFT)) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 4 * PyLong_SHIFT)) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        int val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int) 1) << (sizeof(int) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* FormatTypeName */
#if CYTHON_COMPILING_IN_LIMITED_API
static __Pyx_TypeName
__Pyx_PyType_GetName(PyTypeObject* tp)
{
    PyObject *name = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_n_s_name);
    if (unlikely(name == NULL) || unlikely(!PyUnicode_Check(name))) {
        PyErr_Clear();
        Py_XDECREF(name);
        name = __Pyx_NewRef(__pyx_n_s__79);
    }
    return name;
}
#endif

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(long) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 2 * PyLong_SHIFT)) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 3 * PyLong_SHIFT)) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 4 * PyLong_SHIFT)) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (long) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(long) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(long) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(long) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(long) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(long) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        long val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (long) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (long) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (long) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (long) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(long) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((long) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(long) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((long) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((long) 1) << (sizeof(long) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (long) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = __Pyx_PyType_GetSlot(a, tp_base, PyTypeObject*);
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (cls == a || cls == b) return 1;
    mro = cls->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            PyObject *base = PyTuple_GET_ITEM(mro, i);
            if (base == (PyObject *)a || base == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    if (exc_type1) {
        return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
    } else {
        return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
static unsigned long __Pyx_get_runtime_version(void) {
#if __PYX_LIMITED_VERSION_HEX >= 0x030B00A4
    return Py_Version & ~0xFFUL;
#else
    const char* rt_version = Py_GetVersion();
    unsigned long version = 0;
    unsigned long factor = 0x01000000UL;
    unsigned int digit = 0;
    int i = 0;
    while (factor) {
        while ('0' <= rt_version[i] && rt_version[i] <= '9') {
            digit = digit * 10 + (unsigned int) (rt_version[i] - '0');
            ++i;
        }
        version += factor * digit;
        if (rt_version[i] != '.')
            break;
        digit = 0;
        factor >>= 8;
        ++i;
    }
    return version;
#endif
}
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer) {
    const unsigned long MAJOR_MINOR = 0xFFFF0000UL;
    if ((rt_version & MAJOR_MINOR) == (ct_version & MAJOR_MINOR))
        return 0;
    if (likely(allow_newer && (rt_version & MAJOR_MINOR) > (ct_version & MAJOR_MINOR)))
        return 1;
    {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compile time Python version %d.%d "
                      "of module '%.100s' "
                      "%s "
                      "runtime version %d.%d",
                       (int) (ct_version >> 24), (int) ((ct_version >> 16) & 0xFF),
                       __Pyx_MODULE_NAME,
                       (allow_newer) ? "was newer than" : "does not match",
                       (int) (rt_version >> 24), (int) ((rt_version >> 16) & 0xFF)
       );
        return PyErr_WarnEx(NULL, message, 1);
    }
}

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* VoidPtrExport */
static int __Pyx_ExportVoidPtr(PyObject *name, void *p, const char *sig) {
    PyObject *d;
    PyObject *cobj = 0;
    d = PyDict_GetItem(__pyx_d, __pyx_n_s_pyx_capi);
    Py_XINCREF(d);
    if (!d) {
        d = PyDict_New();
        if (!d)
            goto bad;
        if (__Pyx_PyObject_SetAttrStr(__pyx_m, __pyx_n_s_pyx_capi, d) < 0)
            goto bad;
    }
    cobj = PyCapsule_New(p, sig, 0);
    if (!cobj)
        goto bad;
    if (PyDict_SetItem(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* FunctionExport */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(__pyx_m, (char *)"__pyx_capi__");
    if (!d) {
        PyErr_Clear();
        d = PyDict_New();
        if (!d)
            goto bad;
        Py_INCREF(d);
        if (PyModule_AddObject(__pyx_m, (char *)"__pyx_capi__", d) < 0)
            goto bad;
    }
    tmp.fp = f;
    cobj = PyCapsule_New(tmp.p, sig, 0);
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* InitStrings */
#if PY_MAJOR_VERSION >= 3
static int __Pyx_InitString(__Pyx_StringTabEntry t, PyObject **str) {
    if (t.is_unicode | t.is_str) {
        if (t.intern) {
            *str = PyUnicode_InternFromString(t.s);
        } else if (t.encoding) {
            *str = PyUnicode_Decode(t.s, t.n - 1, t.encoding, NULL);
        } else {
            *str = PyUnicode_FromStringAndSize(t.s, t.n - 1);
        }
    } else {
        *str = PyBytes_FromStringAndSize(t.s, t.n - 1);
    }
    if (!*str)
        return -1;
    if (PyObject_Hash(*str) == -1)
        return -1;
    return 0;
}
#endif
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION >= 3
        __Pyx_InitString(*t, t->p);
        #else
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        #endif
        ++t;
    }
    return 0;
}

#include <string.h>
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s) {
    size_t len = strlen(s);
    if (unlikely(len > (size_t) PY_SSIZE_T_MAX)) {
        PyErr_SetString(PyExc_OverflowError, "byte string is too long");
        return -1;
    }
    return (Py_ssize_t) len;
}
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return __Pyx_PyUnicode_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return PyByteArray_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY && !CYTHON_COMPILING_IN_LIMITED_API) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
    __Pyx_TypeName result_type_name = __Pyx_PyType_GetName(Py_TYPE(result));
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ").  "
                "The ability to return an instance of a strict subclass of int is deprecated, "
                "and may be removed in a future version of Python.",
                result_type_name)) {
            __Pyx_DECREF_TypeName(result_type_name);
            Py_DECREF(result);
            return NULL;
        }
        __Pyx_DECREF_TypeName(result_type_name);
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type " __Pyx_FMT_TYPENAME ")",
                 type_name, type_name, result_type_name);
    __Pyx_DECREF_TypeName(result_type_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(__Pyx_PyLong_IsCompact(b))) {
        return __Pyx_PyLong_CompactValue(b);
    } else {
      const digit* digits = __Pyx_PyLong_Digits(b);
      const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(b);
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


/* #### Code section: utility_code_pragmas_end ### */
#ifdef _MSC_VER
#pragma warning( pop )
#endif



/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
